/*
 *  2011 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2011 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 08/01/2011
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import javax.xml.namespace.QName;

import com.ihr.xbrl.om.XMLFragment;
import com.ihr.xbrl.om.errors.XBRLValidationException;
import com.ihr.xbrl.om.formula.XBRLFPRSCustomFunctionLibraryAbstract;
import com.ihr.xbrl.om.formula.functions.XFFunction;
import com.ihr.xbrl.om.formula.functions.errors.XFIEInvalidExplicitDimensionQName;
import com.ihr.xbrl.om.instance.XBRLFact;
import com.ihr.xbrl.om.instance.XBRLFactItem;
import com.ihr.xbrl.om.instance.XBRLFactNumeric;
import com.ihr.xbrl.om.xdt.XDTDimension;
import com.ihr.xbrl.om.xdt.XDTHypercubeUse;
import com.ihr.xbrl.om.xdt.XDTProcessor;
import com.ihr.xbrl.util.XMLTools;
import com.rs.lic.LicenseVerificationException;

import net.sf.saxon.expr.Container;
import net.sf.saxon.expr.StaticContext;
import net.sf.saxon.expr.StaticProperty;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.functions.StandardFunction;
import net.sf.saxon.functions.StandardFunction.Entry;
import net.sf.saxon.om.FunctionItem;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.Sequence;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.trans.SymbolicName;
import net.sf.saxon.trans.XPathException;
import net.sf.saxon.tree.iter.ListIterator;
import net.sf.saxon.type.BuiltInAtomicType;
import net.sf.saxon.type.ItemType;
import net.sf.saxon.type.Type;
import net.sf.saxon.value.DecimalValue;
import net.sf.saxon.value.QNameValue;
import net.sf.saxon.value.SequenceType;

/**
 * @author Ignacio
 *
 */
public class ConformanceFormulasCustomFunctions extends XBRLFPRSCustomFunctionLibraryAbstract {

	private static final String FUNCTIONLIBRARYURI = "http://www.example.com/wgt-avg/function";

	/* (non-Javadoc)
	 * @see com.ihr.xbrl.om.formula.XBRLRSCustomFunctionLibraryAbstract#getFunction(java.lang.String, int)
	 */
	@Override
	public Entry getFunction(String name, int arity) {
		return ConformanceFormulasCustomFunctionsRegistry.getFunction(name, arity);
	}

	/* (non-Javadoc)
	 * @see com.ihr.xbrl.om.formula.XBRLRSCustomFunctionLibraryAbstract#getFunctionLibraryURI()
	 */
	@Override
	public String getFunctionLibraryURI() {
		return FUNCTIONLIBRARYURI;
	}

	/* (non-Javadoc)
	 * @see com.ihr.xbrl.om.formula.XBRLRSCustomFunctionLibraryAbstract#getFunctionLibraryPrefix()
	 */
	@Override
	public String getFunctionLibraryPrefix() {
		return "my-fn";
	}

	/**
	 * This class contains the implementation for the custom function PDxEV
	 * @author Ignacio
	 *
	 */
	public static class FUNCTIONLIBRARYURI_PDxEV extends XFFunction {

//		Implement this method if the function returns one node or atomic value		
//		/* (non-Javadoc)
//		 * @see net.sf.saxon.expr.Expression#evaluateItem(net.sf.saxon.expr.XPathContext)
//		 */
//		@Override
//		public Item evaluateItem(XPathContext context) throws XPathException {
//
//		}
		
//		Implement this method if the function returns a sequence of elements
		/* (non-Javadoc)
		 * @see net.sf.saxon.expr.Expression#iterate(net.sf.saxon.expr.XPathContext)
		 */
		@Override
		public SequenceIterator iterate(XPathContext context) throws XPathException {
			/*
	        for $pd in $v:PDkids, 
	            $ev in $v:EVkids[
	                xfi:fact-dimension-s-equal2(., $pd, 
	                    QName('http://www.example.com/wgt-avg','ExposuresDimension')) 
	                ]     
	        return $pd * $ev
				 */

			QNameValue qDimension = XMLTools.qnameValue(new QName("http://www.example.com/wgt-avg","ExposuresDimension"));

			// obtain an instance of the xdt processor
			XDTProcessor xdtProc = null;
			try {
				xdtProc = XDTProcessor.create(getDTSContainer());
			} catch (LicenseVerificationException e) {
				e.printStackTrace();
				xdtProc = null;
			} catch (XBRLValidationException e) {
				e.printStackTrace();
				xdtProc = null;
			}			
			XDTDimension dim = xdtProc.getDimension(XMLTools.javaxQName(qDimension));
			
			if (dim == null) {
				throw new XFIEInvalidExplicitDimensionQName(XMLTools.javaxQName(qDimension));
			}			
			QName dq = dim.getItem().getQName();
			
			ArrayList<Item> al = new ArrayList<Item>();			
			Item itemPD = null;
			Item itemEV = null;
			
			SequenceIterator iter1 = argument[0].iterate(context);
			while ((itemPD = iter1.next()) != null) {
				if (itemPD instanceof NodeInfo) {
					XMLFragment xNodePD = (XMLFragment) itemPD;
					XBRLFact xFactPD = (XBRLFact)xNodePD; //getFormulaProcessor().getUnderEvaluationInstance().getInstanceRootNode().getFact(xNodePD.getDocument());
					SequenceIterator iter2 = argument[1].iterate(context);
					while ((itemEV = iter2.next()) != null) {
						if (itemEV instanceof NodeInfo) {
							// not for atomic values
							XMLFragment xNodeEV = (XMLFragment) itemEV;
							XBRLFact xFactEV = (XBRLFact)xNodeEV; //getFormulaProcessor().getUnderEvaluationInstance().getInstanceRootNode().getFact(xNodeEV.getDocument());
							
							if (xFactEV instanceof XBRLFactNumeric && xFactPD instanceof XBRLFactNumeric) {
								
								XBRLFactNumeric xFactNumEV = (XBRLFactNumeric)xFactEV; 
								XBRLFactNumeric xFactNumPD = (XBRLFactNumeric)xFactPD; 
														
								if (fact_dimension_sEqual2(xFactNumEV,xFactNumPD,dq,dim,xdtProc)) {								
								// multiply fact item values
									
									if (xFactNumEV.isNumber() && xFactNumPD.isNumber()) {
									
//									XBRLNumber numberEV = xFactNumEV.getNumericValue();
//									XBRLNumber numberPD = xFactNumPD.getNumericValue();
//									if (numberEV.isNumber() && numberPD.isNumber()) {
										
//										XBRLNumericValue numEV = (XBRLNumericValue)numberEV;
//										XBRLNumericValue numPD = (XBRLNumericValue)numberPD;
										
//										BigDecimal bdEV = numEV.getBigDecimal();
//										BigDecimal bdPD = numPD.getBigDecimal();
										BigDecimal bdEV = xFactNumEV.getBigDecimal();
										BigDecimal bdPD = xFactNumPD.getBigDecimal();
										
										BigDecimal result = bdEV.multiply(bdPD);
										
										al.add(new DecimalValue(result));
									}
								}
							} else {
								if (!(xFactEV instanceof XBRLFactNumeric)) {
									String pfx = xFactEV.getTaxonomy().getPrefix(xFactEV.getTaxonomy().getTargetNamespace());
									String name = (pfx == null || pfx.length() == 0 ? "" : pfx+":") + xFactEV.getElementDefinition().getName();
									throw new XPathException("The node '"+name+"' is not an XBRL Fact item!","XPTY0004");
								} else {
									String pfx = xFactPD.getTaxonomy().getPrefix(xFactPD.getTaxonomy().getTargetNamespace());
									String name = (pfx == null || pfx.length() == 0 ? "" : pfx+":") + xFactPD.getElementDefinition().getName();
									throw new XPathException("The node '"+name+"' is not an XBRL Fact item!","XPTY0004");
								}
							}
						}
					}
				}				
			}
			
			return new ListIterator(al);
		}
		
		private boolean fact_dimension_sEqual2(XBRLFactNumeric fact1,XBRLFactNumeric fact2 ,QName dq,XDTDimension dim, XDTProcessor xdtProc) {

			QName qVal1 = null;
			XMLFragment xVal1 = null;
			Map<QName,XMLFragment> htSeg1 = xdtProc.getDimData((XBRLFactItem)fact1, XDTHypercubeUse.SEGMENT, false);
			if (htSeg1.containsKey(dq)) {				
				if (dim.getDimensionType() == XDTDimension.EXPLICIT)
					qVal1 = htSeg1.get(dq).getQNameValue();
				else
					xVal1 = htSeg1.get(dq);
			} else {
				Map<QName,XMLFragment> htScen = xdtProc.getDimData((XBRLFactItem)fact1, XDTHypercubeUse.SCENARIO, false);
				if (htScen.containsKey(dq)) {
					if (dim.getDimensionType() == XDTDimension.EXPLICIT)
						qVal1 = htScen.get(dq).getQNameValue();
					else
						xVal1 = htScen.get(dq);
				} else {
					qVal1 = (dim.getDefault() == null ? null : dim.getDefault());
				}
			}
			QName qVal2 = null;
			XMLFragment xVal2 = null;
			Map<QName,XMLFragment> htSeg = xdtProc.getDimData((XBRLFactItem)fact2, XDTHypercubeUse.SEGMENT, false);
			if (htSeg.containsKey(dq)) {
				if (dim.getDimensionType() == XDTDimension.EXPLICIT)
					qVal2 = htSeg.get(dq).getQNameValue();
				else
					xVal2 = htSeg.get(dq);
			} else {
				Map<QName,XMLFragment> htScen = xdtProc.getDimData((XBRLFactItem)fact2, XDTHypercubeUse.SCENARIO, false);
				if (htScen.containsKey(dq)) {
					if (dim.getDimensionType() == XDTDimension.EXPLICIT)
						qVal2 = htScen.get(dq).getQNameValue();
					else
						xVal2 = htScen.get(dq);
				} else {
					qVal2 = (dim.getDefault() == null ? null : dim.getDefault());
				}
			}
			
			if (dim.getDimensionType() == XDTDimension.EXPLICIT) {
				if (qVal1 == null && qVal2 == null)
					return true;
				if (qVal1 == null || qVal2 == null)
					return false;
				return qVal1.equals(qVal2);
			} else {
				if (xVal1 == null && xVal2 == null)
					return true;
				if (xVal1 == null || xVal2 == null)
					return false;
				return xVal1.s_equal2(xVal2);
			}			
		}
	}
	
	/**
	 * Contains the custom function definitions and methods to register them and access them.
	 * 
	 * @author Ignacio
	 *
	 */
	private static abstract class ConformanceFormulasCustomFunctionsRegistry {
		
	    private static HashMap<String,Entry> functionTable = new HashMap<String,Entry>(200);

	    /**
	     * All functions are declared here
	     */
	    static {
	        // XBRL Item Types are not available
	        Entry e;
	        e = register("PDxEV", FUNCTIONLIBRARYURI_PDxEV.class, 0, 2, 2, BuiltInAtomicType.DECIMAL, StaticProperty.ALLOWS_ONE_OR_MORE);
	        	arg(e, 0, Type.NODE_TYPE, StaticProperty.ALLOWS_ONE_OR_MORE, null);
	        	arg(e, 1, Type.NODE_TYPE, StaticProperty.ALLOWS_ONE_OR_MORE, null);
	        	
// to add another function just declare it here and implement the function expression in a Java class	        	
//	        e = register("PDxEV", FUNCTIONLIBRARYURI_PDxEV.class, 0, 2, 2, Type.NODE_TYPE, StaticProperty.ALLOWS_ONE_OR_MORE);
//	        	arg(e, 0, Type.NODE_TYPE, StaticProperty.EXACTLY_ONE, null);
//	        	arg(e, 0, Type.NODE_TYPE, StaticProperty.EXACTLY_ONE, null);
	        	
	    }
	    
	    /**
	     * Register a system function in the table of function details.
	     * @param name the function name
	     * @param implementationClass the class used to implement the function
	     * @param opcode identifies the function when a single class implements several functions
	     * @param minArguments the minimum number of arguments required
	     * @param maxArguments the maximum number of arguments allowed
	     * @param itemType the item type of the result of the function
	     * @param cardinality the cardinality of the result of the function
	     * @return the entry describing the function. The entry is incomplete, it does not yet contain information
	     * about the function arguments.
	    */
	    private static Entry register( String name,
	                                 Class<? extends XFFunction> implementationClass,
	                                 int opcode,
	                                 int minArguments,
	                                 int maxArguments,
	                                 ItemType itemType,
	                                 int cardinality ) {
	        Entry e = makeEntry(name, implementationClass, opcode, minArguments, maxArguments, itemType, cardinality);
	        functionTable.put(name, e);
	        return e;
	    }

	    /**
	     * Make a table entry describing the signature of a function, with a reference to the implementation class.
	     * @param name the function name
	     * @param implementationClass the class used to implement the function
	     * @param opcode identifies the function when a single class implements several functions
	     * @param minArguments the minimum number of arguments required
	     * @param maxArguments the maximum number of arguments allowed
	     * @param itemType the item type of the result of the function
	     * @param cardinality the cardinality of the result of the function
	     * @return the entry describing the function. The entry is incomplete, it does not yet contain information
	     * about the function arguments.
	     */
	    public static Entry makeEntry(String name, Class<? extends XFFunction> implementationClass, int opcode,
	                                   int minArguments, int maxArguments, ItemType itemType, int cardinality) {
	        Entry e = new Entry();
	        int hash = name.indexOf('#');
	        if (hash < 0) {
	            e.name = name;
	        } else {
	            e.name = name.substring(0, hash);
	        }
	        e.implementationClass = implementationClass;
	        e.opcode = opcode;
	        e.minArguments = minArguments;
	        e.maxArguments = maxArguments;
	        e.itemType = itemType;
	        e.cardinality = cardinality;
	        if (maxArguments > 100) {
	            // special case for concat()
	            e.argumentTypes = new SequenceType[1];
	            e.resultIfEmpty = new Sequence[1];
	        } else {
	            e.argumentTypes = new SequenceType[maxArguments];
	            e.resultIfEmpty = new Sequence[maxArguments];
	        }
	        return e;
	    }

	    /**
	     * Add information to a function entry about the argument types of the function
	     * @param e the entry for the function
	     * @param a the position of the argument, counting from zero
	     * @param type the item type of the argument
	     * @param cardinality the cardinality of the argument
	     * @param resultIfEmpty the value returned by the function if an empty sequence appears as the value,
	     * when this result is unaffected by any other arguments
	     */
	    // jmf public static void arg(Entry e, int a, ItemType type, int cardinality, Value<?> resultIfEmpty) {
	    public static void arg(Entry e, int a, ItemType type, int cardinality, Sequence resultIfEmpty) {
	        try {
	            e.argumentTypes[a] = SequenceType.makeSequenceType(type, cardinality);
	            e.resultIfEmpty[a] = resultIfEmpty;
	        } catch (ArrayIndexOutOfBoundsException err) {
	            System.err.println("Internal Saxon error: Can't set argument " + a + " of " + e.name);
	        } 
	    }    
		
	    /**
	     * Get the table entry for the function with a given name
	     * @param name the name of the function. This may be an unprefixed local-name for functions in the
	     * system namespace, or may use the conventional prefix "saxon:" in the case of Saxon extension functions
	     * that are specially recognized
	     * @return if the function name is known, an Entry containing information about the function. Otherwise,
	     * null
	     */
	    public static Entry getFunction(String name, int arity) {
	        // try first for an entry of the form name#arity
	        Entry e = (Entry)functionTable.get(name + '#' + arity);
	        if (e != null) {
	            return e;
	        }
	        // try for a generic entry
	        return (Entry)functionTable.get(name);
	    }
	}

	@Override
	public FunctionItem getFunctionItem(SymbolicName arg0, StaticContext arg1, Container arg2) throws XPathException {
		// TODO Auto-generated method stub
		// FIXME: jmf: pendiente implementar
		return null;
	}

	@Override
	public boolean isAvailable(SymbolicName symName) {
		String uri = symName.getComponentName().getURI();
		String local = symName.getComponentName().getLocalPart();
		int arity = symName.getArity();
        if (uri.equals(getFunctionLibraryURI())) {
            StandardFunction.Entry entry = getFunction(local, arity);
            if (entry == null)
                return false;
            if (!(arity == -1 || (arity >= entry.minArguments && arity <= entry.maxArguments)))
                return false;
            return true;
        }
		return false;
	}
}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
