/*
 *  2007 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2007 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 16/11/2007
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.io.File;
import java.net.URI;
import java.util.Iterator;
import java.util.List;

import javax.xml.namespace.QName;
import javax.xml.transform.stream.StreamSource;
import javax.xml.xpath.XPathConstants;

import net.sf.saxon.Configuration;
import net.sf.saxon.lib.Validation;
import net.sf.saxon.om.NamespaceBinding;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.xpath.XPathEvaluator;

import com.ihr.xbrl.comparator.Event;
import com.ihr.xbrl.comparator.Path;
//import com.ihr.xbrl.comparator.Event;
//import com.ihr.xbrl.comparator.Path;
import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRL;
import com.ihr.xbrl.om.XMLDocumentImpl;
import com.ihr.xbrl.om.XMLFragment;
import com.ihr.xbrl.om.exLinks.GenericLinkbase;
import com.ihr.xbrl.om.exLinks.XBRLExtendedLink;
import com.ihr.xbrl.om.exLinks.XBRLLinkbase;
import com.ihr.xbrl.om.instance.XBRLFactTuple;
import com.ihr.xbrl.om.instance.XBRLInstance;
import com.ihr.xbrl.om.taxonomy.XBRLArcroleType;
import com.ihr.xbrl.om.taxonomy.XBRLRoleType;
import com.ihr.xbrl.om.taxonomy.XBRLTaxonomy;
import com.ihr.xbrl.om.taxonomy.XBRLTuple;
import com.ihr.xbrl.util.URITools;
import com.ihr.xbrl.util.XMLTools;
import com.ihr.xbrl.versioning.Action;
import com.ihr.xbrl.versioning.EventGroup;
import com.ihr.xbrl.versioning.ReportWriter;
import com.ihr.xbrl.versioning.VlVersioningProcessor;
import com.saxonica.config.EnterpriseConfiguration;

/**
 * @author Ignacio
 *
 */
public class GenerateOutVersioningConfSuite {

	/**
	 * @param args
	 * @throws Exception 
	 */
	@SuppressWarnings("unchecked")
	public static void main(String[] args) throws Exception {
		// Open the XVT file
		
		File fXvt = new File("../conformance-versioning/xvt.xml");
		
		Configuration cf = new EnterpriseConfiguration();
		cf.setSchemaValidationMode(Validation.STRICT);
		XPathEvaluator xp = new XPathEvaluator(cf);
		
		StreamSource ss = new StreamSource(fXvt);		
		NodeInfo niXvt = cf.buildDocument(ss);
		
		// Initialize a versioning processor
		VlVersioningProcessor vp = new VlVersioningProcessor("../conformance-versioning/lib/versioning-2007-11-02.xsd","../conformance-versioning/lib/versioning-conf-suite-2007-11-02.xsd");
		
		DTSContainer versioningDTS = DTSContainer.newEmptyContainer();
		// "versioning-2007-11-02.xsd"
		XBRLTaxonomy verTx = (XBRLTaxonomy)versioningDTS.load(new URI(vp.getVersioningDTSUri()));
		XBRLInstance instance = new XBRLInstance(versioningDTS);
		NamespaceBinding verNS = new NamespaceBinding("ver", verTx.getTargetNamespace());
		/*
		XBRLContext ctx0d = new XBRLContext(versioningDTS,null,null,null);
		XBRLContext ctx0i = new XBRLContext(versioningDTS,null,null,null);
		*/
		XBRLLinkbase lb = new XBRLLinkbase(versioningDTS, false);
		//URI glbURI = fdDirURI.resolve("labels.xml");
		//URI glbURI = fNameUri.resolve("labels.xml");
		//lb.setURI(glbURI);
		// get the instance-linkbase role type
		XBRLArcroleType instanceLinkbaseArt = verTx.getArcroleTypeByURI("http://xbrl.org/2007/versioning/instance-linkbase");
		// add the linkbase to the instance document
		instance.getInstanceRootNode().addXBRLDocument(lb, instanceLinkbaseArt);
		
		// get the standard role URI
		XBRLRoleType role = versioningDTS.getStaticRoleTypeByURI(XBRLExtendedLink.standard_role_URI);
		// Create a GenericLinkbase that will be attached to the Linkbase
		GenericLinkbase glb = new GenericLinkbase(lb, role);
		
		XBRLTuple eventT = (XBRLTuple)verTx.getElementDefinitionByName("event");
		
		// Obtain the list of launch files in the conf suite.
		List<String> lstLaunchFiles = (List<String>)xp.evaluate("for $i in /testcases/testcase return resolve-uri($i/@uri,base-uri($i))", niXvt, XPathConstants.NODESET);
		Iterator<String> iterStr = lstLaunchFiles.iterator();
		while (iterStr.hasNext()) {
			String testcaseUri = iterStr.next();
			System.out.println(testcaseUri);
			
			// load the test case into the versioning processor.
			URI launchFileUri = new URI(testcaseUri);
			vp.loadLaunchFile(launchFileUri);
			
			int ch = testcaseUri.lastIndexOf("/");
			String fname = testcaseUri.substring(ch+1);
			
			// create the "out" folder in place
			URI outFolder = launchFileUri.resolve("out");
			File fOut = new File(outFolder);
			fOut.mkdirs();
			
			URI outFile = outFolder.resolve(fname);
			
			// modify the launch file to add a reference to the out file
			vp.setResult(URITools.relativizeMore(launchFileUri, outFile));
			vp.saveLaunchFile(versioningDTS, launchFileUri);			

			// now delete actions loaded from the launch file.
			do {
				Iterator<Action> iterA = vp.getActions().iterator();
				if (iterA.hasNext()) {
					Action ac = iterA.next();
					vp.removeAction(ac);
				} else
					break;
			} while (true);
			
			// Generate one virtual action
			Action ac = new Action();
			vp.addAction(ac);
			EventGroup eg = new EventGroup();
			ac.addEventGroup(eg);
			
			// put all events inside the action
			vp.getComparator().compare();
			Iterator<Event> iterE = vp.getComparator().getEvents();
			while (iterE.hasNext()) {
				Event e = iterE.next();
				eg.add(e.getPath());
			}
			
			// create a new JDOM
			XMLDocumentImpl outDoc = new XMLDocumentImpl(cf);
//			Document outDoc = new Document();
			
			// Actions
			Iterator<Action> iterAc = vp.getActions().iterator();
			while (iterAc.hasNext()) {
				ac = iterAc.next();
				
				Iterator<EventGroup> iterEg = ac.getEventGroups();
				while (iterEg.hasNext()) {
					eg = iterEg.next();
					// Create one event tuple container
//					XBRLFactTuple event = new XBRLFactTuple(instance, eventT);
					XBRLFactTuple event = new XBRLFactTuple(instance.getInstanceRootNode(), eventT, true);
					
					// start documenting actions
					Iterator<Path> iterP = eg.iterator();
					while (iterP.hasNext())
						ReportWriter.documentAction(iterP.next(),event,verTx,verNS/*,ctx0d,ctx0i*/,glb,instance,true,outFile);
					
					// now extract the jdom from event
//					Element eventEl = event.generateNode(true, outFile);
//					outDoc.addContent(eventEl);
//					eventEl.addNamespaceDeclaration(verNS);
//					eventEl.addNamespaceDeclaration(xsiNS);
//					eventEl.setAttribute("schemaLocation", verTx.getTargetNamespace()+" "+URITools.relativizeMore(outFile, verTx.getDocumentURI()), xsiNS);
					XMLFragment eventEl = event.generateNode(/*true, outFile*/);
					outDoc.setDocumentElement(eventEl);
					eventEl.addNamespace(verNS);
					eventEl.addAttribute(new QName(XBRL.XSI_NS,"schemaLocation","xsi"), verTx.getTargetNamespace()+" "+URITools.relativizeMore(outFile, verTx.getDocumentURI()));
					
					// now empty the instance fact's container
					instance.getInstanceRootNode().remove(event);
				}
			}

			// save the out file
			XMLTools.saveDocumentFile(new File(outFile), outDoc);
		}
		System.exit(0);
	}
}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
