package com.ihr.xbrl.sample;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Iterator;
import java.util.List;

import javax.xml.transform.stream.StreamSource;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;

import net.sf.saxon.Configuration;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.trans.XPathException;
import net.sf.saxon.xpath.XPathEvaluator;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRLCoreProcessor;
import com.ihr.xbrl.om.XBRLDocument;
import com.ihr.xbrl.om.XBRLValidationReport;
import com.ihr.xbrl.om.errors.DTSDiscoveryException;
import com.ihr.xbrl.om.errors.XBRLRuntimeException;
import com.ihr.xbrl.om.errors.XBRLValidationException;
import com.ihr.xbrl.om.instance.XBRLInstance;

public class LaunchConformanceSuiteTestTuples {
	
	public static void usage() {
		System.out.println("Usage: LaunchConformanceSuite {-f {Conformance suite file}} [-m] [-x {exclusion pattern}]");
		System.exit(-1);
	}

	/**
	 * @param args
	 * 
	 * args[0] = xbrl.xml
	 * args[1] = pattern to exclude tests
	 */
	public static void main(String[] args) {
		try {
			String confSuiteFile = null;
			String excludePatterns = null;
			boolean bShowMemoryUsage = false;
			
			int countError = 0;
			int countOk = 0;
			int countSkiped = 0;
			int testNumber = 0;
			
			// parse arguments
			for (int i=0; i < args.length; ++i) {
				if ("-f".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					confSuiteFile = args[i];
				} else if ("-m".equals(args[i])) {
					bShowMemoryUsage = true;
				} else if ("-x".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					excludePatterns = args[i];
				}
			}
			
			if (confSuiteFile == null)
				usage();
			
			// Load the conformance suite
			Configuration conf = new Configuration();
			XPathEvaluator xp = new XPathEvaluator(conf);
			
			StreamSource fs = new StreamSource(new File(confSuiteFile));
			NodeInfo xbrl = conf.buildDocument(fs);

			long now = System.currentTimeMillis();
			List<?> lst = (List<?>)xp.evaluate("for $i in testcases/testcase return resolve-uri($i/@uri,base-uri($i))",xbrl,XPathConstants.NODESET);
			Iterator<?> iter = lst.iterator();
			while (iter.hasNext()) {
				URI uri = (URI)iter.next();
				//System.out.println(uri);
				StreamSource fstc = new StreamSource(uri.toURL().openStream());
				fstc.setSystemId(uri.toString());
				NodeInfo testcases = conf.buildDocument(fstc);
				List<?> vars = (List<?>)xp.evaluate("/testcase/variation",testcases,XPathConstants.NODESET);
				Iterator<?> iterVars = vars.iterator();
				while (iterVars.hasNext()) {
					NodeInfo var = (NodeInfo)iterVars.next();
					String varF = (String)xp.evaluate("for $i in data/*[@readMeFirst = true()] return resolve-uri($i,base-uri($i))",var,XPathConstants.STRING);
					if (excludePatterns != null && varF.matches(excludePatterns)) {
						countSkiped++;
					} else {
						testNumber++;
						StringBuffer result = new StringBuffer();
						boolean bError = false;
						DTSContainer dts = null;
						try {
							if (bShowMemoryUsage) {
								long total = Runtime.getRuntime().totalMemory();
								long free = Runtime.getRuntime().freeMemory();
								System.out.println("Test number: "+testNumber+" Memory at 1: T="+total+" F="+free+" U="+(total-free));
							}
							dts = DTSContainer.newEmptyContainer();
							URI uriF = new URI(varF);
							XBRLDocument doc = dts.load(uriF);
							if (doc instanceof XBRLInstance) {
								XBRLCoreProcessor proc = XBRLCoreProcessor.create(dts);
								proc.setWhatToValidate(XBRLCoreProcessor.VALIDATE_REPORT_ITEMS_AND_TUPLES);
								XBRLValidationReport report = proc.validate(doc);
								report.throwExceptions();
								report.inconsistenciesAsExceptions();
							}
							XBRLCoreProcessor proc = XBRLCoreProcessor.create(dts);
							XBRLValidationReport report = proc.validate();
							report.throwExceptions();
							report.inconsistenciesAsExceptions();
						} catch (XBRLRuntimeException e) {
							bError = true;
							result.append(" ");
							result.append(e.getClass().getCanonicalName());
							result.append(" ");
							result.append(e.getLocalizedMessage());
							result.append("\n");
						} catch (XBRLValidationException e) {
							bError = true;
							result.append(" ");
							result.append(e.getClass().getCanonicalName());
							result.append(" ");
							result.append(e.getLocalizedMessage());
							result.append("\n");
						} catch (DTSDiscoveryException e) {
							bError = true;
							result.append(" ");
							result.append(e.getClass().getCanonicalName());
							result.append(" ");
							result.append(e.getLocalizedMessage());
							result.append("\n");
						}
						dts.clear();
						dts = null;
						String strRes = (String)xp.evaluate("result/@expected",var,XPathConstants.STRING);
						if ("invalid".equalsIgnoreCase(strRes) && bError) {
							countOk++;
						} else if ("invalid".equalsIgnoreCase(strRes) && !bError) {
							countError++;
						} else if ("valid".equalsIgnoreCase(strRes) && bError) {
							countError++;
						} else if ("valid".equalsIgnoreCase(strRes) && !bError) {
							countOk++;
						} else {
							countError++;
						}
						if (bShowMemoryUsage) {
							System.runFinalization();
							System.gc();
							long total = Runtime.getRuntime().totalMemory();
							long free = Runtime.getRuntime().freeMemory();
							System.out.println("Test number: "+testNumber+" "+"Memory at 2: T="+total+" F="+free+" U="+(total-free));
						}
					}
				}
			}
			long now2 = System.currentTimeMillis();
			System.out.println("Total execution time "+(now2-now)/1000+"s");
			System.out.println(" Count of errors: "+countError);
			System.out.println(" Count of OK: "+countOk);
			System.out.println(" Count of skipped: "+countSkiped);
		} catch (XPathException e) {
			e.printStackTrace();
		} catch (XPathExpressionException e) {
			e.printStackTrace();
		} catch (URISyntaxException e) {
			e.printStackTrace();
		} catch (MalformedURLException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (Throwable e) {
			e.printStackTrace();
		}
		System.exit(0);
	}

}
