package com.ihr.xbrl.sample;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringReader;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.xml.namespace.QName;
import javax.xml.transform.stream.StreamSource;

import net.sf.saxon.expr.StringTokenIterator;
import net.sf.saxon.om.AxisInfo;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.TreeModel;
import net.sf.saxon.s9api.Axis;
import net.sf.saxon.s9api.DocumentBuilder;
import net.sf.saxon.s9api.Processor;
import net.sf.saxon.s9api.SaxonApiException;
import net.sf.saxon.s9api.XPathCompiler;
import net.sf.saxon.s9api.XPathExecutable;
import net.sf.saxon.s9api.XPathSelector;
import net.sf.saxon.s9api.XdmItem;
import net.sf.saxon.s9api.XdmNode;
import net.sf.saxon.s9api.XdmNodeKind;
import net.sf.saxon.s9api.XdmSequenceIterator;
import net.sf.saxon.s9api.XdmValue;
import net.sf.saxon.trans.XPathException;
import net.sf.saxon.tree.NamespaceNode;
import net.sf.saxon.tree.iter.AxisIterator;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.jdom.Attribute;
import org.jdom.CDATA;
import org.jdom.Comment;
import org.jdom.DocType;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.EntityRef;
import org.jdom.JDOMException;
import org.jdom.Namespace;
import org.jdom.ProcessingInstruction;
import org.jdom.Text;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import com.ihr.xbrl.adins.efm.EFMParameterResolver;
import com.ihr.xbrl.adins.efm.EFMProcessor;
import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRL;
import com.ihr.xbrl.om.XBRLCoreProcessor;
import com.ihr.xbrl.om.XBRLDocument;
import com.ihr.xbrl.om.XBRLValidationReport;
import com.ihr.xbrl.om.errors.DTSDiscoveryException;
import com.ihr.xbrl.om.errors.XBRLInconsistency;
import com.ihr.xbrl.om.errors.XBRLInitializationException;
import com.ihr.xbrl.om.errors.XBRLRuntimeException;
import com.ihr.xbrl.om.errors.XBRLValidationException;
import com.ihr.xbrl.om.errors.XBRLValidationReportToStringSerializer;
import com.ihr.xbrl.om.utr.UTRProcessor;
import com.ihr.xbrl.om.xdt.XDTProcessor;
import com.ihr.xbrl.util.TextTools;
import com.ihr.xbrl.util.TimeUtils;
import com.ihr.xbrl.util.XMLTools;
import com.rs.lic.LicenseVerificationException;


/**
 * Executes all test cases in the Edgar Filing Manual using the EFM test suite source files
 * 
 * @author Ignacio
 *
 */
public class LaunchEFMConformanceSuiteXHTML {
	private static final String XHTMLNS = "http://www.w3.org/1999/xhtml";
	
	private static final String EdgarNS = "http://edgar/2009/conformance";
	
	public static void usage() {
		System.out.println("Usage: LaunchEFMConformanceSuite {-f {Conformance suite file}} [-m] [-x {exclusion pattern}] [-o {Output file}] [-t {test number,[test number]}] [-p {propname=propvalue[,propname=propvalue]}] [-d {DEBUG_LEVEL}]\n"+
			"DEBUG LEVEL can be OFF,TRACE,..., default is OFF");
		System.exit(-1);
	}

	private static final Logger log = Logger.getLogger(LaunchEFMConformanceSuiteXHTML.class);
			
	/**
	 * Jdom node set.
	 *
	 * @param descriptionNodes the description nodes
	 * @return the collection
	 */
	public static Collection<?> jdomNodeSet(XdmValue descriptionNodes) {
		return jdomNodeSet(descriptionNodes.iterator());
	}

	/**
	 * Jdom node set.
	 *
	 * @param iter the iter
	 * @return the collection
	 */
	public static Collection<?> jdomNodeSet(XdmSequenceIterator iter) {
		Vector<Object> vEl = new Vector<Object>();
		while (iter.hasNext()) {
			XdmNode node = (XdmNode)iter.next();
			vEl.add(jdomNode(node));
		}
		return vEl;
	}

	/**
	 * Jdom node.
	 *
	 * @param node the node
	 * @return org.jdom.Element or org.jdom.Text depending on the node type
	 */
	public static Object jdomNode(XdmNode node) {
		if (node.getNodeKind() == XdmNodeKind.ELEMENT) {
			return jdomElement(node);
		} else if (node.getNodeKind() == XdmNodeKind.TEXT) {
			return jdomText(node);
		}
		return null;
	}

	/**
	 * Jdom element.
	 *
	 * @param node the node
	 * @return the org.jdom. element
	 */
	private static org.jdom.Element jdomElement(XdmNode node) {
		// convert one XdmNode into a new jdomNode
		org.jdom.Element newEl = new org.jdom.Element(
				node.getNodeName().getLocalName(),			
				node.getNodeName().getPrefix(),
				node.getNodeName().getNamespaceURI());
		// add namespace definitions
		AxisIterator ai = node.getUnderlyingNode().iterateAxis(AxisInfo.NAMESPACE);
		NodeInfo ni = ai.next();
		// jmf while (ai.moveNext()) {
		while ( ni != null ) {
			Object oai = ni;
			if (oai instanceof NamespaceNode) {
				NamespaceNode currentNs = (NamespaceNode)oai;
				String prefix = currentNs.getLocalPart();
				String uri = currentNs.getStringValue();
				if ("xml".equals(prefix) && "http://www.w3.org/XML/1998/namespace".equals(uri))
				{
					ni = ai.next();
					continue;
				}
				net.sf.saxon.s9api.QName qNs = new net.sf.saxon.s9api.QName(prefix,uri,"");
				org.jdom.Namespace ns = jdomNamespace(qNs);
				if (!newEl.getNamespace().equals(ns)) // avoids namespace collision when the element namespace is added
					newEl.addNamespaceDeclaration(ns);
			} else
				throw new RuntimeException("Internal error. Unknown returned object type");
			ni = ai.next();
		}
		// add attributes
		XdmSequenceIterator atIter = node.axisIterator(Axis.ATTRIBUTE);
		while (atIter.hasNext()) {
			XdmNode atNode = (XdmNode)atIter.next();
			org.jdom.Attribute at = jdomAttribute(atNode);
			newEl.setAttribute(at);
		}
		// add children elements
		newEl.addContent(jdomNodeSet(node.axisIterator(Axis.CHILD)));
		return newEl;
	}

	/**
	 * Jdom namespace.
	 *
	 * @param nsNode the ns node
	 * @return the org.jdom. namespace
	 */
	private static org.jdom.Namespace jdomNamespace(XdmNode nsNode) {
		String prefix = nsNode.getUnderlyingNode().getPrefix();
		String uri = nsNode.getUnderlyingNode().getURI();
		return jdomNamespace(new net.sf.saxon.s9api.QName(prefix,uri,""));
	}

	/**
	 * Jdom namespace.
	 *
	 * @param qName the q name
	 * @return the org.jdom. namespace
	 */
	private static org.jdom.Namespace jdomNamespace(net.sf.saxon.s9api.QName qName) {
		return Namespace.getNamespace(qName.getPrefix(), qName.getNamespaceURI());
	}
	
	/**
	 * Jdom attribute.
	 *
	 * @param atNode the at node
	 * @return the org.jdom. attribute
	 */
	private static org.jdom.Attribute jdomAttribute(XdmNode atNode) {
		org.jdom.Attribute newAt = null;
		try {
			String value = atNode.getTypedValue().getUnderlyingValue().head().getStringValue();
			newAt = new org.jdom.Attribute(atNode.getNodeName().getLocalName(),value,jdomNamespace(atNode));			
		} catch (XPathException e) {
			e.printStackTrace();
		} catch (SaxonApiException e) {
			e.printStackTrace();
		}
		return newAt;
	}
	
	/**
	 * Jdom text.
	 *
	 * @param textNode the text node
	 * @return the org.jdom. text
	 */
	private static org.jdom.Text jdomText(XdmNode textNode) {
		org.jdom.Text text = new org.jdom.Text(textNode.getStringValue());
		return text;
	}

	/**
	 * Returns a collection of text and Element nodes transforming the content of the string parameter accordingly.
	 *
	 * @param string the string
	 * @return the collection
	 * @throws JDOMException the JDOM exception
	 */
	public static Collection<?> textToJDomNodes(String string) throws JDOMException {		
		Vector<Object> col = new Vector<Object>();
		
		// parse the string as nodes
		org.jdom.input.SAXBuilder sxb = new SAXBuilder();
		try {
			Document doc = sxb.build(new StringReader("<doc>"+string+"</doc>"));
			Element docElement = doc.getRootElement();
			for (Object o : docElement.getContent()) {
				if (o instanceof Element) {
					Element oel = (Element)o;
					col.add(oel.clone());
				} else if (o instanceof Text) {
					Text oel = (Text)o;
					col.add(oel.clone());
				} else if (o instanceof Comment) {
					Comment oel = (Comment)o;
					col.add(oel.clone());
				} else if (o instanceof ProcessingInstruction) {
					ProcessingInstruction oel = (ProcessingInstruction)o;
					col.add(oel.clone());
				} else if (o instanceof CDATA) {
					CDATA oel = (CDATA)o;
					col.add(oel.clone());
				} else if (o instanceof EntityRef) {
					EntityRef oel = (EntityRef)o;
					col.add(oel.clone());
				}
			}
		} catch (JDOMException e) {
			throw e;
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		return col;
	}

	/**
	 * @param args
	 * 
	 * args[0] = xbrl.xml
	 * args[1] = pattern to exclude tests
	 */
	public static void main(String[] args) {
		try {
			String confSuiteFile = null;
			String excludePatterns = null;
			String debugLevel = "OFF";
			String outputFile = null;
			boolean bShowMemoryUsage = false;
			
			/**
			 * Configures DTS properties. Empty means standard XBRL 2.1 properties.
			 */
			Properties dtsProps = new Properties();
			
			int countError = 0;
			int countOk = 0;
			int countSkiped = 0;
			int testNumber = 0;
			HashSet<Integer> hsIncludedTests = null;
			
			// parse arguments
			for (int i=0; i < args.length; ++i) {
				if ("-f".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					confSuiteFile = args[i];
				} else if ("-m".equals(args[i])) {
					bShowMemoryUsage = true;
				} else if ("-x".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					excludePatterns = args[i];
				} else if ("-o".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					outputFile = args[i];
				} else if ("-p".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					dtsProps.clear();
					i = readProperties(i,args,dtsProps);
				} else if ("-t".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					if (hsIncludedTests == null)
						hsIncludedTests = new HashSet<Integer>();
					i = readList(i,false,args,hsIncludedTests);
				} else if ("-d".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					debugLevel = args[i];					
				} else {
					usage();
				}
			}
			
			// configure log level
//			Logger log1 = Logger.getLogger("com.ihr.xbrl.om.formula");
//			Logger log2 = Logger.getLogger("com.ihr.xbrl.om.assertions");
//			log1.setLevel(Level.toLevel(debugLevel));
//			log2.setLevel(Level.toLevel(debugLevel));
//			log.setLevel(Level.toLevel(debugLevel));
			log.setLevel(Level.ALL);
			dtsProps.put(DTSContainer.LOG_LEVEL, debugLevel);
			
			// this property is fixed
			EFMConformanceParameterResolver parameterResolver = new EFMConformanceParameterResolver(); 
			
			// As this is for running edgar suite, we need to activate edgar processor
			dtsProps.setProperty(DTSContainer.PROCESSORS_SEQUENCE,
					
				XBRLCoreProcessor.PROCESSORNAME+","+
				XDTProcessor.PROCESSORNAME+","+
				UTRProcessor.PROCESSORNAME+","+
				EFMProcessor.PROCESSORNAME);
			
			dtsProps.put(EFMProcessor.EFM_PARAMETER_RESOLVER, parameterResolver);
			
			if (confSuiteFile == null)
				usage();
			
			// Load the conformance suite
			Processor xproc = XMLTools.newProcessor(false);
			XPathCompiler xpc = xproc.newXPathCompiler();
			xpc.declareNamespace("e", EdgarNS);

			DocumentBuilder builder = XMLTools.newDocumentBuilder(xproc, TreeModel.TINY_TREE, null);
			XdmNode xbrl = builder.build(new File(confSuiteFile));

			// The output will be sent to a JDom document
			Document page = new Document();
			page.setDocType(new DocType("html","-//W3C//DTD XHTML 1.0 Strict//EN","http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd"));
			Element html = new Element("html",XHTMLNS);
			page.addContent(html);
			Element head = new Element("head",XHTMLNS);
			html.addContent(head);
			Element title = new Element("title",XHTMLNS);
			title.setText("Compliance report with the XBRL Conformance suite");
			head.addContent(title);
			Element style = new Element("style",XHTMLNS);
			head.addContent(style);
			Attribute styleType = new Attribute("type","text/css");
			style.setAttribute(styleType);
			style.setText("td.testOK { background: lightgreen; }\ntd.testFail { background: red; }");			
			Element body = new Element("body",XHTMLNS);
			html.addContent(body);
			Element pageH2 = new Element("h2",XHTMLNS);
			pageH2.setText("Compliance report with the XBRL Conformance suite");
			body.addContent(pageH2);
			Date d = new Date(System.currentTimeMillis());
			SimpleDateFormat dtf = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSZ");
			Element header3 = new Element("h3",XHTMLNS);
			header3.setText("Date executed: "+dtf.format(d));
			body.addContent(header3);
			Element outTable = new Element("table",XHTMLNS);
			Attribute border1 = new Attribute("border","1");
			outTable.setAttribute(border1);
			body.addContent(outTable);
			
			XPathSelector xps = xpc.compile("for $i in testcases/testcase return resolve-uri($i/@uri,base-uri($i))").load();
			//List<String> lst = (List<String>)xp.evaluate("for $i in testcases/testcase return resolve-uri($i/@uri,base-uri($i))",xbrl,XPathConstants.NODESET);
			xps.setContextItem(xbrl);
			
			XPathExecutable execSelExecName = xpc.compile("/e:testcase/e:name");
			XPathExecutable execSelExecDesc = xpc.compile("/e:testcase/e:description/*");
			XPathExecutable execSelVariation = xpc.compile("/e:testcase/e:variation");
			XPathExecutable execSelId = xpc.compile("@id");
			XPathExecutable execSelDescText = xpc.compile("e:description/text()");
			XPathExecutable execSelNameText = xpc.compile("e:name/text()");
			XPathExecutable execSelFirstFile = xpc.compile("for $i in e:data/*[@readMeFirst = true()] return resolve-uri($i,base-uri($i))");
			XPathExecutable execSelFirstFileAlt = xpc.compile("for $i in e:data/* return resolve-uri($i,base-uri($i))");
			XPathExecutable execSelResult = xpc.compile("e:result/*");
//			XPathExecutable execSelResults = xpc.compile("e:results");
			XPathExecutable execSelParameters = xpc.compile("e:data/e:parameter");
			XPathExecutable execSelpName = xpc.compile("data(@name)");
			XPathExecutable execSelpDatatype = xpc.compile("data(@datatype)");
			XPathExecutable execSelpValue = xpc.compile("data(@value)");
			XPathExecutable execSeverity = xpc.compile("@severity");
			XPathExecutable execErrNum = xpc.compile("@num");
			
			//Iterator<String> iter = lst.iterator();
			Iterator<XdmItem> iter = xps.iterator();
			long now = System.currentTimeMillis();
			while (iter.hasNext()) {
				XdmItem item = iter.next();
				String strURI = item.getStringValue();
				URI uri = new URI(strURI);
				StreamSource fstc = new StreamSource(uri.toURL().openStream());
				fstc.setSystemId(strURI);
				DocumentBuilder builder2 = XMLTools.newDocumentBuilder(xproc, TreeModel.TINY_TREE, null);
				XdmNode testcases = builder2.build(fstc);
				
				XPathSelector selExecName = execSelExecName.load();
				selExecName.setContextItem(testcases);
				XdmItem testcaseNameItem = selExecName.evaluateSingle();
				String testcaseName = (testcaseNameItem != null) ? testcaseName = testcaseNameItem.getStringValue() : "";
				Element tr01 = new Element("tr",XHTMLNS);
				outTable.addContent(tr01);
				Element td01 = new Element("td",XHTMLNS);
				tr01.addContent(td01);
				td01.setAttribute("style", "background: darkgreen; color: white;");
				td01.setText(testcaseName);
				
				XPathSelector selExecDesc = execSelExecDesc.load();
				selExecDesc.setContextItem(testcases);
				XdmValue valueDesc = selExecDesc.evaluate();
				if (valueDesc != null && valueDesc.size() > 0) {
					Element tr0101 = new Element("tr",XHTMLNS);
					outTable.addContent(tr0101);
					Element td0101 = new Element("td",XHTMLNS);
					tr0101.addContent(td0101);
					td0101.setAttribute("style", "background: lightblue; color: black;");
					
					Collection<?> col = jdomNodeSet(valueDesc);
					for (Object o : col)
						if (o instanceof Element)
							fixWrongNamespaces((Element)o);
					td0101.addContent( col );					
				}
				
				XPathSelector selVariation = execSelVariation.load();
				selVariation.setContextItem(testcases);
				Iterator<XdmItem> iterVars = selVariation.iterator(); 
				//Iterator<?> iterVars = vars.iterator();
				while (iterVars.hasNext()) {
					Element tr02 = new Element("tr",XHTMLNS);
					outTable.addContent(tr02);
					Element td02 = new Element("td",XHTMLNS);
					tr02.addContent(td02);
					Element table02 = new Element("table",XHTMLNS);
					td02.addContent(table02);
					Attribute at02 = new Attribute("border","1");
					table02.setAttribute(at02);
					Attribute at03 = new Attribute("width","100%");
					table02.setAttribute(at03);
					Element tr03 = new Element("tr",XHTMLNS);
					table02.addContent(tr03);
					XdmItem var = iterVars.next();
					
					XPathSelector selId = execSelId.load();
					selId.setContextItem(var);
					String id = "";
					XdmItem idItem = selId.evaluateSingle();
					if (idItem != null)
						id = idItem.getStringValue();
					//String id = (String)xp.evaluate("@id",var,XPathConstants.STRING);
					
					XPathSelector selDescText = execSelDescText.load();
					selDescText.setContextItem(var);
					String desc = selDescText.evaluateSingle().getStringValue();
					
					XPathSelector selNameText = execSelNameText.load();
					selNameText.setContextItem(var);
					String name = selNameText.evaluateSingle().getStringValue();
					
					//String desc = (String)xp.evaluate("description/text()",var,XPathConstants.STRING);
					Element td03 = new Element("td",XHTMLNS);
					tr03.addContent(td03);
					td03.setText("Id: "+id+" n:"+(testNumber+countSkiped+1)+" Description: "+desc);
					Element td04 = new Element("td",XHTMLNS);
					td04.setText("Name: "+name);
					tr03.addContent(td04);
					Element tr04 = new Element("tr",XHTMLNS);
					table02.addContent(tr04);
					
					XPathSelector selFirstFile = execSelFirstFile.load();
					selFirstFile.setContextItem(var);
					XdmItem selFirstFileResult = selFirstFile.evaluateSingle();
					String varF = null;
					if (selFirstFileResult != null)
						varF = selFirstFileResult.getStringValue();
					//String varF = (String)xp.evaluate("for $i in data/*[@readMeFirst = true()] return resolve-uri($i,base-uri($i))",var,XPathConstants.STRING);
					if (varF == null || varF.trim().isEmpty()) {
						// may not contain readMeFirst=true attribute, in this case only one element is allowed
						XPathSelector selFirstFileAlt = execSelFirstFileAlt.load();
						selFirstFileAlt.setContextItem(var);
						XdmValue lstX = selFirstFileAlt.evaluate();						
						//List lstX = (List)xp.evaluate("for $i in data/* return resolve-uri($i,base-uri($i))",var,XPathConstants.NODESET);
						if (lstX.size() != 1) {
							System.err.println("Test number: "+testNumber+" conformance suite error. Multiple files exist in the test case and none of them are marked as readMeFirst=true()");
							varF = "";
						} else
							varF = lstX.itemAt(0).getStringValue();
					}
					if ((excludePatterns != null && varF.matches(excludePatterns)) || (hsIncludedTests != null && !hsIncludedTests.contains(testNumber+countSkiped+1))) {
						Element td05 = new Element("td",XHTMLNS);
						td05.setText("Execution excluded");
						tr04.addContent(td05);
						countSkiped++;
					} else {
						testNumber++;
						StringBuffer result = new StringBuffer();
						boolean bError = false;
						boolean bInconsistencies = false;
						int errNum = -1;
						DTSContainer dts = null;
						XBRLDocument doc = null;
						XBRLValidationReport report = null;
						long dtsCreationTime = 0l;
						long dtsLoadingTime = 0l;
						long dtsValidationTime = 0l;
						try {
							if (bShowMemoryUsage) {
								//System.out.println("Total memory at 1:"+Runtime.getRuntime().freeMemory())
								long total = Runtime.getRuntime().totalMemory();
								long free = Runtime.getRuntime().freeMemory();
								log.info("Test number: "+(testNumber+countSkiped)+" Memory at 1: T="+total+" F="+free+" U="+(total-free)+" L=\""+varF+"\"");
							}
							
							// obtain test case parameters
							XPathSelector selParameters = execSelParameters.load();
							selParameters.setContextItem(var);
							XdmValue allParameters = selParameters.evaluate();
							XdmSequenceIterator paramIter = allParameters.iterator();
							
							parameterResolver.clear();
							
							// process varF and extracts variation name that will be parameter value for parameter cikNameList when no other value is set.
							int idx = varF.lastIndexOf('/');
							if (idx < 0)
								idx = 0;
							int idx2 = varF.indexOf('.', idx);
							if (idx2 < 0)
								idx2 = varF.length();
							int idx3 = varF.indexOf('-', idx);
							if (idx3 < 0)
								idx3 = varF.length();
							if (idx3 < idx2)
								idx2 = idx3;
							idx3 = varF.indexOf('n', idx); // remove also good and no good marks
							if (idx3 < 0)
								idx3 = varF.length();
							if (idx3 < idx2)
								idx2 = idx3;
							idx3 = varF.indexOf('g', idx);
							if (idx3 < 0)
								idx3 = varF.length();
							if (idx3 < idx2)
								idx2 = idx3;
							
							parameterResolver.setDefaultTestCaseName(varF.substring(idx+1, idx2));
							
							while (paramIter.hasNext()) {
								XdmNode nodeParam = (XdmNode) paramIter.next();
								
								XPathSelector selExecSelpName = execSelpName.load();
								XPathSelector selExecSelpDatatype = execSelpDatatype.load();
								XPathSelector selExecSelpValue = execSelpValue.load();

								selExecSelpName.setContextItem(nodeParam);
								String pName = selExecSelpName.evaluateSingle().getStringValue();
								
								selExecSelpDatatype.setContextItem(nodeParam);
								QName pType =  XMLTools.javaxQName( selExecSelpDatatype.evaluateSingle().getStringValue(), nodeParam );
								
								selExecSelpValue.setContextItem(nodeParam);
								String pValue = selExecSelpValue.evaluateSingle().getStringValue();
								
								// extract parameter name, type and value.
								parameterResolver.addParameter(pName,pType,pValue,nodeParam);
							}
							URI uriF = new URI(varF);
							
							long lap0 = System.currentTimeMillis();
							dts = DTSContainer.newEmptyContainer(dtsProps);
							long lap1 = System.currentTimeMillis();
							doc = dts.load(uriF);
							
//							String statistics = Statistics.getReport("XMLDocumentURI.createDocument");
//							System.out.println(statistics);
//							String statistics = Statistics.getReport("loadIntoConfiguration");
//							System.out.println(statistics);
//							String statistics = Statistics.getReport("execLoadFinal");
//							log.info(statistics);
							
							long lap2 = System.currentTimeMillis();
							// XBRLCoreProcessor proc = XBRLCoreProcessor.create(dts);
							// report = proc.validate();
							report = dts.validate();
							long lap3 = System.currentTimeMillis();
							
							dtsCreationTime = lap1 - lap0;
							dtsLoadingTime = lap2 - lap1;
							dtsValidationTime = lap3 - lap2;

//							report.throwExceptions();
//						} catch (XBRLValidationException e) {
//							bError = true;
//							result.append(" ");
//							result.append(XBRLValidationException.documentException(e));
//							result.append("\n");
//							errNum = e.getCode();

							bError = report.getExceptionsCount() > 0;							
							bInconsistencies = report.getInconsistenciesCount() > 0;
						} catch (XBRLRuntimeException e) {
							bError = true;
							result.append(" ");
							result.append(XBRLValidationException.documentException(e));
							result.append("\n");
							errNum = e.getCode();
						} catch (DTSDiscoveryException e) {
							bError = true;
							result.append(" ");
							result.append(XBRLValidationException.documentException(e));
							result.append("\n");
							errNum = e.getCode();
						} catch (LicenseVerificationException e) {
							e.printStackTrace();
							System.exit(-1);
						} catch (XBRLInitializationException e) {
							e.printStackTrace();
							System.exit(-1);
						}
						// jmf: probando un caso de numero "NaN" ____
						catch (NumberFormatException e) {
							bError = true;
							result.append(" ");
							result.append(XBRLValidationException.documentException(e));
							result.append("\n");
							errNum = 9999;//e.getCode();
						}						
//						try {
//							if (report != null)
//								report.inconsistenciesAsExceptions();
//						} catch (XBRLValidationException e) {
//							bInconsistencies = true;
//							errNum = e.getCode();
//						}
						XPathSelector selResult = execSelResult.load();
						selResult.setContextItem(var);
						XdmNode resultNode = (XdmNode)selResult.evaluateSingle(); // this may evaluate to a sequence... but this is not happening in the entire conf suite.
						if (resultNode == null) {
							// result is OK
							String strRes = "valid";
							if ("valid".equalsIgnoreCase(strRes) && (bError || bInconsistencies)) {
								Element td05 = new Element("td",XHTMLNS);
								Attribute at06 = new Attribute("class","testFail");
								td05.setAttribute(at06);
								td05.setText("NOT PASS. Expected valid - result is invalid");
								tr04.addContent(td05);
								countError++;
								if (report != null) {
									XBRLValidationReportToStringSerializer serializer = new XBRLValidationReportToStringSerializer();
									report.serialize(serializer);
									result.append( serializer.getReportString() );
								}
							} else if ("valid".equalsIgnoreCase(strRes) && !bError && !bInconsistencies) {
								Element td05 = new Element("td",XHTMLNS);
								Attribute at06 = new Attribute("class","testOK");
								td05.setAttribute(at06);
								td05.setText("OK. Expected valid - result is valid");
								tr04.addContent(td05);
								countOk++;
							} else {
								Element td05 = new Element("td",XHTMLNS);
								td05.setText("UNKNOWN. Expected "+strRes+" result is "+bError);
								tr04.addContent(td05);
								countError++;
							}
						} else {
							if ("assert".equals(resultNode.getNodeName().getLocalName()) && EdgarNS.equals(resultNode.getNodeName().getNamespaceURI())) {
																
								XPathSelector selSeverity = execSeverity.load();
								selSeverity.setContextItem(resultNode);
								String strRes = selSeverity.evaluateSingle().getStringValue();
								
								XPathSelector selErrNum = execErrNum.load();
								selErrNum.setContextItem(resultNode);
								int compareErrNum = Integer.parseInt(selErrNum.evaluateSingle().getStringValue());
								
								//String strRes = (String)xp.evaluate("result/@expected",var,XPathConstants.STRING);
								if ("err".equalsIgnoreCase(strRes) && bError && reportContainsError(report,compareErrNum, errNum, result)) {
									Element td05 = new Element("td",XHTMLNS);
									Attribute at06 = new Attribute("class","testOK");
									td05.setAttribute(at06);
									td05.setText("OK. Expected invalid - result is invalid");
									tr04.addContent(td05);
									countOk++;
								} else if ("err".equalsIgnoreCase(strRes) && bError) {
									Element td05 = new Element("td",XHTMLNS);
									Attribute at06 = new Attribute("class","testOK");
									td05.setAttribute(at06);
									td05.setText("OK. Expected error number "+compareErrNum+" cannot be found, but errors was detected");
									tr04.addContent(td05);
									countOk++;
								} else if ("err".equalsIgnoreCase(strRes) && !bError) {
									Element td05 = new Element("td",XHTMLNS);
									Attribute at06 = new Attribute("class","testFail");
									td05.setAttribute(at06);
									td05.setText("NOT PASS. Expected invalid - result is valid");
									tr04.addContent(td05);
									countError++;
								} else if ("wrn".equalsIgnoreCase(strRes) && bInconsistencies && reportContainsWarning(report,compareErrNum, errNum, result)) {
									Element td05 = new Element("td",XHTMLNS);
									Attribute at06 = new Attribute("class","testOK");
									td05.setAttribute(at06);
									td05.setText("OK. Expected warning - result is warnings generated");
									tr04.addContent(td05);
									countOk++;
								} else if ("wrn".equalsIgnoreCase(strRes) && bInconsistencies) {
									Element td05 = new Element("td",XHTMLNS);
									Attribute at06 = new Attribute("class","testOK");
									td05.setAttribute(at06);
									td05.setText("OK. Expected warning number "+compareErrNum+" cannot be found, but warning was detected");
									tr04.addContent(td05);
									countOk++;
								} else if ("wrn".equalsIgnoreCase(strRes) && !bInconsistencies) {
									Element td05 = new Element("td",XHTMLNS);
									Attribute at06 = new Attribute("class","testFail");
									td05.setAttribute(at06);
									td05.setText("NOT PASS. Expected warning - result is warning not generated");
									tr04.addContent(td05);
									countError++;
								} else {
									Element td05 = new Element("td",XHTMLNS);
									td05.setText("UNKNOWN. Expected "+strRes+" result is "+bError);
									tr04.addContent(td05);
									countError++;
								}
							} else {
								// process as unknown node element in testcase
								Element td05 = new Element("td",XHTMLNS);
								td05.setText("UNKNOWN node in conformance suite."+resultNode.getNodeName()+" validation result is "+bError);
								tr04.addContent(td05);
								countError++;
							}
						}
						
						Element td06 = new Element("td",XHTMLNS);
						try {
							Collection<?> col = textToJDomNodes(TextTools.textToHTML(result.toString()));
							for (Object o : col) {
								if (o instanceof Element) {
									fixWrongNamespaces((Element)o);
								}
							}
							td06.setContent( col ) ;
						} catch (JDOMException e) {
							td06.setText(result.toString());
						}
						tr04.addContent(td06);
						Element tr05 = new Element("tr",XHTMLNS);
						table02.addContent(tr05);
						Element td07 = new Element("td",XHTMLNS);
						Attribute at07 = new Attribute("colspan","2");
						td07.setAttribute(at07);
						td07.setText(varF);
						tr05.addContent(td07);
						
						if (bShowMemoryUsage) {
							System.gc();
							long total = Runtime.getRuntime().totalMemory();
							long free = Runtime.getRuntime().freeMemory();
							log.info("Test number: "+(testNumber+countSkiped)+" "+"Memory at 2: T="+total+" F="+free+" U="+(total-free));
							log.info("DTS Creation time: "+TimeUtils.toTime(dtsCreationTime)+" DTS Loading time: "+TimeUtils.toTime(dtsLoadingTime)+" DTS Validation time: "+TimeUtils.toTime(dtsValidationTime));
						}
						try {
							if (doc != null)
								doc.clear();
							if (dts != null)
								dts.clear();
							if (report != null)
								report.clear();
						} catch (Throwable e) {
							log.error("Exception thrown during memory cleanup on test #"+(testNumber+countSkiped)+"!");
						}
						doc = null;
						dts = null;
						report = null;
					}
				} // Shall we close stream here?
			}
			Element table03 = new Element("table",XHTMLNS);
			body.addContent(table03);
			Element tr05 = new Element("tr",XHTMLNS);
			table03.addContent(tr05);
			Element td07 = new Element("td",XHTMLNS);
			tr05.addContent(td07);
			td07.setText("Not passed:");
			Element td08 = new Element("td",XHTMLNS);
			tr05.addContent(td08);
			td08.setText((new Integer(countError)).toString());
			Element tr06 = new Element("tr",XHTMLNS);
			table03.addContent(tr06);
			Element td09 = new Element("td",XHTMLNS);
			tr06.addContent(td09);
			td09.setText("Passed:");
			Element td10 = new Element("td",XHTMLNS);
			tr06.addContent(td10);
			td10.setText((new Integer(countOk)).toString());
			Element tr07 = new Element("tr",XHTMLNS);
			table03.addContent(tr07);
			Element td11 = new Element("td",XHTMLNS);
			tr07.addContent(td11);
			td11.setText("Not executed:");
			Element td12 = new Element("td",XHTMLNS);
			tr07.addContent(td12);
			td12.setText((new Integer(countSkiped)).toString());

			// produce the output to stdout or file
			Format ft = Format.getPrettyFormat();
			ft.setEncoding("ISO-8859-1");
			XMLOutputter outputter = new XMLOutputter(ft);
			String s = outputter.outputString(page);
			File f = new File(outputFile);
			if (outputFile != null) {
				FileWriter fw = new FileWriter(f);
				fw.write(s);
				fw.flush();
				fw.close();
			} else {
				System.out.println(s);
			}			
			long now2 = System.currentTimeMillis();
			log.info("Total execution time "+TimeUtils.toTime(now2-now));
			log.info("Log witten to file "+f.toURI().toString()+"\"");
			f = null;
		} catch (URISyntaxException e) {
			e.printStackTrace();
		} catch (MalformedURLException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (SaxonApiException e) {
			e.printStackTrace();
		}
		System.exit(0);
	}

	/**
	 * @param report
	 * @param compareErrNum
	 * @return
	 */
	private static boolean reportContainsWarning(XBRLValidationReport report, int compareErrNum, int errNum, StringBuffer result) {
		if (report == null && errNum >= 0) {
			return true; // error already detected in a previous catch.
		}
		if (report == null && errNum < 0 && compareErrNum >= 0) {
			result.append("Validation report is empty, no error was detected and expected error is "+compareErrNum);
			return false; // do not process the report. shall return false...
		}
		if (report == null) {
			result.append("Validation report is empty, detected error is "+errNum+" expected error is "+compareErrNum);
			return false; // do not process the report. shall return false...
		} else {
			Iterator<XBRLInconsistency> iter = report.getInconsistencies();
			StringBuffer sbAllErrors = new StringBuffer();
			while (iter.hasNext()) {
				XBRLInconsistency e = iter.next();
				if (e.getCode() == compareErrNum) {
					result.append(e.getMessage());
					return true;
				}
				if (sbAllErrors.length() > 0)
					sbAllErrors.append(" \n");
				sbAllErrors.append(e.getMessage());
			}
			result.append("None of the errors matches the expected error!\n");
			result.append(sbAllErrors);
		}
		return false;
	}

	/**
	 * @param report
	 * @param compareErrNum
	 * @return
	 */
	private static boolean reportContainsError(XBRLValidationReport report,	int compareErrNum, int errNum, StringBuffer result) {
		if (report == null && errNum >= 0) {
			return true; // error already detected in a previous catch.
		}
		if (report == null && errNum < 0 && compareErrNum >= 0) {
			result.append("Validation report is empty, no error was detected and expected error is "+compareErrNum);
			return false; // do not process the report. shall return false...
		}
		if (report == null) {
			result.append("Validation report is empty, detected error is "+errNum+" expected error is "+compareErrNum);
			return false; // do not process the report. shall return false...
		} else {
			Iterator<XBRLValidationException> iter = report.getExceptions();
			StringBuffer sbAllErrors = new StringBuffer();
			while (iter.hasNext()) {
				XBRLValidationException e = iter.next();
				if (e.getCode() == compareErrNum) {
					result.append(XBRLValidationException.documentException(e));
					return true;
				}
				if (sbAllErrors.length() > 0)
					sbAllErrors.append(" \n");
				sbAllErrors.append(XBRLValidationException.documentException(e));
			}
			result.append("None of the errors matches the expected error!\n");
			result.append(sbAllErrors);
		}
		return false;
	}

	/**
	 * @param el
	 */
	private static Element fixWrongNamespaces(Element el) {
		// the returned element is is the wrong namespace (People creating this conformance suite don't know what a namespace is!)
		// so, remove the default namespace of EdgarNS and move them to the NO NAMESPACE for HTML nodes.
		if (el.getNamespaceURI().equals(EdgarNS)) {
			Namespace xhtmlNs = Namespace.getNamespace("", XHTMLNS);
			el.setNamespace(xhtmlNs);
		} else if (el.getNamespaceURI().equals("")) { // elements at the no namespace belongs to the XHTML namespace.
			Namespace xhtmlNs = Namespace.getNamespace("", XHTMLNS);
			el.setNamespace(xhtmlNs);
		}
		@SuppressWarnings("unchecked")
		List<?> lst = new ArrayList<Object>(el.getAdditionalNamespaces());		
		for (Object oNs : lst) {
			if (oNs instanceof Namespace)
				el.removeNamespaceDeclaration((Namespace)oNs);
		}
		for (Object oChild : el.getChildren()) {
			if (oChild instanceof Element) {
				fixWrongNamespaces((Element)oChild);
			}
		}
		return el;
	}

	/**
	 * @param i
	 * @param args
	 * @param dtsProps 
	 * @return
	 */
	private static int readProperties(int i, String[] args, Properties dtsProps) {
		if (args.length == i)
			return i;
		StringBuffer sb = new StringBuffer();
		int j=i;
		for (;j<args.length;++j) {
			String argument = args[j].trim();
			if (argument.startsWith("-")) {
				break;
			} else {
				sb.append(argument);
			}
		}		
		i = j-1;
		if (sb.length() == 0)
			usage();
		StringTokenIterator sti = new StringTokenIterator(sb.toString(), ",");
		Item itm = null;
		while ((itm = sti.next()) != null) {
			// find equal sign
			String token = itm.getStringValue();
			int idx = token.indexOf("=");
			if (idx < 0)
				usage();
			String propName = token.substring(0, idx);
			String propValue = token.substring(idx+1);
			dtsProps.setProperty(propName, propValue);
		}
		return i;
	}
	
	/**
	 * reads arguments starting at argument i. if it is a number then add the number to the list of
	 * test cases to process. continue until a non comma is found.
	 * 
	 * @param i
	 * @returns the next parameter to read
	 */
	private static int readList(int i, boolean bSkipComma,String args[],HashSet<Integer> hs) {
		if (args.length == i) {
			return i;
		}
		String whatToProcess = args[i].trim();
		if (!bSkipComma) {
			if (whatToProcess.startsWith(","))
				whatToProcess = whatToProcess.substring(1).trim();
			
			if (whatToProcess.isEmpty()) {
				return readList(++i,true,args,hs);
			}
		}
		try {
			StringTokenIterator sti = new StringTokenIterator(whatToProcess, ",");
			Item itm = null;
			while ((itm = sti.next())!= null) {
				int test = Integer.parseInt(itm.getStringValue());
				hs.add(test);
			}
		} catch (NumberFormatException e) {
			return i-1;
		}
		return readList(++i,true,args,hs);
	}
}

/**
 * The EFM processor will be asking for parameters and parameter values when needed.
 * 
 * @author Ignacio
 */
class EFMConformanceParameterResolver implements EFMParameterResolver {
	
	static class ClsTypeAndValue {
		QName pType;
		String pValue;
		XdmNode sourceNode; // used only when pValue requires namespaces.
		
		/**
		 * @return
		 */
		public Object getParameterValue() {
			// convert to the proper data type. Implement at new data types are found
			if ( (new QName(XBRL.XML_schemaNS,"string")).equals(pType) )
				return pValue;
			if ( (new QName(XBRL.XML_schemaNS,"boolean")).equals(pType) )
				return Boolean.parseBoolean(pValue);
			
			throw new UnsupportedOperationException("Parameter conversion to "+pType.getLocalPart()+" not implemented.");
		}
	}
	
	Hashtable<String,ClsTypeAndValue> database = new Hashtable<String,ClsTypeAndValue>();
	
	private boolean useDefaults = true;
	private String defaultTestCaseName = null;

	private boolean bCikListAdded = false;
	private boolean bCikNameListAdded = false;
	private Vector<String> vCIKs = null;
	private Vector<String> vNAMEs = null;

	/* (non-Javadoc)
	 * @see com.ihr.xbrl.adins.efm.EFMParameterResolver#getParameterValue(java.lang.String)
	 */
	@Override
	public Object getParameterValue(String parameterName) {
		ClsTypeAndValue c = database.get(parameterName);
		if (c != null)
			return c.getParameterValue();
		if (useDefaults) {
			if ("submissionType".equals(parameterName))
				return "8-K";
			else if ("CIK".equals(parameterName))
				return "9876543210";
			else if ("9876543210".equals(parameterName))
				return defaultTestCaseName;
			else if ("forceUtrValidation".equals(parameterName))
				return Boolean.FALSE;
		}
		return null;
	}

	/* (non-Javadoc)
	 * @see com.ihr.xbrl.adins.efm.EFMParameterResolver#isDefaultParameterValue(java.lang.String, java.lang.Object)
	 */
	@Override
	public boolean isDefaultParameterValue(String parameterName, Object parameterValue) {
		Object oToCompare = getParameterValue(parameterName);
		if (oToCompare == null)
			return parameterValue == null;
		else 
			return oToCompare.equals(parameterValue);
	}
	
	/**
	 * @param substring
	 */
	public void setDefaultTestCaseName(String substring) {
		defaultTestCaseName = substring;
	}

	/**
	 * @param pName
	 * @param pType
	 * @param pValue
	 * @param nodeParam
	 */
	public void addParameter(String pName, QName pType, String pValue, XdmNode nodeParam) {
		ClsTypeAndValue value = new ClsTypeAndValue();
		if ("cikNameList".equals(pName)) {
			addCikNameList(pValue);
			return;
		} else if ("cikList".equals(pName)) {
			addCikList(pValue);
			return;
		} else {
			value.pType = pType;
			value.pValue = pValue;
			value.sourceNode = nodeParam;
		}
		database.put(pName, value);
	}

	/**
	 * @param pValue
	 */
	private void addCikList(String pValue) {
		StringTokenizer st = new StringTokenizer(pValue, ",");		
		vCIKs = new Vector<String>(st.countTokens());
		while (st.hasMoreTokens()) {
			vCIKs.add(st.nextToken().trim());
		}
		bCikListAdded = true;
		processCiks();
	}

	/**
	 * @param pValue
	 */
	private void addCikNameList(String pValue) {
		vNAMEs = new Vector<String>();
		for (String val: pValue.split("\\|Edgar\\|")) {
			vNAMEs.add(val.trim());
		}
		bCikNameListAdded = true;
		processCiks();
	}

	/**
	 * 
	 */
	private void processCiks() {
		if (bCikListAdded && bCikNameListAdded) {
			// process only the first matching elements on both vectors.
			
			Iterator<String> iterNames = vNAMEs.iterator();
			Iterator<String> iterCIK = vCIKs.iterator();
			while (iterNames.hasNext() && iterCIK.hasNext()) {
				String name = iterNames.next();
				String cik = iterCIK.next();
				
				ClsTypeAndValue value = new ClsTypeAndValue();
				value.pType = new QName(XBRL.XML_schemaNS,"string");
				value.pValue = name;
				
				database.put(cik, value);
			}
		}
	}

	/**
	 * 
	 */
	public void clear() {
		database.clear();
		defaultTestCaseName = null;
		bCikListAdded = false;
		bCikNameListAdded = false;
		if (vCIKs != null)
			vCIKs.clear();
		vCIKs = null;
		if (vNAMEs != null)
			vNAMEs.clear();
		vNAMEs = null;
	}
}
