package com.ihr.xbrl.sample;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Properties;

import javax.xml.namespace.QName;
import javax.xml.transform.stream.StreamSource;

import net.sf.saxon.expr.StringTokenIterator;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.Sequence;
import net.sf.saxon.om.TreeModel;
import net.sf.saxon.s9api.DocumentBuilder;
import net.sf.saxon.s9api.Processor;
import net.sf.saxon.s9api.SaxonApiException;
import net.sf.saxon.s9api.XPathCompiler;
import net.sf.saxon.s9api.XPathExecutable;
import net.sf.saxon.s9api.XPathSelector;
import net.sf.saxon.s9api.XdmAtomicValue;
import net.sf.saxon.s9api.XdmItem;
import net.sf.saxon.s9api.XdmNode;
import net.sf.saxon.s9api.XdmValue;
import net.sf.saxon.value.QNameValue;

import org.jdom.Attribute;
import org.jdom.DocType;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRLCoreProcessor;
import com.ihr.xbrl.om.XBRLDocument;
import com.ihr.xbrl.om.XBRLValidationReport;
import com.ihr.xbrl.om.errors.DTSDiscoveryException;
import com.ihr.xbrl.om.errors.XBRLInitializationException;
import com.ihr.xbrl.om.errors.XBRLRuntimeException;
import com.ihr.xbrl.om.errors.XBRLValidationException;
import com.ihr.xbrl.om.xdt.XDTProcessor;
import com.ihr.xbrl.util.GeneralIterable;
import com.ihr.xbrl.util.XMLTools;
import com.rs.lic.LicenseVerificationException;

public class LaunchExtensibleEnumerationsConformanceSuite {
	private static final String XHTMLNS = "http://www.w3.org/1999/xhtml";
	
	public static void usage() {
		System.out.println("Usage: LaunchExtensibleEnumerationsConformanceSuite {-f {Conformance suite file}} [-m] [-z] [-x {exclusion pattern}] [-o {Output file}] [-t {test number,[test number]}] [-p {propname=propvalue[,propname=propvalue]}]");
		System.exit(-1);
	}

	/**
	 * @param args
	 * 
	 * args[0] = xbrl.xml
	 * args[1] = pattern to exclude tests
	 */
	public static void main(String[] args) {
		try {
			String confSuiteFile = null;
			String excludePatterns = null;
			String outputFile = null;
			boolean bShowMemoryUsage = false;
			
			Properties dtsProps = new Properties();
			dtsProps.setProperty(DTSContainer.PROCESSORS_SEQUENCE,
				XBRLCoreProcessor.PROCESSORNAME+","+
				XDTProcessor.PROCESSORNAME
			);
			
			int countError = 0;
			int countOk = 0;
			int countSkiped = 0;
			int testNumber = 0;
			HashSet<Integer> hsIncludedTests = null;
			boolean inThread = false;
			
			// parse arguments
			for (int i=0; i < args.length; ++i) {
				if ("-f".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					confSuiteFile = args[i];
					continue;
				} else if ("-m".equals(args[i])) {
					bShowMemoryUsage = true;
					continue;
				} else if ("-x".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					excludePatterns = args[i];
					continue;
				} else if ("-o".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					outputFile = args[i];
					continue;
				} else if ("-p".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					dtsProps.clear();
					i = readProperties(i,args,dtsProps);
					continue;
				} else if ("-t".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					if (hsIncludedTests == null)
						hsIncludedTests = new HashSet<Integer>();
					i = readList(i,false,args,hsIncludedTests);
					continue;
				} else if ("-z".equals(args[i])) {
					if (args.length < i)
						usage();
					inThread = true;
					continue;
				} else {
					System.out.println(args[i]+" Unknown argument!");
					usage();
				}
			}
			
			if (confSuiteFile == null)
				usage();
			
			// Load the conformance suite
			Processor xproc = XMLTools.newProcessor(true);
			XPathCompiler xpc = xproc.newXPathCompiler();
			xpc.setSchemaAware(true);
			
			DocumentBuilder builder = XMLTools.newDocumentBuilder(xproc, TreeModel.TINY_TREE, null);
			XdmNode xbrl = builder.build(new File(confSuiteFile));

			// The output will be sent to a JDom document
			Document page = new Document();
			page.setDocType(new DocType("html","-//W3C//DTD XHTML 1.0 Strict//EN","http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd"));
			Element html = new Element("html",XHTMLNS);
			page.addContent(html);
			Element head = new Element("head",XHTMLNS);
			html.addContent(head);
			Element title = new Element("title",XHTMLNS);
			title.setText("Compliance report with the XBRL Extensible Enumerations Conformance suite");
			head.addContent(title);
			Element style = new Element("style",XHTMLNS);
			head.addContent(style);
			Attribute styleType = new Attribute("type","text/css");
			style.setAttribute(styleType);
			style.setText("td.testOK { background: lightgreen; }\ntd.testFail { background: red; }");			
			Element body = new Element("body",XHTMLNS);
			html.addContent(body);
			Element pageH2 = new Element("h2",XHTMLNS);
			pageH2.setText("Compliance report with the XBRL Extensible Enumerations Conformance suite");
			body.addContent(pageH2);
			Date d = new Date(System.currentTimeMillis());
			SimpleDateFormat dtf = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSZ");
			Element header3 = new Element("h3",XHTMLNS);
			header3.setText("Date executed: "+dtf.format(d));
			body.addContent(header3);
			Element outTable = new Element("table",XHTMLNS);
			Attribute border1 = new Attribute("border","1");
			outTable.setAttribute(border1);
			body.addContent(outTable);
			
			XPathSelector xps = xpc.compile("for $i in testcases/testcase return resolve-uri($i/@uri,base-uri($i))").load();
			//List<String> lst = (List<String>)xp.evaluate("for $i in testcases/testcase return resolve-uri($i/@uri,base-uri($i))",xbrl,XPathConstants.NODESET);
			xps.setContextItem(xbrl);
			
//			XPathSelector selExecDescription = xpc.compile("/testcase/@description").load();
//			XPathSelector selVariation = xpc.compile("/testcase/variation").load();
//			XPathSelector selId = xpc.compile("@id").load();
//			XPathSelector selDescText = xpc.compile("description/text()").load();
//			XPathSelector selFirstFile = xpc.compile("for $i in data/*[@readMeFirst = true()] return resolve-uri($i,base-uri($i))").load();
//			XPathSelector selFirstFileAlt = xpc.compile("for $i in data/* return resolve-uri($i,base-uri($i))").load();
//			XPathSelector selResult = xpc.compile("result/@expected").load();
			xpc.declareNamespace("t", "http://xbrl.org/2005/conformance");
			XPathExecutable execSelExecDescription = xpc.compile("/t:testcase/@description");
			XPathExecutable execSelVariation = xpc.compile("/t:testcase/t:variation");
			XPathExecutable execSelId = xpc.compile("@id");
			XPathExecutable execSelDescText = xpc.compile("t:description/text()");
			XPathExecutable execSelFirstFile = xpc.compile("for $i in t:data/*[@readMeFirst = true()] return resolve-uri($i,base-uri($i))");
			XPathExecutable execSelFirstFileAlt = xpc.compile("for $i in t:data/* return resolve-uri($i,base-uri($i))");
			XPathExecutable execSelResult = xpc.compile("if (exists(t:result/t:error)) then t:result/t:error else 'valid'");
			
			//Iterator<String> iter = lst.iterator();
			Iterator<XdmItem> iter = xps.iterator();
			long now = System.currentTimeMillis();
			while (iter.hasNext()) {
				XdmItem item = iter.next();
				String strURI = item.getStringValue();
				URI uri = new URI(strURI);
				StreamSource fstc = new StreamSource(uri.toURL().openStream());
				fstc.setSystemId(strURI);
				DocumentBuilder db = XMLTools.newDocumentBuilder(xproc, TreeModel.TINY_TREE, xproc.getSchemaManager().newSchemaValidator());
				XdmNode testcases = db.build(fstc);
				//String description = (String)xp.evaluate("/testcase/@description",testcases,XPathConstants.STRING);
				
//				XPathSelector selExecDescription = xpc.compile("/testcase/@description").load();
//				XPathSelector selVariation = xpc.compile("/testcase/variation").load();
//				XPathSelector selId = xpc.compile("@id").load();
//				XPathSelector selDescText = xpc.compile("description/text()").load();
//				XPathSelector selFirstFile = xpc.compile("for $i in data/*[@readMeFirst = true()] return resolve-uri($i,base-uri($i))").load();
//				XPathSelector selFirstFileAlt = xpc.compile("for $i in data/* return resolve-uri($i,base-uri($i))").load();
//				XPathSelector selResult = xpc.compile("result/@expected").load();
				
				XPathSelector selExecDescription = execSelExecDescription.load();
				selExecDescription.setContextItem(testcases);
				String description = selExecDescription.evaluateSingle().getStringValue();
				Element tr01 = new Element("tr",XHTMLNS);
				outTable.addContent(tr01);
				Element td01 = new Element("td",XHTMLNS);
				tr01.addContent(td01);
				td01.setAttribute("style", "background: orange");
				td01.setText(description);
				//List<?> vars = (List<?>)xp.evaluate("/testcase/variation",testcases,XPathConstants.NODESET);
				
				XPathSelector selVariation = execSelVariation.load();
				selVariation.setContextItem(testcases);
				Iterator<XdmItem> iterVars = selVariation.iterator(); 
				//Iterator<?> iterVars = vars.iterator();
				while (iterVars.hasNext()) {
					Element tr02 = new Element("tr",XHTMLNS);
					outTable.addContent(tr02);
					Element td02 = new Element("td",XHTMLNS);
					tr02.addContent(td02);
					Element table02 = new Element("table",XHTMLNS);
					td02.addContent(table02);
					Attribute at02 = new Attribute("border","1");
					table02.setAttribute(at02);
					Attribute at03 = new Attribute("width","100%");
					table02.setAttribute(at03);
					Element tr03 = new Element("tr",XHTMLNS);
					table02.addContent(tr03);
					XdmItem var = iterVars.next();
					
					XPathSelector selId = execSelId.load();
					selId.setContextItem(var);
					String id = "";
					XdmItem idItem = selId.evaluateSingle();
					if (idItem != null)
						id = idItem.getStringValue();
					//String id = (String)xp.evaluate("@id",var,XPathConstants.STRING);
					
					XPathSelector selDescText = execSelDescText.load();
					selDescText.setContextItem(var);
					String desc = selDescText.evaluateSingle().getStringValue();
					//String desc = (String)xp.evaluate("description/text()",var,XPathConstants.STRING);
					Element td03 = new Element("td",XHTMLNS);
					tr03.addContent(td03);
					td03.setText("Id: "+id+" n:"+(testNumber+countSkiped+1));
					Element td04 = new Element("td",XHTMLNS);
					td04.setText("Description: "+desc);
					tr03.addContent(td04);
					Element tr04 = new Element("tr",XHTMLNS);
					table02.addContent(tr04);
					
					XPathSelector selFirstFile = execSelFirstFile.load();
					selFirstFile.setContextItem(var);
					XdmItem selFirstFileResult = selFirstFile.evaluateSingle();
					String varF = null;
					if (selFirstFileResult != null)
						varF = selFirstFileResult.getStringValue();
					//String varF = (String)xp.evaluate("for $i in data/*[@readMeFirst = true()] return resolve-uri($i,base-uri($i))",var,XPathConstants.STRING);
					if (varF == null || varF.trim().isEmpty()) {
						// may not contain readMeFirst=true attribute, in this case only one element is allowed
						XPathSelector selFirstFileAlt = execSelFirstFileAlt.load();
						selFirstFileAlt.setContextItem(var);
						XdmValue lstX = selFirstFileAlt.evaluate();						
						//List lstX = (List)xp.evaluate("for $i in data/* return resolve-uri($i,base-uri($i))",var,XPathConstants.NODESET);
						if (lstX.size() != 1) {
							System.err.println("Test number: "+testNumber+" conformance suite error. Multiple files exist in the test case and none of them are marked as readMeFirst=true()");
							varF = "";
						} else
							varF = lstX.itemAt(0).getStringValue();
					}
					if ((excludePatterns != null && varF.matches(excludePatterns)) || (hsIncludedTests != null && !hsIncludedTests.contains(testNumber+countSkiped+1))) {
						Element td05 = new Element("td",XHTMLNS);
						td05.setText("Execution excluded");
						tr04.addContent(td05);
						countSkiped++;
					} else {
						testNumber++;
						// the only one reason for doing this in a separate thread is a bug in version 9.4.0.3 in Saxon processor that creates
						// thread specific variables that consume all memory if Processors are created in the same thread.
						ExtensibleEnumerationTestCase tc = new ExtensibleEnumerationTestCase(bShowMemoryUsage,testNumber,countSkiped,dtsProps,varF);
						if (inThread) {
							Thread th = new Thread(tc);
							th.start();
							try {
								th.join();
							} catch (InterruptedException e1) {							
							}
						} else {
							tc.run();
						}
						XPathSelector selResult = execSelResult.load();
						selResult.setContextItem(var);
						XdmItem resultItem = selResult.evaluateSingle();
						
						String strRes = "";
						QName errCode = null;
						if (resultItem instanceof XdmAtomicValue) {
							strRes = ((XdmAtomicValue)resultItem).getStringValue();
						} else if (resultItem instanceof XdmNode) {
							Sequence v = ((XdmNode)resultItem).getTypedValue().getUnderlyingValue();
							errCode = XMLTools.javaxQName((QNameValue)v);
						}
						
						//String strRes = (String)xp.evaluate("result/@expected",var,XPathConstants.STRING);
						if (errCode != null && tc.isError(errCode)) {
							Element td05 = new Element("td",XHTMLNS);
							Attribute at06 = new Attribute("class","testOK");
							td05.setAttribute(at06);
							td05.setText("OK. Expected invalid - result is invalid");
							tr04.addContent(td05);
							countOk++;
						} else if (errCode != null && !tc.isError(errCode)) {
							Element td05 = new Element("td",XHTMLNS);
							Attribute at06 = new Attribute("class","testFail");
							td05.setAttribute(at06);
							td05.setText("NOT PASS. Expected invalid - result is valid");
							tr04.addContent(td05);
							countError++;
						} else if ("valid".equalsIgnoreCase(strRes) && tc.bError) {
							Element td05 = new Element("td",XHTMLNS);
							Attribute at06 = new Attribute("class","testFail");
							td05.setAttribute(at06);
							td05.setText("NOT PASS. Expected valid - result is invalid");
							tr04.addContent(td05);
							countError++;
						} else if ("valid".equalsIgnoreCase(strRes) && !tc.bError) {
							Element td05 = new Element("td",XHTMLNS);
							Attribute at06 = new Attribute("class","testOK");
							td05.setAttribute(at06);
							td05.setText("OK. Expected valid - result is valid");
							tr04.addContent(td05);
							countOk++;
						} else {
							Element td05 = new Element("td",XHTMLNS);
							td05.setText("UNKNOWN. Expected "+strRes+" result is "+tc.bError);
							tr04.addContent(td05);
							countError++;
						}
						Element td06 = new Element("td",XHTMLNS);
						td06.setText(tc.result.toString());
						tr04.addContent(td06);
						Element tr05 = new Element("tr",XHTMLNS);
						table02.addContent(tr05);
						Element td07 = new Element("td",XHTMLNS);
						Attribute at07 = new Attribute("colspan","2");
						td07.setAttribute(at07);
						td07.addContent(createLinkTo(varF));
//						td07.setText(varF);
						tr05.addContent(td07);
					}
				} // Shall we close stream here?
//				if (bShowMemoryUsage) {
//					//System.runFinalization();
//					System.gc();
//					long total = Runtime.getRuntime().totalMemory();
//					long free = Runtime.getRuntime().freeMemory();
//					System.out.println("Test number: "+testNumber+" "+"Memory at 2: T="+total+" F="+free+" U="+(total-free));
//				}
			}
			Element table03 = new Element("table",XHTMLNS);
			body.addContent(table03);
			Element tr05 = new Element("tr",XHTMLNS);
			table03.addContent(tr05);
			Element td07 = new Element("td",XHTMLNS);
			tr05.addContent(td07);
			td07.setText("Not passed:");
			Element td08 = new Element("td",XHTMLNS);
			tr05.addContent(td08);
			td08.setText((new Integer(countError)).toString());
			Element tr06 = new Element("tr",XHTMLNS);
			table03.addContent(tr06);
			Element td09 = new Element("td",XHTMLNS);
			tr06.addContent(td09);
			td09.setText("Passed:");
			Element td10 = new Element("td",XHTMLNS);
			tr06.addContent(td10);
			td10.setText((new Integer(countOk)).toString());
			Element tr07 = new Element("tr",XHTMLNS);
			table03.addContent(tr07);
			Element td11 = new Element("td",XHTMLNS);
			tr07.addContent(td11);
			td11.setText("Not executed:");
			Element td12 = new Element("td",XHTMLNS);
			tr07.addContent(td12);
			td12.setText((new Integer(countSkiped)).toString());

			// produce the output to stdout or file
			Format ft = Format.getPrettyFormat();
			ft.setEncoding("ISO-8859-1");
			XMLOutputter outputter = new XMLOutputter(ft);
			String s = outputter.outputString(page);
			if (outputFile != null) {
				File f = new File(outputFile);
				FileWriter fw = new FileWriter(f);
				fw.write(s);
				fw.flush();
				fw.close();
			} else {
				System.out.println(s);
			}
			long now2 = System.currentTimeMillis();
			System.out.println("Total execution time "+(now2-now)/1000+"s");
		} catch (URISyntaxException e) {
			e.printStackTrace();
		} catch (MalformedURLException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (SaxonApiException e) {
			e.printStackTrace();
		}
		System.exit(0);
	}

	private static Element createLinkTo(String varF) {
		Element link = new Element("a",XHTMLNS);		
		Attribute at = new Attribute("href",varF);
		link.setAttribute(at);
		link.setText(varF);
		return link;
	}

	/**
	 * @param i
	 * @param args
	 * @param dtsProps 
	 * @return
	 */
	private static int readProperties(int i, String[] args, Properties dtsProps) {
		if (args.length == i)
			return i;
		StringBuffer sb = new StringBuffer();
		int j=i;
		for (;j<args.length;++j) {
			String argument = args[j].trim();
			if (argument.startsWith("-")) {
				break;
			} else {
				sb.append(argument);
			}
		}		
		i = j-1;
		if (sb.length() == 0)
			usage();
		StringTokenIterator sti = new StringTokenIterator(sb.toString(), ",");
		Item itm = null;
		while ((itm = sti.next()) != null) {
			// find equal sign
			String token = itm.getStringValue();
			int idx = token.indexOf("=");
			if (idx < 0)
				usage();
			String propName = token.substring(0, idx);
			String propValue = token.substring(idx+1);
			dtsProps.setProperty(propName, propValue);
		}
		return i;
	}
	
	/**
	 * reads arguments starting at argument i. if it is a number then add the number to the list of
	 * test cases to process. continue until a non comma is found.
	 * 
	 * @param i
	 * @returns the next parameter to read
	 */
	private static int readList(int i, boolean bSkipComma,String args[],HashSet<Integer> hs) {
		if (args.length == i) {
			return i;
		}
		String whatToProcess = args[i].trim();
		if (!bSkipComma) {
			if (whatToProcess.startsWith(","))
				whatToProcess = whatToProcess.substring(1).trim();
			
			if (whatToProcess.isEmpty()) {
				return readList(++i,true,args,hs);
			}
		}
		try {
			StringTokenIterator sti = new StringTokenIterator(whatToProcess, ",");
			Item itm = null;
			while ((itm = sti.next())!= null) {
				int test = Integer.parseInt(itm.getStringValue());
				hs.add(test);
			}
		} catch (NumberFormatException e) {
			return i;
		}
		return readList(++i,true,args,hs);
	}
}

class ExtensibleEnumerationTestCase implements Runnable {
	
	StringBuffer result = new StringBuffer();
	boolean bError = false;
	boolean bShowMemoryUsage;
	int testNumber;
	int countSkiped;
	Properties dtsProps;
	String varF;
	URI uriF;
	XBRLValidationReport report = null;
	
	ExtensibleEnumerationTestCase(boolean bShowMemoryUsage, int testNumber, int countSkiped, Properties dtsProps, String varF) throws URISyntaxException {
		this.bShowMemoryUsage = bShowMemoryUsage;
		this.testNumber = testNumber;
		this.countSkiped = countSkiped;
		this.dtsProps = dtsProps;
		this.varF = varF;
		this.uriF = new URI(varF);
	}

	/**
	 * @param errCode
	 * @return
	 */
	public boolean isError(QName errCode) {
		if (report == null)
			return false;
		for (XBRLValidationException err : GeneralIterable.asIterator(report.getExceptions())) {
			if (errCode.equals( err.getQNameError()))
				return true;
		}
		return false;
	}

	/* (non-Javadoc)
	 * @see java.lang.Runnable#run()
	 */
	@Override
	public void run() {
		DTSContainer dts = null;
		XBRLDocument doc = null;
		try {
			if (bShowMemoryUsage) {
				//System.out.println("Total memory at 1:"+Runtime.getRuntime().freeMemory())
				long total = Runtime.getRuntime().totalMemory();
				long free = Runtime.getRuntime().freeMemory();
				System.out.println("Test number: "+(testNumber+countSkiped)+" Memory at 1: T="+total+" F="+free+" U="+(total-free)+" L=\""+varF+"\"");
			}
			dts = DTSContainer.newEmptyContainer(dtsProps);
			
			doc = dts.load(uriF);
			// XBRLCoreProcessor proc = XBRLCoreProcessor.create(dts);
			// report = proc.validate();
			report = dts.validate();
			report.throwExceptions();
			report.inconsistenciesAsExceptions();
		} catch (XBRLValidationException e) {
			bError = true;
			result.append(" ");
//			result.append(e.getClass().getCanonicalName());
//			result.append(" ");
			result.append(XBRLValidationException.documentException(e));
			result.append("\n");
		} catch (XBRLRuntimeException e) {
			bError = true;
			result.append(" ");
//			result.append(e.getClass().getCanonicalName());
//			result.append(" ");
			result.append(XBRLValidationException.documentException(e));
			result.append("\n");
		} catch (DTSDiscoveryException e) {
			bError = true;
			result.append(" ");
//			result.append(e.getClass().getCanonicalName());
//			result.append(" ");
			result.append(XBRLValidationException.documentException(e));
			result.append("\n");
		} catch (LicenseVerificationException e) {
			e.printStackTrace();
			System.exit(-1);
		} catch (XBRLInitializationException e) {
			e.printStackTrace();
			System.exit(-1);
		}
		try {
			if (doc != null)
				doc.clear();
			if (dts != null)
				dts.clear();
//			if (report != null)
//				report.clear();
		} catch (Throwable e) {
			System.out.println("Exception thrown during memory cleanup on test #"+(testNumber+countSkiped)+"!");
		}
		doc = null;
		dts = null;
//		report = null;
		if (bShowMemoryUsage) {
//			System.runFinalization();
			System.gc();
			long total = Runtime.getRuntime().totalMemory();
			long free = Runtime.getRuntime().freeMemory();
			System.out.println("Test number: "+(testNumber+countSkiped)+" "+"Memory at 2: T="+total+" F="+free+" U="+(total-free));
		}
	}
}

