package com.ihr.xbrl.sample;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Properties;
import java.util.StringTokenizer;

import javax.xml.namespace.QName;
import javax.xml.transform.stream.StreamSource;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.jdom.Attribute;
import org.jdom.Content;
import org.jdom.DocType;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.Text;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRLValidationReport;
import com.ihr.xbrl.om.errors.DTSDiscoveryException;
import com.ihr.xbrl.om.errors.XBRLValidationException;
import com.ihr.xbrl.om.quality.QualityCheckInconsistency;
import com.ihr.xbrl.om.quality.QualityProcessor;
import com.ihr.xbrl.om.quality.rs.RSCHTQFactory;
import com.ihr.xbrl.util.XMLTools;

import net.sf.saxon.expr.StringTokenIterator;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.TreeModel;
import net.sf.saxon.s9api.DocumentBuilder;
import net.sf.saxon.s9api.Processor;
import net.sf.saxon.s9api.SaxonApiException;
import net.sf.saxon.s9api.XPathCompiler;
import net.sf.saxon.s9api.XPathExecutable;
import net.sf.saxon.s9api.XPathSelector;
import net.sf.saxon.s9api.XdmItem;
import net.sf.saxon.s9api.XdmNode;
import net.sf.saxon.s9api.XdmValue;

/**
 * In principle this is a copy of the LaunchFormulaConformanceSuiteXHTML class. We would review later if sharing classes or code would make sense.
 * 
 * @author Ignacio
 *
 */
public class LaunchFRTAConformanceSuiteXHTML {
	
	// when test cases in the FRTA conformance suite fails other rules not under validation
	// this array documents what rule shall be ignored when testing for another rule. The
	// array is used to configure the processor properties adding a ignore rule property
	private static Object[][] testCaseProperties = new Object[][] {
		{  60, QualityProcessor.getPropertyKey("2.3.9"),  "false"},
		{  61, QualityProcessor.getPropertyKey("2.3.9"),  "false"},
		{  65, QualityProcessor.getPropertyKey("2.3.9"),  "false"},
		{  72, DTSContainer.DTS_DISCOVERY_REPLACE_OFFICIAL_SCHEMAS_SCHEMALOCATION, "true"},
		{  88, QualityProcessor.getPropertyKey("2.1.10"), "false"},
		{  88, QualityProcessor.getPropertyKey("2.1.11"), "false"},
		{  89, QualityProcessor.getPropertyKey("2.1.10"), "false"},
		{  89, QualityProcessor.getPropertyKey("2.1.11"), "false"},
		{ 112, QualityProcessor.getPropertyKey("2.1.5"),  "false"},
		{ 122, QualityProcessor.getPropertyKey("3.1.7"),  "false"},
		{ 123, QualityProcessor.getPropertyKey("3.1.7"),  "false"},
		{ 124, QualityProcessor.getPropertyKey("3.1.7"),  "false"},
		{ 125, QualityProcessor.getPropertyKey("3.1.7"),  "false"},
		{ 126, QualityProcessor.getPropertyKey("3.1.7"),  "false"},
		{ 133, QualityProcessor.getPropertyKey("4.3.2"),  "false"},
		{ 134, QualityProcessor.getPropertyKey("4.3.2"),  "false"},
		{ 136, QualityProcessor.getPropertyKey("4.3.2"),  "false"}
	};
	
	// test case 72 of the FRTA conformance suite contains <import namespace="http://www.xbrl.org/2003/instance" schemaLocation="xbrl-instance-2003-12-31.xsd"/>
	// which is wrong because that resolves locally to the same directory where the container schema is located. RS processor
	// raises a DTSDiscoveryException by default. but, if configured with DTSContainer.DTS_DISCOVERY_REPLACE_OFFICIAL_SCHEMAS_SCHEMALOCATION = true
	// the processor fix the wrong location by the official location and adds a warning message to the validation result.
	// this array holds test case numbers and messages that must be ignored to properly detect the rule being checked
	static Object[][] ignoreErrors = new Object[][] {
		{  8, 620 },
		{ 60, 620 },
		{ 61, 620 },
		{ 65, 620 },
		{ 66, 620 }, // import with namespace and no schemaLocation
		{ 72, 620 } // official schema with relative location to the container schema
	};
	
//	/**
//	 * @author Jose Maria
//	 */
//	public static class ConformanceParametersFactory implements HasCreateFormulaParameterResolver {
//		
//		private XdmValue value;
//
//		/**
//		 * @param evaluate
//		 */
//		public ConformanceParametersFactory(XdmValue value) {
//			this.value = value;
//		}
//
//		/* (non-Javadoc)
//		 * @see com.ihr.xbrl.om.HasCreateFormulaParameterResolver#createAndRegisterFormulaParameterResolver(com.ihr.xbrl.om.DTSContainer)
//		 */
//		@Override
//		public void createAndRegisterFormulaParameterResolver(DTSContainer dts) {
//			ConformanceParameters ccf = new ConformanceParameters(value);			
//			dts.installFormulaParameterResolver(ccf);			
//			value = null;
//		}
//	}
//
	private static final String XHTMLNS = "";
	private static final String CONFORMANCE_NAMESPACE = "";
	
	private static final Logger log = Logger.getLogger(LaunchFRTAConformanceSuiteXHTML.class);
	
	public static void usage() {
		System.out.println("Usage: LaunchFRTAConformanceSuiteXHTML {-f {Conformance suite file}} [-m] [-x {exclusion pattern}] [-r {starting test case number} {end test case number}] [-n {test case number}] [-o {Output file}] [-d {DEBUG_LEVEL}]\n" +
				"DEBUG LEVEL can be OFF,TRACE,...");
		System.exit(-1);
	}
	
	@SuppressWarnings("unused")
	private static boolean debug = false;
	 		
	/**
	 * Valid (mostly for generic linkbases)
	 */
	private static int VALID = 0x08;
		
	/**
	 * Invalid (mostly for generic linkbases)
	 */
	private static int INVALID = 0x10;
	
	/**
	 * @param args
	 */
	@SuppressWarnings("unused")
	public static void main(String[] args) {
		try {
			LinkedList<Integer> errors = new LinkedList<>();
			String confSuiteFile = null;
			String excludePatterns = null;
			String outputFile = null;
			String debugLevel = "OFF";
			boolean bShowMemoryUsage = false;
			Properties dtsProps = new Properties();
			
			int countError = 0;
			int countOk = 0;
			int countSkiped = 0;
			int testNumber = 0;
			int startAt = 0;
			int endsAt = Integer.MAX_VALUE;
			
			ArrayList<Integer> testsToRun = new ArrayList<Integer>();
			
			// parse arguments
			for (int i=0; i < args.length; ++i) {
				if ("-f".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					confSuiteFile = args[i];
				} else if ("-m".equals(args[i])) {
					bShowMemoryUsage = true;
				} else if ("-x".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					excludePatterns = args[i];
				} else if ("-o".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					outputFile = args[i];
				} else if ("-r".equals(args[i])) {
					if (args.length < i+1)
						usage();
					++i;
					startAt = Integer.parseInt(args[i])-1;
					++i;
					endsAt = Integer.parseInt(args[i])-1;
				} else if ("-p".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					i = readProperties(i, args, dtsProps);
					continue;
				} else if ("-d".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					debugLevel = args[i];					
				} else if ("-n".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					testsToRun.add(Integer.parseInt(args[i])-1);
				}
			}
			
			if (confSuiteFile == null)
				usage();
			
			// configure log level
			Logger log1 = Logger.getLogger("com.ihr.xbrl.om");
			log1.setLevel(Level.toLevel(debugLevel));
			log.setLevel(Level.toLevel(debugLevel));
			dtsProps.setProperty(DTSContainer.LOG_LEVEL, debugLevel);
			
			dtsProps.setProperty(DTSContainer.PROCESSORS_SEQUENCE,
//					XBRLCoreProcessor.PROCESSORNAME     +","+
//					XDTProcessor.PROCESSORNAME          +","+
//					XBRLTableProcessor.PROCESSORNAME    +","+
					QualityProcessor.PROCESSORNAME
					);
			
			RSCHTQFactory rsRulesFactory = new RSCHTQFactory();
			for (String ruleCode : rsRulesFactory.getRuleCodes())
				dtsProps.setProperty(QualityProcessor.getPropertyKey(ruleCode), "false");
			
			// Load the conformance suite
			Processor xproc = XMLTools.newProcessor(false);
			XPathCompiler xpc = xproc.newXPathCompiler();
			xpc.declareNamespace("c", CONFORMANCE_NAMESPACE);
			
			DocumentBuilder db = XMLTools.newDocumentBuilder(xproc, TreeModel.TINY_TREE, null);
			// C:\xbrl\conformances\conformance-frta\index.xml ____
			XdmNode xbrl = db.build(new File(confSuiteFile));

			// The output will be sent to a JDom document
			Document page = new Document();
			page.setDocType(new DocType("html"));
			Element html = new Element("html",XHTMLNS);
			page.addContent(html);
			Element head = new Element("head",XHTMLNS);
			html.addContent(head);
			Element title = new Element("title",XHTMLNS);
			title.setText("Compliance report with the FRTA 1.0 Conformance Suite");
			head.addContent(title);
			Element style = new Element("style",XHTMLNS);
			head.addContent(style);
			Attribute styleType = new Attribute("type","text/css");
			style.setAttribute(styleType);
			style.setText("td.testOK { background: lightgreen; }\ntd.testFail { background: red; }");			
			Element body = new Element("body",XHTMLNS);
			html.addContent(body);
			Element pageH2 = new Element("h2",XHTMLNS);
			pageH2.setText("Compliance report with the FRTA 1.0 Conformance suite");
			body.addContent(pageH2);
			Date d = new Date(System.currentTimeMillis());
			SimpleDateFormat dtf = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSZ");
			Element header3 = new Element("h3",XHTMLNS);
			header3.setText("Date executed: "+dtf.format(d));
			body.addContent(header3);
			Element outTable = new Element("table",XHTMLNS);
			Attribute border1 = new Attribute("border","0");
			outTable.setAttribute(border1);
			body.addContent(outTable);
			
			XPathSelector xps = xpc.compile("for $i in testcases/testcase return resolve-uri($i/@uri,base-uri($i))").load();
			xps.setContextItem(xbrl);
			
			XPathExecutable execSelExecDescription = xpc.compile("/c:testcase/@description");
			XPathExecutable execSelVariation = xpc.compile("/c:testcase/c:variation");
			XPathExecutable execSelId = xpc.compile("@id");
			XPathExecutable execSelDescText = xpc.compile("c:description/text()");
			XPathExecutable execSelExecVariationName = xpc.compile("c:name/text()");
			XPathExecutable execSelFirstFile = xpc.compile("for $i in c:data/*[@readMeFirst = true()] return resolve-uri($i,base-uri($i))");
			XPathExecutable execSelParamData = xpc.compile("c:data/c:parameter");
			XPathExecutable execSelFirstFileAlt = xpc.compile("for $i in c:data/* return resolve-uri($i,base-uri($i))");
//			XPathExecutable execSelResult = xpc.compile("c:result/*");
//			XPathExecutable execSelResult = xpc.compile("c:result/@expected");
			XPathExecutable execSelResult2 = xpc.compile("if (exists(c:result/@expected)) then true() else false()");
			XPathExecutable execSelResult3 = xpc.compile("if (c:result/@expected eq 'valid') then true() else false()");
			XPathExecutable execSelInstance = xpc.compile("doc(resolve-uri(.,base-uri(.)))");
			XPathExecutable execSelTable = xpc.compile("doc(resolve-uri(.,base-uri(.)))");
			XPathExecutable execSelCountSatisfied = xpc.compile("string(./@countSatisfied)");
			XPathExecutable execSelCountNotSatisfied = xpc.compile("string(./@countNotSatisfied)");
			XPathExecutable execSelAssertionId = xpc.compile("string(./@assertionID)");
			
			Iterator<XdmItem> iter = xps.iterator();
			long now = System.currentTimeMillis();
			while (iter.hasNext()) {
				XdmItem item = iter.next();
				String strURI = item.getStringValue();
				// file:/C:/xbrl/conformances/conformance-frta/200-Concepts/202-ElementId/202-TestCase-ElementId.xml ____
				URI uri = new URI(strURI);
				StreamSource fstc = null;
				try {
					fstc = new StreamSource(uri.toURL().openStream());
				} catch (FileNotFoundException e) {
					log.warn("File: "+uri.toURL().toString()+" cannot be oppened!\nSource document: "+xbrl.getDocumentURI().toString());
					continue; // skip this file
				}
				fstc.setSystemId(strURI);
				DocumentBuilder db2 = XMLTools.newDocumentBuilder(xproc, TreeModel.TINY_TREE, null);
				XdmNode testcases = db2.build(fstc);
				
				XPathSelector selExecDescription = execSelExecDescription.load();
				selExecDescription.setContextItem(testcases);
				String description;
				XdmItem xDescription = selExecDescription.evaluateSingle();
				if (xDescription != null)
					description = xDescription.getStringValue();
				else
					description = "unknown";
				Element tr01 = new Element("tr",XHTMLNS);
				outTable.addContent(tr01);
				Element td01 = new Element("td",XHTMLNS);
				tr01.addContent(td01);
				td01.setAttribute("style", "background: orange");
				Element link = new Element("a",XHTMLNS);
				link.setText(description);
				link.setAttribute("href", uri.toURL().toString());
				td01.addContent(link);
				
				XPathSelector selVariation = execSelVariation.load();
				selVariation.setContextItem(testcases);
				Iterator<XdmItem> iterVars = selVariation.iterator(); 
				while (iterVars.hasNext()) {
					Element tr02 = new Element("tr",XHTMLNS);
					outTable.addContent(tr02);
					Element td02 = new Element("td",XHTMLNS);
					tr02.addContent(td02);
					Element table02 = new Element("table",XHTMLNS);
					td02.addContent(table02);
					Attribute at02 = new Attribute("border","1");
					table02.setAttribute(at02);
					Attribute at03 = new Attribute("width","100%");
					table02.setAttribute(at03);
					Element tr03 = new Element("tr",XHTMLNS);
					table02.addContent(tr03);
					XdmItem var = iterVars.next();
					
					XPathSelector selId = execSelId.load();
					selId.setContextItem(var);
					String id = "";
					XdmItem idItem = selId.evaluateSingle();
					if (idItem != null)
						id = idItem.getStringValue();
					
					XPathSelector selDescText = execSelDescText.load();
					selDescText.setContextItem(var);
					String desc = selDescText.evaluateSingle().getStringValue();
					
					XPathSelector selVarNameText = execSelExecVariationName.load();
					selVarNameText.setContextItem(var);
					XdmItem nameNode = selVarNameText.evaluateSingle();
					String name = null;
					if (nameNode != null) 
						name = nameNode.getStringValue();
					
					XPathSelector selFirstFile = execSelFirstFile.load();
					selFirstFile.setContextItem(var);
					XdmItem selFirstFileResult = selFirstFile.evaluateSingle();
					// file:/C:/xbrl/conformances/conformance-frta/200-Concepts/202-ElementId/202-00-ProperID-valid.xsd ____
					Element td03 = new Element("td",XHTMLNS);
					tr03.addContent(td03);
					Element href = new Element("a",XHTMLNS);
					href.setAttribute("name", ""+(testNumber+countSkiped+1));
					td03.addContent(href);
					td03.addContent(new Text("Id: "+id+" n:"+(testNumber+countSkiped+1)+(name != null ? (" Name:"+name) : "")));
					Element hrefDown = new Element("a",XHTMLNS);
					hrefDown.setAttribute("href", "#down");
					hrefDown.setText("(dn)");
					td03.addContent(hrefDown);
					Element td04 = new Element("td",XHTMLNS);
					td04.setText("Description: "+desc);
					tr03.addContent(td04);
					Element tr04 = new Element("tr",XHTMLNS);
					table02.addContent(tr04);
					
					// read execution expected result
					boolean bValidInvalidOk = false;
										
					XPathSelector selResult2 = execSelResult2.load();
					selResult2.setContextItem(var);
					XPathSelector selResult3 = execSelResult3.load();
					selResult3.setContextItem(var);
					int testType = selResult3.effectiveBooleanValue() ? VALID : INVALID;

					String varF = null;
					if (selFirstFileResult != null)
						varF = selFirstFileResult.getStringValue();
					// varF = file:/C:/xbrl/conformances/conformance-frta/200-Concepts/202-ElementId/202-00-ProperID-valid.xsd
					if (varF == null || varF.trim().isEmpty()) {
						// may not contain readMeFirst=true attribute, in this case only one element is allowed
						XPathSelector selFirstFileAlt = execSelFirstFileAlt.load();
						selFirstFileAlt.setContextItem(var);
						XdmValue lstX = selFirstFileAlt.evaluate();						
						if (lstX.size() != 1) {
							log.error("Test number: "+(testNumber+countSkiped)+" conformance suite error. Multiple files exist in the test case and none of them are marked as readMeFirst=true()");
							varF = "";
						} else
							varF = lstX.itemAt(0).getStringValue();
					}
					if ((excludePatterns != null && varF.matches(excludePatterns)) || ((testNumber+countSkiped) < startAt || (testNumber+countSkiped) > endsAt) || (testsToRun.size() > 0 && !testsToRun.contains(testNumber+countSkiped))) {
						Element td05 = new Element("td", XHTMLNS);
						td05.setText("Execution excluded");
						tr04.addContent(td05);
						countSkiped++;
					} else {
						testNumber++;
						// build up the test case properties
						Properties testcaseProps = new Properties();
						testcaseProps.putAll(dtsProps); // copy all properties form dtsProps
						for (Object[] sub : testCaseProperties)
							if ((int)sub[0] == testNumber + countSkiped)
								testcaseProps.setProperty(sub[1].toString(), sub[2].toString());
						StringBuilder resultText = new StringBuilder();
						FRTATestCase tc = new FRTATestCase(bShowMemoryUsage, testNumber, countSkiped, testcaseProps, varF, resultText);
						tc.run();
						int exceptionCode = tc.getExceptionCode();
						Element td05 = new Element("td", XHTMLNS);
						boolean bTestOk = false;
						String message = "unknown";
						if (testType == INVALID) {
							if (resultText.length() == 0 || exceptionCode == XBRLValidationException.ERROR_OK) {
								bTestOk = false;
								message = "NOT PASS. Expected invalid - result is valid";
								countError++;
								errors.add(testNumber+countSkiped);
							} else if (exceptionCode != QualityCheckInconsistency.FRTA_INCONSISTENCY_BASE) {
								bTestOk = true;
								message = "POSSIBLY PASS. Expected invalid - result is invalid with another processor";
								countOk++;
							} else {
								bTestOk = true;
								message = "OK. Expected invalid - result is invalid";
								countOk++;
							}
						}
						else if (testType == VALID) {
							if (resultText.length() > 0 || exceptionCode != XBRLValidationException.ERROR_OK) {
								bTestOk = false;
								message = "NOT PASS. Expected valid - result is invalid";
								countError++;
								errors.add(testNumber+countSkiped);
							} else {
								bTestOk = true;
								message = "OK. Expected valid - result is valid";
								countOk++;
							}
						} else {
							System.out.println("Unexpected testType code");
							System.exit(-1);
						}
						Attribute at06 = new Attribute("class", bTestOk ? "testOK" : "testFail");
						td05.setAttribute(at06);
												
						td05.setText(message);
						tr04.addContent(td05);
						
						Element td06 = new Element("td",XHTMLNS);
						td06.setText(resultText.toString().trim());
						tr04.addContent(td06);
					}
					Element tr05 = new Element("tr",XHTMLNS);
					table02.addContent(tr05);
					Element td07 = new Element("td",XHTMLNS);
					Attribute at07 = new Attribute("colspan","2");
					td07.setAttribute(at07);
					tr05.addContent(td07);
					Element link03 = new Element("a",XHTMLNS);
					link03.setText(varF);
					link03.setAttribute("href", varF.substring(varF.indexOf(':')+1));
					td07.addContent(link03);
				} // Shall we close stream here?
			}
			Element table03 = new Element("table",XHTMLNS);
			body.addContent(table03);
			Element tr05 = new Element("tr",XHTMLNS);
			table03.addContent(tr05);
			Element td07 = new Element("td",XHTMLNS);
			tr05.addContent(td07);
			td07.setText("Not passed:");
			Element td08 = new Element("td",XHTMLNS);
			tr05.addContent(td08);
			Element linkDownPair = new Element("a",XHTMLNS);
			linkDownPair.setAttribute("name", "down");
			td08.addContent(linkDownPair);
			td08.addContent(new Text((new Integer(countError)).toString()));
			td08.addContent(getErrorsCollection(errors));
			Element tr06 = new Element("tr",XHTMLNS);
			table03.addContent(tr06);
			Element td09 = new Element("td",XHTMLNS);
			tr06.addContent(td09);
			td09.setText("Passed:");
			Element td10 = new Element("td",XHTMLNS);
			tr06.addContent(td10);
			td10.setText((new Integer(countOk)).toString());
			Element tr07 = new Element("tr",XHTMLNS);
			table03.addContent(tr07);
			Element td11 = new Element("td",XHTMLNS);
			tr07.addContent(td11);
			td11.setText("Not executed:");
			Element td12 = new Element("td",XHTMLNS);
			tr07.addContent(td12);
			td12.setText((new Integer(countSkiped)).toString());

			// produce the output to stdout or file
			Format ft = Format.getPrettyFormat();
			ft.setEncoding("ISO-8859-1");
			XMLOutputter outputter = new XMLOutputter(ft);
			String s = outputter.outputString(page);
			if (outputFile != null) {
				File f = new File(outputFile);
				FileWriter fw = new FileWriter(f);
				fw.write(s);
				fw.flush();
				fw.close();
			} else
				System.out.println(s);
			long now2 = System.currentTimeMillis();
			log.info("Total execution time "+(now2-now)/1000+"s");
		} catch (URISyntaxException e) {
			e.printStackTrace();
		} catch (MalformedURLException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (SaxonApiException e) {
			e.printStackTrace();
		}
		System.exit(0);
	}

	/**
	 * @param errors
	 * @return
	 */
	private static Collection<Content> getErrorsCollection(LinkedList<Integer> errors) {
		ArrayList<Content> al = new ArrayList<>(errors.size());
		for (Integer error : errors) {
			al.add(new Text(" "));
			Element link = new Element("a", XHTMLNS);
			link.setAttribute("href", "#"+error);
			link.setText(""+error);
			al.add(link);
		}
		return al;
	}
	
	/**
	 * @param i
	 * @param args
	 * @param dtsProps 
	 * @return
	 */
	private static int readProperties(int i, String[] args, Properties dtsProps) {
		if (args.length == i)
			return i;
		StringBuffer sb = new StringBuffer();
		int j=i;
		for (;j<args.length;++j) {
			String argument = args[j].trim();
			if (argument.startsWith("-"))
				break;
			sb.append(argument);
		}		
		i = j-1;
		if (sb.length() == 0)
			usage();
		StringTokenIterator sti = new StringTokenIterator(sb.toString(), ",");
		Item itm = null;
		while ((itm = sti.next()) != null) {
			// find equal sign
			String token = itm.getStringValue();
			int idx = token.indexOf("=");
			if (idx < 0)
				usage();
			String propName = token.substring(0, idx);
			String propValue = token.substring(idx+1);
			dtsProps.setProperty(propName, propValue);
		}
		return i;
	}
}

class FRTATestCase implements Runnable {

	boolean bShowMemoryUsage;
	int testNumber;
	int countSkiped;
	Properties dtsProps;
	String varF;
	URI uriF;
	StringBuilder resultText;
	private int exceptionCode;
	
	public int getExceptionCode() {
		return exceptionCode;
	}
	
	FRTATestCase(boolean bShowMemoryUsage, int testNumber, int countSkiped, Properties dtsProps, String varF, StringBuilder resultText) throws URISyntaxException {
		this.bShowMemoryUsage = bShowMemoryUsage;
		this.testNumber = testNumber;
		this.countSkiped = countSkiped;
		this.dtsProps = dtsProps;
		this.varF = varF;
		this.uriF = new URI(varF);
		this.resultText = resultText;
		this.exceptionCode = -1;
	}

	@Override
	public void run() {
		boolean bOnInconsistencies = false; // Quality Check processor does not raise XBRLExceptions. it only raises inconsistencies
		try {
			if (bShowMemoryUsage) {
				long total = Runtime.getRuntime().totalMemory();
				long free = Runtime.getRuntime().freeMemory();
				System.out.println("Test number: "+(testNumber+countSkiped)+" Memory at 1: T="+total+" F="+free+" U="+(total-free)+" L=\""+varF+"\"");
			}

			DTSContainer dts = DTSContainer.newEmptyContainer(dtsProps);
			dts.load(uriF);
			XBRLValidationReport report = dts.validate();
			report.throwExceptions();
			bOnInconsistencies = true;
			report.inconsistenciesAsExceptions();
			
			dts.clear();
			this.exceptionCode = XBRLValidationException.ERROR_OK;
//		} catch (FRTAValidationException e) {
//			//e.printStackTrace();
//			exceptionCode = e.getCode();
//			resultText.append(XBRLValidationException.documentException(e));
		} catch (XBRLValidationException e) {
			//e.printStackTrace();
//			exceptionCode = e.getCode();
			if (bOnInconsistencies) {
				QName inconsistencyCodes = e.getQNameError(); // error
				// pick the first one
				String commaSeparatedCodes = inconsistencyCodes.getLocalPart();
				StringTokenizer st = new StringTokenizer(commaSeparatedCodes, ",");
				while (st.hasMoreTokens()) {
					String code = st.nextToken().trim();
					int icode = Integer.parseInt(code);
					boolean bIgnoreError = false;
					for (Object[] ignore : LaunchFRTAConformanceSuiteXHTML.ignoreErrors)
						if ((int)ignore[0] == testNumber+countSkiped && icode == (int)ignore[1]) {
							bIgnoreError = true;
							break;
						}
					if (!bIgnoreError)
						exceptionCode = icode;
				}
			}
			if (exceptionCode == -1)
				exceptionCode = XBRLValidationException.ERROR_OK;
			else
				resultText.append(XBRLValidationException.documentException(e));
		} catch (DTSDiscoveryException e) {
			//e.printStackTrace();
			exceptionCode = e.getCode();
			resultText.append(XBRLValidationException.documentException(e));
		} catch (Throwable e) {
			// possibly caused by dts.clear() ____
			e.printStackTrace();
			System.exit(-1);
		}
		if (bShowMemoryUsage) {
			System.gc();
			long total = Runtime.getRuntime().totalMemory();
			long free = Runtime.getRuntime().freeMemory();
			System.out.println("Test number: "+(testNumber+countSkiped)+" "+"Memory at 2: T="+total+" F="+free+" U="+(total-free));
		}
	}
}
