package com.ihr.xbrl.sample;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Properties;
import java.util.Vector;

import javax.xml.namespace.QName;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.jdom.Attribute;
import org.jdom.Content;
import org.jdom.DocType;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.Namespace;
import org.jdom.Text;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRL;
import com.ihr.xbrl.om.XBRLDocument;
import com.ihr.xbrl.om.XMLDocumentURI;
import com.ihr.xbrl.om.XMLFragment;
import com.ihr.xbrl.om.errors.DTSDiscoveryException;
import com.ihr.xbrl.om.errors.XBRLRuntimeException;
import com.ihr.xbrl.om.errors.XBRLValidationException;
import com.ihr.xbrl.om.formula.XBRLFormulaProcessor;
import com.ihr.xbrl.om.formula.errors.XFValidationException;
import com.ihr.xbrl.om.instance.XBRLInstance;
import com.ihr.xbrl.om.taxonomy.XBRLTaxonomy;
import com.ihr.xbrl.util.URITools;
import com.ihr.xbrl.util.XMLTools;
import com.rs.lic.LicenseVerificationException;

import net.sf.saxon.expr.StringTokenIterator;
import net.sf.saxon.om.AtomicSequence;
import net.sf.saxon.om.AxisInfo;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.TreeModel;
import net.sf.saxon.s9api.Axis;
import net.sf.saxon.s9api.DocumentBuilder;
import net.sf.saxon.s9api.Processor;
import net.sf.saxon.s9api.SaxonApiException;
import net.sf.saxon.s9api.SaxonApiUncheckedException;
import net.sf.saxon.s9api.XPathCompiler;
import net.sf.saxon.s9api.XPathExecutable;
import net.sf.saxon.s9api.XPathSelector;
import net.sf.saxon.s9api.XdmAtomicValue;
import net.sf.saxon.s9api.XdmItem;
import net.sf.saxon.s9api.XdmNode;
import net.sf.saxon.s9api.XdmNodeKind;
import net.sf.saxon.s9api.XdmSequenceIterator;
import net.sf.saxon.s9api.XdmValue;
import net.sf.saxon.trans.XPathException;
import net.sf.saxon.tree.NamespaceNode;
import net.sf.saxon.tree.iter.AxisIterator;
import net.sf.saxon.type.SchemaType;

public class LaunchFunctionsConformanceSuiteXHTML {
	private static final String XHTMLNS = "http://www.w3.org/1999/xhtml";
	
	private static final Logger log = Logger.getLogger(LaunchFunctionsConformanceSuiteXHTML.class);
	
	@SuppressWarnings("unused")
	private static boolean debug = false;

	public static void usage() {
		System.out.println(
			"Usage: LaunchFunctionsConformanceSuiteXHTML {-f {Conformance suite file}} [-m] [-x {exclusion pattern}] [-r {starting test case number} {end test case number}] [-t {test number,[test number]}] [-s {status filter REC|IWD}] [-o {Output file}] [-d DEBUG_LEVEL].\n"+
				"DEBUG LEVEL can be OFF,TRACE,... and applies to com.ihr.xbrl.om.formula");
		
		System.exit(-1);
	}
		
	private static XMLFragment xdmNodeToXMLFragment(DTSContainer dts, XdmNode node) {
		if (node.getNodeKind() != XdmNodeKind.ELEMENT)
			throw new RuntimeException("Internal error. Expedted ELEMENT node type");

		net.sf.saxon.s9api.QName nodeName = node.getNodeName();
		XMLFragment newFragment = new XMLFragment(dts, nodeName.getLocalName(), nodeName.getPrefix(), nodeName.getNamespaceURI());
		newFragment.setBase(node.getBaseURI());
		SchemaType schemaType = node.getUnderlyingNode().getSchemaType();
		newFragment.setTypeAnnotation(schemaType.getFingerprint() | NodeInfo.IS_NILLED);

		// add namespace definitions ____
		XdmSequenceIterator namespaceIter = node.axisIterator(Axis.NAMESPACE);
		while (namespaceIter.hasNext()) {
			XdmNode namespaceNode = (XdmNode)namespaceIter.next();
			nodeName = namespaceNode.getNodeName();
			if (nodeName != null) {
				String attrPrefix = nodeName.getLocalName();
				String attrUri = namespaceNode.getStringValue();
				if (attrUri.length() > 0)
					newFragment.addNamespace(attrPrefix, attrUri);
			}
		}
		
		// add attributes ____
		XdmSequenceIterator attrIter = node.axisIterator(Axis.ATTRIBUTE);
		while (attrIter.hasNext()) {
			XdmNode attrNode = (XdmNode)attrIter.next();
			try {
				schemaType = attrNode.getUnderlyingNode().getSchemaType();
				String attrValue = attrNode.getTypedValue().getUnderlyingValue().head().getStringValue();
				newFragment.addAttribute(schemaType, attrNode.getNodeName().getLocalName(), attrValue, attrNode.getUnderlyingNode().getPrefix(), attrNode.getUnderlyingNode().getURI());
			} catch (XPathException e) {
				throw new RuntimeException("Internal error. XPathException: "+e);
			} catch (SaxonApiException e) {
				throw new RuntimeException("Internal error. SaxonApiException: "+e);
			}
		}
		
		// add children elements ____
		XdmSequenceIterator childIter = node.axisIterator(Axis.CHILD);
		while (childIter.hasNext()) {
			XdmNode childNode = (XdmNode)childIter.next();
			if (childNode.getNodeKind() == XdmNodeKind.ELEMENT) {
				XMLFragment nf = xdmNodeToXMLFragment(dts, childNode);
				newFragment.insertChildren(new NodeInfo[] {nf}, false, true);
			} else
				newFragment.insertChildren(new NodeInfo[] {childNode.getUnderlyingNode()}, true, true);
		}
		return newFragment;
	}

	/**
	 * @param args
	 * 
	 * args[0] = xbrl.xml
	 * args[1] = pattern to exclude tests
	 */
	public static void main(String[] args) {
		try {
			LinkedList<Integer> errors = new LinkedList<>();
			String confSuiteFile = null;
			String excludePatterns = null;
			String outputFile = null;
			String status = "REC"; // Default is REC status functions only
			boolean bShowMemoryUsage = false;
			String debugLevel = "OFF";
			
			int countError = 0;
			int countOk = 0;
			int countSkiped = 0;
			int testNumber = 0;
			int startAt = 0;
			int endsAt = Integer.MAX_VALUE;
			HashSet<Integer> hsIncludedTests = null;
			ArrayList<Integer> testsToRun = new ArrayList<Integer>();
			
			// parse arguments
			for (int i=0; i < args.length; ++i) {
				if ("-f".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					confSuiteFile = args[i];
				} else if ("-m".equals(args[i])) {
					bShowMemoryUsage = true;
				} else if ("-x".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					excludePatterns = args[i];
				} else if ("-o".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					outputFile = args[i];
				} else if ("-r".equals(args[i])) {
					if (args.length < i+1)
						usage();
					++i;
					startAt = Integer.parseInt(args[i])-1;
					++i;
					endsAt = Integer.parseInt(args[i])-1;
					hsIncludedTests = null;
				} else if ("-d".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					debugLevel = args[i];					
				} else if ("-s".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					status = args[i];
				} else if ("-t".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					startAt = 0;
					endsAt = Integer.MAX_VALUE;
					if (hsIncludedTests == null)
						hsIncludedTests = new HashSet<Integer>();
					i = readList(i, false, args, hsIncludedTests);
				} else if ("-n".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					testsToRun.add(Integer.parseInt(args[i])-1);
				}
			}
			
			if (confSuiteFile == null)
				usage();
			
			// configure log level
			Logger log1 = Logger.getLogger("com.ihr.xbrl.om.formula");
			Logger log2 = Logger.getLogger("com.ihr.xbrl.om.assertions");
			log1.setLevel(Level.toLevel(debugLevel));
			log2.setLevel(Level.toLevel(debugLevel));
			log.setLevel(Level.toLevel(debugLevel));
			
			// Load the conformance suite
			Processor xproc = XMLTools.newProcessor(true);
			File fConfSuite = new File(confSuiteFile);
			DocumentBuilder db = XMLTools.newDocumentBuilder(xproc, TreeModel.TINY_TREE, null);
			XdmNode xbrl = db.build(fConfSuite);
			
			XPathCompiler xpc = xproc.newXPathCompiler();
			xpc.declareNamespace("r", "http://xbrl.org/2008/registry");
			xpc.declareNamespace("n", "http://xbrl.org/2008/function");
			xpc.declareNamespace("c", "http://xbrl.org/2008/conformance");
			xpc.declareNamespace("xlink", XBRL.XLINK_NS);
			xpc.declareNamespace("xhtml", "http://www.w3.org/1999/xhtml");
			xpc.declareNamespace("cfcn", "http://xbrl.org/2008/conformance/function");			

			// The output will be sent to a JDom document
			Document page = new Document();
			page.setDocType(new DocType("html","-//W3C//DTD XHTML 1.0 Strict//EN","http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd"));
			Element html = new Element("html",XHTMLNS);
			page.addContent(html);
			Element head = new Element("head",XHTMLNS);
			html.addContent(head);
			Element title = new Element("title",XHTMLNS);
			title.setText("Compliance report with the XBRL Functions Conformance suite");
			head.addContent(title);
			Element style = new Element("style",XHTMLNS);
			head.addContent(style);
			Attribute styleType = new Attribute("type","text/css");
			style.setAttribute(styleType);
			style.setText("td.testOK { background: lightgreen; }\ntd.testFail { background: red; }");			
			Element body = new Element("body",XHTMLNS);
			html.addContent(body);
			Element pageH2 = new Element("h2",XHTMLNS);
			pageH2.setText("Compliance report with the XBRL Functions Conformance suite");
			body.addContent(pageH2);
			Date d = new Date(System.currentTimeMillis());
			SimpleDateFormat dtf = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSZ");
			Element header3 = new Element("h3",XHTMLNS);
			header3.setText("Date executed: "+dtf.format(d));
			body.addContent(header3);
			Element outTable = new Element("table",XHTMLNS);
			Attribute border1 = new Attribute("border","1");
			outTable.setAttribute(border1);
			body.addContent(outTable);
			
			XPathSelector xps = xpc.compile("for $i in /r:registry/r:entry[r:status=\""+status+"\"]/r:url return doc(resolve-uri($i/@xlink:href,base-uri($i)))").load();
			xps.setContextItem(xbrl);
			XPathSelector selConformanceTest = xpc.compile("for $i in /n:function/n:conformanceTest return doc(resolve-uri($i/@xlink:href,base-uri($i)))").load();
			XPathExecutable execSelExecDescription = xpc.compile("/c:testcase/c:documentation/xhtml:*");
			XPathExecutable execSelVariation = xpc.compile("/c:testcase/c:variation");
			XPathExecutable execSelId = xpc.compile("@id");
			XPathExecutable execSelDescText = xpc.compile("c:documentation/xhtml:*");
			XPathExecutable execSelFirstFile = xpc.compile("for $i in c:inputs/c:*[@readMeFirst = true()] return resolve-uri($i/@xlink:href,base-uri($i))");
			XPathExecutable execSelCall = xpc.compile("c:inputs/cfcn:call/text()");
			XPathExecutable execSelCallContext = xpc.compile("for $i in c:inputs/c:*[@id = ../cfcn:call/@file] return resolve-uri($i/@xlink:href,base-uri($i))");
			XPathExecutable execSelResult = xpc.compile("c:outputs");
			XPathExecutable execSelError = xpc.compile("c:error");
			XPathExecutable execSelTest = xpc.compile("cfcn:test/text()");
			
			Iterator<XdmItem> iter = xps.iterator();
			long now = System.currentTimeMillis();
			try {
				while (iter.hasNext()) {
					XdmItem functionDoc = iter.next();
					selConformanceTest.setContextItem(functionDoc);
					XdmNode testcases = (XdmNode)selConformanceTest.evaluate();
					
					XPathSelector selExecDescription = execSelExecDescription.load();
					selExecDescription.setContextItem(testcases);
					XdmValue descriptionNodes = selExecDescription.evaluate();
					Element tr01 = new Element("tr",XHTMLNS);
					outTable.addContent(tr01);
					Element td01 = new Element("td",XHTMLNS);
					tr01.addContent(td01);
					td01.setAttribute("style", "background: lightblue");
					td01.addContent(jdomNodeSet(descriptionNodes));
					Element p01 = new Element("p",XHTMLNS);
					td01.addContent(p01);
					p01.setText("Test cases file: "+
						URITools.decodeUri(						
							URITools.relativizeMore(((XdmNode)functionDoc).getDocumentURI(), testcases.getDocumentURI()).toString()
						)
					);
					
					XPathSelector selVariation = execSelVariation.load();
					selVariation.setContextItem(testcases);
					Iterator<XdmItem> iterVars = selVariation.iterator(); 
					while (iterVars.hasNext()) {
						Element tr02 = new Element("tr",XHTMLNS);
						outTable.addContent(tr02);
						Element td02 = new Element("td",XHTMLNS);
						tr02.addContent(td02);
						Element table02 = new Element("table",XHTMLNS);
						td02.addContent(table02);
						Attribute at02 = new Attribute("border","1");
						table02.setAttribute(at02);
						Attribute at03 = new Attribute("width","100%");
						table02.setAttribute(at03);
						Element tr03 = new Element("tr",XHTMLNS);
						table02.addContent(tr03);
						XdmItem var = iterVars.next();
						
						XPathSelector selId = execSelId.load();
						selId.setContextItem(var);
						String id = "";
						XdmItem idItem = selId.evaluateSingle();
						if (idItem != null)
							id = idItem.getStringValue();
						
						XPathSelector selDescText = execSelDescText.load();
						selDescText.setContextItem(var);
						XdmValue desc = selDescText.evaluate();
						Element td03 = new Element("td",XHTMLNS);
						tr03.addContent(td03);
						Element href = new Element("a",XHTMLNS);
						href.setAttribute("name", ""+(testNumber+countSkiped+1));
						td03.addContent(href);
						td03.addContent(new Text("Id: "+id+" n:"+(testNumber+countSkiped+1)/*+(name != null ? (" Name:"+name) : "")*/));
						Element hrefDown = new Element("a",XHTMLNS);
						hrefDown.setAttribute("href", "#down");
						hrefDown.setText("(dn)");
						td03.addContent(hrefDown);
						Element td04 = new Element("td",XHTMLNS);
						td04.addContent(jdomNodeSet(desc));
						tr03.addContent(td04);
						Element tr04 = new Element("tr",XHTMLNS);
						table02.addContent(tr04);
						
						// read execution expected result
						QName errorCode = null;
						XPathSelector selResult = execSelResult.load();
						selResult.setContextItem(var);
						XdmNode resultNode = (XdmNode)selResult.evaluateSingle();
						// check if result is an error
						XPathSelector selError = execSelError.load();
						selError.setContextItem(resultNode);
						XdmNode errorNode = (XdmNode)selError.evaluateSingle();
						String testExpression = null;
						if (errorNode instanceof XdmNode) {
							// read errorCode
							AtomicSequence qError = errorNode.getUnderlyingNode().atomize();
							String sVal = qError.getStringValue();
							int idx = sVal.indexOf(':');
							String pfx = sVal.substring(0, idx).trim();
							String local = sVal.substring(idx+1).trim();
							String ns = XMLTools.getNamespaceURIforPrefix(pfx, errorNode);
							errorCode = new QName(ns,local,pfx);
						} else {						
							XPathSelector selTest = execSelTest.load();
							selTest.setContextItem(resultNode);
							XdmValue testValue = selTest.evaluate();
							StringBuffer testExpressionBuffer = new StringBuffer();
							/*
							if (testItem == null) {
								log.warn("Test number: "+(testNumber+countSkiped)+" there is an invalid definition of output test. Obtaining expression from input node.");
								testExpression = resultNode.getStringValue();
							} else {
								testExpression = testItem.getStringValue().trim();
							}
							*/
							for (XdmItem testItem : testValue) {
								if (testExpressionBuffer.length() > 0)
									testExpressionBuffer.append(" and ");
								testExpressionBuffer.append(testItem.getStringValue());
							}
							testExpression = testExpressionBuffer.toString();
						}
						
						XPathSelector selFirstFile = execSelFirstFile.load();
						selFirstFile.setContextItem(var);
						XdmItem selFirstFileResult = selFirstFile.evaluateSingle();
						String varF = null;
						if (selFirstFileResult != null)
							varF = selFirstFileResult.getStringValue();
						QName qE = null;
						if ((excludePatterns != null && varF.matches(excludePatterns)) || ((testNumber+countSkiped) < startAt || (testNumber+countSkiped) > endsAt) || (hsIncludedTests != null && !hsIncludedTests.contains(testNumber+countSkiped+1)) || (testsToRun.size() > 0 && !testsToRun.contains(testNumber+countSkiped))) {
							Element td05 = new Element("td",XHTMLNS);
							td05.setText("Execution excluded");
							tr04.addContent(td05);
							countSkiped++;
						} else {
							testNumber++;
							debug = (testNumber == 342);
							StringBuffer resultText = new StringBuffer();
							StringBuffer result = new StringBuffer();
							boolean testPassed = false;
							DTSContainer dts = null;
							try {
								if (bShowMemoryUsage) {
									long total = Runtime.getRuntime().totalMemory();
									long free = Runtime.getRuntime().freeMemory();
									log.info("Test number: "+(testNumber+countSkiped)+" Memory at 1: T="+total+" F="+free+" U="+(total-free)+" L=\""+varF+"\"");
								}
								Properties props = new Properties();
								
//								props.setProperty(DTSContainer.XEQUAL_MODE, DTSContainer.USE_XEQUAL_AS_FORMULA_COMPLIANT);
//								props.setProperty(DTSContainer.SEQUAL_MODE, DTSContainer.USE_SEQUAL_AS_FORMULA_COMPLIANT);
								props.setProperty(DTSContainer.XEQUAL_MODE, DTSContainer.USE_XEQUAL_AS_XBRL_COMPLIANT);
								props.setProperty(DTSContainer.SEQUAL_MODE, DTSContainer.USE_SEQUAL_AS_XBRL_COMPLIANT);
								
								dts = DTSContainer.newCompatibleContainer(xproc, props);
								URI uriF = new URI(varF);
								XBRLDocument doc = dts.load(uriF);
//								XBRLCoreProcessor proc = new XBRLCoreProcessor(dts);
//								XBRLValidationReport report = proc.validate();								
//								report.throwExceptions();
//								report.inconsistenciesAsExceptions();
//								XDTProcessor xdt = XDTProcessor.create(dts);
//								report.merge(xdt.validate());
//								report.throwExceptions();
//								report.inconsistenciesAsExceptions();
								XBRLFormulaProcessor fp = XBRLFormulaProcessor.create(dts);
//								report.merge(fp.validate());
//								report.throwExceptions();
//								report.inconsistenciesAsExceptions();
								// invoke the function call
								
								// Obtain the node that acts as the context node
								XPathSelector selCallContext = execSelCallContext.load();
								selCallContext.setContextItem(var);
								XdmItem contextItem = selCallContext.evaluateSingle();
								
								XMLDocumentURI cacheDocument = null;
								if (contextItem != null) {
									String contextFileName = contextItem.getStringValue();
									URI fContextURI = URI.create(contextFileName);
									cacheDocument = XBRLDocument.createDocument(dts, fContextURI, null, false, null);
								}
								
								XPathSelector selCall = execSelCall.load();
								selCall.setContextItem(var);
								String xpathCall = selCall.evaluateSingle().getStringValue().trim();
								
								XMLFragment varFragment = xdmNodeToXMLFragment(dts, (XdmNode) var);
								
								XPathCompiler fpXPathCompiler = fp.getFPXPathCompiler(varFragment.getXPathNamespaceContext(), varFragment, true);
								XPathExecutable varExecutable = fpXPathCompiler.compile(xpathCall);
								XPathSelector selFunction = varExecutable.load();
								if (cacheDocument != null)
									selFunction.setContextItem(cacheDocument.getDocument());
								else {
									// warning, there is an error in the test file (ids).
									log.warn("Test number: "+(testNumber+countSkiped)+" "+varF+" contains a variation with unset context.");
									// try to recover the error
									if (doc instanceof XBRLInstance || doc instanceof XBRLTaxonomy) {
										log.warn("Test number: "+(testNumber+countSkiped)+" context set to document node on file: "+doc.getDocumentURI());
										selFunction.setContextItem(doc.getDocument());
									} else {
										log.warn("Test number: "+(testNumber+countSkiped)+" context set to null.");
										selFunction.setContextItem(null);
									}
								}
								if (doc instanceof XBRLInstance)
									dts.setUnderEvaluationInstance((XBRLInstance)doc);
								else
									dts.setUnderEvaluationInstance(null);
								
								// Evaluates the function expression								
								XdmValue functionCallResult = selFunction.evaluate();
								//System.out.println("functionCallResult: "+functionCallResult);
								
								// Execute comparison between the function call result and
								// the expected result
								if (testExpression != null) {
									fpXPathCompiler.declareVariable(new net.sf.saxon.s9api.QName("result"));
									XPathExecutable execTest = fpXPathCompiler.compile(testExpression);
									XPathSelector selTest = execTest.load();
									selTest.setVariable(new net.sf.saxon.s9api.QName("result"), functionCallResult);
									if (cacheDocument != null)
										selTest.setContextItem(cacheDocument.getDocument());
									XdmValue val = selTest.evaluate();
									//System.out.println("testExpression: "+testExpression.trim()+" = "+ val+"\n\n");
									if (val instanceof XdmAtomicValue) {
										if (((XdmAtomicValue)val).getBooleanValue()) {
											countOk++;
											testPassed = true;
										} else {
											countError++;
											errors.add(testNumber+countSkiped);
										}
									} else {
										countError++;
										errors.add(testNumber+countSkiped);
									}
								}
								dts.clear();
							} catch (NullPointerException e) {
								if (log.isTraceEnabled())
									e.printStackTrace();
								countError++;
								errors.add(testNumber+countSkiped);
								resultText.append(" ");
								resultText.append(XBRLValidationException.documentException(e));
								resultText.append("\n");
							} catch (SaxonApiException e) {
								Throwable th = e.getCause();
								if (th instanceof XPathException) {
									XPathException xpe = (XPathException)th;
									if (errorCode != null) {
										if (xpe.getErrorCodeLocalPart() != null) {
											qE = new QName(xpe.getErrorCodeNamespace(), xpe.getErrorCodeLocalPart());
											if (qE.equals(errorCode)) {
												countOk++;
												testPassed = true;
											} else {
												countError++;
												errors.add(testNumber+countSkiped);
											}
										} else {
											countError++;
											errors.add(testNumber+countSkiped);
										}
									} else {
										countError++;
										errors.add(testNumber+countSkiped);
									}
									if (!testPassed && log.isTraceEnabled()) {
										e.printStackTrace();
										th.printStackTrace();
									}
								} else {
									countError++;
									errors.add(testNumber+countSkiped);
									if (log.isTraceEnabled())
										e.printStackTrace();
								}
								resultText.append(" ");
								resultText.append(XBRLValidationException.documentException(e));
								resultText.append("\n");
							} catch (XFValidationException e) {
								qE = e.getQNameError();
								if (errorCode == null || !qE.equals(errorCode)) {
									countError++;
									errors.add(testNumber+countSkiped);
								} else
									countOk++;
								resultText.append(" ");
								resultText.append(XBRLValidationException.documentException(e));
								resultText.append("\n");
								result.append(qE.getPrefix());
								result.append(":");
								result.append(qE.getLocalPart());
							} catch (XBRLValidationException e) {
								if (log.isTraceEnabled())
									e.printStackTrace();
								countError++;
								errors.add(testNumber+countSkiped);
								resultText.append(" ");
								resultText.append(XBRLValidationException.documentException(e));
								resultText.append("\n");
							} catch (XBRLRuntimeException e) {
								if (log.isTraceEnabled())
									e.printStackTrace();
								countError++;
								errors.add(testNumber+countSkiped);
								resultText.append(" ");
								resultText.append(XBRLValidationException.documentException(e));
								resultText.append("\n");
							} catch (DTSDiscoveryException e) {
								if (log.isTraceEnabled())
									e.printStackTrace();
								countError++;
								errors.add(testNumber+countSkiped);
								resultText.append(" ");
								resultText.append(XBRLValidationException.documentException(e));
								resultText.append("\n");
							} catch (LicenseVerificationException e) {
								e.printStackTrace();
								System.exit(-1);
							} catch (Throwable e) {
								if (log.isTraceEnabled())
									e.printStackTrace();
								countError++;
								errors.add(testNumber+countSkiped);
								resultText.append(" ");
								resultText.append(XBRLValidationException.documentException(e));
								resultText.append("\n");
							}
//							finally {
//								if (dts != null)
//									try {
//										dts.clear();
//									} catch (Throwable e) {
//										e.printStackTrace();
//									}
//							}
							if (bShowMemoryUsage) {
								System.gc();
								long total = Runtime.getRuntime().totalMemory();
								long free = Runtime.getRuntime().freeMemory();
								log.info("Test number: "+(testNumber+countSkiped)+" "+"Memory at 2: T="+total+" F="+free+" U="+(total-free));
							}
							Element td05 = new Element("td",XHTMLNS);
							boolean bTestOk = (errorCode == null && qE == null && testPassed)	||
											(errorCode != null && errorCode.equals(qE));
							Attribute at06 = new Attribute("class", bTestOk ? "testOK" : "testFail");
							td05.setAttribute(at06);
							td05.setText("Result unknown at this time.");
							td05.setText(
								bTestOk ? 
									"OK. Expected "+ (errorCode == null ? "valid" : errorCode.getPrefix()+":"+errorCode.getLocalPart()) +
									" - result is "+ (result.length() == 0 ? "valid" : result.toString()) +". "+resultText.toString() :
									"NOT PASS. Expected "+ (errorCode == null ? "valid" : errorCode.getPrefix()+":"+errorCode.getLocalPart()) + 
									" - result is "+ (result.length() == 0 ? "Test failed" : result.toString()) +". "+resultText.toString());
							tr04.addContent(td05);
							
							Element td06 = new Element("td",XHTMLNS);
							td06.setText(resultText.toString());
							tr04.addContent(td06);
							Element tr05 = new Element("tr",XHTMLNS);
							table02.addContent(tr05);
							Element td07 = new Element("td",XHTMLNS);
							Attribute at07 = new Attribute("colspan","2");
							td07.setAttribute(at07);
							tr05.addContent(td07);
//							td07.setText(varF);
							Element link03 = new Element("a",XHTMLNS);
							link03.setText(varF);
							link03.setAttribute("href", "xbrl:"+varF.substring(varF.indexOf(':')+1));
							td07.addContent(link03);
						}
					} // Shall we close stream here?
	//				if (bShowMemoryUsage) {
	//					//System.runFinalization();
	//					System.gc();
	//					long total = Runtime.getRuntime().totalMemory();
	//					long free = Runtime.getRuntime().freeMemory();
	//					System.out.println("Test number: "+testNumber+" "+"Memory at 2: T="+total+" F="+free+" U="+(total-free));
	//				}
				}
			} catch (SaxonApiUncheckedException e) {
				// Error in the conformance suite.
				Element tr01 = new Element("tr",XHTMLNS);
				outTable.addContent(tr01);
				Element td01 = new Element("td",XHTMLNS);
				tr01.addContent(td01);
				td01.setAttribute("style", "background: orange");
				td01.setText(e.getMessage());
			}
			Element table03 = new Element("table",XHTMLNS);
			body.addContent(table03);
			Element tr05 = new Element("tr",XHTMLNS);
			table03.addContent(tr05);
			Element td07 = new Element("td",XHTMLNS);
			tr05.addContent(td07);
			td07.setText("Not passed:");
			Element td08 = new Element("td",XHTMLNS);
			Element linkDownPair = new Element("a",XHTMLNS);
			linkDownPair.setAttribute("name", "down");
			td08.addContent(linkDownPair);
			td08.addContent(new Text((new Integer(countError)).toString()));
			td08.addContent(getErrorsCollection(errors));
			tr05.addContent(td08);
			Element tr06 = new Element("tr",XHTMLNS);
			table03.addContent(tr06);
			Element td09 = new Element("td",XHTMLNS);
			tr06.addContent(td09);
			td09.setText("Passed:");
			Element td10 = new Element("td",XHTMLNS);
			tr06.addContent(td10);
			td10.setText((new Integer(countOk)).toString());
			Element tr07 = new Element("tr",XHTMLNS);
			table03.addContent(tr07);
			Element td11 = new Element("td",XHTMLNS);
			tr07.addContent(td11);
			td11.setText("Not executed:");
			Element td12 = new Element("td",XHTMLNS);
			tr07.addContent(td12);
			td12.setText((new Integer(countSkiped)).toString());

			// produce the output to stdout or file
			Format ft = Format.getPrettyFormat();
			ft.setEncoding("ISO-8859-1");
			XMLOutputter outputter = new XMLOutputter(ft);
			String s = outputter.outputString(page);
			if (outputFile != null) {
				File f = new File(outputFile);
				FileWriter fw = new FileWriter(f);
				fw.write(s);
				fw.flush();
				fw.close();
			} else
				System.out.println(s);
			long now2 = System.currentTimeMillis();
			log.info("Total execution time "+(now2-now)/1000+"s");
		} catch (MalformedURLException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (SaxonApiException e) {
			e.printStackTrace();
		} catch (XPathException e) {
			e.printStackTrace();
		}
		System.exit(0);
	}

	/**
	 * @param errors
	 * @return
	 */
	private static Collection<Content> getErrorsCollection(LinkedList<Integer> errors) {
		ArrayList<Content> al = new ArrayList<>(errors.size());
		for (Integer error : errors) {
			al.add(new Text(" "));
			Element link = new Element("a",XHTMLNS);
			link.setAttribute("href", "#"+error);
			link.setText(""+error);
			al.add(link);
		}
		return al;
	}

	/**
	 * reads arguments starting at argument i. if it is a number then add the number to the list of
	 * test cases to process. continue until a non comma is found.
	 * 
	 * @param i
	 * @returns the next parameter to read
	 */
	private static int readList(int i, boolean bSkipComma, String args[], HashSet<Integer> hs) {
		if (args.length == i)
			return i;
		String whatToProcess = args[i].trim();
		if (!bSkipComma) {
			if (whatToProcess.startsWith(","))
				whatToProcess = whatToProcess.substring(1).trim();
			
			if (whatToProcess.isEmpty())
				return readList(++i, true, args, hs);
		}
		try {
			StringTokenIterator sti = new StringTokenIterator(whatToProcess, ",");
			Item itm = null;
			while ((itm = sti.next())!= null) {
				int test = Integer.parseInt(itm.getStringValue());
				hs.add(test);
			}
		} catch (NumberFormatException e) {
			return --i;
		}
		return readList(++i, true, args, hs);
	}

	/**
	 * Jdom node set.
	 *
	 * @param descriptionNodes the description nodes
	 * @return the collection
	 */
	public static Collection<?> jdomNodeSet(XdmValue descriptionNodes) {
		return jdomNodeSet(descriptionNodes.iterator());
	}

	/**
	 * Jdom node set.
	 *
	 * @param iter the iter
	 * @return the collection
	 */
	public static Collection<?> jdomNodeSet(XdmSequenceIterator iter) {
		Vector<Object> vEl = new Vector<Object>();
		while (iter.hasNext()) {
			XdmNode node = (XdmNode)iter.next();
			vEl.add(jdomNode(node));
		}
		return vEl;
	}

	/**
	 * Jdom node.
	 *
	 * @param node the node
	 * @return org.jdom.Element or org.jdom.Text depending on the node type
	 */
	public static Object jdomNode(XdmNode node) {
		if (node.getNodeKind() == XdmNodeKind.ELEMENT)
			return jdomElement(node);
		else if (node.getNodeKind() == XdmNodeKind.TEXT)
			return jdomText(node);
		return null;
	}

	/**
	 * Jdom text.
	 *
	 * @param textNode the text node
	 * @return the org.jdom. text
	 */
	private static org.jdom.Text jdomText(XdmNode textNode) {
		org.jdom.Text text = new org.jdom.Text(textNode.getStringValue());
		return text;
	}

	/**
	 * Jdom element.
	 *
	 * @param node the node
	 * @return the org.jdom. element
	 */
	private static org.jdom.Element jdomElement(XdmNode node) {
		// convert one XdmNode into a new jdomNode
		org.jdom.Element newEl = new org.jdom.Element(
				node.getNodeName().getLocalName(),			
				node.getNodeName().getPrefix(),
				node.getNodeName().getNamespaceURI());
		// add namespace definitions
		AxisIterator ai = node.getUnderlyingNode().iterateAxis(AxisInfo.NAMESPACE);
		NodeInfo ni = ai.next();
		// jmf while (ai.moveNext()) {
		while ( ni != null ) {
			Object oai = ni;
			if (oai instanceof NamespaceNode) {
				NamespaceNode currentNs = (NamespaceNode)oai;
				String prefix = currentNs.getLocalPart();
				String uri = currentNs.getStringValue();
				if ("xml".equals(prefix) && "http://www.w3.org/XML/1998/namespace".equals(uri))
				{
					ni = ai.next();
					continue;
				}
				net.sf.saxon.s9api.QName qNs = new net.sf.saxon.s9api.QName(prefix,uri,"");
				org.jdom.Namespace ns = jdomNamespace(qNs);
				if (!newEl.getNamespace().equals(ns)) // avoids namespace collision when the element namespace is added
					newEl.addNamespaceDeclaration(ns);
			} else
				throw new RuntimeException("Internal error. Unknown returned object type");
			ni = ai.next();
		}
		// add attributes
		XdmSequenceIterator atIter = node.axisIterator(Axis.ATTRIBUTE);
		while (atIter.hasNext()) {
			XdmNode atNode = (XdmNode)atIter.next();
			org.jdom.Attribute at = jdomAttribute(atNode);
			newEl.setAttribute(at);
		}
		// add children elements
		newEl.addContent(jdomNodeSet(node.axisIterator(Axis.CHILD)));
		return newEl;
	}

	/**
	 * Jdom namespace.
	 *
	 * @param nsNode the ns node
	 * @return the org.jdom. namespace
	 */
	private static org.jdom.Namespace jdomNamespace(XdmNode nsNode) {
		String prefix = nsNode.getUnderlyingNode().getPrefix();
		String uri = nsNode.getUnderlyingNode().getURI();
		return jdomNamespace(new net.sf.saxon.s9api.QName(prefix, uri, ""));
	}

	/**
	 * Jdom namespace.
	 *
	 * @param qName the q name
	 * @return the org.jdom. namespace
	 */
	private static org.jdom.Namespace jdomNamespace(net.sf.saxon.s9api.QName qName) {
		return Namespace.getNamespace(qName.getPrefix(), qName.getNamespaceURI());
	}
	
	/**
	 * Jdom attribute.
	 *
	 * @param atNode the at node
	 * @return the org.jdom. attribute
	 */
	private static org.jdom.Attribute jdomAttribute(XdmNode atNode) {
		org.jdom.Attribute newAt = null;
		try {
			String value = atNode.getTypedValue().getUnderlyingValue().head().getStringValue();
			newAt = new org.jdom.Attribute(atNode.getNodeName().getLocalName(), value, jdomNamespace(atNode));			
		} catch (XPathException e) {
			e.printStackTrace();
		} catch (SaxonApiException e) {
			e.printStackTrace();
		}
		return newAt;
	}
}
