package com.ihr.xbrl.sample;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Properties;

import javax.xml.transform.ErrorListener;
import javax.xml.transform.Source;
import javax.xml.transform.TransformerException;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.jdom.Attribute;
import org.jdom.Content;
import org.jdom.DocType;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.Text;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRL;
import com.ihr.xbrl.om.XMLDocumentURI;
import com.ihr.xbrl.om.errors.DTSDiscoveryException;
import com.ihr.xbrl.om.instance.XBRLInstance;
import com.ihr.xbrl.om.ixbrl.IIXBRLProcessor;
import com.ihr.xbrl.om.ixbrl.ixt.IXT1FunctionLibrary;
import com.ihr.xbrl.om.ixbrl.ixt.IXT2FunctionLibrary;
import com.ihr.xbrl.om.ixbrl.ixt.IXT3FunctionLibrary;
import com.ihr.xbrl.util.XMLTools;

import net.sf.saxon.expr.StringTokenIterator;
import net.sf.saxon.functions.FunctionLibrary;
import net.sf.saxon.functions.FunctionLibraryList;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NamespaceResolver;
import net.sf.saxon.om.TreeModel;
import net.sf.saxon.s9api.Axis;
import net.sf.saxon.s9api.DocumentBuilder;
import net.sf.saxon.s9api.Processor;
import net.sf.saxon.s9api.SaxonApiException;
import net.sf.saxon.s9api.SchemaValidator;
import net.sf.saxon.s9api.XPathCompiler;
import net.sf.saxon.s9api.XPathExecutable;
import net.sf.saxon.s9api.XPathSelector;
import net.sf.saxon.s9api.XdmItem;
import net.sf.saxon.s9api.XdmNode;
import net.sf.saxon.s9api.XdmSequenceIterator;
import net.sf.saxon.sxpath.IndependentContext;
import net.sf.saxon.tree.NamespaceNode;
import net.sf.saxon.value.AtomicValue;
import net.sf.saxon.value.QNameValue;

/**
 * In principle this is a copy of the LaunchFRTAConformanceSuiteXHTML class. We would review later if sharing classes or code would make sense.
 * 
 * @author Ignacio
 *
 */
public class LaunchIXBRLTRv2ConformanceSuiteXHTML {
	
	static Object[][] ignoreErrors = new Object[][] { };

	private static class IXBRLTransformationRegistry2TestCase {
		
		boolean bShowMemoryUsage;
		int testNumber;
		int countSkiped;
		XdmItem callNode;
		XdmItem functionName;
		XdmItem input;
		XdmItem output;
		XdmItem outputType;
		XdmItem variationResult;
		StringBuffer resultText;
		Processor xproc;
		
		IXBRLTransformationRegistry2TestCase(
			Processor xproc, 
			boolean bShowMemoryUsage, 
			int testNumber, 
			int countSkiped, 
			XdmItem itemVariation, 
			XdmItem functionName, 
			XdmItem outputType, 
			XdmItem input, 
			XdmItem output, 
			XdmItem variationResult,
			StringBuffer resultText
		) {
			this.bShowMemoryUsage = bShowMemoryUsage;
			this.testNumber = testNumber;
			this.countSkiped = countSkiped;
			this.callNode = itemVariation;
			this.functionName = functionName;
			this.input = input;
			this.output = output;
			this.resultText = resultText;
			this.xproc = xproc;
			this.outputType = outputType;
			this.variationResult = variationResult;
		}
	
		public boolean runTest() {
			boolean bPass = false;
			try {
				if (bShowMemoryUsage) {
					long total = Runtime.getRuntime().totalMemory();
					long free = Runtime.getRuntime().freeMemory();
					System.out.println("Test number: "+(testNumber+countSkiped)+" Memory at 1: T="+total+" F="+free+" U="+(total-free));
				}

				Processor proc = DTSContainer.getCompatibleProcessor(xproc, true);
				// ignore XPath compilation errors that we know will be produced. Add code in debug sessions when needed
				proc.getUnderlyingConfiguration().setErrorListener(new ErrorListener() {

					@Override
					public void warning(TransformerException exception) throws TransformerException {
						log.error(exception);
					}

					@Override
					public void error(TransformerException exception) throws TransformerException {
						log.error(exception);
					}

					@Override
					public void fatalError(TransformerException exception) throws TransformerException {
						log.error(exception);
					}});
				XPathCompiler xpc = proc.newXPathCompiler();
				xpc.setSchemaAware(true);
				FunctionLibrary env = xpc.getUnderlyingStaticContext().getFunctionLibrary();
				if (env instanceof FunctionLibraryList) {
					IIXBRLProcessor fakeProcessor = new IIXBRLProcessor() {
						
						@Override
						public boolean isFormatValueValid(QNameValue format, AtomicValue value) {
							return false;
						}
						
						@Override
						public XMLDocumentURI getXBRLInstanceSource(Source sourceTree) throws DTSDiscoveryException {
							return null;
						}
						
						@Override
						public DTSContainer getContainer() {
							return null;
						}

						@Override
						public boolean isFormatDefined(QNameValue format) {
							return false;
						}

						@Override
						public URI resolveToURI(String targetNamespace) {
							// TODO Auto-generated method stub
							return null;
						}

						@Override
						public String[] getTargets() {
							// TODO Auto-generated method stub
							return null;
						}

						@Override
						public XBRLInstance getInstance(String target) {
							// TODO Auto-generated method stub
							return null;
						}
					};
					((FunctionLibraryList) env).addFunctionLibrary(new IXT1FunctionLibrary(fakeProcessor));
					((FunctionLibraryList) env).addFunctionLibrary(new IXT2FunctionLibrary(fakeProcessor));
					((FunctionLibraryList) env).addFunctionLibrary(new IXT3FunctionLibrary(fakeProcessor));
				} else
					throw new RuntimeException("Standard function library is not a library list ??");

				IndependentContext ic = (IndependentContext)xpc.getUnderlyingStaticContext();
				ic.setNamespaceResolver(new NamespaceResolver() {

					@Override
					public String getURIForPrefix(String prefix, boolean useDefault) {
						XdmSequenceIterator sq = ((XdmNode) callNode).axisIterator(Axis.NAMESPACE);
						while (sq.hasNext()) {
							XdmItem namespaceItem = sq.next();
							NamespaceNode nsn = (NamespaceNode)namespaceItem.getUnderlyingValue();
							if (nsn.getLocalPart().equals(prefix))
								return nsn.getStringValue();
						}
						return "";
					}

					@Override
					public Iterator<String> iteratePrefixes() {
						return null;
					}
				});
				// format: ixt:booleanfalse("Anydata")
				xpc.declareNamespace("xs", XBRL.XML_schemaNS);
				xpc.declareVariable(new net.sf.saxon.s9api.QName("input"));
				String xpathInput;
				if (variationResult.getStringValue().equals("valid"))
					xpathInput = String.format("every $result in (%1$s($input)) satisfies $result eq %2$s(\"%3$s\")", functionName.getStringValue(),checkOutputType(outputType.getStringValue()),output.getStringValue());
				else
					xpathInput = String.format("%1$s($input)", functionName.getStringValue());
				XPathExecutable exec = xpc.compile(xpathInput);
				XPathSelector sel = exec.load();
				sel.setVariable(new net.sf.saxon.s9api.QName("input"), input);
				if (variationResult.getStringValue().equals("valid"))
					bPass = sel.effectiveBooleanValue();
				else {
					sel.evaluateSingle(); // shall raise an error
					bPass = true; // if not an error. result is true so error will be detected
				}
				// format: $result eq "false"
//				String xpathOutput = "$result eq ($output as "+outputType.getStringValue()+")";
//				xpc.declareVariable(new net.sf.saxon.s9api.QName("result"));
//				XPathExecutable testExec = xpc.compile(xpathOutput);
//				XPathSelector sel2 = testExec.load();
//				sel2.setVariable(new net.sf.saxon.s9api.QName("result"), res);
//				sel2.setVariable(new net.sf.saxon.s9api.QName("output"), output);
//				
//				bPass = sel2.effectiveBooleanValue();
			} catch (SaxonApiException e) {
				resultText.append(e.getMessage());
			} catch (Throwable e) {
				resultText.append(e.getMessage());
				e.printStackTrace();
//				System.exit(-1);
			} finally {
				if (bShowMemoryUsage) {
					System.gc();
					long total = Runtime.getRuntime().totalMemory();
					long free = Runtime.getRuntime().freeMemory();
					System.out.println("Test number: "+(testNumber+countSkiped)+" "+"Memory at 2: T="+total+" F="+free+" U="+(total-free));
				}
			}
			return bPass;
		}

		/**
		 * @param stringValue
		 * @return
		 */
		private String checkOutputType(String type) {
			if ("ixt:nocontentType".equals(type))
				return "xs:string";
			if ("ixt:nonNegativeDecimalType".equals(type))
				return "xs:decimal";
			if ("ixt:zeroIntegerType".equals(type))
				return "xs:integer";
			return type;
		}
	}

	private static final String XHTMLNS = "";
	private static final String CONFORMANCE_NAMESPACE = "http://xbrl.org/2011/conformance-rendering/transforms";
	
	private static final Logger log = Logger.getLogger(LaunchIXBRLTRv2ConformanceSuiteXHTML.class);
	
	public static void usage() {
		System.out.println("Usage: LaunchIXBRLTRv2ConformanceSuiteXHTML {-f {Conformance suite file}} [-m] [-r {starting test case number} {end test case number}] [-n {test case number}] [-o {Output file}] [-d {DEBUG_LEVEL}]\n" +
				"DEBUG LEVEL can be OFF,TRACE,...");
		System.exit(-1);
	}
	
	@SuppressWarnings("unused")
	private static boolean debug = false;
	 
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		try {
			LinkedList<Integer> errors = new LinkedList<>();
			String confSuiteFile = null;
			String outputFile = null;
			String debugLevel = "OFF";
			boolean bShowMemoryUsage = false;
			Properties dtsProps = new Properties();
			
			int countError = 0;
			int countOk = 0;
			int countSkiped = 0;
			int testNumber = 0;
			int startAt = 0;
			int endsAt = Integer.MAX_VALUE;
			
			ArrayList<Integer> testsToRun = new ArrayList<Integer>();
			
			// parse arguments
			for (int i=0; i < args.length; ++i) {
				if ("-f".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					confSuiteFile = args[i];
				} else if ("-m".equals(args[i])) {
					bShowMemoryUsage = true;
				} else if ("-o".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					outputFile = args[i];
				} else if ("-r".equals(args[i])) {
					if (args.length < i+1)
						usage();
					++i;
					startAt = Integer.parseInt(args[i])-1;
					++i;
					endsAt = Integer.parseInt(args[i])-1;
				} else if ("-p".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					i = readProperties(i, args, dtsProps);
					continue;
				} else if ("-d".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					debugLevel = args[i];					
				} else if ("-n".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					testsToRun.add(Integer.parseInt(args[i])-1);
				}
			}
			
			if (confSuiteFile == null)
				usage();
			
			// configure log level
			Logger log1 = Logger.getLogger("com.ihr.xbrl.om");
			log1.setLevel(Level.toLevel(debugLevel));
			log.setLevel(Level.toLevel(debugLevel));
			dtsProps.setProperty(DTSContainer.LOG_LEVEL, debugLevel);
			
			// Load the conformance suite
			Processor xproc = XMLTools.newProcessor(true);
			
			final String sTitle = "Compliance report with the Inline XBRL Transformation Registry 2 Conformance Suite";
			// The output will be sent to a JDom document
			Document page = new Document();
			page.setDocType(new DocType("html"));
			Element html = new Element("html", XHTMLNS);
			page.addContent(html);
			Element head = new Element("head", XHTMLNS);
			html.addContent(head);
			Element title = new Element("title", XHTMLNS);
			title.setText(sTitle);
			head.addContent(title);
			Element style = new Element("style", XHTMLNS);
			head.addContent(style);
			Attribute styleType = new Attribute("type", "text/css");
			style.setAttribute(styleType);
			style.setText("td.testOK { background: lightgreen; }\ntd.testFail { background: red; }");			
			Element body = new Element("body", XHTMLNS);
			html.addContent(body);
			Element pageH2 = new Element("h2", XHTMLNS);
			pageH2.setText(sTitle);
			body.addContent(pageH2);
			Date d = new Date(System.currentTimeMillis());
			SimpleDateFormat dtf = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSZ");
			Element header3 = new Element("h3", XHTMLNS);
			header3.setText("Date executed: "+dtf.format(d));
			body.addContent(header3);
			Element outTable = new Element("table", XHTMLNS);
			Attribute border1 = new Attribute("border", "0");
			outTable.setAttribute(border1);
			body.addContent(outTable);
			
			SchemaValidator scv = xproc.getSchemaManager().newSchemaValidator();
			DocumentBuilder db = XMLTools.newDocumentBuilder(xproc, TreeModel.TINY_TREE, scv);
			
			XPathCompiler xpc = xproc.newXPathCompiler();
			xpc.setSchemaAware(true);
			xpc.declareNamespace("c", CONFORMANCE_NAMESPACE);
			XdmNode xdmTestcase = db.build(new File(confSuiteFile));

			XPathSelector selAllTransforms = xpc.compile("for $i in /c:testcase/c:transform return $i").load();
			XPathSelector selVariationTransform = xpc.compile("for $i in ./c:variation return $i").load();
			XPathSelector selTransformName = xpc.compile("data(@name)").load();
			XPathSelector selVariationInput = xpc.compile("data(@input)").load();
			XPathSelector selVariationOutput = xpc.compile("data(@output)").load();
			XPathSelector selVariationResult = xpc.compile("data(@result)").load();
			XPathSelector selVariationOutputType = xpc.compile("data(@outputType)").load();

			selAllTransforms.setContextItem(xdmTestcase);
			Iterator<XdmItem> iterTransform = selAllTransforms.iterator();
			long now = System.currentTimeMillis();
			while (iterTransform.hasNext()) {
				XdmItem itemTransform = iterTransform.next();
				selTransformName.setContextItem(itemTransform);
				XdmItem functionName = selTransformName.evaluateSingle();
				selVariationOutputType.setContextItem(itemTransform);
				XdmItem outputType = selVariationOutputType.evaluateSingle();

				Element tr01 = new Element("tr", XHTMLNS);
				outTable.addContent(tr01);
				Element td01 = new Element("td", XHTMLNS);
				tr01.addContent(td01);
				td01.setAttribute("style", "background: orange");
				Element link = new Element("a", XHTMLNS);
				link.setText(functionName.getStringValue());
				td01.addContent(link);

				selVariationTransform.setContextItem(itemTransform);
				Iterator<XdmItem> iterVariation = selVariationTransform.iterator();
				while (iterVariation.hasNext()) {
					XdmItem itemVariation = iterVariation.next();
					selVariationInput.setContextItem(itemVariation);
					XdmItem variationInput = selVariationInput.evaluateSingle();
					selVariationOutput.setContextItem(itemVariation);
					XdmItem variationOutput = selVariationOutput.evaluateSingle();
					selVariationResult.setContextItem(itemVariation);
					XdmItem variationResult = selVariationResult.evaluateSingle();

					Element tr02 = new Element("tr", XHTMLNS);
					outTable.addContent(tr02);
					Element td02 = new Element("td", XHTMLNS);
					tr02.addContent(td02);
					Element table02 = new Element("table", XHTMLNS);
					td02.addContent(table02);
					Attribute at02 = new Attribute("border", "1");
					table02.setAttribute(at02);
					Attribute at03 = new Attribute("width", "100%");
					table02.setAttribute(at03);
					Element tr03 = new Element("tr", XHTMLNS);
					table02.addContent(tr03);

					Element td03 = new Element("td", XHTMLNS);
					tr03.addContent(td03);
					Element href = new Element("a", XHTMLNS);
					href.setAttribute("name", ""+(testNumber+countSkiped+1));
					td03.addContent(href);
					td03.addContent(new Text("N: "+(testNumber+countSkiped+1)));
					Element hrefDown = new Element("a", XHTMLNS);
					hrefDown.setAttribute("href", "#down");
					hrefDown.setText("(dn)");
					td03.addContent(hrefDown);
					Element td04 = new Element("td", XHTMLNS);
					td04.setText("input=\""+variationInput+"\" output=\""+variationOutput+"\" result=\""+variationResult+"\"");
					tr03.addContent(td04);
					Element tr04 = new Element("tr", XHTMLNS);
					table02.addContent(tr04);

					if (((testNumber+countSkiped) < startAt || (testNumber+countSkiped) > endsAt) || (testsToRun.size() > 0 && !testsToRun.contains(testNumber+countSkiped))) {
						Element td05 = new Element("td", XHTMLNS);
						td05.setText("Execution excluded");
						tr04.addContent(td05);
						countSkiped++;
					} else {
						testNumber++;

						StringBuffer resultText = new StringBuffer();
						IXBRLTransformationRegistry2TestCase tc = new IXBRLTransformationRegistry2TestCase(
								xproc, bShowMemoryUsage, 
								testNumber, countSkiped, 
								itemVariation, functionName, outputType, variationInput, variationOutput, variationResult, resultText);
						boolean bPass = tc.runTest();
						Element td05 = new Element("td", XHTMLNS);
						String message = "unknown";
						boolean passAttr;
						if (bPass) {
							if (variationResult.getStringValue().equals("valid")) {
								message = "OK. Expected valid - result is valid";
								countOk++;
								passAttr = true;
							} else {
								message = "NOT PASS. Expected invalid - result is valid";
								countError++;
								errors.add(testNumber+countSkiped);
								passAttr = false;
							}
						} else {
							if (variationResult.getStringValue().equals("invalid")) {
								message = "OK. Expected invalid - result is invalid";
								countOk++;
								passAttr = true;
							} else {
								message = "NOT PASS. Expected valid - result is invalid";
								countError++;
								errors.add(testNumber+countSkiped);
								passAttr = false;
							}
						}
						Attribute at06 = new Attribute("class", passAttr ? "testOK" : "testFail");
						td05.setAttribute(at06);
						td05.setText(message);
						tr04.addContent(td05);
						Element td06 = new Element("td", XHTMLNS);
						td06.setText(resultText.toString().trim());
						tr04.addContent(td06);
					}
				}
			}
			Element table03 = new Element("table", XHTMLNS);
			body.addContent(table03);
			Element tr05 = new Element("tr", XHTMLNS);
			table03.addContent(tr05);
			Element td07 = new Element("td", XHTMLNS);
			tr05.addContent(td07);
			td07.setText("Not passed:");
			Element td08 = new Element("td", XHTMLNS);
			tr05.addContent(td08);
			Element linkDownPair = new Element("a", XHTMLNS);
			linkDownPair.setAttribute("name", "down");
			td08.addContent(linkDownPair);
			td08.addContent(new Text((new Integer(countError)).toString()));
			td08.addContent(getErrorsCollection(errors));
			Element tr06 = new Element("tr", XHTMLNS);
			table03.addContent(tr06);
			Element td09 = new Element("td", XHTMLNS);
			tr06.addContent(td09);
			td09.setText("Passed:");
			Element td10 = new Element("td", XHTMLNS);
			tr06.addContent(td10);
			td10.setText((new Integer(countOk)).toString());
			Element tr07 = new Element("tr", XHTMLNS);
			table03.addContent(tr07);
			Element td11 = new Element("td", XHTMLNS);
			tr07.addContent(td11);
			td11.setText("Not executed:");
			Element td12 = new Element("td", XHTMLNS);
			tr07.addContent(td12);
			td12.setText((new Integer(countSkiped)).toString());

			// produce the output to stdout or file
			Format ft = Format.getPrettyFormat();
			ft.setEncoding("ISO-8859-1");
			XMLOutputter outputter = new XMLOutputter(ft);
			String s = outputter.outputString(page);
			if (outputFile != null) {
				File f = new File(outputFile);
				FileWriter fw = new FileWriter(f);
				fw.write(s);
				fw.flush();
				fw.close();
			} else
				System.out.println(s);
			long now2 = System.currentTimeMillis();
			log.info("Total execution time "+(now2-now)/1000+"s");
		} catch (MalformedURLException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (SaxonApiException e) {
			e.printStackTrace();
		}
		System.exit(0);
	}

	/**
	 * @param errors
	 * @return
	 */
	private static Collection<Content> getErrorsCollection(LinkedList<Integer> errors) {
		ArrayList<Content> al = new ArrayList<>(errors.size());
		for (Integer error : errors) {
			al.add(new Text(" "));
			Element link = new Element("a", XHTMLNS);
			link.setAttribute("href", "#"+error);
			link.setText(""+error);
			al.add(link);
		}
		return al;
	}
	
	/**
	 * @param i
	 * @param args
	 * @param dtsProps 
	 * @return
	 */
	private static int readProperties(int i, String[] args, Properties dtsProps) {
		if (args.length == i)
			return i;
		StringBuffer sb = new StringBuffer();
		int j=i;
		for (;j<args.length;++j) {
			String argument = args[j].trim();
			if (argument.startsWith("-"))
				break;
			sb.append(argument);
		}		
		i = j-1;
		if (sb.length() == 0)
			usage();
		StringTokenIterator sti = new StringTokenIterator(sb.toString(), ",");
		Item itm = null;
		while ((itm = sti.next()) != null) {
			// find equal sign
			String token = itm.getStringValue();
			int idx = token.indexOf("=");
			if (idx < 0)
				usage();
			String propName = token.substring(0, idx);
			String propValue = token.substring(idx+1);
			dtsProps.setProperty(propName, propValue);
		}
		return i;
	}
}
