package com.ihr.xbrl.sample;

import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.net.URI;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.LinkedList;
import java.util.Properties;

import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.jdom.Attribute;
import org.jdom.Content;
import org.jdom.DocType;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.Text;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.ixbrl.IXBRLTransformerFactory;

import net.sf.saxon.expr.StringTokenIterator;
import net.sf.saxon.om.Item;

/**
 * @author Ignacio
 */
public class LaunchIXBRLv10ConformanceSuiteXHTML {
	
	static Object[][] ignoreErrors = new Object[][] { };
	
	final static String confSuiteDirectory = "inline-1.0-conformance";
	final static String outDirectory = "xbrl-temp-xslt-1.0";

	private static class IXBRL10TestCase {
	
		boolean bShowMemoryUsage;
		int testNumber;
		int countSkiped;
		File fd;
		StringBuilder log;
		
		IXBRL10TestCase(boolean bShowMemoryUsage, int testNumber, int countSkiped, File fd, StringBuilder log) {
			this.bShowMemoryUsage = bShowMemoryUsage;
			this.testNumber = testNumber;
			this.countSkiped = countSkiped;
			this.fd = fd;
			this.log = log;
		}
		
		public boolean runTest() {
			boolean bIsOK = false;
			try {
				if (bShowMemoryUsage) {
					long total = Runtime.getRuntime().totalMemory();
					long free = Runtime.getRuntime().freeMemory();
					System.out.println("Test number: "+(testNumber+countSkiped)+" "+fd.getName()+" Memory at 1: T="+total+" F="+free+" U="+(total-free));
				}

				boolean isError = fd.getName().startsWith("FAIL");
				IXBRLTransformerFactory factory = new IXBRLTransformerFactory();
				factory.setAttribute(IXBRLTransformerFactory.SPECIFICATION_VERSION_ATTRIBUTE, "1.0");
				Transformer tr = factory.newTransformer();
				FileReader fr = new FileReader(fd);
				Source ss = new StreamSource(fr);
				ss.setSystemId(fd.toURI().toString());
				Path outPath = Paths.get(URI.create(fd.toURI().toString().replace(confSuiteDirectory, outDirectory).replace(".html", ".xbrl")));
				outPath.toFile().getParentFile().mkdirs();
				FileOutputStream fout = new FileOutputStream(outPath.toFile());
				Result rr = new StreamResult(fout);
				try {
					tr.transform(ss, rr);
					fout.close();
					fout = null;
					fr.close();
					fr = null;
					if (!isError) {
						bIsOK = true;
						log.append("OK - Check output file!");
					} else
						log.append("ERROR - Error not detected");
				} catch (TransformerException e) {
					if (isError) {
						bIsOK = true;
						log.append(e.getMessage());
					} else
						log.append(e.getMessage());
					if (fout != null)
						fout.close();
					if (fr != null)
						fr.close();
				}
				if (bShowMemoryUsage) {
					System.gc();
					long total = Runtime.getRuntime().totalMemory();
					long free = Runtime.getRuntime().freeMemory();
					System.out.println("Test number: "+(testNumber+countSkiped)+" "+fd.getName()+" "+"Memory at 2: T="+total+" F="+free+" U="+(total-free));
				}
			} catch (Throwable e) {
				// possibly caused by dts.clear() ____
				e.printStackTrace();
				System.exit(-1);
			}
			return bIsOK;
		}
	}

	private static final String XHTMLNS = "";
	
	private static final Logger log = Logger.getLogger(LaunchIXBRLv10ConformanceSuiteXHTML.class);
	
	public static void usage() {
		System.out.println("Usage: LaunchIXBRLv1ConformanceSuiteXHTML [-m] [-r {starting test case number} {end test case number}] [-n {test case number}] [-o {Output file}] [-d {DEBUG_LEVEL}]\n" +
				"DEBUG LEVEL can be OFF,TRACE,...");
		System.exit(-1);
	}
	
	@SuppressWarnings("unused")
	private static boolean debug = false;
	 
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		try {
			LinkedList<Integer> errors = new LinkedList<>();
			String outputFile = null;
			String debugLevel = "OFF";
			boolean bShowMemoryUsage = false;
			Properties dtsProps = new Properties();
			
			int countError = 0;
			int countOk = 0;
			int countSkiped = 0;
			int testNumber = 0;
			int startAt = 0;
			int endsAt = Integer.MAX_VALUE;
			
			ArrayList<Integer> testsToRun = new ArrayList<Integer>();
			
			// parse arguments
			for (int i=0; i < args.length; ++i) {
				if ("-m".equals(args[i])) {
					bShowMemoryUsage = true;
				} else if ("-o".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					outputFile = args[i];
				} else if ("-r".equals(args[i])) {
					if (args.length < i+1)
						usage();
					++i;
					startAt = Integer.parseInt(args[i])-1;
					++i;
					endsAt = Integer.parseInt(args[i])-1;
				} else if ("-p".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					i = readProperties(i, args, dtsProps);
					continue;
				} else if ("-d".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					debugLevel = args[i];					
				} else if ("-n".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					testsToRun.add(Integer.parseInt(args[i])-1);
				}
			}
			
			if (confSuiteDirectory == null)
				usage();
			
			// configure log level
			Logger log1 = Logger.getLogger("com.ihr.xbrl.om");
			log1.setLevel(Level.toLevel(debugLevel));
			log.setLevel(Level.toLevel(debugLevel));
			dtsProps.setProperty(DTSContainer.LOG_LEVEL, debugLevel);
			
			final String sTitle = "Compliance report with the Inline XBRL Processor 1.0 Conformance Suite";
			// The output will be sent to a JDom document
			Document page = new Document();
			page.setDocType(new DocType("html"));
			Element html = new Element("html", XHTMLNS);
			page.addContent(html);
			Element head = new Element("head", XHTMLNS);
			html.addContent(head);
			Element title = new Element("title", XHTMLNS);
			title.setText(sTitle);
			head.addContent(title);
			Element style = new Element("style", XHTMLNS);
			head.addContent(style);
			Attribute styleType = new Attribute("type", "text/css");
			style.setAttribute(styleType);
			style.setText("td.testOK { background: lightgreen; }\ntd.testFail { background: red; }");			
			Element body = new Element("body", XHTMLNS);
			html.addContent(body);
			Element pageH2 = new Element("h2", XHTMLNS);
			pageH2.setText(sTitle);
			body.addContent(pageH2);
			Date d = new Date(System.currentTimeMillis());
			SimpleDateFormat dtf = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSZ");
			Element header3 = new Element("h3", XHTMLNS);
			header3.setText("Date executed: "+dtf.format(d));
			body.addContent(header3);
			Element outTable = new Element("table", XHTMLNS);
			Attribute border1 = new Attribute("border", "0");
			outTable.setAttribute(border1);
			body.addContent(outTable);
			
			long now = System.currentTimeMillis();

			File directory = new File(confSuiteDirectory);
			File[] dirFiles = directory.listFiles();
			if (dirFiles != null) {
				for (final File fd : dirFiles) {
					if ((fd.getName().startsWith("FAIL-") || fd.getName().startsWith("PASS-"))&& fd.getName().endsWith(".html")) {
						String description = fd.getAbsolutePath();
						
						Element tr01 = new Element("tr", XHTMLNS);
						outTable.addContent(tr01);
						Element td01 = new Element("td", XHTMLNS);
						tr01.addContent(td01);
						td01.setAttribute("style", "background: orange");
						Element link = new Element("a", XHTMLNS);
						link.setText(fd.getName());
						td01.addContent(link);
						
						Element tr02 = new Element("tr", XHTMLNS);
						outTable.addContent(tr02);
						Element td02 = new Element("td", XHTMLNS);
						tr02.addContent(td02);
						Element table02 = new Element("table", XHTMLNS);
						td02.addContent(table02);
						Attribute at02 = new Attribute("border", "1");
						table02.setAttribute(at02);
						Attribute at03 = new Attribute("width", "100%");
						table02.setAttribute(at03);
						Element tr03 = new Element("tr", XHTMLNS);
						table02.addContent(tr03);
	
						Element td03 = new Element("td", XHTMLNS);
						tr03.addContent(td03);
						Element href = new Element("a", XHTMLNS);
						href.setAttribute("name", ""+(testNumber+countSkiped+1));
						td03.addContent(href);
						td03.addContent(new Text("N: "+(testNumber+countSkiped+1)));
						Element hrefDown = new Element("a", XHTMLNS);
						hrefDown.setAttribute("href", "#down");
						hrefDown.setText("(dn)");
						td03.addContent(hrefDown);
						
						Element td04 = new Element("td", XHTMLNS);
						Element href2 = new Element("a", XHTMLNS);
						href2.setAttribute("href", "file:///"+description);
						href2.addContent(description);
						td04.addContent(href2);
						tr03.addContent(td04);
						Element tr04 = new Element("tr", XHTMLNS);
						table02.addContent(tr04);
						
						if (((testNumber+countSkiped) < startAt || (testNumber+countSkiped) > endsAt) || (testsToRun.size() > 0 && !testsToRun.contains(testNumber+countSkiped))) {
							Element td05 = new Element("td", XHTMLNS);
							td05.setText("Execution excluded");
							tr04.addContent(td05);
							countSkiped++;
						} else {
							testNumber++;
							
							StringBuilder log = new StringBuilder();
							IXBRL10TestCase tc = new IXBRL10TestCase(bShowMemoryUsage, testNumber, countSkiped, fd, log);
							boolean pass = tc.runTest();
							Element td05 = new Element("td", XHTMLNS);
							String message = "unknown";
							if (!pass) {
								message = "NOT PASS.";
								countError++;
								errors.add(testNumber+countSkiped);
							} else {
								message = "OK.";
								countOk++;
							}
							Attribute at06 = new Attribute("class", pass ? "testOK" : "testFail");
							td05.setAttribute(at06);
							td05.setText(message);
							tr04.addContent(td05);
							Element td06 = new Element("td", XHTMLNS);
							td06.setText(log.toString());
							tr04.addContent(td06);
						}
					}
				}
			}
			Element table03 = new Element("table", XHTMLNS);
			body.addContent(table03);
			Element tr05 = new Element("tr", XHTMLNS);
			table03.addContent(tr05);
			Element td07 = new Element("td", XHTMLNS);
			tr05.addContent(td07);
			td07.setText("Not passed:");
			Element td08 = new Element("td", XHTMLNS);
			tr05.addContent(td08);
			Element linkDownPair = new Element("a", XHTMLNS);
			linkDownPair.setAttribute("name", "down");
			td08.addContent(linkDownPair);
			td08.addContent(new Text((new Integer(countError)).toString()));
			td08.addContent(getErrorsCollection(errors));
			Element tr06 = new Element("tr", XHTMLNS);
			table03.addContent(tr06);
			Element td09 = new Element("td", XHTMLNS);
			tr06.addContent(td09);
			td09.setText("Passed:");
			Element td10 = new Element("td", XHTMLNS);
			tr06.addContent(td10);
			td10.setText((new Integer(countOk)).toString());
			Element tr07 = new Element("tr", XHTMLNS);
			table03.addContent(tr07);
			Element td11 = new Element("td", XHTMLNS);
			tr07.addContent(td11);
			td11.setText("Not executed:");
			Element td12 = new Element("td", XHTMLNS);
			tr07.addContent(td12);
			td12.setText((new Integer(countSkiped)).toString());
			// produce the output to stdout or file
			Format ft = Format.getPrettyFormat();
			ft.setEncoding("ISO-8859-1");
			XMLOutputter outputter = new XMLOutputter(ft);
			String s = outputter.outputString(page);
			if (outputFile != null) {
				File f = new File(outputFile);
				FileWriter fw = new FileWriter(f);
				fw.write(s);
				fw.flush();
				fw.close();
			} else
				System.out.println(s);
			long now2 = System.currentTimeMillis();
			log.info("Total execution time "+(now2-now)/1000+"s");
		} catch (IOException e) {
			e.printStackTrace();
//		} catch (SaxonApiException e) {
//			e.printStackTrace();
		}
		System.exit(0);
	}

	/**
	 * @param errors
	 * @return
	 */
	private static Collection<Content> getErrorsCollection(LinkedList<Integer> errors) {
		ArrayList<Content> al = new ArrayList<>(errors.size());
		for (Integer error : errors) {
			al.add(new Text(" "));
			Element link = new Element("a", XHTMLNS);
			link.setAttribute("href", "#"+error);
			link.setText(""+error);
			al.add(link);
		}
		return al;
	}
	
	/**
	 * @param i
	 * @param args
	 * @param dtsProps 
	 * @return
	 */
	private static int readProperties(int i, String[] args, Properties dtsProps) {
		if (args.length == i)
			return i;
		StringBuffer sb = new StringBuffer();
		int j=i;
		for (;j<args.length;++j) {
			String argument = args[j].trim();
			if (argument.startsWith("-"))
				break;
			sb.append(argument);
		}		
		i = j-1;
		if (sb.length() == 0)
			usage();
		StringTokenIterator sti = new StringTokenIterator(sb.toString(), ",");
		Item itm = null;
		while ((itm = sti.next()) != null) {
			// find equal sign
			String token = itm.getStringValue();
			int idx = token.indexOf("=");
			if (idx < 0)
				usage();
			String propName = token.substring(0, idx);
			String propValue = token.substring(idx+1);
			dtsProps.setProperty(propName, propValue);
		}
		return i;
	}
}
