package com.ihr.xbrl.sample;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.math.BigDecimal;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import javax.xml.namespace.QName;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.jdom.Attribute;
import org.jdom.Content;
import org.jdom.DocType;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.Text;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;
import org.w3c.dom.Attr;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xmlunit.builder.DiffBuilder;
import org.xmlunit.diff.Comparison;
import org.xmlunit.diff.Comparison.Detail;
import org.xmlunit.diff.ComparisonController;
import org.xmlunit.diff.ComparisonResult;
import org.xmlunit.diff.ComparisonType;
import org.xmlunit.diff.DefaultNodeMatcher;
import org.xmlunit.diff.Diff;
import org.xmlunit.diff.Difference;
import org.xmlunit.diff.DifferenceEvaluator;
import org.xmlunit.diff.DifferenceEvaluators;
import org.xmlunit.diff.ElementSelector;
import org.xmlunit.diff.ElementSelectors;
import org.xmlunit.util.Nodes;
import org.xmlunit.util.Predicate;
import org.xmlunit.xpath.JAXPXPathEngine;
import org.xmlunit.xpath.XPathEngine;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRL;
import com.ihr.xbrl.om.XMLFragment;
import com.ihr.xbrl.om.exLinks.FootnoteLinkbase;
import com.ihr.xbrl.om.exLinks.XBRLArcroleRef;
import com.ihr.xbrl.om.exLinks.XBRLLocator;
import com.ihr.xbrl.om.exLinks.XBRLRelationship;
import com.ihr.xbrl.om.exLinks.XBRLRoleRef;
import com.ihr.xbrl.om.exLinks.XLinkElement;
import com.ihr.xbrl.om.instance.XBRLContext;
import com.ihr.xbrl.om.instance.XBRLFactItem;
import com.ihr.xbrl.om.instance.XBRLFactNumeric;
import com.ihr.xbrl.om.instance.XBRLInstance;
import com.ihr.xbrl.om.instance.XBRLSchemaRef;
import com.ihr.xbrl.om.instance.XBRLUnit;
import com.ihr.xbrl.om.ixbrl.IXBRLProcessorBase;
import com.ihr.xbrl.om.ixbrl.IXBRLProcessorBase.WrapStreamSource;
import com.ihr.xbrl.om.ixbrl.IXBRLTransformerFactory;
import com.ihr.xbrl.om.taxonomy.XMLElementDefinition;
import com.ihr.xbrl.util.TextTools;

import net.sf.saxon.expr.StringTokenIterator;
import net.sf.saxon.om.Item;

/**
 * @author Ignacio
 */
public class LaunchIXBRLv11ConformanceSuiteXHTML {
	
	static String confSuiteDirectory = "inline-1.1-conformance";
	final static String outDirectory = "xbrl-temp-xslt-1.1";

	static int testNumber = 0;
	static int countSkiped = 0;
	static int startAt = 0;
	static int endsAt = Integer.MAX_VALUE;
	static int countError = 0;
	static int countOk = 0;
	static Object[][] ignoreErrors = new Object[][] { };
	static ArrayList<Integer> testsToRun = new ArrayList<Integer>();
	static LinkedList<Integer> errors = new LinkedList<>();
	static boolean bShowMemoryUsage = false;
	

	private static class IXBRL11TestCase {
	
		File fd;
		StringBuilder log;
		Path confPredictedPath = null;
		Path outPath = null;
		boolean isError;
		
		IXBRL11TestCase(File fd, StringBuilder log) {
			this.fd = fd;
			this.log = log;
		}
		
		public boolean runTest() {
			boolean bIsOK = false;
			try {
				if (bShowMemoryUsage) {
					long total = Runtime.getRuntime().totalMemory();
					long free = Runtime.getRuntime().freeMemory();
					System.out.println("Test number: "+(testNumber+countSkiped)+" "+fd.getName()+" Memory at 1: T="+total+" F="+free+" U="+(total-free));
				}

				confPredictedPath = Paths.get(URI.create(fd.toURI().toString().replace(".html", ".predictedOutput")));
				isError = fd.getName().startsWith("FAIL");
				IXBRLTransformerFactory factory = new IXBRLTransformerFactory();
				factory.setAttribute(IXBRLTransformerFactory.SPECIFICATION_VERSION_ATTRIBUTE, "1.1");
				Transformer tr = factory.newTransformer();
				FileInputStream fr = new FileInputStream(fd);
				Source ss = new StreamSource(fr);
				ss.setSystemId(fd.toURI().toString());
				outPath = Paths.get(URI.create(fd.toURI().toString().replace(confSuiteDirectory, outDirectory).replace(".html", ".xbrl")));
				outPath.toFile().getParentFile().mkdirs();
				FileOutputStream fout = new FileOutputStream(outPath.toFile());
				Result rr = new StreamResult(fout);
				try {
					tr.transform(ss, rr);
					fout.close();
					fout = null;
					fr.close();
					fr = null;
					if (!isError) {
						// only ok if the result is semantically equals to conformance .predictedOutput file
						StringBuffer sbLog = new StringBuffer();
						boolean bEquals = /*true; */ comparePathInstances(confPredictedPath, outPath, sbLog); // work in progress
						if (bEquals) {
							bIsOK = true;
							log.append("OK - Comparison OK");
						} else {
							log.append("ERROR - Not equals to predicted output: "+sbLog.toString());
						}
					} else
						log.append("ERROR - Error not detected");
				} catch (TransformerException e) {
					if (isError) {
						bIsOK = true;
						log.append(e.getMessage());
					} else
						log.append("ERROR - Should be OK: "+e.getMessage());
					if (fout != null)
						fout.close();
					if (fr != null)
						fr.close();
				}
				if (bShowMemoryUsage) {
					System.gc();
					long total = Runtime.getRuntime().totalMemory();
					long free = Runtime.getRuntime().freeMemory();
					System.out.println("Test number: "+(testNumber+countSkiped)+" "+fd.getName()+" "+"Memory at 2: T="+total+" F="+free+" U="+(total-free));
				}
			} catch (Throwable e) {
				// possibly caused by dts.clear() ____
				e.printStackTrace();
				System.exit(-1);
			}
			return bIsOK;
		}
		
		/**
		 * @return the confPredictedPath
		 */
		public Path getConfPredictedPath() {
			return confPredictedPath;
		}
		
		/**
		 * @return the outPath
		 */
		public Path getOutPath() {
			return outPath;
		}
		
		/**
		 * @return the isError
		 */
		public boolean isError() {
			return isError;
		}
	}

	private static final String XHTMLNS = "";
	
	private static final Logger log = Logger.getLogger(LaunchIXBRLv11ConformanceSuiteXHTML.class);
	
	public static void usage() {
		System.out.println("Usage: LaunchIXBRLv1ConformanceSuiteXHTML [-m] [-r {starting test case number} {end test case number}] [-n {test case number}] [-o {Output file}] [-d {DEBUG_LEVEL}]\n" +
				"DEBUG LEVEL can be OFF,TRACE,...");
		System.exit(-1);
	}
	
	@SuppressWarnings("unused")
	private static boolean debug = false;
	 
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		try {
			String outputFile = null;
			String debugLevel = "OFF";
			Properties dtsProps = new Properties();
			
			// parse arguments
			for (int i=0; i < args.length; ++i) {
				if ("-f".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					confSuiteDirectory = args[i];
				} else if ("-m".equals(args[i])) {
					bShowMemoryUsage = true;
				} else if ("-o".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					outputFile = args[i];
				} else if ("-r".equals(args[i])) {
					if (args.length < i+1)
						usage();
					++i;
					startAt = Integer.parseInt(args[i])-1;
					++i;
					endsAt = Integer.parseInt(args[i])-1;
				} else if ("-p".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					i = readProperties(i, args, dtsProps);
					continue;
				} else if ("-d".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					debugLevel = args[i];					
				} else if ("-n".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					testsToRun.add(Integer.parseInt(args[i])-1);
				}
			}
			
			// configure log level
			Logger log1 = Logger.getLogger("com.ihr.xbrl.om");
			log1.setLevel(Level.toLevel(debugLevel));
			log.setLevel(Level.toLevel(debugLevel));
			dtsProps.setProperty(DTSContainer.LOG_LEVEL, debugLevel);
			
			final String sTitle = "Compliance report with the Inline XBRL Processor 1.1 Conformance Suite";
			
			Document page = new Document();
			page.setDocType(new DocType("html"));
			Element html = new Element("html", XHTMLNS);
			page.addContent(html);
			Element head = new Element("head", XHTMLNS);
			html.addContent(head);
			Element title = new Element("title", XHTMLNS);
			title.setText(sTitle);
			head.addContent(title);
			Element style = new Element("style", XHTMLNS);
			head.addContent(style);
			Attribute styleType = new Attribute("type", "text/css");
			style.setAttribute(styleType);
			style.setText("td.testOK { background: lightgreen; }\ntd.testFail { background: red; }");			
			Element body = new Element("body", XHTMLNS);
			html.addContent(body);
			Element pageH2 = new Element("h2", XHTMLNS);
			pageH2.setText(sTitle);
			body.addContent(pageH2);
			Date d = new Date(System.currentTimeMillis());
			SimpleDateFormat dtf = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSZ");
			Element header3 = new Element("h3", XHTMLNS);
			header3.setText("Date executed: "+dtf.format(d));
			body.addContent(header3);
			Element outTable = new Element("table", XHTMLNS);
			Attribute border1 = new Attribute("border", "0");
			outTable.setAttribute(border1);
			body.addContent(outTable);
			
			long now = System.currentTimeMillis();

			File directory = new File(confSuiteDirectory);
			processDirectory(directory, outTable);

			Element table03 = new Element("table", XHTMLNS);
			body.addContent(table03);
			Element tr05 = new Element("tr", XHTMLNS);
			table03.addContent(tr05);
			Element td07 = new Element("td", XHTMLNS);
			tr05.addContent(td07);
			td07.setText("Not passed:");
			Element td08 = new Element("td", XHTMLNS);
			tr05.addContent(td08);
			Element linkDownPair = new Element("a", XHTMLNS);
			linkDownPair.setAttribute("name", "down");
			td08.addContent(linkDownPair);
			td08.addContent(new Text((new Integer(countError)).toString()));
			td08.addContent(getErrorsCollection(errors));
			Element tr06 = new Element("tr", XHTMLNS);
			table03.addContent(tr06);
			Element td09 = new Element("td", XHTMLNS);
			tr06.addContent(td09);
			td09.setText("Passed:");
			Element td10 = new Element("td", XHTMLNS);
			tr06.addContent(td10);
			td10.setText((new Integer(countOk)).toString());
			Element tr07 = new Element("tr", XHTMLNS);
			table03.addContent(tr07);
			Element td11 = new Element("td", XHTMLNS);
			tr07.addContent(td11);
			td11.setText("Not executed:");
			Element td12 = new Element("td", XHTMLNS);
			tr07.addContent(td12);
			td12.setText((new Integer(countSkiped)).toString());
			// produce the output to stdout or file
			Format ft = Format.getPrettyFormat();
			ft.setEncoding("ISO-8859-1");
			XMLOutputter outputter = new XMLOutputter(ft);
			String s = outputter.outputString(page);
			if (outputFile != null) {
				File f = new File(outputFile);
				FileWriter fw = new FileWriter(f);
				fw.write(s);
				fw.flush();
				fw.close();
			} else
				System.out.println(s);
			long now2 = System.currentTimeMillis();
			log.info("Total execution time "+(now2-now)/1000+"s");
		} catch (IOException e) {
			e.printStackTrace();
		}
		System.exit(0);
	}

	/**
	 * @param errors
	 * @return
	 */
	private static Collection<Content> getErrorsCollection(LinkedList<Integer> errors) {
		ArrayList<Content> al = new ArrayList<>(errors.size());
		for (Integer error : errors) {
			al.add(new Text(" "));
			Element link = new Element("a", XHTMLNS);
			link.setAttribute("href", "#"+error);
			link.setText(""+error);
			al.add(link);
		}
		return al;
	}
	
	/**
	 * @param i
	 * @param args
	 * @param dtsProps 
	 * @return
	 */
	private static int readProperties(int i, String[] args, Properties dtsProps) {
		if (args.length == i)
			return i;
		StringBuffer sb = new StringBuffer();
		int j=i;
		for (;j<args.length;++j) {
			String argument = args[j].trim();
			if (argument.startsWith("-"))
				break;
			sb.append(argument);
		}		
		i = j-1;
		if (sb.length() == 0)
			usage();
		StringTokenIterator sti = new StringTokenIterator(sb.toString(), ",");
		Item itm = null;
		while ((itm = sti.next()) != null) {
			// find equal sign
			String token = itm.getStringValue();
			int idx = token.indexOf("=");
			if (idx < 0)
				usage();
			String propName = token.substring(0, idx);
			String propValue = token.substring(idx+1);
			dtsProps.setProperty(propName, propValue);
		}
		return i;
	}

	private static void processFile(File fd, Element outTable) {
		String description = fd.getAbsolutePath();
		
		Element tr01 = new Element("tr", XHTMLNS);
		outTable.addContent(tr01);
		Element td01 = new Element("td", XHTMLNS);
		tr01.addContent(td01);
		td01.setAttribute("style", "background: orange");
		Element link = new Element("a", XHTMLNS);
		link.setText(fd.getName());
		td01.addContent(link);
		
		Element tr02 = new Element("tr", XHTMLNS);
		outTable.addContent(tr02);
		Element td02 = new Element("td", XHTMLNS);
		tr02.addContent(td02);
		Element table02 = new Element("table", XHTMLNS);
		td02.addContent(table02);
		Attribute at02 = new Attribute("border", "1");
		table02.setAttribute(at02);
		Attribute at03 = new Attribute("width", "100%");
		table02.setAttribute(at03);
		Element tr03 = new Element("tr", XHTMLNS);
		table02.addContent(tr03);

		Element td03 = new Element("td", XHTMLNS);
		tr03.addContent(td03);
		Element href = new Element("a", XHTMLNS);
		href.setAttribute("name", ""+(testNumber+countSkiped+1));
		td03.addContent(href);
		td03.addContent(new Text("N: "+(testNumber+countSkiped+1)));
		Element hrefDown = new Element("a", XHTMLNS);
		hrefDown.setAttribute("href", "#down");
		hrefDown.setText("(dn)");
		td03.addContent(hrefDown);
		
		Element td04 = new Element("td", XHTMLNS);
		td04.addContent("Source: ");
		Element href2 = new Element("a", XHTMLNS);
		href2.setAttribute("href", "file:///"+description);
		href2.addContent(description);
		td04.addContent(href2);
		tr03.addContent(td04);
		Element tr04 = new Element("tr", XHTMLNS);
		table02.addContent(tr04);

		if (((testNumber+countSkiped) < startAt || (testNumber+countSkiped) > endsAt) || (testsToRun.size() > 0 && !testsToRun.contains(testNumber+countSkiped))) {
			Element td05 = new Element("td", XHTMLNS);
			td05.setText("Execution excluded");
			tr04.addContent(td05);
			countSkiped++;
		} else {
			testNumber++;
			
			StringBuilder log = new StringBuilder();
			IXBRL11TestCase tc = new IXBRL11TestCase(fd, log);
			boolean pass = tc.runTest();
			Element td05 = new Element("td", XHTMLNS);
			String message = "unknown";
			if (!pass) {
				message = "NOT PASS.";
				countError++;
				errors.add(testNumber+countSkiped);
			} else {
				message = "OK.";
				countOk++;
			}
			Attribute at06 = new Attribute("class", pass ? "testOK" : "testFail");
			td05.setAttribute(at06);
			td05.setText(message);
			tr04.addContent(td05);
			Element td06 = new Element("td", XHTMLNS);
			td06.setText(log.toString());
			tr04.addContent(td06);
			
			// add extra href on the header pointing to the generated output and
			// the reference 
			
			if (!tc.isError()) {
				Path predicted = tc.getConfPredictedPath();
				if (predicted != null) {
					Element br = new Element("br", XHTMLNS);
					td04.addContent(br);
					td04.addContent("Reference: ");
					Element href3 = new Element("a", XHTMLNS);
					href3.setAttribute("href", "file:///"+predicted.toString());
					href3.addContent(predicted.toString());
					td04.addContent(href3);
				}
	
				Path outPath = tc.getOutPath();
				if (outPath != null) {
					Element br = new Element("br", XHTMLNS);
					td04.addContent(br);
					td04.addContent("Generated: ");
					Element href4 = new Element("a", XHTMLNS);
					href4.setAttribute("href", "file:///"+outPath.toString());
					href4.addContent(outPath.toString());
					td04.addContent(href4);
				}
			}
		
		}
	}

	static void processDirectory(File directory, Element outTable) {
		File[] dirFiles = directory.listFiles();
		if (dirFiles != null)
			for (final File fd : dirFiles) {
				if (fd.isDirectory())
					processDirectory(fd, outTable);
				else {
					if ((fd.getName().startsWith("FAIL-") || fd.getName().startsWith("PASS-"))&& fd.getName().endsWith(".html"))
						processFile(fd, outTable);
				}
			}
	}

	private static boolean comparePathInstances(Path confPredictedPath, Path generatedPath, StringBuffer sbLog) throws MalformedURLException {
		URL url1 = confPredictedPath.toUri().toURL();		
		URL url2 = generatedPath.toUri().toURL();
		try {
			org.xmlunit.diff.Diff d1 = makeXBRLDiff(url1, url2);
			if (d1.hasDifferences()) {
				sbLog.append(d1.toString());
				return false;
			}
		} catch (org.xmlunit.XMLUnitException | IOException e) {
			sbLog.append("Unexpected "+e.getClass().getSimpleName()+": "+e.getMessage());
			return false;
		}
		return true;
	}

//	private static QName getNodeName(Node el) {
//		String uri = el.getNamespaceURI();
//		return new QName(uri == null ? "" : uri,el.getLocalName());
//	}
	
//	/**
//	 * Compare xlink:href attributes on both nodes for equality after resolution against xml:base
//	 * 
//	 * @param n1
//	 * @param n2
//	 * @param n2URL 
//	 * @param n1URL 
//	 * @return
//	 * @throws URISyntaxException 
//	 */
//	private static boolean absoluteHrefAreEqual(Node n1, Node n2, URL n1URL, URL n2URL) {
//		try {
//			String baseOnN1 = n1.getBaseURI();
//			if (baseOnN1 == null)
//				baseOnN1 = n1URL.toURI().toString();
//			String baseOnN2 = n2.getBaseURI();
//			if (baseOnN2 == null)
//				baseOnN2 = n2URL.toURI().toString();
//			String hrefOnN1 = getHRef(n1);
//			String hrefOnN2 = getHRef(n2);
//			
//			URI absUri1 = URI.create(baseOnN1).resolve(hrefOnN1);
//			URI absUri2 = URI.create(baseOnN2).resolve(hrefOnN2);
//			
//			return absUri1.equals(absUri2);
//		} catch (URISyntaxException e) {
//			return false;
//		}
//	}
	
//	private static String getHRef(Node n1) {
//		return n1.getAttributes().getNamedItemNS(XBRL.XLINK_NS, "href").getTextContent();
//	}
	
//	private static boolean isFactItem(Node n1) {
//		return n1.getAttributes().getNamedItem("contextRef") != null;
//	}
//	
//	private static boolean isNumericFactItem(Node n1) {
//		return n1.getAttributes().getNamedItem("unitRef") != null;
//	}
	
	private static Predicate<org.w3c.dom.Element> isNumericFactItem = new Predicate<org.w3c.dom.Element>() {
		
		@Override
		public boolean test(org.w3c.dom.Element toTest) {
			return toTest.hasAttribute("unitRef");
		}
	};
	
	private static Predicate<org.w3c.dom.Element> isFactItem = new Predicate<org.w3c.dom.Element>() {
		
		@Override
		public boolean test(org.w3c.dom.Element toTest) {
			return toTest.hasAttribute("contextRef");
		}
	};
	
	private static Predicate<org.w3c.dom.Element> isFactTuple = new Predicate<org.w3c.dom.Element>() {
		
		@Override
		public boolean test(org.w3c.dom.Element toTest) {
			return 
				!toTest.getNamespaceURI().equals(XBRL.XBRL_instanceNS) &&
				!toTest.getNamespaceURI().equals(XBRL.XBRL_linkbaseNS) &&
				parentIsXBRLorTuple(toTest) &&
				!toTest.hasAttribute("contextRef");
		}

		private boolean parentIsXBRLorTuple(org.w3c.dom.Element toTest) {
			Node pNode = toTest.getParentNode();
			if (pNode instanceof org.w3c.dom.Element) {
				if (Nodes.getQName(pNode).equals(XBRLInstance.docNodeName) )
					return true;
				return (isFactTuple.test((org.w3c.dom.Element) pNode));
			}
			return false;
		}
	};
	
//    /**
//     * Elements with the same local name (and namespace URI - if any)
//     * and nested text (if any) can be compared.
//     */
//    private static final ElementSelector byTrimText = new ElementSelector() {
//		@Override
//		public boolean canBeCompared(org.w3c.dom.Element controlElement, org.w3c.dom.Element testElement) {
//            String v1 = Nodes.getMergedNestedText(controlElement);
//            if (!v1.isEmpty())
//            	v1 = v1.trim();
//            String v2 = Nodes.getMergedNestedText(testElement);
//            if (!v2.isEmpty())
//            	v2 = v2.trim();
//            return TextTools.equalsNullSafe(v1,v2);
//		}
//    };
    
    /**
     * Elements with the same local name (and namespace URI - if any)
     * and nested text (if any) can be compared.
     */
    private static final ElementSelector byNumericValue = new ElementSelector() {
		@Override
		public boolean canBeCompared(org.w3c.dom.Element controlElement, org.w3c.dom.Element testElement) {
            String v1 = Nodes.getMergedNestedText(controlElement);
            if (!v1.isEmpty())
            	v1 = v1.trim();
            String v2 = Nodes.getMergedNestedText(testElement);
            if (!v2.isEmpty())
            	v2 = v2.trim();
            try {
            	if (v1.isEmpty())
            		return v2.isEmpty();
            	else if (v2.isEmpty())
            		return false;
	            BigDecimal bd1 = new BigDecimal(v1);
	            BigDecimal bd2 = new BigDecimal(v2);
	            return bd1.compareTo(bd2) == 0;
            } catch (NumberFormatException e) {
            	return false;
            }
		}
    };
    
    private static final org.w3c.dom.Element findNodeById(org.w3c.dom.Element el, String id) {
	    	org.w3c.dom.Element parentForLoop = upTo(el,XBRLInstance.docNodeName);
	    	NodeList nl = parentForLoop.getChildNodes();
	    	int size = nl.getLength();
	    	for (int i=0; i<size; ++i) {
	    		Node ni = nl.item(i);
	    		if (ni instanceof org.w3c.dom.Element) {
	    			org.w3c.dom.Element elni = (org.w3c.dom.Element)ni;
	    			String idVal = elni.getAttribute("id");
	    			if (id.equals(idVal))
	    				return elni;
	    		}
	    	}
	    	return null;
    }
    
    /**
	 * @param el
	 * @param docnodename
	 * @return
	 */
	private static org.w3c.dom.Element upTo(org.w3c.dom.Element el, QName stopNodeName) {
		Node parent = el.getParentNode();
		if (parent instanceof org.w3c.dom.Element) {
			QName nodeName = new QName(parent.getNamespaceURI(),parent.getLocalName(),parent.getPrefix());
			if (stopNodeName.equals(nodeName))
				return (org.w3c.dom.Element)parent;
			return upTo((org.w3c.dom.Element)parent, stopNodeName);
		}
		return null;
	}

	private static final ElementSelector bySEqualChildren = new ElementSelector() {
	    	public boolean canBeCompared(org.w3c.dom.Element controlElement, org.w3c.dom.Element testElement) {
	    		return ElementSelectors.and
	    		(
	       			ElementSelectors.byNameAndText,
	    			ElementSelectors.byXPath("*", bySEqualChildren2)
	    		)
	    		.canBeCompared(controlElement, testElement);
	    	}
    };
    
    private static final ElementSelector byAttributesExceptNamespaceDeclarations = new ElementSelector() {

		@Override
		public boolean canBeCompared(org.w3c.dom.Element controlElement, org.w3c.dom.Element testElement) {
            Map<QName, String> cAttrs = removeNSDeclarations(Nodes.getAttributes(controlElement));
            Map<QName, String> tAttrs = removeNSDeclarations(Nodes.getAttributes(testElement));
            if (cAttrs.size() != tAttrs.size()) {
                return false;
            }
            return mapsEqualForKeys(cAttrs, tAttrs, cAttrs.keySet());
		}

    };
    private static final ElementSelector byTargetOrDefault = new ElementSelector() {

		@Override
		public boolean canBeCompared(org.w3c.dom.Element controlElement, org.w3c.dom.Element testElement) {
			String defaultStr = controlElement.getAttribute("default");
			if ("yes".equals(defaultStr)) {
				return "yes".equals(testElement.getAttribute("default"));
			}
			String controlTarget = controlElement.getAttribute("target");
			String testTarget = testElement.getAttribute("target");
			return TextTools.equalsNullSafe(controlTarget, testTarget);
		}

    };
    private static boolean mapsEqualForKeys(Map<QName, String> control,
            Map<QName, String> test,
            Iterable<QName> keys) {
		for (QName q : keys) {
			if (!TextTools.equalsNullSafe(control.get(q), test.get(q))) {
				return false;
			}
		}
		return true;
	}
	private static Map<QName, String> removeNSDeclarations(Map<QName, String> attributes) {
		LinkedList<QName> toRemove = new LinkedList<>();
		for (QName key : attributes.keySet()) {
			if (key.getNamespaceURI().equals("http://www.w3.org/2000/xmlns/"))
				toRemove.add(key);
		}
		for (QName key : toRemove)
			attributes.remove(key);
		return attributes;
	}
    		
    private static final ElementSelector bySEqualChildren2 = new ElementSelector() {
	    	public boolean canBeCompared(org.w3c.dom.Element controlElement, org.w3c.dom.Element testElement) {
	    		return ElementSelectors.and
	    		(
	       			byAttributesExceptNamespaceDeclarations,
	       			ElementSelectors.byNameAndText,
	    			ElementSelectors.byXPath("*", bySEqualChildren2)
	    		)
	    		.canBeCompared(controlElement, testElement);
	    	}
    };
    		
    /**
     * Elements with the same local name (and namespace URI - if any)
     * and nested text (if any) can be compared.
     */
    private static final ElementSelector byCEqual = new ElementSelector() {

		@Override
		public boolean canBeCompared(org.w3c.dom.Element controlElement, org.w3c.dom.Element testElement) {
			org.w3c.dom.Element controlContext = getContext(controlElement);
			org.w3c.dom.Element testContext = getContext(testElement);
			return bySEqualChildren.canBeCompared(controlContext, testContext);
		}
		
		private org.w3c.dom.Element getContext(org.w3c.dom.Element el) {
			String contextRef = el.getAttribute("contextRef").trim();
			return findNodeById(el, contextRef);
		}
    };
    
    /**
     * Elements with the same local name (and namespace URI - if any)
     * and nested text (if any) can be compared.
     */
    private static final ElementSelector byNil = new ElementSelector() {

		@Override
		public boolean canBeCompared(org.w3c.dom.Element controlElement, org.w3c.dom.Element testElement) {
			String controlNil = controlElement.getAttributeNS(XBRL.XSI_NS, "nil");
			boolean controlIsNil = (controlNil == null || controlNil.isEmpty()) ? false : "1".equals(controlNil.trim()) || "true".equals(controlNil.trim());
			String testNil = testElement.getAttributeNS(XBRL.XSI_NS, "nil");
			boolean testIsNil = (testNil == null || testNil.isEmpty()) ? false : "1".equals(testNil.trim()) || "true".equals(testNil.trim());
			return controlIsNil == testIsNil;
		}
    };
    
    /**
     * Elements with the same local name (and namespace URI - if any)
     * and nested text (if any) can be compared.
     */
    private static final ElementSelector byUEqual = new ElementSelector() {

		@Override
		public boolean canBeCompared(org.w3c.dom.Element controlElement, org.w3c.dom.Element testElement) {
			org.w3c.dom.Element controlUnit = getUnit(controlElement);
			org.w3c.dom.Element testUnit = getUnit(testElement);
			return bySEqualChildren.canBeCompared(controlUnit, testUnit);
		}
		
		private org.w3c.dom.Element getUnit(org.w3c.dom.Element el) {
			String unitRef = el.getAttribute("unitRef").trim();
			return findNodeById(el, unitRef);
		}
    };
    
    /**
     * Elements with the same local name (and namespace URI - if any)
     * and nested text (if any) can be compared.
     */
    private static final ElementSelector byLangEqual = new ElementSelector() {

		@Override
		public boolean canBeCompared(org.w3c.dom.Element controlElement, org.w3c.dom.Element testElement) {
			String controlLang = getLang(controlElement);
			String testLang = getLang(testElement);
			return TextTools.equalsNullSafe(controlLang, testLang);
		}
		
		private String getLang(org.w3c.dom.Element el) {
			String lang = el.getAttributeNS(XBRL.XML_baseNS, "lang");
			if (lang == null) {
				org.w3c.dom.Node parent = el.getParentNode();
				if (parent instanceof org.w3c.dom.Element)
					return getLang((org.w3c.dom.Element)parent);
			}
			return null;
		}
    };
    
	/**
	 * 
	 * @author Ignacio
	 *
	 */
	private static class CollectedFootnoteInfo {
		String footnoteLinkRole = null;
		String arcArcrole = null;
		org.w3c.dom.Element target = null;
		
		/* (non-Javadoc)
		 * @see java.lang.Object#hashCode()
		 */
		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((arcArcrole == null) ? 0 : arcArcrole.hashCode());
			result = prime * result + ((footnoteLinkRole == null) ? 0 : footnoteLinkRole.hashCode());
			return result;
		}
		
		/* (non-Javadoc)
		 * @see java.lang.Object#equals(java.lang.Object)
		 */
		@Override
		public boolean equals(Object obj) {
			if (this == obj) {
				return true;
			}
			if (obj == null) {
				return false;
			}
			if (!(obj instanceof CollectedFootnoteInfo)) {
				return false;
			}
			CollectedFootnoteInfo other = (CollectedFootnoteInfo) obj;
			if (arcArcrole == null) {
				if (other.arcArcrole != null) {
					return false;
				}
			} else if (!arcArcrole.equals(other.arcArcrole)) {
				return false;
			}
			if (footnoteLinkRole == null) {
				if (other.footnoteLinkRole != null) {
					return false;
				}
			} else if (!footnoteLinkRole.equals(other.footnoteLinkRole)) {
				return false;
			}
			if (target == null) {
				if (other.target != null) {
					return false;
				}
			} else if (!ElementSelectors.and(
					ElementSelectors.byNameAndText,
					ElementSelectors.byNameAndAttributes(XLinkElement.role),
					byLangEqual
					).canBeCompared(target,other.target)) {
				return false;
			}
			return true;
		} 
		
		
	}
	
   /**
     * Elements with the same local name (and namespace URI - if any)
     * and nested text (if any) can be compared.
     */
    private static final ElementSelector bySameFootnotes = new ElementSelector() {
    	

		@Override
		public boolean canBeCompared(org.w3c.dom.Element controlElement, org.w3c.dom.Element testElement) {
			Set<CollectedFootnoteInfo> controlFootnoteInfo = getFootnoteInfo(controlElement);
			Set<CollectedFootnoteInfo> testFootnoteInfo = getFootnoteInfo(testElement);
			return controlFootnoteInfo.equals(testFootnoteInfo);
		}
		
		private Set<CollectedFootnoteInfo> getFootnoteInfo(org.w3c.dom.Element el) {
			// collect footnote information in a class and add it to a set.
			// only equal sets can be compared
			HashSet<CollectedFootnoteInfo> retValue = new HashSet<>();
			String id = el.getAttribute("id").trim();
			if (id != null) {
				// we have no XBRL API available. let's use XPath to locate footnote information to be collected
				// we are interested only in the semantics
				
				// find locators with an href of the form #id in the same file
				HashMap<String,String> map = new HashMap<>();
				map.put("xlink", XBRL.XLINK_NS);
				final XPathEngine engine = new JAXPXPathEngine();
	            engine.setNamespaceContext(map);
				Iterable<Node> locators = engine.selectNodes("//*[@xlink:href='#"+id+"']", el); // sadly "eq" operator cannot be used on this XPath compiler
				for (Node locator : locators) {
					Node extendedLink = locator.getParentNode();
					String footnoteLinkRole = extendedLink.getAttributes().getNamedItemNS(XBRL.XLINK_NS, "role").getNodeValue();
					// no information from the locator will be collected, there are no semantics on locator nodes
					String labelFrom = locator.getAttributes().getNamedItemNS(XBRL.XLINK_NS, "label").getNodeValue();
					Iterable<Node> arcsFrom = engine.selectNodes("*[@xlink:from='"+labelFrom+"']", extendedLink);
					for (Node arc : arcsFrom) {
						String toLabel = arc.getAttributes().getNamedItemNS(XBRL.XLINK_NS, "to").getNodeValue();
						Iterable<Node> arcsTo = engine.selectNodes("*[@xlink:to='"+toLabel+"']", extendedLink);
						for (Node target : arcsTo) {
							CollectedFootnoteInfo ci = new CollectedFootnoteInfo();
							ci.footnoteLinkRole = footnoteLinkRole;
							ci.arcArcrole = arc.getAttributes().getNamedItemNS(XBRL.XLINK_NS, "arcrole").getNodeValue();
							ci.target = (org.w3c.dom.Element)target;
							retValue.add(ci);
						}
					}
				}
			}			
			return retValue;
		}
    };
    
//    private static class TuplesBeingCompared {
//
//    		org.w3c.dom.Element controlElement;
//    		org.w3c.dom.Element testElement;
//
//		/**
//		 * @param controlElement
//		 * @param testElement
//		 */
//		public TuplesBeingCompared(org.w3c.dom.Element controlElement, org.w3c.dom.Element testElement) {
//			super();
//			this.controlElement = controlElement;
//			this.testElement = testElement;
//		}
//
//		/* (non-Javadoc)
//		 * @see java.lang.Object#equals(java.lang.Object)
//		 */
//		@Override
//		public boolean equals(Object obj) {
//			if (this == obj) {
//				return true;
//			}
//			if (obj == null) {
//				return false;
//			}
//			if (!(obj instanceof TuplesBeingCompared)) {
//				return false;
//			}
//			TuplesBeingCompared other = (TuplesBeingCompared) obj;
//			if (controlElement == null) {
//				if (other.controlElement != null) {
//					return false;
//				}
//			} else if (!controlElement.equals(other.controlElement)) {
//				return false;
//			}
//			if (testElement == null) {
//				if (other.testElement != null) {
//					return false;
//				}
//			} else if (!testElement.equals(other.testElement)) {
//				return false;
//			}
//			return true;
//		}
//    }
    
//    private static final Stack<TuplesBeingCompared> tuplesBeingChecked = new Stack<>();
    
    /**
     * Elements with the same local name (and namespace URI - if any)
     * and nested text (if any) can be compared.
     */
    private static final ElementSelector byPEqual = new ElementSelector() {

		@Override
		public boolean canBeCompared(org.w3c.dom.Element controlElement, org.w3c.dom.Element testElement) {
			Node controlPNode = controlElement.getParentNode();
			Node testPNode = testElement.getParentNode();
			if (controlPNode == null)
				return testPNode == null;
			else if (testPNode == null)
				return false;
			if (controlPNode instanceof org.w3c.dom.Element && testPNode instanceof org.w3c.dom.Element) {
				return ElementSelectors.and(
					ElementSelectors.byName,
					byPEqual).canBeCompared((org.w3c.dom.Element)controlPNode, (org.w3c.dom.Element)testPNode);
			}
			if (controlPNode instanceof org.w3c.dom.Document && testPNode instanceof org.w3c.dom.Document)
				return true;
			return false;
		}		
    };
    
    private static final ElementSelector byBaseElementSelectorChildren = new ElementSelector() {

		@Override
		public boolean canBeCompared(org.w3c.dom.Element controlElement, org.w3c.dom.Element testElement) {
//			TuplesBeingCompared tc = new TuplesBeingCompared(controlElement, testElement);
//			if (tuplesBeingChecked.contains(tc))
//				return true;
//			tuplesBeingChecked.push(tc);
//			try {
				Node controlChild = controlElement.getFirstChild();
				Node testChild = controlElement.getFirstChild();
				if (controlChild == null && testChild == null)
					return true;
				while (controlChild != null && testChild != null) {
					if (controlChild instanceof org.w3c.dom.Element &&
						testChild instanceof org.w3c.dom.Element)
						if (!base.canBeCompared((org.w3c.dom.Element)controlChild, (org.w3c.dom.Element)testChild))
							return false;
					controlChild = controlChild.getNextSibling();
					testChild = testChild.getNextSibling();
				}
				if (controlChild != null || testChild != null)
					return false;
				return true;
//			} finally {
//				tuplesBeingChecked.pop();
//			}
		}
    };
	    
    private final static QName multipleOutputQName = new QName("","MultipleOutputWrapper");
    private final static QName singleOutputQName = new QName("","output");
    
    private static ElementSelector base = ElementSelectors.conditionalBuilder()
			.whenElementIsNamed(multipleOutputQName).thenUse(ElementSelectors.byName)
			.whenElementIsNamed(singleOutputQName).thenUse(byTargetOrDefault)
			.whenElementIsNamed(XBRLInstance.docNodeName).thenUse(ElementSelectors.byName)
			.whenElementIsNamed(XBRLSchemaRef.standardSchemaRefQName).thenUse(byAttributesExceptNamespaceDeclarations)
			.whenElementIsNamed(XBRLArcroleRef.standardArcroleRefQName).thenUse(ElementSelectors.byNameAndAttributes(XBRLArcroleRef.arcroleURI))
			.whenElementIsNamed(XBRLRoleRef.standardRoleRefQName).thenUse(ElementSelectors.byNameAndAttributes(XBRLRoleRef.roleURI))
			.when(isNumericFactItem).thenUse(
				ElementSelectors.and(
					ElementSelectors.byName,
					byCEqual,
					byUEqual,
					byPEqual,
					byNumericValue,
					byNil,
					bySameFootnotes))
			.when(isFactItem).thenUse(
				ElementSelectors.and(
					ElementSelectors.byNameAndText,
					byCEqual,
					byPEqual,
					byNil,
					bySameFootnotes))
			.when(isFactTuple).thenUse(
				ElementSelectors.and(
					ElementSelectors.byName,
					byBaseElementSelectorChildren,
					byPEqual,
					byNil,
					bySameFootnotes))
			.whenElementIsNamed(XBRLContext.contextNode).thenUse(bySEqualChildren)
			.whenElementIsNamed(XBRLUnit.unitQNameNode).thenUse(bySEqualChildren)
			.whenElementIsNamed(XBRLLocator.standardLocQName).thenUse(
				ElementSelectors.byNameAndAttributes(XLinkElement.href,XLinkElement.type))
			.whenElementIsNamed(FootnoteLinkbase.standard_arc).thenUse(
					ElementSelectors.byNameAndAttributes(XLinkElement.arcrole,XLinkElement.type))
			.whenElementIsNamed(FootnoteLinkbase.lbResource).thenUse(
					ElementSelectors.and(
							ElementSelectors.byNameAndText,
							ElementSelectors.byNameAndAttributes(XLinkElement.type,XLinkElement.role)
					)
			)
			.elseUse(
				ElementSelectors.byNameAndAttributes(
					XLinkElement.from,
					XLinkElement.to,
					XLinkElement.arcrole,
					XLinkElement.role,
					XLinkElement.label,
					XMLElementDefinition.name,
					XBRLRelationship.orderQName,
					XBRLFactItem.contextRef,
					XBRLFactNumeric.unitRef,
					XBRLFactNumeric.decimals,
					XBRLFactNumeric.precision
			)).build();
    
	/**
	 * Compares two XBRL Instance files for XML "similar".
	 * This comparison is specifically for this conformance purpose.
	 * 
	 * @param control
	 * @param test
	 * @return
	 * @throws IOException 
	 */
	public static Diff makeXBRLDiff(final URL controlURL, final URL testURL) throws IOException {
		// wrap the controlURL into a Source with a fake root node
        InputStream controlIn = controlURL.openStream();
        StreamSource controlStreamSource = new StreamSource(controlIn);
        IXBRLProcessorBase.WrapStreamSource controlSource = new WrapStreamSource(controlStreamSource, multipleOutputQName);
		
        InputStream testIn = testURL.openStream();
        StreamSource testStreamSource = new StreamSource(testIn);
        IXBRLProcessorBase.WrapStreamSource testSource = new WrapStreamSource(testStreamSource, multipleOutputQName);
        
		final HashMap<String,String> nsMap = new HashMap<>();
		nsMap.put("xbrli", XBRL.XBRL_instanceNS);
		return DiffBuilder
			.compare(controlSource)
			.withTest(testSource)
			.ignoreWhitespace()
			.withNodeMatcher(new DefaultNodeMatcher(base))
			.withAttributeFilter(new Predicate<Attr>() {	
				@Override
				public boolean test(Attr toTest) {
					if (XLinkElement.href.equals(Nodes.getQName(toTest)))
						return false;
					if (XMLFragment.baseQName.equals(Nodes.getQName(toTest)))
						return false;
					if (XLinkElement.label.equals(Nodes.getQName(toTest)))
						return false;
					if (XLinkElement.from.equals(Nodes.getQName(toTest)))
						return false;
					if (XLinkElement.to.equals(Nodes.getQName(toTest)))
						return false;
					if (XMLFragment.langQName.equals(Nodes.getQName(toTest)))
						return false;
					if (XBRLFactItem.contextRef.equals(Nodes.getQName(toTest)))
						return false;
					if (XBRLFactNumeric.unitRef.equals(Nodes.getQName(toTest)))
						return false;
					if (XMLFragment.nil.equals(Nodes.getQName(toTest)))
						return false;
					return true;
				}})
			.withNodeFilter(new Predicate<Node>() {
				@Override
				public boolean test(Node toTest) {
					if (toTest.getNodeType() == Node.COMMENT_NODE)
						return false;
					if (toTest.getNodeType() == Node.ELEMENT_NODE) {
						if (XBRLContext.contextNode.equals(Nodes.getQName(toTest)))
							return false; // test case 154 contains duplicate contexts. Comparison of generated contexts shall not be required because comparison of facts are already considering CEqual condition
						if (XBRLUnit.unitQNameNode.equals(Nodes.getQName(toTest)))
							return false;
						if (FootnoteLinkbase.lbType.equals(Nodes.getQName(toTest)))
							return false;
						if (FootnoteLinkbase.lbResource.equals(Nodes.getQName(toTest)))
							return false;
						if (FootnoteLinkbase.standard_arc.equals(Nodes.getQName(toTest)))
							return false;
					}
					return true;
				}})
			.withComparisonController(
				new ComparisonController() {
					@Override
					public boolean stopDiffing(Difference difference) {
						Comparison comp = difference.getComparison();
						ComparisonType cType = comp.getType();
						if (cType == ComparisonType.CHILD_NODELIST_SEQUENCE)
							return false;
						return true;
					}
				})
			.withDifferenceEvaluator(
				DifferenceEvaluators.chain(
					DifferenceEvaluators.Default,
					new DifferenceEvaluator() {

						@Override
						public ComparisonResult evaluate(Comparison comparison, ComparisonResult outcome) {
							if (outcome == ComparisonResult.DIFFERENT) {
								switch (comparison.getType()) {
								case TEXT_VALUE:
									// try comparing numbers
									Detail controlDetail = comparison.getControlDetails();
									Detail testDetail = comparison.getTestDetails();
									String controlValue = controlDetail.getValue().toString();
									String testValue = testDetail.getValue().toString();
									try {
										BigDecimal controlBd = new BigDecimal(controlValue);
										BigDecimal testBd = new BigDecimal(testValue);
										if (controlBd.compareTo(testBd) == 0)
											outcome = ComparisonResult.SIMILAR;
									} catch (NumberFormatException e) {
									}
									break;
								case CHILD_NODELIST_SEQUENCE:
								case CHILD_NODELIST_LENGTH:
									return ComparisonResult.EQUAL; // ignore this difference because some test cases contains duplicate contexts in output and our API does not allow this to happen
								default:
									break;
								}
							}
							return outcome;
						}
					}
				))
			.checkForSimilar()
			.build();
	}
}
