package com.ihr.xbrl.sample;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Properties;
import java.util.Vector;

import javax.xml.namespace.QName;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.jdom.Attribute;
import org.jdom.Content;
import org.jdom.DocType;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.Text;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.HasCreateFormulaParameterResolver;
import com.ihr.xbrl.om.errors.XBRLRuntimeException;
import com.ihr.xbrl.om.errors.XBRLValidationException;
import com.ihr.xbrl.om.resolvers.TaxonomyPackageURIResolver;
import com.ihr.xbrl.om.resolvers.errors.TPValidationException;
import com.ihr.xbrl.util.XMLTools;

import net.sf.saxon.expr.StringTokenIterator;
import net.sf.saxon.om.AtomicSequence;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.TreeModel;
import net.sf.saxon.s9api.DocumentBuilder;
import net.sf.saxon.s9api.Processor;
import net.sf.saxon.s9api.SaxonApiException;
import net.sf.saxon.s9api.XPathCompiler;
import net.sf.saxon.s9api.XPathExecutable;
import net.sf.saxon.s9api.XPathSelector;
import net.sf.saxon.s9api.XdmItem;
import net.sf.saxon.s9api.XdmNode;
import net.sf.saxon.s9api.XdmValue;
import net.sf.saxon.trans.XPathException;

/**
 * In principle this is a copy of the LaunchFormulaConformanceSuiteXHTML class. We would review later if sharing classes or code would make sense.
 * 
 * @author Ignacio
 *
 */
public class LaunchTaxonomyPackageConformanceSuiteXHTML {
	
//	private static HashMap<Integer,Properties> specialPropertiesByTestCase = new HashMap<>();
	
//	// The conformance suite is wrong and contains errors in the http://xbrl.org/WGWD/YYYY-MM-DD/taxonomy-package/errors namespace
//	// our implementation produces errors in the http://xbrl.org/2016/taxonomy-package/errors namespace so here we will correct invalid codes
//	private static QName getCorrectedErrorCode(QName originalCode) {
//		if ("http://xbrl.org/WGWD/YYYY-MM-DD/taxonomy-package/errors".equals(originalCode.getNamespaceURI()))
//			return new QName("http://xbrl.org/2016/taxonomy-package/errors",originalCode.getLocalPart(),originalCode.getPrefix());
//		return originalCode;
//	}
	
	static {
	}
	
	
	
	/**
	 * @author Ignacio
	 *
	 */
	public static class ConformanceParametersFactory implements HasCreateFormulaParameterResolver {
		
		private XdmValue value;

		/**
		 * @param evaluate
		 */
		public ConformanceParametersFactory(XdmValue value) {
			this.value = value;
		}

		/* (non-Javadoc)
		 * @see com.ihr.xbrl.om.HasCreateFormulaParameterResolver#createAndRegisterFormulaParameterResolver(com.ihr.xbrl.om.DTSContainer)
		 */
		@Override
		public void createAndRegisterFormulaParameterResolver(DTSContainer dts) {
			ConformanceParameters ccf = new ConformanceParameters(value);			
			dts.installFormulaParameterResolver(ccf);			
			value = null;
		}
	}

	//private static final String XHTMLNS = "http://www.w3.org/1999/xhtml";	
	private static final String XHTMLNS = "";
	private static final String CONFORMANCE_NAMESPACE = "";
	
	private static final net.sf.saxon.s9api.QName errorQName = new net.sf.saxon.s9api.QName(CONFORMANCE_NAMESPACE,"c:error");
	private static final net.sf.saxon.s9api.QName assetionTestQName = new net.sf.saxon.s9api.QName(CONFORMANCE_NAMESPACE,"c:assertionTests");
	private static final net.sf.saxon.s9api.QName instanceTestQName = new net.sf.saxon.s9api.QName(CONFORMANCE_NAMESPACE,"c:instance");
//	private static final net.sf.saxon.s9api.QName tableTestQName = new net.sf.saxon.s9api.QName(CONFORMANCE_NAMESPACE,"t:table");
	
	private static final Logger log = Logger.getLogger(LaunchTaxonomyPackageConformanceSuiteXHTML.class);
	
	public static void usage() {
		System.out.println("Usage: LaunchTaxonomyPackageConformanceSuiteXHTML {-f {Conformance suite file}} [-m] [-x {exclusion pattern}] [-r {starting test case number} {end test case number}] [-n {test case number}] [-o {Output file}] [-d {DEBUG_LEVEL}]\n" +
				"DEBUG LEVEL can be OFF,TRACE,... and applies to com.ihr.xbrl.om.table");
		System.exit(-1);
	}
	
	@SuppressWarnings("unused")
	private static boolean debug = false;
	
	 		
	/**
	 * one error code should be generated
	 */
	private static int ERROR_CODE = 0x01;
		
	/**
	 * formulas that returns an instance document to compare
	 */
	private static int FORMULA = 0x02;
		
	/**
	 * Assertion test that contains number of satisfied and not satisfied tests
	 */
	private static int ASSERTION = 0x04;
	
	/**
	 * This code is used when the test case output contains both, the assertion test and one instance document
	 */
	private static int ASSERTION_PLUS_INSTANCE = FORMULA | ASSERTION;
	
	/**
	 * Valid (mostly for generic linkbases)
	 */
	private static int VALID = 0x08;
		
	/**
	 * Invalid (mostly for generic linkbases)
	 */
	private static int INVALID = 0x10;
	
	/**
	 * Table model is written to a file that must be loaded and compared with output
	 */
	private static int TABLE = 0x20;
	
	/**
	 * not specified or unknown
	 */
	private static int UNKNOWN = 0x000;
	
//	private static final String TableModelNS = "http://xbrl.org/2014/table/model";
	
	/**
	 * @param args
	 * 
	 * args[0] = xbrl.xml
	 * args[1] = pattern to exclude tests
	 */
	public static void main(String[] args) {
		try {
			LinkedList<Integer> errors = new LinkedList<>();
			String confSuiteFile = null;
			String excludePatterns = null;
			String outputFile = null;
			String debugLevel = "OFF";
			boolean bShowMemoryUsage = false;
			Properties dtsProps = new Properties();
			
			int countError = 0;
			int countOk = 0;
			int countSkiped = 0;
			int testNumber = 0;
			int startAt = 0;
			int endsAt = Integer.MAX_VALUE;
			int testType = UNKNOWN;
			
			ArrayList<Integer> testsToRun = new ArrayList<Integer>();
			
			// parse arguments
			for (int i=0; i < args.length; ++i) {
				if ("-f".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					confSuiteFile = args[i];
				} else if ("-m".equals(args[i])) {
					bShowMemoryUsage = true;
				} else if ("-x".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					excludePatterns = args[i];
				} else if ("-o".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					outputFile = args[i];
				} else if ("-r".equals(args[i])) {
					if (args.length < i+1)
						usage();
					++i;
					startAt = Integer.parseInt(args[i])-1;
					++i;
					endsAt = Integer.parseInt(args[i])-1;
				} else if ("-p".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					dtsProps.clear();
					i = readProperties(i,args,dtsProps);
					continue;
				} else if ("-d".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					debugLevel = args[i];					
				} else if ("-n".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					testsToRun.add(Integer.parseInt(args[i])-1);
				}
			}
			
			if (confSuiteFile == null)
				usage();
			
			// configure log level
			Logger log1 = Logger.getLogger("com.ihr.xbrl.om.resolvers");
			log1.setLevel(Level.toLevel(debugLevel));
			log.setLevel(Level.toLevel(debugLevel));
			
			// Load the conformance suite
			Processor xproc = XMLTools.newProcessor(false);
			XPathCompiler xpc = xproc.newXPathCompiler();
			xpc.declareNamespace("c", CONFORMANCE_NAMESPACE);
			
			DocumentBuilder db = XMLTools.newDocumentBuilder(xproc, TreeModel.TINY_TREE, null);
			XdmNode xbrl = db.build(new File(confSuiteFile));

			// The output will be sent to a JDom document
			Document page = new Document();
//			page.setDocType(new DocType("html","-//W3C//DTD XHTML 1.0 Strict//EN","http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd"));
			page.setDocType(new DocType("html"));
			Element html = new Element("html",XHTMLNS);
			page.addContent(html);
			Element head = new Element("head",XHTMLNS);
			html.addContent(head);
			Element title = new Element("title",XHTMLNS);
			title.setText("Compliance report with the XBRL Taxonomy Packages Conformance Suite");
			head.addContent(title);
			Element style = new Element("style",XHTMLNS);
			head.addContent(style);
			Attribute styleType = new Attribute("type","text/css");
			style.setAttribute(styleType);
			style.setText("td.testOK { background: lightgreen; }\ntd.testFail { background: red; }");			
			Element body = new Element("body",XHTMLNS);
			html.addContent(body);
			Element pageH2 = new Element("h2",XHTMLNS);
			pageH2.setText("Compliance report with the XBRL Table Rendering Conformance suite");
			body.addContent(pageH2);
			Date d = new Date(System.currentTimeMillis());
			SimpleDateFormat dtf = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSZ");
			Element header3 = new Element("h3",XHTMLNS);
			header3.setText("Date executed: "+dtf.format(d));
			body.addContent(header3);
			Element outTable = new Element("table",XHTMLNS);
			Attribute border1 = new Attribute("border","0");
			outTable.setAttribute(border1);
			body.addContent(outTable);
			
			XPathSelector xps = xpc.compile("for $i in testcases/testcase return resolve-uri($i/@uri,base-uri($i))").load();
			xps.setContextItem(xbrl);
			
//			XPathExecutable execSelExecDescription = xpc.compile("/c:testcase/c:description");
			XPathExecutable execSelVariation = xpc.compile("/c:testcase/c:variation");
			XPathExecutable execSelId = xpc.compile("@id");
			XPathExecutable execSelDescText = xpc.compile("c:description/text()");
			XPathExecutable execSelExecVariationName = xpc.compile("c:name/text()");
			XPathExecutable execSelFirstFile = xpc.compile("for $i in c:data/*[@readMeFirst = true()] return resolve-uri($i,base-uri($i))");
			XPathExecutable execSelParamData = xpc.compile("c:data/c:parameter");
			XPathExecutable execSelFirstFileAlt = xpc.compile("for $i in c:data/* return resolve-uri($i,base-uri($i))");
			XPathExecutable execSelResult = xpc.compile("c:result/*");
			XPathExecutable execSelResult2 = xpc.compile("if (exists(c:result/@expected)) then true() else false()");
			XPathExecutable execSelResult3 = xpc.compile("if (c:result/@expected eq 'valid') then true() else false()");
			XPathExecutable execSelInstance = xpc.compile("doc(resolve-uri(.,base-uri(.)))");
//			XPathExecutable execSelTable = xpc.compile("doc(resolve-uri(.,base-uri(.)))");
			XPathExecutable execSelCountSatisfied = xpc.compile("string(./@countSatisfied)");
			XPathExecutable execSelCountNotSatisfied = xpc.compile("string(./@countNotSatisfied)");
			XPathExecutable execSelAssertionId = xpc.compile("string(./@assertionID)");
			
//			Iterator<XdmItem> iter = xps.iterator();
			long now = System.currentTimeMillis();
//			while (iter.hasNext()) {
//				XdmItem item = iter.next();
//				String strURI = item.getStringValue();
//				URI uri = new URI(strURI);
//				StreamSource fstc = null;
//				try {
//					fstc = new StreamSource(uri.toURL().openStream());
//				} catch (FileNotFoundException e) {
//					log.warn("File: "+uri.toURL().toString()+" cannot be oppened!\nSource document: "+xbrl.getDocumentURI().toString());
//					continue; // skip this file
//				}
//				fstc.setSystemId(strURI);
//				DocumentBuilder db2 = XMLTools.newDocumentBuilder(xproc, TreeModel.TINY_TREE, null, false);
//				XdmNode testcases = db2.build(fstc);
//				
//				XPathSelector selExecDescription = execSelExecDescription.load();
//				selExecDescription.setContextItem(testcases);
//				String description = selExecDescription.evaluateSingle().getStringValue();
//				Element tr01 = new Element("tr",XHTMLNS);
//				outTable.addContent(tr01);
//				Element td01 = new Element("td",XHTMLNS);
//				tr01.addContent(td01);
//				td01.setAttribute("style", "background: orange");
//				Element link = new Element("a",XHTMLNS);
//				link.setText(description);
//				link.setAttribute("href", uri.toURL().toString());
//				td01.addContent(link);
//				//List<?> vars = (List<?>)xp.evaluate("/testcase/variation",testcases,XPathConstants.NODESET);
//				
				XPathSelector selVariation = execSelVariation.load();
//				selVariation.setContextItem(testcases);
				selVariation.setContextItem(xbrl);
				Iterator<XdmItem> iterVars = selVariation.iterator(); 
				//Iterator<?> iterVars = vars.iterator();
				while (iterVars.hasNext()) {
					Element tr02 = new Element("tr",XHTMLNS);
					outTable.addContent(tr02);
					Element td02 = new Element("td",XHTMLNS);
					tr02.addContent(td02);
					Element table02 = new Element("table",XHTMLNS);
					td02.addContent(table02);
					Attribute at02 = new Attribute("border","1");
					table02.setAttribute(at02);
					Attribute at03 = new Attribute("width","100%");
					table02.setAttribute(at03);
					Element tr03 = new Element("tr",XHTMLNS);
					table02.addContent(tr03);
					XdmItem var = iterVars.next();
					
					XPathSelector selId = execSelId.load();
					selId.setContextItem(var);
					String id = "";
					XdmItem idItem = selId.evaluateSingle();
					if (idItem != null)
						id = idItem.getStringValue();
					//String id = (String)xp.evaluate("@id",var,XPathConstants.STRING);
					
					XPathSelector selDescText = execSelDescText.load();
					selDescText.setContextItem(var);
					String desc = selDescText.evaluateSingle().getStringValue();
					
					XPathSelector selVarNameText = execSelExecVariationName.load();
					selVarNameText.setContextItem(var);
					XdmItem nameNode = selVarNameText.evaluateSingle();
					String name = null;
					if (nameNode != null) 
						name = nameNode.getStringValue();
					
					XPathSelector selFirstFile = execSelFirstFile.load();
					selFirstFile.setContextItem(var);
					XdmItem selFirstFileResult = selFirstFile.evaluateSingle();
					
					//String desc = (String)xp.evaluate("description/text()",var,XPathConstants.STRING);
					Element td03 = new Element("td",XHTMLNS);
					tr03.addContent(td03);
					Element href = new Element("a",XHTMLNS);
					href.setAttribute("name", ""+(testNumber+countSkiped+1));
					td03.addContent(href);
					td03.addContent(new Text("Id: "+id+" n:"+(testNumber+countSkiped+1)+(name != null ? (" Name:"+name) : "")));
					Element hrefDown = new Element("a",XHTMLNS);
					hrefDown.setAttribute("href", "#down");
					hrefDown.setText("(dn)");
					td03.addContent(hrefDown);
					Element td04 = new Element("td",XHTMLNS);
					td04.setText("Description: "+desc);
					tr03.addContent(td04);
					Element tr04 = new Element("tr",XHTMLNS);
					table02.addContent(tr04);
					
					// read execution expected result
					boolean assertionOk = false;
					boolean bValidInvalidOk = false;
					QName errorCode = null;
					XPathSelector selResult = execSelResult.load();
					selResult.setContextItem(var);
					XdmValue resultNode = selResult.evaluate();
					testType = UNKNOWN;
					XdmNode rNode = null;
					XdmNode instanceNode = null;
//					XdmNode tableNode = null;
					class ClsAssertionInfo {
						long assertionCountSatisfied = 0;
						long assertionCountNotSatisfied = 0;
						long assertionCountSatisfiedReport = 0;
						long assertionCountNotSatisfiedReport = 0;
						String assertionId = "";
					}
					Vector<ClsAssertionInfo> htAssertions = new Vector<ClsAssertionInfo>();
										
					XPathSelector selResult2 = execSelResult2.load();
					selResult2.setContextItem(var);
					if (selResult2.effectiveBooleanValue()) {
						XPathSelector selResult3 = execSelResult3.load();
						selResult3.setContextItem(var);
						testType |= selResult3.effectiveBooleanValue() ? VALID : INVALID;
					}
					for (XdmItem itemR : resultNode) {
						rNode = (XdmNode)itemR;					
						if (rNode.getNodeName().equals(errorQName)) {
							testType |= ERROR_CODE;
							XdmNode errorNode = rNode;						
							AtomicSequence qError = errorNode.getUnderlyingNode().atomize();
							String sVal = qError.getStringValue();
							int idx = sVal.indexOf(':');
							String pfx = sVal.substring(0, idx).trim();
							String local = sVal.substring(idx+1).trim();
							String ns = XMLTools.getNamespaceURIforPrefix(pfx, errorNode);
//							errorCode = getCorrectedErrorCode(new QName(ns,local,pfx));
							errorCode = new QName(ns,local,pfx);
						} else if (rNode.getNodeName().equals(assetionTestQName)) {
							testType |= ASSERTION;
							XPathSelector selCountSatisfied = execSelCountSatisfied.load();
							XPathSelector selCountNotSatisfied = execSelCountNotSatisfied.load();
							XPathSelector selAssertionId = execSelAssertionId.load();
							selCountSatisfied.setContextItem(rNode);
							selCountNotSatisfied.setContextItem(rNode);
							selAssertionId.setContextItem(rNode);
							
							ClsAssertionInfo asi = new ClsAssertionInfo();
							asi.assertionCountSatisfied = Integer.parseInt(selCountSatisfied.evaluateSingle().getStringValue());
							asi.assertionCountNotSatisfied = Integer.parseInt(selCountNotSatisfied.evaluateSingle().getStringValue());
							asi.assertionId = selAssertionId.evaluateSingle().getStringValue();
							htAssertions.add(asi);
						} else if (rNode.getNodeName().equals(instanceTestQName)) {
							testType |= FORMULA;
							XPathSelector selInstance = execSelInstance.load();
							selInstance.setContextItem(rNode);
							instanceNode = (XdmNode)selInstance.evaluateSingle();
//						} else if (rNode.getNodeName().equals(tableTestQName)) {
//							testType |= TABLE;
//							XPathSelector selTable = execSelTable.load();
//							selTable.setContextItem(rNode);
//							tableNode = (XdmNode)selTable.evaluateSingle();
						}
					}

					//String strRes = (String)xp.evaluate("result/@expected",var,XPathConstants.STRING);
					
					String varF = null;
					if (selFirstFileResult != null)
						varF = selFirstFileResult.getStringValue();
					//String varF = (String)xp.evaluate("for $i in data/*[@readMeFirst = true()] return resolve-uri($i,base-uri($i))",var,XPathConstants.STRING);
					if (varF == null || varF.trim().isEmpty()) {
						// may not contain readMeFirst=true attribute, in this case only one element is allowed
						XPathSelector selFirstFileAlt = execSelFirstFileAlt.load();
						selFirstFileAlt.setContextItem(var);
						XdmValue lstX = selFirstFileAlt.evaluate();						
						//List lstX = (List)xp.evaluate("for $i in data/* return resolve-uri($i,base-uri($i))",var,XPathConstants.NODESET);
						if (lstX.size() != 1) {
							log.error("Test number: "+(testNumber+countSkiped)+" conformance suite error. Multiple files exist in the test case and none of them are marked as readMeFirst=true()");
							varF = "";
						} else
							varF = lstX.itemAt(0).getStringValue();
					}
					QName qE = null;
					if ((excludePatterns != null && varF.matches(excludePatterns)) || ((testNumber+countSkiped) < startAt || (testNumber+countSkiped) > endsAt) || (testsToRun.size() > 0 && !testsToRun.contains(testNumber+countSkiped))) {
						Element td05 = new Element("td",XHTMLNS);
						td05.setText("Execution excluded");
						tr04.addContent(td05);
						countSkiped++;
					} else {
						testNumber++;
						debug = (testNumber == 41);
						StringBuilder resultText = new StringBuilder();
						StringBuilder result = new StringBuilder();
						//boolean bError = false;
						boolean instanceEquals = false;
						boolean tableEquals = false;
						try {
							if (bShowMemoryUsage) {
								//System.out.println("Total memory at 1:"+Runtime.getRuntime().freeMemory())
								long total = Runtime.getRuntime().totalMemory();
								long free = Runtime.getRuntime().freeMemory();
								log.info("Test number: "+(testNumber+countSkiped)+" Memory at 1: T="+total+" F="+free+" U="+(total-free)+" L=\""+varF+"\"");
							}
							XPathSelector selParameters = execSelParamData.load();
							selParameters.setContextItem(var);
//							Properties props = new Properties();
//							props.putAll(dtsProps);
//							props.setProperty(DTSContainer.PROCESSORS_SEQUENCE, 
//									XBRLCoreProcessor.PROCESSORNAME+","+
//									XDTProcessor.PROCESSORNAME+","+
//									XBRLFormulaProcessor.PROCESSORNAME+","+
//									XBRLTableProcessor.PROCESSORNAME);
//							props.put(DTSContainer.FORMULA_PROCESSOR_PARAMETER_RESOLVER, new ConformanceParametersFactory(selParameters.evaluate()));
//							
//							/**
//							 * IMPORTANT. This flag is set because we pass the conformance suite directly from the
//							 * files downloaded from the SVN repository. The conformance suite is using local references
//							 * to the official schemas rather than the official copies on the official locations. The
//							 * processor cannot handle the same namespace loaded from different URLs and defining the
//							 * same data types (this is an XML error). But the processor handles the same namespace
//							 * on different URLs when the two schemas "cooperate" on include statements.
//							 */
//							props.setProperty(DTSContainer.DTS_DISCOVERY_TRUST_NAMESPACES_ONLY, "true");
//							/**
//							 * In the formula comformance suite there are files with wrong values in the schemaLocation attribute on
//							 * import statements. This directive ignores the error and loads the fine from the official URL which does
//							 * not produce any other errors.
//							 */
//							props.setProperty(DTSContainer.DTS_DISCOVERY_ON_DIFFERENT_FILE_NAMES, DTSContainer.IGNORE_PROPERTY_VALUE);
//							/**
//							 * The formula specification is not compatible with the latest version of the XBRL 2.1 specification and
//							 * the conformance suite test cases contains tests that requires old specification version to be used 
//							 */
//							props.setProperty(DTSContainer.XBRL_VERSION, DTSContainer.XBRL_VERSION_2_JUL_2008);
//							
//							props.setProperty(DTSContainer.FORMULA_PROCESSOR_RETAIN_AFTER_FORMULA_INTANCE_ON_ASSERTION_RESULTS, "true");
////							props.setProperty(DTSContainer.FORMULA_PROCESSOR_RETAIN_INPUT_INTANCE_ON_ASSERTION_RESULTS, "true");
//							
//							Properties special = specialPropertiesByTestCase.get(testNumber+countSkiped); 
//							if (special != null) {
//								props.putAll(special);
//							}
							
//							DTSContainer dts = DTSContainer.newEmptyContainer(props);
//							XBRLInstance loadedInstance = null;
							// Get parameters							
//							ConformanceParameters cfs = new ConformanceParameters(selParameters.evaluate());
//							ConformanceFormulasCustomFunctions cff = new ConformanceFormulasCustomFunctions();
//							XBRLFormulaProcessor fp = XBRLFormulaProcessor.create(dts);
//							fp.setParametersResolver(cfs);
//							fp.addCustomFunctionLibrary(cff);
//							URI uriF = new URI(varF);
							File zipF = new File(new URI(varF));
							
							/*TaxonomyPackage tp = */TaxonomyPackageURIResolver.getPackage(zipF);
							
//							XBRLDocument loadedDocument = dts.load(uriF);
//							if (loadedDocument instanceof XBRLInstance)
//								loadedInstance = (XBRLInstance)loadedDocument;
//							XBRLValidationReport report = dts.validate();
							
							if (errorCode != null) {
								TPValidationException error = TaxonomyPackageURIResolver.getTaxonomyPackageError(zipF);
								if (error != null)
									throw error;
								countError++;
								errors.add(testNumber+countSkiped);

								// select from all errors in the validation report
								// the one that matches the errorCode, if any exist
//								for (XBRLValidationException ex : GeneralIterable.asIterator( report.getExceptions())) {
//									if (ex.getCode() == XBRLValidationException.ENCAPSULATE_XPATH_EXCEPTION) {
//										Throwable ex2 = ex.getCause();
//										if (ex2 instanceof XFException) {
//											XFException xfe = (XFException)ex2;
//											if (errorCode.equals( xfe.getQNameError()))
//												throw xfe;
//										}
//										if (ex2 instanceof XFValidationException) {
//											XFValidationException xfe = (XFValidationException)ex2;
//											if (errorCode.equals( xfe.getQNameError()))
//												throw xfe;
//										}
//									}
//									if (errorCode.equals( ex.getQNameError()))
//										throw ex;
//								}
//							}
//							report.throwExceptions();
//							report.inconsistenciesAsExceptions();
							
//							XBRLTableProcessor tProc = XBRLTableProcessor.create(dts);
//							tProc.setUnderEvaluationInstance(loadedInstance);
//							try {
//								Iterator<XBRLTable> tables = tProc.getTables();
//								for (XBRLTable table : GeneralIterable.asIterator(tables))
//									table.prepare(null, null);
//							} finally {
//								tProc.setUnderEvaluationInstance(null);
//							}
//							
//							report.throwExceptions();
//							report.inconsistenciesAsExceptions();
//							
//							XBRLValidationReport report2 = null;
//							XBRLCoreProcessor proc = XBRLCoreProcessor.create(dts);
//							XDTProcessor xdt = XDTProcessor.create(dts);
//							switch (testType) {
							
							// Consider each case individually							
//							if (testType == ASSERTION_PLUS_INSTANCE) {
//								// validate the returned instance
//								if (report.assertionResultsAvailable()) {
//									boolean bInsist;
//									for (String asId : report.getAssertionResultIDs()) {
//										bInsist = false;
//										XBRLAssertionResult asr[] = report.getAssertionResult(asId);
//										if (asr.length > 0) {
//											XBRLInstance instance = asr[0].getAfterFormulaInstance();
//											if (instance != null) {
//												report2 = proc.validate(asr[0].getAfterFormulaInstance());
//												report2.throwExceptions();
//												report2.inconsistenciesAsExceptions();
//												report2.merge(xdt.validate(asr[0].getAfterFormulaInstance()));
//												report2.throwExceptions();
//												report2.inconsistenciesAsExceptions();
//												instanceEquals = compareInstances(asr[0].getAfterFormulaInstance(),instanceNode,resultText);
//											} else
//												bInsist = true;
//										} else
//											bInsist = true;
//										if (!bInsist)
//											break;
//									}
//								} else {
//									report2 = proc.validate(report.getFormulaProcessorGeneratedInstance());
//									report2.throwExceptions();
//									report2.inconsistenciesAsExceptions();
//									report2.merge(xdt.validate(report.getFormulaProcessorGeneratedInstance()));
//									report2.throwExceptions();
//									report2.inconsistenciesAsExceptions();
//									instanceEquals = compareInstances(report.getFormulaProcessorGeneratedInstance(),instanceNode,resultText);
//								}
//
////								if (instanceEquals) {
//								boolean bAllOk = true;
//								for (ClsAssertionInfo as : htAssertions) {
//									as.assertionCountNotSatisfiedReport = report.getCountAssertionsNotSatisfied(as.assertionId);
//									as.assertionCountSatisfiedReport = report.getCountAssertionsSatisfied(as.assertionId);
//									if (as.assertionCountSatisfiedReport != as.assertionCountSatisfied || 
//											as.assertionCountNotSatisfiedReport != as.assertionCountNotSatisfied) {
//										bAllOk = false;
//										break; // for
//									}
//								}
//								if (bAllOk || instanceEquals) {
//									countOk++;
//									assertionOk = true;										
//								} else {
//									assertionOk = false;
//									countError++;										
//									errors.add(testNumber+countSkiped);
//								}
//							} else if (testType == FORMULA) {
//								if (report.assertionResultsAvailable()) {
//									String asId = report.getAssertionResultIDs().iterator().next();
//									XBRLAssertionResult asr[] = report.getAssertionResult(asId);
//									if (asr.length > 0) {
//										report2 = proc.validate(asr[0].getAfterFormulaInstance());
//										report2.throwExceptions();
//										report2.inconsistenciesAsExceptions();
//										report2.merge(xdt.validate(asr[0].getAfterFormulaInstance()));
//										report2.throwExceptions();
//										report2.inconsistenciesAsExceptions();
//										instanceEquals = compareInstances(asr[0].getAfterFormulaInstance(),instanceNode,resultText);
//										if (instanceEquals) {
//											countOk++;
//										} else {
//											countError++;
//											errors.add(testNumber+countSkiped);
//										}
//									} else {
//										countError++;
//										errors.add(testNumber+countSkiped);
//									}
//								} else {
//									report2 = proc.validate(report.getFormulaProcessorGeneratedInstance());
//									report2.throwExceptions();
//									report2.inconsistenciesAsExceptions();
//									report2.merge(xdt.validate(report.getFormulaProcessorGeneratedInstance()));
//									report2.throwExceptions();
//									report2.inconsistenciesAsExceptions();
//									instanceEquals = compareInstances(report.getFormulaProcessorGeneratedInstance(),instanceNode,resultText);
//									if (instanceEquals) {
//										countOk++;
//									} else {
//										countError++;
//										errors.add(testNumber+countSkiped);
//									}
//								}
//							} else if (testType == TABLE) {
//								tableEquals = compareTableModel(tableNode, dts, loadedInstance, resultText); 
//								if (tableEquals) {
//									countOk++;
//								} else {
//									countError++;
//									errors.add(testNumber+countSkiped);
//								}
//							} else if ((testType & ERROR_CODE) == ERROR_CODE) {
//								if (report.getFormulaProcessorGeneratedInstance() != null) {
//									report2 = proc.validate(report.getFormulaProcessorGeneratedInstance());
//									report2.throwExceptions();
//									report2.inconsistenciesAsExceptions();
//									report2.merge(xdt.validate(report.getFormulaProcessorGeneratedInstance()));
//									report2.throwExceptions();
//									report2.inconsistenciesAsExceptions();
//									
//									log.trace("Instance document generated when error code was expected");
//									
//									XBRLInstance cpInstance = null;
//									try {
//										if (instanceNode != null) {
//											cpInstance = new XBRLInstance(report.getFormulaProcessorGeneratedInstance().getContainer(),instanceNode.asSource(),true);
//											if (log.isTraceEnabled()) {
//												log.trace("Comparing instances:");
//												log.trace("conformance suite instance:");
//												log.trace(new String(XmlFormater.format(cpInstance.generateDocNode(false, null)),"UTF-8").toString());
//											}
//										} else {
//											resultText.append("instance node is null maybe this is an aspect test that did not generate an XBRL report");
//										}
//									} catch (SaxonApiException e) {
//										resultText.append("There is an error in the instance document in the conformance suite.\n"+instanceNode.getDocumentURI()+"\n"+e.getMessage());
//									} catch (XBRLSerializationException e) {
//										resultText.append("There is an error in the instance document in the conformance suite.\n"+instanceNode.getDocumentURI()+"\n"+e.getMessage());
//									} catch (NullPointerException e) {
//										// TODO:
//										resultText.append("There is an error in the instance document in the conformance suite.\n"+e.getMessage());
//									}
//									
//								} else {
//									log.trace("No instance document generated. May occur if the processor has validated a taxonomy");
//								}
//								countError++;
//								errors.add(testNumber+countSkiped);
//							} else if (testType == ASSERTION) {
//								boolean bAllOk = true;
//								for (ClsAssertionInfo as : htAssertions) {
//									as.assertionCountNotSatisfiedReport = report.getCountAssertionsNotSatisfied(as.assertionId);
//									as.assertionCountSatisfiedReport = report.getCountAssertionsSatisfied(as.assertionId);
//									if (as.assertionCountSatisfiedReport != as.assertionCountSatisfied || 
//											as.assertionCountNotSatisfiedReport != as.assertionCountNotSatisfied) {
//										bAllOk = false;
//										break; // for
//									}
//								}
//								if (bAllOk) {
//									countOk++;
//									assertionOk = true;										
//								} else {
//									assertionOk = false;
//									countError++;										
//									errors.add(testNumber+countSkiped);
//								}
							} else if (testType == VALID) {
								TPValidationException error = TaxonomyPackageURIResolver.getTaxonomyPackageError(zipF);
								if (error == null) {
									countOk++; // no other errors, so this is OK.
									bValidInvalidOk = true;
								} else {
									countError++;
									errors.add(testNumber+countSkiped);
								}
							} else if (testType == INVALID) {
								TPValidationException error = TaxonomyPackageURIResolver.getTaxonomyPackageError(zipF);
								if (error == null) {
									countError++;
									errors.add(testNumber+countSkiped);
								} else {
									countOk++; // no other errors, so this is OK.
									bValidInvalidOk = true;
								}
							} else if (testType == UNKNOWN) {
								log.error("Unknown test case type: "+(testNumber+countSkiped));
								instanceEquals = false;
								countError++;
								errors.add(testNumber+countSkiped);
							} else {
								countError++;
								errors.add(testNumber+countSkiped);
								bValidInvalidOk = false;								
							}
						} catch (TPValidationException e) {
							qE = e.getQNameError();
							if (errorCode == null || !qE.equals(errorCode)) {
								countError++;
								errors.add(testNumber+countSkiped);
								bValidInvalidOk = false;
							} else {
								countOk++;
								bValidInvalidOk = true;
							}
							resultText.append(" ");
							resultText.append(XBRLValidationException.documentException(e));
							resultText.append("\n");
							result.append(qE.getPrefix());
							result.append(":");
							result.append(qE.getLocalPart());
//						} catch (XFValidationException e) {
//							qE = e.getQNameError();
//							if (errorCode == null || !qE.equals(errorCode)) {
//								countError++;
//								errors.add(testNumber+countSkiped);
//							} else {
//								countOk++;
//							}
//							resultText.append(" ");
//							resultText.append(XBRLValidationException.documentException(e));
//							resultText.append("\n");
//							result.append(qE.getPrefix());
//							result.append(":");
//							result.append(qE.getLocalPart());
//						} catch (XDTValidationException e) {
//							qE = e.getQNameError();
//							if (errorCode == null || !qE.equals(errorCode)) {
//								countError++;
//								errors.add(testNumber+countSkiped);
//							} else {
//								countOk++;
//							}
//							resultText.append(" ");
//							resultText.append(XBRLValidationException.documentException(e));
//							resultText.append("\n");
//							result.append(qE.getPrefix());
//							result.append(":");
//							result.append(qE.getLocalPart());
//						} catch (XBRLValidationException e) {
//							qE = e.getQNameError();
//							if (errorCode == null || qE == null || !qE.equals(errorCode)) {
//								if ((testType & INVALID) == INVALID) {
//									countOk++;
//									bValidInvalidOk = true;								
//								} else {
//									countError++;
//									errors.add(testNumber+countSkiped);
//									bValidInvalidOk = false;								
//								}
//							} else {
//								countOk++;
//								bValidInvalidOk = true;								
//							}
//							resultText.append(" ");
//							resultText.append(XBRLValidationException.documentException(e));
//							resultText.append("\n");
//							if (qE != null) {
//								result.append(qE.getPrefix());
//								result.append(":");
//								result.append(qE.getLocalPart());
//							}
//						} catch (XFException e) {
//							qE = e.getQNameError();
//							if (errorCode == null || qE == null || !qE.equals(errorCode)) {
//								if ((testType & INVALID) == INVALID) {
//									countOk++;
//									bValidInvalidOk = true;								
//								} else {
//									countError++;
//									errors.add(testNumber+countSkiped);
//									bValidInvalidOk = false;								
//								}
//							} else {
//								countOk++;
//								bValidInvalidOk = true;								
//							}
//							resultText.append(" ");
//							resultText.append(XBRLValidationException.documentException(e));
//							resultText.append("\n");
//							if (qE != null) {
//								result.append(qE.getPrefix());
//								result.append(":");
//								result.append(qE.getLocalPart());
//							}
						} catch (XBRLRuntimeException e) {
							if (errorCode == null /*|| !qE.equals(errorCode)*/) {
								if ((testType & INVALID) == INVALID) {
									countOk++;
									bValidInvalidOk = true;								
								} else {
									countError++;
									errors.add(testNumber+countSkiped);
									bValidInvalidOk = false;								
								}
							} else {
								countOk++;
								bValidInvalidOk = true;
							}
							resultText.append(" ");
							resultText.append(XBRLValidationException.documentException(e));
							resultText.append("\n");
							result.append(qE == null ? "" : qE.getPrefix());
							result.append(":");
							result.append(qE == null ? XBRLValidationException.documentException(e) : qE.getLocalPart());
//						} catch (DTSDiscoveryException e) {
//							//qE = e.getQNameError();
//							if (errorCode == null /*|| !qE.equals(errorCode)*/) {
//								if ((testType & INVALID) == INVALID) {
//									countOk++;
//									bValidInvalidOk = true;								
//								} else {
//									countError++;
//									errors.add(testNumber+countSkiped);
//									bValidInvalidOk = false;								
//								}
//							} else {
//								countOk++;
//								bValidInvalidOk = true;								
//							}
//							resultText.append(" ");
//							resultText.append(XBRLValidationException.documentException(e));
//							resultText.append("\n");
//							result.append(resultText.toString());
////							result.append(qE.getPrefix());
////							result.append(":");
////							result.append(qE.getLocalPart());
//						} catch (LicenseVerificationException e) {
//							e.printStackTrace();
//							System.exit(-1);
//						} catch (XBRLInitializationException e) {
//							e.printStackTrace();
//							System.exit(-1);
						} catch (NullPointerException e) {
							resultText.append("NullPointerException received.");
							result.append(resultText.toString());
							countError++;
							errors.add(testNumber+countSkiped);
							if (log.isTraceEnabled())
								e.printStackTrace();
						} catch (RuntimeException e) {
							resultText.append("RuntimeException received.");
							resultText.append(XBRLValidationException.documentException(e));
							result.append(resultText.toString());
							countError++;
							errors.add(testNumber+countSkiped);
							if (log.isTraceEnabled())
								e.printStackTrace();
						}
						if (bShowMemoryUsage) {
							System.gc();
							long total = Runtime.getRuntime().totalMemory();
							long free = Runtime.getRuntime().freeMemory();
							log.info("Test number: "+(testNumber+countSkiped)+" "+"Memory at 2: T="+total+" F="+free+" U="+(total-free));
						}
						Element td05 = new Element("td",XHTMLNS);
						boolean bTestOk = false;
						String message = null;
//						switch (testType) {
//						case ASSERTION_PLUS_INSTANCE:
							
						if (( testType & ASSERTION_PLUS_INSTANCE) == ASSERTION_PLUS_INSTANCE) {
							bTestOk = assertionOk;
//							message = 							bTestOk ? 
//									"Formula and assertion OK. Expected instance" +
//									" - result is comparison OK "+result.toString()+". "+resultText.toString() :
//									"Formula NOT PASS. Expected instance" + 
//									" - result is "+ (instanceEquals ? "comparison OK" : "comparison Fail") +". "+resultText.toString();
							message = "";
							for (ClsAssertionInfo as : htAssertions) {
//								message = message + (message.length() > 0 ? "<br/>" : ""); 
								message = message +					(bTestOk ? 
									"Assertion OK. Expected "+as.assertionId+ " Count satisfied=" + as.assertionCountSatisfied + " Count not satisfied=" + as.assertionCountNotSatisfied +
									" - result is "+ as.assertionId + " Count satisfied="+as.assertionCountSatisfiedReport + " Count not satisfied="+ as.assertionCountNotSatisfiedReport +". "+resultText.toString() :
									"Assertion NOT PASS. Expected "+as.assertionId+ " Count satisfied=" + as.assertionCountSatisfied + " Count not satisfied=" + as.assertionCountNotSatisfied +
									" - result is "+ as.assertionId + " Count satisfied="+as.assertionCountSatisfiedReport + " Count not satisfied="+ as.assertionCountNotSatisfiedReport +". "+resultText.toString());
							}
//							break;
						}
						if (( testType & FORMULA) == FORMULA) {
//						case FORMULA:
							bTestOk = instanceEquals;
							message = 							bTestOk ? 
									"Formula OK. Expected "+ (errorCode == null ? (instanceNode == null ? "valid" : "instance") : errorCode.getPrefix()+":"+errorCode.getLocalPart()) +
									" - result is "+ (result.length() == 0 ? (instanceEquals ? "comparison OK" : "valid") : result.toString()) +". "+resultText.toString().trim() :
									"Formula NOT PASS. Expected "+ (errorCode == null ? (instanceNode == null ? "valid" : "instance") : errorCode.getPrefix()+":"+errorCode.getLocalPart()) + 
									" - result is "+ (result.length() == 0 ? (instanceEquals ? "comparison OK" : "comparison Fail") : result.toString()) +". "+resultText.toString().trim();
//							break;
						}
						if (( testType & ERROR_CODE) == ERROR_CODE) {
//						case ERROR_CODE:
							bTestOk = errorCode.equals(qE);
							message = 							bTestOk ? 
									"Error code OK. Expected "+ (errorCode == null ? (instanceNode == null ? "valid" : "instance") : errorCode.getPrefix()+":"+errorCode.getLocalPart()) +
									" - result is "+ (result.length() == 0 ? (instanceEquals ? "comparison OK" : "valid") : result.toString()) +". "+resultText.toString().trim() :
									"Error code NOT PASS. Expected "+ (errorCode == null ? (instanceNode == null ? "valid" : "instance") : errorCode.getPrefix()+":"+errorCode.getLocalPart()) + 
									" - result is "+ (result.length() == 0 ? (instanceEquals ? "comparison OK" : "No error code detected.") : result.toString()) +". "+resultText.toString().trim();
//							break;
						}
						if (( testType & ASSERTION) == ASSERTION) {
//						case ASSERTION:
							bTestOk = assertionOk;
							message = "";
							for (ClsAssertionInfo as : htAssertions) {
//								message = message + (message.length() > 0 ? "<br/>" : ""); 
								message = message + (							bTestOk ? 
									"Assertion OK. Expected "+as.assertionId+ " Count satisfied=" + as.assertionCountSatisfied + " Count not satisfied=" + as.assertionCountNotSatisfied +
									" - result is "+ as.assertionId + " Count satisfied="+as.assertionCountSatisfiedReport + " Count not satisfied="+ as.assertionCountNotSatisfiedReport +". "+resultText.toString().trim() :
									"Assertion NOT PASS. Expected "+as.assertionId+ " Count satisfied=" + as.assertionCountSatisfied + " Count not satisfied=" + as.assertionCountNotSatisfied +
									" - result is "+ as.assertionId + " Count satisfied="+as.assertionCountSatisfiedReport + " Count not satisfied="+ as.assertionCountNotSatisfiedReport +". "+resultText.toString().trim());
							}
//							break;
						}
						if (( testType & TABLE) == TABLE) {
							bTestOk = tableEquals;
							message = bTestOk ?
									"Table OK. Expected model and produced model matches" :
									"Table NOT PASS. Expected model and produced model mistmatch"; // add message from table model comparison
						}
						if (( testType == UNKNOWN)) {
//						case UNKNOWN:
							bTestOk = false;
							message = "UNKNOWN test type.";
//							break;
						}
						if ((testType & INVALID) == INVALID) {
							bTestOk = bValidInvalidOk;
							if (message == null) {
								message = resultText.length() > 0 ? resultText.toString() : "Test case detected invalid";
							}
						}
						if ((testType & VALID) == VALID) {
							bTestOk = bValidInvalidOk;
							if (message == null) {
								message = resultText.length() > 0 ? resultText.toString() : "Test case is valid";
							}							
						}
						Attribute at06 = new Attribute("class", bTestOk ? "testOK" : "testFail");
						td05.setAttribute(at06);
												
						td05.setText(message);
//							bTestOk ? 
//								"OK. Expected "+ (errorCode == null ? (instanceNode == null ? "valid" : "instance") : errorCode.getPrefix()+":"+errorCode.getLocalPart()) +
//								" - result is "+ (result.length() == 0 ? (instanceEquals ? "comparison OK" : "valid") : result.toString()) +". "+resultText.toString() :
//								"NOT PASS. Expected "+ (errorCode == null ? (instanceNode == null ? "valid" : "instance") : errorCode.getPrefix()+":"+errorCode.getLocalPart()) + 
//								" - result is "+ (result.length() == 0 ? (instanceEquals ? "comparison OK" : "comparison Fail") : result.toString()) +". "+resultText.toString());
						tr04.addContent(td05);
						
						Element td06 = new Element("td",XHTMLNS);
						td06.setText(resultText.toString().trim());
						tr04.addContent(td06);
					}
					Element tr05 = new Element("tr",XHTMLNS);
					table02.addContent(tr05);
					Element td07 = new Element("td",XHTMLNS);
					Attribute at07 = new Attribute("colspan","2");
					td07.setAttribute(at07);
					tr05.addContent(td07);
//					td07.setText(varF);
					Element link03 = new Element("a",XHTMLNS);
					link03.setText(varF);
					link03.setAttribute("href", "xbrl:"+varF.substring(varF.indexOf(':')+1));
					td07.addContent(link03);
				} // Shall we close stream here?
//				if (bShowMemoryUsage) {
//					//System.runFinalization();
//					System.gc();
//					long total = Runtime.getRuntime().totalMemory();
//					long free = Runtime.getRuntime().freeMemory();
//					System.out.println("Test number: "+testNumber+" "+"Memory at 2: T="+total+" F="+free+" U="+(total-free));
//				}
//			}
			Element table03 = new Element("table",XHTMLNS);
			body.addContent(table03);
			Element tr05 = new Element("tr",XHTMLNS);
			table03.addContent(tr05);
			Element td07 = new Element("td",XHTMLNS);
			tr05.addContent(td07);
			td07.setText("Not passed:");
			Element td08 = new Element("td",XHTMLNS);
			tr05.addContent(td08);
			Element linkDownPair = new Element("a",XHTMLNS);
			linkDownPair.setAttribute("name", "down");
			td08.addContent(linkDownPair);
			td08.addContent(new Text((new Integer(countError)).toString()));
			td08.addContent(getErrorsCollection(errors));
			Element tr06 = new Element("tr",XHTMLNS);
			table03.addContent(tr06);
			Element td09 = new Element("td",XHTMLNS);
			tr06.addContent(td09);
			td09.setText("Passed:");
			Element td10 = new Element("td",XHTMLNS);
			tr06.addContent(td10);
			td10.setText((new Integer(countOk)).toString());
			Element tr07 = new Element("tr",XHTMLNS);
			table03.addContent(tr07);
			Element td11 = new Element("td",XHTMLNS);
			tr07.addContent(td11);
			td11.setText("Not executed:");
			Element td12 = new Element("td",XHTMLNS);
			tr07.addContent(td12);
			td12.setText((new Integer(countSkiped)).toString());

			// produce the output to stdout or file
			Format ft = Format.getPrettyFormat();
			ft.setEncoding("ISO-8859-1");
			XMLOutputter outputter = new XMLOutputter(ft);
			String s = outputter.outputString(page);
			if (outputFile != null) {
				File f = new File(outputFile);
				FileWriter fw = new FileWriter(f);
				fw.write(s);
				fw.flush();
				fw.close();
			} else {
				System.out.println(s);
			}
			long now2 = System.currentTimeMillis();
			log.info("Total execution time "+(now2-now)/1000+"s");
		} catch (URISyntaxException e) {
			e.printStackTrace();
		} catch (MalformedURLException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (SaxonApiException e) {
			e.printStackTrace();
		} catch (XPathException e) {
			e.printStackTrace();
		}
		System.exit(0);
	}

	/**
	 * @param errors
	 * @return
	 */
	private static Collection<Content> getErrorsCollection(LinkedList<Integer> errors) {
		ArrayList<Content> al = new ArrayList<>(errors.size());
		for (Integer error : errors) {
			al.add(new Text(" "));
			Element link = new Element("a",XHTMLNS);
			link.setAttribute("href", "#"+error);
			link.setText(""+error);
			al.add(link);
		}
		return al;
	}
	
	/**
	 * @param i
	 * @param args
	 * @param dtsProps 
	 * @return
	 */
	private static int readProperties(int i, String[] args, Properties dtsProps) {
		if (args.length == i)
			return i;
		StringBuffer sb = new StringBuffer();
		int j=i;
		for (;j<args.length;++j) {
			String argument = args[j].trim();
			if (argument.startsWith("-")) {
				break;
			} else {
				sb.append(argument);
			}
		}		
		i = j-1;
		if (sb.length() == 0)
			usage();
		StringTokenIterator sti = new StringTokenIterator(sb.toString(), ",");
		Item itm = null;
		while ((itm = sti.next()) != null) {
			// find equal sign
			String token = itm.getStringValue();
			int idx = token.indexOf("=");
			if (idx < 0)
				usage();
			String propName = token.substring(0, idx);
			String propValue = token.substring(idx+1);
			dtsProps.setProperty(propName, propValue);
		}
		return i;
	}
	
//	/**
//	 * @param tableNode
//	 * @param dts
//	 * @param resultText 
//	 * @param loadedInstance 
//	 * @return
//	 * @throws XBRLValidationException 
//	 * @throws LicenseVerificationException 
//	 * @throws XFException 
//	 */
//	private static boolean compareTableModel(XdmNode tableNode, DTSContainer dts, XBRLInstance loadedInstance, StringBuilder resultText) throws LicenseVerificationException, XBRLValidationException, XFException {
//		XBRLTableProcessor tableProc = XBRLTableProcessor.create(dts);
//		if (tableProc == null) {
//			resultText.insert(0,"No table processor instance created ");
//			return false;
//		}
//		
//		tableProc.setUnderEvaluationInstance(loadedInstance);
//
//		// If there is nothing to compare with and no tables are generated, then this is OK
//		if (tableNode == null && !tableProc.getTables().hasNext()) {
//			resultText.insert(0, "Nothing generated and nothing to compare with ");
//			return true;
//		}
//		
//		// tableNode shall be loaded into a new Processor that is compatible with the DTS and schema aware
//		Processor proc2 = DTSContainer.getCompatibleProcessor(dts.getConfiguration(), true);
//		try {
//			proc2.getSchemaManager().load(dts.getConfiguration().getUnderlyingConfiguration().getURIResolver().resolve(XBRL.XBRL_tableModelURL, null));
//		} catch (SaxonApiException e1) {
//			resultText.append(e1.getMessage());
//			e1.printStackTrace();
//		} catch (TransformerException e1) {
//			resultText.append(e1.getMessage());
//			e1.printStackTrace();
//		}
//		
//		DocumentBuilder db2 = XMLTools.newDocumentBuilder(proc2, TreeModel.TINY_TREE, proc2.getSchemaManager().newSchemaValidator(), false);
//		try {
//			tableNode = db2.build(tableNode.asSource());
//		} catch (SaxonApiException e1) {
//			resultText.append(e1.getMessage());
//			e1.printStackTrace();
//		}
//		
//		// be aware, just one table will be executed
//		for (XBRLTable t : GeneralIterable.asIterator(tableProc.getTables())) {
//			XMLDocumentImpl docModel = tableProc.createModel(t);
//			docModel.moveNamespacesUp();
//			DOMSource jdomSource = new DOMSource();
//			DocumentBuilder db = XMLTools.newDocumentBuilder(proc2, TreeModel.TINY_TREE, proc2.getSchemaManager().newSchemaValidator(), true);
//			try {
//				XdmNode newDoc = db.build(docModel.getDocumentImpl());
//				return compareTableModel(tableNode, newDoc, resultText, dts);
//			} catch (SaxonApiException e) {
//				// let's not use schema validation
//				resultText.append(e.getMessage());
//				Processor proc = DTSContainer.getCompatibleProcessor(dts.getConfiguration(), false);
//				DocumentBuilder db3 = XMLTools.newDocumentBuilder(proc, TreeModel.TINY_TREE, null, false);
//				try {					
//					XdmNode newDoc = db3.build(jdomSource);
//					return compareTableModel(tableNode, newDoc, resultText, dts);
//				} catch (SaxonApiException e1) {
//					resultText.append(e1.getMessage());
//					e1.printStackTrace();
//				}
//				return false;
//			} finally {
//				tableProc.setUnderEvaluationInstance(null);
//			}
//		}
//		return false;
//	}
//
//	/**
//	 * @param tableNode
//	 * @param producedModel
//	 * @param resultText 
//	 * @return
//	 */
//	private static boolean compareTableModel(XdmNode tableNode, XdmNode producedModel, StringBuilder resultText, DTSContainer dts) {
//		boolean bRetVal = false;
//		
//		if (tableNode == null) {
//			bRetVal = producedModel == null;
//			if (bRetVal)
//				resultText.insert(0,"Both documents are null ");
//			else
//				resultText.insert(0,(tableNode == null ? "Conformance model" : producedModel == null ? "Produced model" : "XX")+" is null ");
//			return bRetVal;
//		}
//		
//		XdmNode nodeTable = skipComments(rootNode(tableNode));
//		XdmNode nodeProduced = rootNode(producedModel);
//		
//		bRetVal = compareTableModelNodes(nodeTable, nodeProduced, resultText, dts);
//		if (bRetVal)
//			resultText.insert(0, "Equal models ");
//		return bRetVal;
//	}
//
//	/**
//	 * @param tableNode
//	 * @return
//	 */
//	private static XdmNode rootNode(XdmNode node) {
//		if (node.getNodeKind() == XdmNodeKind.DOCUMENT) {
//			node = (XdmNode)node.axisIterator(net.sf.saxon.s9api.Axis.DESCENDANT, new net.sf.saxon.s9api.QName(TableModelNS, "tableModel")).next();
//		}
//		return node;
//	}
//
//	/**
//	 * @param tableNode
//	 * @return
//	 */
//	private static XdmNode skipComments(XdmNode tableNode) {
//		while (tableNode != null && tableNode.getNodeKind() != XdmNodeKind.ELEMENT && tableNode.axisIterator(net.sf.saxon.s9api.Axis.FOLLOWING_SIBLING).hasNext()) {			
//			tableNode = (XdmNode)tableNode.axisIterator(net.sf.saxon.s9api.Axis.FOLLOWING_SIBLING).next();
//		}
//		return tableNode;
//	}
//
//	/**
//	 * @param tableNode
//	 * @param producedModel
//	 * @param resultText
//	 */
//	private static boolean compareTableModelNodes(XdmNode tableNode, XdmNode producedModel, StringBuilder resultText, DTSContainer dts) {
//		if (!(tableNode.getNodeName().equals(producedModel.getNodeName()))) {
//			resultText.insert(0,"Nodes mistmatch ");
//			return false;
//		}
//		// tableSets
//		XdmSequenceIterator iterT = tableNode.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "tableSet"));
//		XdmSequenceIterator iterP = producedModel.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "tableSet"));
//		while (iterT.hasNext() && iterP.hasNext()) {
//			// Deal with not ordered tableSet nodes...
//			XdmNode setTNode = (XdmNode)iterT.next();
//			XdmNode setPNode = (XdmNode)iterP.next();
//			if (!compareTableSets(setTNode,setPNode,resultText,dts)) {
//				resultText.insert(0,"TableSet mistmatch ");
//				return false;
//			}
//		}
//		if (iterT.hasNext()) {
//			resultText.insert(0,"Conformance model has more tableSet nodes ");
//			return false;
//		}
//		if (iterP.hasNext()) {
//			resultText.insert(0,"Produced model has more tableSet nodes ");
//			return false;
//		}
//		return true;
//	}
//
//	/**
//	 * @param next
//	 * @param next2
//	 * @param resultText
//	 */
//	private static boolean compareTableSets(XdmNode tableNode, XdmNode producedModel, StringBuilder resultText, DTSContainer dts) {
//		if (!(tableNode.getNodeName().equals(producedModel.getNodeName()))) {
//			resultText.insert(0,"Nodes mistmatch ");
//			return false;
//		}
//		if (!compareLabels(tableNode,producedModel,resultText)) {
//			return false;
//		}
//		if (!compareTable(tableNode,producedModel,resultText,dts)) {
//			return false;
//		}
//
//		return true;
//	}
//
//	/**
//	 * @param tableNode
//	 * @param producedModel
//	 * @param resultText
//	 * @return
//	 */
//	private static boolean compareTable(XdmNode tableNode, XdmNode producedModel, StringBuilder resultText,DTSContainer dts) {
//		XdmSequenceIterator iterT = tableNode.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "table"));
//		XdmSequenceIterator iterP = producedModel.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "table"));
//		while (iterT.hasNext() && iterP.hasNext()) {
//			// Deal with not ordered tableSet nodes...
//			XdmNode tableT = (XdmNode)iterT.next();
//			XdmNode tableP = (XdmNode)iterP.next();
//			if (!compareTableNodes(tableT,tableP,resultText,dts)) {
//				resultText.insert(0,"Table mistmatch ");
//				return false;
//			}
//		}
//		if (iterT.hasNext()) {
//			resultText.insert(0,"Conformance model has more table nodes ");
//			return false;
//		}
//		if (iterP.hasNext()) {
//			resultText.insert(0,"Produced model has more table nodes ");
//			return false;
//		}
//		return true;
//	}
//
//	/**
//	 * @param next
//	 * @param next2
//	 * @param resultText
//	 * @return
//	 */
//	private static boolean compareTableNodes(XdmNode tableNode, XdmNode producedModel, StringBuilder resultText, DTSContainer dts) {
//		if (!(tableNode.getNodeName().equals(producedModel.getNodeName()))) {
//			resultText.insert(0,"Nodes mistmatch ");
//			return false;
//		}
//
//		for (Axis ax : new Axis[]{Axis.z, Axis.y, Axis.x}) {
//			if (!compareHeaders(ax,tableNode,producedModel,resultText,dts)) {
//				resultText.insert(0,"Axis header mistmatch '"+ax+"' ");
//				return false;
//			}
//		}
//		if (!compareCells(tableNode,producedModel,resultText)) {
//			resultText.insert(0,"cells mistmatch ");
//			return false;
//		}
//		
//		return true;
//	}
//
//	/**
//	 * @param tableNode
//	 * @param producedModel
//	 * @param resultText
//	 * @return
//	 */
//	private static boolean compareCells(XdmNode tableNode, XdmNode producedModel, StringBuilder resultText) {
//		// this might be challenging because there is no way to detect the z axis in the conformance suite
//		// is not consistent
//		// so, for the time being. just count number of rows and number of columns and facts cell by cell
//		XdmSequenceIterator iterT = tableNode.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "cells"));		
//		XdmSequenceIterator iterP = producedModel.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "cells"));
//		
//		// skip Z in all cases
//		XdmNode cellsT = (XdmNode)iterT.next();
//		if (cellsT.getAttributeValue(new net.sf.saxon.s9api.QName("axis")).equals("z"))
//			cellsT = (XdmNode)cellsT.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "cells")).next();
//		
//		XdmNode cellsP = (XdmNode)iterP.next();
//		if (cellsP.getAttributeValue(new net.sf.saxon.s9api.QName("axis")).equals("z"))
//			cellsP = (XdmNode)cellsP.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "cells")).next();
//		
//		return compareYCells(cellsT,cellsP,resultText);
//	}
//
//	/**
//	 * @param cellsT
//	 * @param cellsP
//	 * @param resultText
//	 * @return
//	 */
//	private static boolean compareYCells(XdmNode cellsT, XdmNode cellsP, StringBuilder resultText) {
//		XdmSequenceIterator iterT = cellsT.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "cells"));		
//		XdmSequenceIterator iterP = cellsP.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "cells"));
//		if (!iterT.hasNext()) {
//			if (!iterP.hasNext()) {
//				return compareXCells(cellsT, cellsP, resultText);
//			} else {
//				resultText.append(" produced model has more cells");
//				return false;
//			}
//		}
//		if (!iterP.hasNext()) {
//			resultText.append(" table model has more cells");
//			return false;
//		}
//		XdmNode cellTX = (XdmNode)iterT.next();
//		XdmNode cellPX = (XdmNode)iterP.next();
//		
//		return compareXCells(cellTX,cellPX,resultText);
//	}
//
//	/**
//	 * @param cellTX
//	 * @param cellPX
//	 * @return
//	 */
//	private static boolean compareXCells(XdmNode cellTX, XdmNode cellPX,StringBuilder resultText) {
//		XdmSequenceIterator iterT = cellTX.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "cell"));		
//		XdmSequenceIterator iterP = cellPX.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "cell"));
//		while (iterT.hasNext() && iterP.hasNext()) {
//			if (!compareCell((XdmNode)iterT.next(),(XdmNode)iterP.next(),resultText)) {
//				resultText.insert(0,"Cell mistmatch ");
//				return false;
//			}
//		}
//		if (iterT.hasNext()) {
//			resultText.insert(0,"Conformance model has more cell nodes ");
//			return false;
//		}
//		if (iterP.hasNext()) {
//			resultText.insert(0,"Produced model has more cell nodes ");
//			return false;
//		}
//		return true;
//	}
//
//	/**
//	 * @param next
//	 * @param next2
//	 * @param resultText
//	 * @return
//	 */
//	private static boolean compareCell(XdmNode cellT, XdmNode cellP, StringBuilder resultText) {
//		XdmSequenceIterator iterT = cellT.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "fact"));		
//		XdmSequenceIterator iterP = cellP.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "fact"));
//		while (iterT.hasNext() && iterP.hasNext()) {
//			if (!comparefact((XdmNode)iterT.next(),(XdmNode)iterP.next(),resultText)) {
//				resultText.insert(0,"Fact mistmatch ");
//				return false;
//			}
//		}
//		if (iterT.hasNext()) {
//			resultText.insert(0,"Conformance model has more fact nodes ");
//			return false;
//		}
//		if (iterP.hasNext()) {
//			resultText.insert(0,"Produced model has more fact nodes ");
//			return false;
//		}
//		return true;
//	}
//
//	/**
//	 * @param next
//	 * @param next2
//	 * @param resultText
//	 * @return
//	 */
//	private static boolean comparefact(XdmNode factT, XdmNode factP, StringBuilder resultText) {
//		String sT = trimPath(factT.getStringValue());
//		String sP = trimPath(factP.getStringValue());
//		if (sT.equals(sP))
//			return true;
//		resultText.insert(0, "Facts mistmatch");
//		return false;
//	}
//
//	/**
//	 * @param stringValue
//	 * @return
//	 */
//	private static String trimPath(String stringValue) {
//		int idx = stringValue.lastIndexOf('/');
//		return stringValue.substring(idx+1);
//	}
//
//	/**
//	 * @param ax
//	 * @param tableNode
//	 * @param producedModel
//	 * @param resultText
//	 * @return
//	 */
//	private static boolean compareHeaders(Axis ax, XdmNode tableNode, XdmNode producedModel, StringBuilder resultText, DTSContainer dts) {
//		// find nodes for the same axis
//		XdmSequenceIterator iterT = tableNode.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "headers"));
//		XdmSequenceIterator iterP = producedModel.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "headers"));
//		XdmNode headerT = findHeader(ax, iterT);
//		XdmNode headerP = findHeader(ax, iterP);
//		if (headerT == null && headerP == null)
//			return true; // axis does not exists 
//		if (headerP == null) {
//			resultText.insert(0,"header node not found for "+ax+" axis in produced model");
//			return false;
//		}
//		if (headerT == null) {
//			resultText.insert(0,"header node not found for "+ax+" axis in conformace model");
//			return false;
//		}
//		if (!compareHeaderNodes(headerT,headerP,resultText,dts)) {
//			resultText.insert(0,"header nodes mistmatch ");
//			return false;
//		}
//		return true;
//	}
//
//	/**
//	 * @param headerT
//	 * @param headerP
//	 * @param resultText
//	 * @return
//	 */
//	private static boolean compareHeaderNodes(XdmNode tableNode, XdmNode producedModel,	StringBuilder resultText,DTSContainer dts) {
//		// compare groups
//		XdmSequenceIterator iterT = tableNode.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "group"));
//		XdmSequenceIterator iterP = producedModel.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "group"));
//		int groupNumber = 1;
//		while (iterT.hasNext() && iterP.hasNext()) {
//			// Deal with not ordered tableSet nodes...
//			XdmNode groupT = (XdmNode)iterT.next();
//			XdmNode groupP = (XdmNode)iterP.next();
//			if (!compareGroupNodes(groupT,groupP,resultText,dts)) {
//				resultText.insert(0,"Group "+groupNumber+" mistmatch ");
//				return false;
//			}
//			groupNumber++;
//		}
//		if (iterT.hasNext()) {
//			resultText.insert(0,"Conformance model has more group nodes ");
//			return false;
//		}
//		if (iterP.hasNext()) {
//			resultText.insert(0,"Produced model has more group nodes ");
//			return false;
//		}
//		return true;
//	}
//
//	/**
//	 * @param next
//	 * @param next2
//	 * @param resultText
//	 * @return
//	 */
//	private static boolean compareGroupNodes(XdmNode tableNode, XdmNode producedModel, StringBuilder resultText,DTSContainer dts) {
//		if (!compareLabels(tableNode, producedModel, resultText)) {
//			return false;
//		}
//		XdmSequenceIterator iterT = tableNode.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "header"));
//		XdmSequenceIterator iterP = producedModel.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "header"));
//		int headerNumber = 1;
//		while (iterT.hasNext() && iterP.hasNext()) {
//			// Deal with not ordered tableSet nodes...
//			XdmNode headerT = (XdmNode)iterT.next();
//			XdmNode headerP = (XdmNode)iterP.next();
//			if (!compareGroupHeaderNodes(headerT,headerP,resultText,dts)) {
//				resultText.insert(0,"Group header "+headerNumber+" mistmatch ");
//				return false;
//			}
//			headerNumber++;
//		}
//		if (iterT.hasNext()) {
//			resultText.insert(0,"Conformance model has more group header nodes ");
//			return false;
//		}
//		if (iterP.hasNext()) {
//			resultText.insert(0,"Produced model has more group header nodes ");
//			return false;
//		}
//		return true;
//	}
//
//	/**
//	 * @param next
//	 * @param next2
//	 * @param resultText
//	 * @return
//	 */
//	private static boolean compareGroupHeaderNodes(XdmNode tableNode, XdmNode producedModel, StringBuilder resultText,DTSContainer dts) {
//		XdmSequenceIterator iterT = tableNode.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "cell"));
//		XdmSequenceIterator iterP = producedModel.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "cell"));
//		int cellNumber = 1;
//		while (iterT.hasNext() && iterP.hasNext()) {
//			// Deal with not ordered tableSet nodes...
//			if (!compareCellNodes((XdmNode)iterT.next(),(XdmNode)iterP.next(),resultText,dts)) {
//				resultText.insert(0,"Cell "+cellNumber+" mistmatch ");
//				return false;
//			}
//			cellNumber++;
//		}
//		if (iterT.hasNext()) {
//			resultText.insert(0,"Conformance model has more cell nodes ");
//			return false;
//		}
//		if (iterP.hasNext()) {
//			resultText.insert(0,"Produced model has more cell nodes ");
//			return false;
//		}
//		return true;
//	}
//
//	/**
//	 * @param next
//	 * @param next2
//	 * @param resultText
//	 * @return
//	 */
//	private static boolean compareCellNodes(XdmNode tableNode, XdmNode producedModel, StringBuilder resultText, DTSContainer dts) {
//		if (!compareLabels(tableNode, producedModel, resultText)) {
//			return false;
//		}
//		XdmSequenceIterator iterT = tableNode.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "constraint"));
//		int countT = 0;
//		int maxP = 0;
//		while (iterT.hasNext()) {
//			countT++;
//			XdmNode constraintT = (XdmNode)iterT.next();
//			XdmSequenceIterator iterP = producedModel.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "constraint"));
//			boolean bOneFound = false;
//			int countP = 0;
//			while (iterP.hasNext() && !bOneFound) {
//				countP++;
//				if (countP > maxP)
//					maxP = countP;
//				// Deal with not ordered tableSet nodes...
//				bOneFound = compareConstraintNodes(constraintT,(XdmNode)iterP.next(),resultText,dts);
//			}
//			if (!bOneFound) {
//				resultText.insert(0,"Constraint missing "+constraintT.toString());
//				return false;
//			}
//		}
//		if (countT > maxP) {
//			resultText.insert(0,"Conformance model has more constraint nodes ");
//			return false;
//		}
//		if (maxP > countT) {
//			resultText.insert(0,"Produced model has more constraint nodes ");
//			return false;
//		}
//		return true;
//	}
//
//	/**
//	 * @param next
//	 * @param next2
//	 * @param resultText
//	 * @return
//	 */
//	private static boolean compareConstraintNodes(XdmNode tableNode, XdmNode producedModel, StringBuilder resultText,DTSContainer dts) {
//		XdmSequenceIterator iterT = tableNode.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "aspect"));
//		XdmSequenceIterator iterP = producedModel.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "aspect"));
//		while (iterT.hasNext() && iterP.hasNext()) {
//			// Deal with not ordered tableSet nodes...
//			if (!compareAspectNodes((XdmNode)iterT.next(), (XdmNode)iterP.next(), resultText, dts)) {
//				resultText.insert(0,"aspect mistmatch ");
//				return false;
//			}
//		}
//		if (iterT.hasNext()) {
//			resultText.insert(0,"Conformance model has more aspect nodes ");
//			return false;
//		}
//		if (iterP.hasNext()) {
//			resultText.insert(0,"Produced model has more aspect nodes ");
//			return false;
//		}
//		iterT = tableNode.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "value"));
//		iterP = producedModel.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "value"));
//		while (iterT.hasNext() && iterP.hasNext()) {
//			// FIXME: Deal with not ordered tableSet nodes...
//			XdmNode val1 = (XdmNode)iterT.next();
//			XdmNode val2 = (XdmNode)iterP.next();
//			if (!compareValueNodes(val1, val2, resultText,dts)) {
//				resultText.insert(0,"aspect mistmatch ");
//				return false;
//			}
//		}
//		if (iterT.hasNext()) {
//			resultText.insert(0,"Conformance model has more aspect nodes ");
//			return false;
//		}
//		if (iterP.hasNext()) {
//			resultText.insert(0,"Produced model has more aspect nodes ");
//			return false;
//		}
//		return true;
//	}
//
//	/**
//	 * @param next
//	 * @param next2
//	 * @param resultText
//	 * @return
//	 */
//	private static boolean compareValueNodes(XdmNode tableNode, XdmNode producedModel, StringBuilder resultText, DTSContainer dts) {
//		return XMLFragment.equalXdmValues(tableNode, producedModel, dts.getAtomicComparator());
//	}
//
//	/**
//	 * @param next
//	 * @param next2
//	 * @param resultText
//	 * @return
//	 */
//	private static boolean compareAspectNodes(XdmNode tableNode, XdmNode producedModel, StringBuilder resultText, DTSContainer dts) {
//		return XMLFragment.equalXdmValues(tableNode, producedModel, dts.getAtomicComparator());
//	}
//
//	/**
//	 * @param ax
//	 * @param iterT
//	 * @return
//	 */
//	private static XdmNode findHeader(Axis ax, XdmSequenceIterator iter) {
//		while (iter.hasNext()) {
//			XdmNode nodeHeader = (XdmNode)iter.next();
//			if (nodeHeader.getAttributeValue(new net.sf.saxon.s9api.QName("axis")).equals(ax.toString()))
//				return nodeHeader;
//		}
//		return null;
//	}
//
//	/**
//	 * @param tableNode
//	 * @param producedModel
//	 * @param resultText
//	 * @return
//	 */
//	private static boolean compareLabels(XdmNode tableNode,	XdmNode producedModel, StringBuilder resultText) {
//		XdmSequenceIterator iterT = tableNode.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "label"));
//		XdmSequenceIterator iterP = producedModel.axisIterator(net.sf.saxon.s9api.Axis.CHILD, new net.sf.saxon.s9api.QName(TableModelNS, "label"));
//		HashSet<String> labelsT = getLabels(iterT);
//		HashSet<String> labelsP = getLabels(iterP);
//		if (labelsT.size() != labelsP.size()) {
//			resultText.insert(0,"Number of labels mistmatch ");
//			return false;
//		}
//		for (String label1 : labelsT) {
//			if (!(labelsP.contains(label1))) {
//				resultText.insert(0,"Missing label ");
//				return false;
//			}
//		}
//		
//		return true;
//	}
//
//	/**
//	 * @param iterT
//	 * @return
//	 */
//	private static HashSet<String> getLabels(XdmSequenceIterator iterT) {
//		HashSet<String> retValue = new HashSet<String>();
//		while (iterT.hasNext()) {
//			XdmItem label = iterT.next();
//			retValue.add(label.getStringValue());
//		}
//		return retValue;
//	}
//
//	/**
//	 * @param formulaProcessorGeneratedInstance
//	 * @param instanceNode
//	 * @param resultText 
//	 * @return
//	 */
//	private static boolean compareInstances(XBRLInstance inputInstance, XdmNode instanceNode, StringBuilder resultText) {
//		// create a new instance in the DTS from fpInstance using nodes from instanceNode
//		try {
//			URI inpInstanceURI = DTSContainer.getTempURL().resolve("tempInstance1.xbrl");
//			inputInstance.setURI(inpInstanceURI);
//			String ss = new String(XmlFormater.format(inputInstance.generateDocNode(false, null)));
//			StreamSource ss2 = new StreamSource(new StringReader(ss));
//			ss2.setSystemId(instanceNode.asSource().getSystemId());
//			XBRLInstance cpInstance = null;
//			try {
//				cpInstance = new XBRLInstance(inputInstance.getContainer(), instanceNode.asSource(), true);
//			} catch (SaxonApiException e) {
//				resultText.append("There is an error in the instance document in the conformance suite.\n"+instanceNode.getDocumentURI()+"\n"+e.getMessage());
//				return false;
//			}
//			XBRLInstance fpInstance = new XBRLInstance(inputInstance.getContainer(), ss2, true);
//			
//			if (log.isTraceEnabled()) {
//				log.trace("Comparing instances:");
//				log.trace("conformance suite instance:");
//				log.trace(new String(XmlFormater.format(cpInstance.generateDocNode(false, null)), "UTF-8").toString());
//				log.trace("Generated instance:");
//				log.trace(new String(XmlFormater.format(fpInstance.generateDocNode(false, null)), "UTF-8").toString());
//			}
//			
//			try {
//				cpInstance.readInfoset();
//				fpInstance.readInfoset();
//			} catch (XBRLValidationException e) {
//				e.printStackTrace();
//				resultText.append("Error reading infoset.");
//				return false;
//			}
//			
//			// check number of facts
//			if (cpInstance.size() != fpInstance.size()) {
//				resultText.append("Number of concepts in instances does not match.");
//				return false;
//			}
//			
//			// now compare the two instances...
//			
//			// fact comparison
//			Iterator<XBRLFact> iterF = fpInstance.iterator();
//			while (iterF.hasNext()) {
//				XBRLFact fpFact = iterF.next();
//				boolean bFound = false;
//				Iterator<XBRLFact> iterF2 = cpInstance.iterator();
//				while (iterF2.hasNext() && ! bFound) {
//					XBRLFact cpFact = iterF2.next();
//					
//					if (compareTwoFacts(fpFact, cpFact, resultText)) {
//						bFound = true;
//						resultText.setLength(0);
//					}
//				}
//				if (!bFound) {
//					resultText.append("Fact "+fpFact.getNodeName().toString()+" not found.");
//					return false;
//				}
//			}
//			
//			return true;
//		} catch (SaxonApiException e) {
//			e.printStackTrace();			
//		} catch (IOException e) {
//			e.printStackTrace();
//		} catch (XBRLSerializationException e) {
//			e.printStackTrace();
//		} catch (DTSDiscoveryException e) {
//			e.printStackTrace();
//		}
//		return false;
//	}
//
//	/**
//	 * Compare two facts and return true if they are semantically equivalent
//	 * Note: semantically equivalent facts has not been defined in any XBRL specification.
//	 * 
//	 * @param fpFact
//	 * @param cpFact
//	 * @param resultText 
//	 * @return boolean value
//	 */
//	private static boolean compareTwoFacts(XBRLFact fpFact, XBRLFact cpFact, StringBuilder resultText) {
//		if (!fpFact.getNodeName().equals(cpFact.getNodeName()))
//			return false;
//		if (fpFact instanceof XBRLFactItem && cpFact instanceof XBRLFactItem)
//			return compareTwoItemFacts((XBRLFactItem)fpFact,(XBRLFactItem)cpFact,resultText);
//		else if (fpFact instanceof XBRLFactTuple && cpFact instanceof XBRLFactTuple)
//			return compareTwoTupleFacts((XBRLFactTuple)fpFact,(XBRLFactTuple)cpFact,resultText);
//		return true;
//	}
//
//	/**
//	 * @param fpFact
//	 * @param cpFact
//	 * @param resultText 
//	 * @return
//	 */
//	private static boolean compareTwoTupleFacts(XBRLFactTuple fpFact, XBRLFactTuple cpFact, StringBuilder resultText) {
//		if (fpFact.size() != cpFact.size()) {
//			resultText.append("tuple has different number of elements");
//			return false;
//		}
//		Iterator<XBRLFact> iter1 = fpFact.iterator();
//		Iterator<XBRLFact> iter2 = cpFact.iterator();
//		while (iter1.hasNext()) {
//			XBRLFact fact1 = iter1.next();
//			XBRLFact fact2 = iter2.next();
//			if (!compareTwoFacts(fact1,fact2,resultText))
//				return false;
//		}
//		return false;
//	}
//
//	/**
//	 * @param fpFact
//	 * @param cpFact
//	 * @param resultText 
//	 * @return
//	 */
//	private static boolean compareTwoItemFacts(XBRLFactItem fpFact, XBRLFactItem cpFact, StringBuilder resultText) {
//		// compare contexts
//		XBRLContext ctx1 = fpFact.getContext();
//		XBRLContext ctx2 = cpFact.getContext();
//		if (!compareTwoContexts(ctx1,ctx2,resultText)) {
//			resultText.append(" not equal contexts");
//			return false;
//		}
//		if (fpFact instanceof XBRLFactNumeric && cpFact instanceof XBRLFactNumeric)
//			return compareTwoNumericFacts((XBRLFactNumeric)fpFact,(XBRLFactNumeric)cpFact,resultText);
//		else if (fpFact instanceof XBRLFactNonNumeric && cpFact instanceof XBRLFactNonNumeric)
//			return compareTwoNonNumericFacts((XBRLFactNonNumeric)fpFact,(XBRLFactNonNumeric)cpFact,resultText);
//		return true;
//	}
//
//	/**
//	 * @param ctx1
//	 * @param ctx2
//	 * @return
//	 */
//	private static boolean compareTwoContexts(XBRLContext ctx1, XBRLContext ctx2,StringBuilder resultText) {
////		return ctx1.equals(ctx2);
////		if (ctx1.hashCode() != ctx2.hashCode())
////			return false;
////		return true;
//		
//		// implement our own context comparison algorithm
//		XBRLEntity entity1 = ctx1.getEntity();
//		XBRLEntity entity2 = ctx2.getEntity();
//		if (!compareTwoEntities(entity1,entity2,resultText)) {
//			resultText.append(" not equal entities");
//			return false;
//		}
//		
//		XBRLPeriod p1 = ctx1.getPeriod();
//		XBRLPeriod p2 = ctx2.getPeriod();
//		if (!p1.equals(p2)) {
//			resultText.append(" not equal periods");
//			return false;
//		}
//		
//		XBRLScenario s1 = ctx1.getScenario();
//		XBRLScenario s2 = ctx2.getScenario();
//		
//		if (!compareTwoDimensionContainers(s1, s2, resultText))
//			return false;
//		return true;
//	}
//
//	/**
//	 * @param entity1
//	 * @param entity2
//	 * @return
//	 */
//	private static boolean compareTwoEntities(XBRLEntity entity1, XBRLEntity entity2, StringBuilder sb) {
//		if (!entity1.getScheme().equals(entity2.getScheme()))
//			return false;
//		if (!entity1.getIdentifier().equals(entity2.getIdentifier()))
//			return false;	
//		if (!compareTwoDimensionContainers(entity1.getSegment(), entity2.getSegment(), sb))
//			return false;
//		return true;
//	}
//
//	/**
//	 * @param segment
//	 * @param segment2
//	 * @return
//	 */
//	private static boolean compareTwoDimensionContainers(
//			XBRLDimensionsContainer dc1,
//			XBRLDimensionsContainer dc2,
//			StringBuilder sb) {
//		if ((dc1 == null || dc1.isEmpty()) && (dc2 == null || dc2.isEmpty()))
//			return true;
//		if (dc1.getNumberOfElements() != dc2.getNumberOfElements()) {
//			sb.append("number of children elements in the context container does not match");
//			return false;
//		}
//		Iterator<XMLFragment> iter1 = dc1.getElements();
//		int position = 0;
//		int foundAtPosition = -1;
//		while (iter1.hasNext()) {
//			XMLFragment dimElement = iter1.next();
//			// three possibilities: (1) a typedDimension container, (2) an explicitDimension container, (3) a node
//			if (XDTProcessor.typedMemberContainerQName.equals(dimElement.getNodeName())) {
//				// search for the same node in the target container. position is not important
//				if (findElement(dimElement,dc2,0) < 0)
//					return false;
//			} else if (XDTProcessor.explicitMemberContainerQName.equals(dimElement.getNodeName())) {
//				// search for the same node in the target container. position is not important
//				if (findElement(dimElement, dc2, 0) < 0)
//					return false;
//			} else {				
//				// a node
//				foundAtPosition = findElement(dimElement,dc2,position);
//				// search for the same node in the target container. position is important.
//				if (foundAtPosition == -1)
//					return false;
//				position = foundAtPosition+1; // first node to compare on next loop
//			}
//		}
//		return true;
//	}
//
//	/**
//	 * @param dimElement
//	 * @param dc2
//	 * @param position
//	 * @return
//	 */
//	private static int findElement(XMLFragment el,	XBRLDimensionsContainer dc2, int position) {
//		for (int i=position; i < dc2.getNumberOfElements(); ++i) {
//			XMLFragment child = dc2.getNthElement(i);
//			if (child.equals(el))
//				return i;
////			if (child.s_equal2(el)) {
////				return i;
////			}				
//		}
//		return -1;
//	}
//
//	/**
//	 * @param fpFact
//	 * @param cpFact
//	 * @param resultText 
//	 * @return
//	 */
//	@SuppressWarnings("unchecked")
//	private static boolean compareTwoNonNumericFacts(XBRLFactNonNumeric fpFact, XBRLFactNonNumeric cpFact, StringBuilder resultText) {
//		XdmValue v1 = fpFact.getValue();
//		XdmValue v2 = cpFact.getValue();
//		Sequence vr1 = v1.getUnderlyingValue();
//		Sequence vr2 = v2.getUnderlyingValue();
//
//		if (vr1 instanceof AtomicValue && vr2 instanceof AtomicValue)
//			return ((AtomicValue)vr1).getSchemaComparable().compareTo(((AtomicValue)vr2).getSchemaComparable()) == 0;
//		if (vr1 == EmptySequence.getInstance() && vr2 == EmptySequence.getInstance())
//			return true;
//		resultText.append("not equal fact values");
//		return false;
//	}
//
//	/**
//	 * @param unit
//	 * @param unit2
//	 * @return
//	 */
//	private static boolean compareTwoUnits(XBRLUnit unit1, XBRLUnit unit2) {
//		if (unit1.hashCode() != unit1.hashCode())
//			return false;
//		return true;
//	}
//
//	/**
//	 * @param fpFact
//	 * @param cpFact
//	 * @param resultText 
//	 * @return
//	 */
//	private static boolean compareTwoNumericFacts(XBRLFactNumeric fpFact, XBRLFactNumeric cpFact, StringBuilder resultText) {
//		if (!compareTwoUnits(fpFact.getUnit(),cpFact.getUnit())) {
//			resultText.append("not equal units");
//			return false;
//		}
//		if (!compareAccuracy(fpFact,cpFact)) {
//			resultText.append("not equal accuracy");
//			return false;
//		}
//		if (cpFact.isNil())
//			return fpFact.isNil();
//		else if (fpFact.isNil())
//			return false;
//		XBRLNumber v1 = cpFact.getNumericValue();
//		XBRLNumber v2 = fpFact.getNumericValue();
//		if (v1 == null)
//			return v2 == null;
//		if (v2 == null)
//			return v1 == null;
////		if (v1 instanceof XBRLNumericValue && v2 instanceof XBRLNumericValue) {
////			XBRLNumericValue nv1 = (XBRLNumericValue)v1;
////			XBRLNumericValue nv2 = (XBRLNumericValue)v2;
////			
////			return nv1.getBigDecimal().equals(nv2.getBigDecimal());
////		}
//		if (cpFact.isNumber() && fpFact.isNumber()) {
//			return cpFact.getBigDecimal().equals(fpFact.getBigDecimal());
//		}
//		
//		Sequence val1 = cpFact.getValue().getUnderlyingValue();
//		Sequence val2 = fpFact.getValue().getUnderlyingValue();
//		if (val1 == EmptySequence.getInstance() && val2 == EmptySequence.getInstance())
//			return true;
//		if (val1 instanceof NumericValue && (!(val2 instanceof NumericValue) || !val1.equals(val2))) {
//			resultText.append("not equal fact values");
//			return false;
//		}
//		if (val2 instanceof NumericValue && (!(val1 instanceof NumericValue) || !val2.equals(val1))) {
//			resultText.append("not equal fact values");
//			return false;
//		}
//		if (!((AtomicSequence) val2).getStringValue().equals(((AtomicSequence) val1).getStringValue())) {
//			resultText.append("not equal fact values");
//			return false;
//		}
//		return true;
//	}
//
//	/**
//	 * @param fpFact
//	 * @param cpFact
//	 * @return
//	 */
//	private static boolean compareAccuracy(XBRLFactNumeric fpFact,	XBRLFactNumeric cpFact) {
//		String decimals1 = fpFact.getAttributeValue(new QName("decimals"));
//		String precision1 = fpFact.getAttributeValue(new QName("precision"));
//		String decimals2 = cpFact.getAttributeValue(new QName("decimals"));
//		String precision2 = cpFact.getAttributeValue(new QName("precision"));
//		if (decimals1 != null && !decimals1.equals(decimals2))
//			return false;
//		if (precision1 != null && !precision1.equals(precision2))
//			return false;
//		return true;
//	}
}