/*
 *  2009 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2009 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 29/04/2009
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.util.Iterator;

import org.hibernate.Query;
import org.hibernate.Session;

import com.ihr.xbrl.database.common.XBRLDatabaseAccessApp;
import com.ihr.xbrl.pojo.XBRLInstance;
//import com.ihr.xbrl.pojo.XBRLInstance;

/**
 * This is a sample class that demonstrate how to connect to the
 * XBRL database and execute a Query.
 * 
 * The first thing you need to know is that your application object must
 * extends XBRLDatabaseAccessApp. This will automatically add the logic in
 * order to access to the connection properties and will provide you
 * helpful methods in order to execute queries to the database.
 * 
 * @author Ignacio
 *
 */
public class OpenXBRLDatabaseAndRunAQuery  extends XBRLDatabaseAccessApp {

	/**
	 * You could add new tables managed by your application and you could
	 * change the properties to connect to the database. If you just want
	 * to use the defaults then call the super constructor with null parameters. 
	 * 
	 * @param registerExtraClasses
	 * @param extraProperties
	 */
	public OpenXBRLDatabaseAndRunAQuery() {
		super(null, null, true, false);
	}

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		OpenXBRLDatabaseAndRunAQuery application = new OpenXBRLDatabaseAndRunAQuery();
		
		// request to the application to show the URI of instance documents in the database
		application.showInstanceDocuments();
	}

	/**
	 * Executes a query to the database. Note that the query result are objects. 
	 */
	private void showInstanceDocuments() {
		// Obtain a session with the database
		Session session = getDbUtil().getSessionFactory(null).openSession();

		// this is the query that will be executed. The query returns all
		// instance documents in the database. They are instances of the
		// com.ihr.xbrl.pojo.XBRLInstance object
		// see documentation about hibernate for more information
		Query query = session.createQuery("from XBRLInstance");
		
		// queries can be executed inside transactions
		session.beginTransaction();
		
		// Executes the query and returns back an Iterator object
		Iterator<?> iter = query.iterate();
		while (iter.hasNext()) {
			Object report = iter.next();
			if (report instanceof XBRLInstance) {
				XBRLInstance xReport = (XBRLInstance)report;
				// print the report URI to the console and the number of fact items.
				System.out.println(
					xReport.getURI()+
					" facts="+
					xReport.getFacts().size()
				);
			}
		}
		session.getTransaction().commit();
	}

}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
