/*
 *  2009 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2009 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 12/05/2009
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.Iterator;

import javax.xml.namespace.QName;

import net.sf.saxon.value.DoubleValue;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRLAttribute;
import com.ihr.xbrl.om.XMLFragment;
import com.ihr.xbrl.om.errors.DTSDiscoveryException;
import com.ihr.xbrl.om.errors.XBRLSerializationException;
import com.ihr.xbrl.om.errors.XBRLValidationException;
import com.ihr.xbrl.om.exLinks.CalculationLinkbase;
import com.ihr.xbrl.om.exLinks.DTSBase;
import com.ihr.xbrl.om.exLinks.XBRLExtendedLink;
import com.ihr.xbrl.om.exLinks.XBRLLinkbase;
import com.ihr.xbrl.om.exLinks.XBRLRelationship;
import com.ihr.xbrl.om.taxonomy.XBRLArcroleType;
import com.ihr.xbrl.om.taxonomy.XBRLRoleType;
import com.ihr.xbrl.om.taxonomy.XBRLTaxonomy;
import com.rs.lic.LicenseVerificationException;

/**
 * @author Ignacio
 *
 */
public class SampleAddRelationship {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		try {
			DTSContainer dts = DTSContainer.newEmptyContainer();
			XBRLTaxonomy tx = (XBRLTaxonomy)dts.load(new URI("taxonomy.xsd"));
			
			// let's find an existing calculation extended link
			XBRLArcroleType arcrole = dts.getArcroleType(CalculationLinkbase.summation_item_arcrole_URI).firstElement();
			XBRLRoleType roleLink = dts.getRoleType(XBRLExtendedLink.standard_role_URI).firstElement();
			DTSBase base = dts.getBase(CalculationLinkbase.lbType, roleLink);

			// Obtain the calculation extended link
			CalculationLinkbase cl = null;
			Iterator<XBRLExtendedLink> iterL = base.getExtendedLinks();
			if (!iterL.hasNext()) {
				XBRLLinkbase lb = new XBRLLinkbase(dts, false);
				// there are no existing calculation extended links so a new Calculation Extended link must be created
				cl = new CalculationLinkbase(lb, roleLink);
				lb.setURI(new URI("calc.xml"));
			} else {
				// just use the first one
				cl = (CalculationLinkbase)iterL.next();
			}
			
			// Obtain concept definitions
			String txNs = tx.getTargetNamespace();
			XMLFragment conceptA = dts.getConcept(new QName(txNs, "conceptA"));
			XMLFragment sales = dts.getConcept(new QName(txNs, "totalSales"));
			
			// Create a new relationship
			XBRLRelationship.make(
					cl,
					CalculationLinkbase.standard_arc,
					conceptA,
					sales,
					arcrole,
					new XBRLAttribute[] {new XBRLAttribute(new QName("weight"), new DoubleValue(1).getStringValue())},
					true);

			// save the taxonomy and linkbases back to disk
			dts.save(/*true*/);
			
			System.exit(0);
			
		} catch (DTSDiscoveryException e) {
			e.printStackTrace();
		} catch (LicenseVerificationException e) {
			e.printStackTrace();
		} catch (URISyntaxException e) {
			e.printStackTrace();
		} catch (XBRLSerializationException e) {
			e.printStackTrace();
		} catch (XBRLValidationException e) {
			e.printStackTrace();
		}
	}
}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
