/*
 *  2009 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2009 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 04/08/2009
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.io.File;
import java.net.URI;
import java.util.Properties;

import javax.xml.transform.stream.StreamSource;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRLCoreProcessor;
import com.ihr.xbrl.om.XBRLValidationReport;
import com.ihr.xbrl.om.instance.XBRLInstance;

/**
 * This example demonstrates how to use Reporting Standard XBRL API for doing fast XBRL Validation.
 * 
 * In Fast Validation mode, the taxonomy is not loaded each time a document must be validated. It is
 * loaded once and it is reused for the validation of documents using the same DTS. If the DTS is not
 * changed, then new documents can be validated without having to load the DTS again.
 * 
 * @author Ignacio
 *
 */
public class SampleBatchValidation {

	/**
	 * @param args
	 * @throws Exception 
	 */
	public static void main(String[] args) throws Exception {
		
		Properties props = new Properties();
		
		// Configure what processors shall be executed and in which order		
		props.setProperty(DTSContainer.PROCESSORS_SEQUENCE, XBRLCoreProcessor.PROCESSORNAME);
		
		// Load the taxonomy first.
		DTSContainer dts = DTSContainer.newEmptyContainer(props);
		dts.load(new URI("sampleTaxonomy.xsd")); // change this to the official URL of the taxonomy used in the instance document.

		// load and validate instance.xbrl 
		loadAndValidate("instance.xbrl", dts);
		
		// load and validate instance1.xbrl
		loadAndValidate("instance1.xbrl", dts);
	}

	static XBRLValidationReport loadAndValidate(String fileName, DTSContainer dts) throws Exception {
		File f = new File(fileName);
		StreamSource source = new StreamSource(f);
		XBRLInstance instance = new XBRLInstance(dts,source,false);		
		return dts.validate(instance);
	}
}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
