/*
 *  2009 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2009 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 05/05/2009
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.net.URI;

import javax.xml.namespace.QName;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRL;
import com.ihr.xbrl.om.taxonomy.XBRLItem;
import com.ihr.xbrl.om.taxonomy.XBRLTaxonomy;
import com.ihr.xbrl.om.taxonomy.XBRLTuple;
import com.ihr.xbrl.om.taxonomy.XBRLXSDTypeDefinition;
import com.ihr.xbrl.om.taxonomy.XMLElementDefinition;
import com.ihr.xbrl.om.taxonomy.XSDAttributeDefinition;
import com.ihr.xbrl.om.taxonomy.XSDComposition;
import com.ihr.xbrl.om.taxonomy.XSDElementRef;

/**
 * Example about how to create a tuple definition in a taxonomy.
 * 
 * @author Ignacio
 *
 */
public class SampleCreateAndReadTaxonomyProgramaticallyWithTuple {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		try {
			DTSContainer dts = DTSContainer.newEmptyContainer();
			
			// Due to the reason we will not do dts.load(...) of any other taxonomies 
			// we can call recompile now so the elements defined in the XBRL schemas 
			//are available for further schema validation.
			dts.recompile();
			
			// Create the taxonomy first.
			XBRLTaxonomy tx = new XBRLTaxonomy(dts, true);
			
			// populate the taxonomy with default minimum information
			String targetNamespace = "http://www.example.com/newTaxonomyWithTuple";
			tx.setURI(new URI("taxonomyWithTuple.xsd"));
			tx.setTargetNamespace(targetNamespace);
			tx.getDocumentRootNode().addNamespace("tx", targetNamespace);
			
			// This is required when new documents are created programatically because the load method
			// automatically adds all discovered documents to the DTS but new documents created are not
			// added automatically. Documents are indexed in the DTS by the document key. The document key
			// is build from the document URI. so call this method after setting a URI to the new document.
			dts.addDocumentToDTS(tx);
			
			// Obtain the monetaryItemType from the DTS in order to define two monetary items
			XBRLXSDTypeDefinition monetaryType = dts.getTypeDefinition(
					new QName(XBRL.XBRL_instanceNS, "monetaryItemType"));
			
			// Create conceptA (item)
			XBRLItem conceptA = new XBRLItem(
					tx,
					"conceptA",
					null,
					monetaryType,
					XBRLItem.substitutionGroup,
					XBRLItem.PeriodType.INSTANT,
					XBRLItem.BalanceType.NO_BALANCE,
					false,
					true,
					XMLElementDefinition.UNDEFINED,
					XMLElementDefinition.UNDEFINED,
					null);
			
			// Create conceptB (item)
			XBRLItem conceptB = new XBRLItem(
					tx,
					"conceptB",
					null,
					monetaryType,
					XBRLItem.substitutionGroup,
					XBRLItem.PeriodType.INSTANT,
					XBRLItem.BalanceType.NO_BALANCE,
					false,
					true,
					XMLElementDefinition.UNDEFINED,
					XMLElementDefinition.UNDEFINED,
					null);
			
			// Create the tuple content model. In this case it is a SEQUENCE
			XSDComposition comp = new XSDComposition(
					XSDComposition.SEQUENCE,1,1);
			
			// Add conceptA and conceptB to the content model of the tuple
			comp.add(new XSDElementRef(conceptA,1,1));
			comp.add(new XSDElementRef(conceptB,1,XSDComposition.UNBOUNDED));
			
			// Best practices recommends to define an attribute called id of xs:ID type
			comp.addAttribute(new XSDAttributeDefinition("id","optional",new QName(XBRL.XML_schemaNS,"ID","xs"),null,null,null,null,null));
			
			// Create a tuple with the content model defined above
			/* XBRLTuple t = */new XBRLTuple(
					tx,"SampleTuple",XBRLTuple.substitutionGroup,false,false,
					XMLElementDefinition.UNDEFINED,XMLElementDefinition.UNDEFINED,
					comp);
			
			// the content model can still be modified after the tuple has been created
			
            // Save the whole DTS
            dts.save(/*true*/);
            
            DTSContainer dts2 = DTSContainer.newEmptyContainer();
            dts2.load(new URI("taxonomyWithTuple.xsd"));
            dts2.save(/*true*/);
            
			System.exit(0);
			
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
