/*
 *  2009 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2009 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 05/08/2009
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.net.URI;

import javax.xml.namespace.QName;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRL;
import com.ihr.xbrl.om.exLinks.DefinitionLinkbase;
import com.ihr.xbrl.om.exLinks.XBRLExtendedLink;
import com.ihr.xbrl.om.exLinks.XBRLLinkbase;
import com.ihr.xbrl.om.exLinks.XBRLRelationship;
import com.ihr.xbrl.om.taxonomy.XBRLArcroleType;
import com.ihr.xbrl.om.taxonomy.XBRLImport;
import com.ihr.xbrl.om.taxonomy.XBRLItem;
import com.ihr.xbrl.om.taxonomy.XBRLRoleType;
import com.ihr.xbrl.om.taxonomy.XBRLTaxonomy;
import com.ihr.xbrl.om.taxonomy.XBRLXSDTypeDefinition;
import com.ihr.xbrl.om.taxonomy.XMLElementDefinition;

/**
 * @author Ignacio
 *
 */
public class SampleCreateDefinitionLinkbase {

	/**
	 * @param args
	 * @throws Exception in case of any errors
	 */
	public static void main(String[] args) throws Exception {
		DTSContainer dts = DTSContainer.newEmptyContainer();

		// Due to the reason we will not do dts.load(...) of any other taxonomies 
		// we can call recompile now so the elements defined in the XBRL schemas 
		// are available for further schema validation.
		dts.recompile();
		
		// Create a new taxonomy
		XBRLTaxonomy tx = new XBRLTaxonomy(dts, true);
		
		String targetNamespace = "http://www.example.com/newTaxonomyWithDefinitionLinkbase";
		tx.setURI(new URI("taxonomyWithDefinition.xsd"));
		tx.setTargetNamespace(targetNamespace);
		tx.getDocumentRootNode().addNamespace("tx", targetNamespace);
		
		// This is required when new documents are created programatically because the load method
		// automatically adds all discovered documents to the DTS but new documents created are not
		// added automatically. Documents are indexed in the DTS by the document key. The document key
		// is build from the document URI. so call this method after setting a URI to the new document.
		dts.addDocumentToDTS(tx);
		
		// Import the XBRL Instance schema
		XBRLImport.make(tx, XBRL.XBRL_instanceNS, XBRL.XBRL_instanceURL);
//		tx.addImport(new XBRLImport(tx, XBRL.XBRL_instanceNS, XBRL.XBRL_instanceURL));
		
		// create two concepts
		XBRLXSDTypeDefinition monetaryType = 
			new XBRLXSDTypeDefinition(
					dts,
					new QName(XBRL.XBRL_instanceNS, "monetaryItemType")
			);
		
		QName itemSg = new QName(XBRL.XBRL_instanceNS, "item");
		
		XBRLItem conceptA = new XBRLItem(
				tx,
				"conceptA",
				null,
				monetaryType,
				itemSg,
				XBRLItem.PeriodType.INSTANT,
				XBRLItem.BalanceType.NO_BALANCE,
				false,
				true,
				XMLElementDefinition.UNDEFINED,
				XMLElementDefinition.UNDEFINED,
				null);
		
		XBRLItem conceptB = new XBRLItem(
				tx,
				"conceptB",
				null,
				monetaryType,
				itemSg,
				XBRLItem.PeriodType.INSTANT,
				XBRLItem.BalanceType.NO_BALANCE,
				false,
				true,
				XMLElementDefinition.UNDEFINED,
				XMLElementDefinition.UNDEFINED,
				null);
		
		// Create the linkbase container and attach it to the taxonomy
		XBRLLinkbase linkBase = new XBRLLinkbase(dts, true);
		// we need a role type for any extended link. Get the standard role from the DTS
		XBRLRoleType roleLink = dts.getRoleType(XBRLExtendedLink.standard_role_URI).firstElement();
		// Create a new reference linkbase in the DTS
		DefinitionLinkbase lb = new DefinitionLinkbase(linkBase, roleLink);
		
		// set a uri for the new reference linkbase
		linkBase.setURI(new URI("definition.xml"));
		dts.addDocumentToDTS(linkBase);
		
		// get the essence_alias arcrole
		XBRLArcroleType essence_alias_arcrole = dts.getArcroleType(DefinitionLinkbase.essence_alias_arcrole_URI).firstElement();
		// Create the relationship linking the concept with the resource
		XBRLRelationship.make(lb, DefinitionLinkbase.standard_arc, conceptA, conceptB, essence_alias_arcrole, null, true);
		
		dts.save(/*true*/);
	}
}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
