/*
 *  2008 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2008 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 26/01/2008
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.net.URI;

import javax.xml.namespace.QName;

import net.sf.saxon.value.StringValue;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRLCoreProcessor;
import com.ihr.xbrl.om.XBRLValidationReport;
import com.ihr.xbrl.om.exLinks.FootnoteLinkbase;
import com.ihr.xbrl.om.exLinks.XBRLExtendedLink;
import com.ihr.xbrl.om.exLinks.XBRLRelationship;
import com.ihr.xbrl.om.exLinks.XBRLResource;
import com.ihr.xbrl.om.instance.XBRLContext;
import com.ihr.xbrl.om.instance.XBRLEntity;
import com.ihr.xbrl.om.instance.XBRLFactNonNumeric;
import com.ihr.xbrl.om.instance.XBRLInstance;
import com.ihr.xbrl.om.instance.XBRLPeriod;
import com.ihr.xbrl.om.taxonomy.XBRLArcroleType;
import com.ihr.xbrl.om.taxonomy.XBRLItem;

/**
 * 
 * Demonstration about how an application can create an instance document
 * and include a footnote for a fact item.
 * 
 * @author Ignacio
 *
 */
public class SampleCreateFactFootnote {
	
	private final static String ns = "http://www.reportingstandard.com/sampleTaxonomy"; 

	public static void main(String args[]) throws Exception {
		
		// Loads the sample taxonomy
		DTSContainer dts = DTSContainer.newEmptyContainer();
		URI uri = new URI("sampleTaxonomy.xsd");
		dts.load(uri);
		
		// Create the instance document
		XBRLInstance instance = new XBRLInstance(dts);
		
		// create a context for the instant "Begning of year 2008"
		XBRLEntity ent = new XBRLEntity(dts,"http://www.xbrl.org/scheme","Sample Company",null);
		XBRLPeriod p = new XBRLPeriod(dts,"2008-01-01");
		XBRLContext ctx = new XBRLContext(instance,ent,p,null);
		
		// Obtain the concept definition of concept "A" from the DTS
		XBRLItem itemA = (XBRLItem)dts.getConcept(new QName(ns,"A"));
		
		// create the Non Numeric (Text) fact item
		XBRLFactNonNumeric fA = new XBRLFactNonNumeric(instance.getInstanceRootNode(), ctx, itemA, true);
		// fA is nil until a value is assigned
		fA.setValue(new StringValue("test"));

		// Create a container for footnotes (Footnote Extended Link)
		FootnoteLinkbase fl = new FootnoteLinkbase(instance, 
				dts.getStaticRoleTypeByURI(XBRLExtendedLink.standard_role_URI));
		
		// Create a Footnote resource and add it to the fl container
		XBRLResource footnoteRes = new XBRLResource(fl, FootnoteLinkbase.lbResource, true);
		// As footnotes can have simpleType content and complexType content (xhtml) the default is
		// complexType content. So we can force the resource to have simple type content by changing
		// that bit.
		footnoteRes.setSimpleType();
		footnoteRes.setValue(new StringValue("Footnote content"));
		footnoteRes.setLang("en");
		
		// Create a relationship linking the fact with the footnote.
		XBRLArcroleType fact_footnote = dts.getStaticArcroleTypeByURI(FootnoteLinkbase.fact_footnote_arcrole_URI);
		XBRLRelationship.make(fl, FootnoteLinkbase.standard_arc, fA, footnoteRes, fact_footnote, null, true);

		// Validate the report and obtain the results in the report object
		XBRLValidationReport rep = instance.validate(XBRLCoreProcessor.create(dts));
		
		// if the report is invalid, then throw XBRLExceptions...
		rep.throwExceptions();
		// do the same for calculation inconsistencies
		rep.inconsistenciesAsExceptions();
		
		// if everything was right, then serialize the report to disk.
		instance.setURI(new URI("instance.xbrl"));
		instance.save(/*true*/);
		
		System.exit(0);
	}
}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
