/*
 *  2009 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2009 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 05/08/2009
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.net.URI;

import javax.xml.namespace.QName;

import net.sf.saxon.value.StringValue;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRL;
import com.ihr.xbrl.om.XMLFragment;
import com.ihr.xbrl.om.exLinks.ReferenceLinkbase;
import com.ihr.xbrl.om.exLinks.XBRLExtendedLink;
import com.ihr.xbrl.om.exLinks.XBRLLinkbase;
import com.ihr.xbrl.om.exLinks.XBRLRelationship;
import com.ihr.xbrl.om.exLinks.XBRLResource;
import com.ihr.xbrl.om.taxonomy.XBRLArcroleType;
import com.ihr.xbrl.om.taxonomy.XBRLImport;
import com.ihr.xbrl.om.taxonomy.XBRLItem;
import com.ihr.xbrl.om.taxonomy.XBRLRoleType;
import com.ihr.xbrl.om.taxonomy.XBRLTaxonomy;
import com.ihr.xbrl.om.taxonomy.XBRLXSDTypeDefinition;
import com.ihr.xbrl.om.taxonomy.XMLElementDefinition;

/**
 * @author Ignacio
 *
 */
public class SampleCreateReferenceLinkbase {

	/**
	 * @param args
	 * @throws Exception in case of any errors
	 */
	public static void main(String[] args) throws Exception {
		DTSContainer dts = DTSContainer.newEmptyContainer();

		// Due to the reason we will not do dts.load(...) of any other taxonomies 
		// we can call recompile now so the elements defined in the XBRL schemas 
		//are available for further schema validation.
		dts.recompile();
		
		// Create a new taxonomy
		XBRLTaxonomy tx = new XBRLTaxonomy(dts, true);
		
		String targetNamespace = "http://www.example.com/newTaxonomyWithReferenceLinkbase";
		tx.setURI(new URI("taxonomyWithReference.xsd"));
		tx.setTargetNamespace(targetNamespace);
		tx.getDocumentRootNode().addNamespace("tx", targetNamespace);
		
		// This is required when new documents are created programatically because the load method
		// automatically adds all discovered documents to the DTS but new documents created are not
		// added automatically. Documents are indexed in the DTS by the document key. The document key
		// is build from the document URI. so call this method after setting a URI to the new document.
		dts.addDocumentToDTS(tx);
		
		// Import the XBRL Instance schema
		tx.addImport(new XBRLImport(tx, XBRL.XBRL_instanceNS,XBRL.XBRL_instanceURL));
		// Import one of the schemas that already defines elements in the
		// link:part substitutionGroup. This is not a requirement.
		// According to the XBRL 2.1 specification a reference resource may contain
		// any element in the link:part substitutionGroup so you could define
		// your own elements in your taxonomy for the purpose of modeling
		// your own references
		tx.addImport(new XBRLImport(tx, XBRL.XBRL_ref_2006_02_27_NS,XBRL.XBRL_ref_2006_02_27));
		
		// create a concept
		XBRLXSDTypeDefinition monetaryType = 
			new XBRLXSDTypeDefinition(
					dts,
					new QName(XBRL.XBRL_instanceNS,"monetaryItemType")
			);
		
		QName itemSg = new QName(XBRL.XBRL_instanceNS,"item");
		
		XBRLItem conceptA = new XBRLItem(
				tx,
				"conceptA",
				null,
				monetaryType,
				itemSg,
				XBRLItem.PeriodType.INSTANT,
				XBRLItem.BalanceType.NO_BALANCE,
				false,
				true,
				XMLElementDefinition.UNDEFINED,
				XMLElementDefinition.UNDEFINED,
				null);
		
		// Create the linkbase container and attach it to the taxonomy
		XBRLLinkbase linkBase = new XBRLLinkbase(dts, false);
		// we need a role type for any extended link. Get the standard role from the DTS
		XBRLRoleType roleLink = dts.getRoleType(XBRLExtendedLink.standard_role_URI).firstElement();
		// Create a new reference linkbase in the DTS
		ReferenceLinkbase lb = new ReferenceLinkbase(linkBase, roleLink);
		
		// set a uri for the new reference linkbase
		linkBase.setURI(new URI("reference.xml"));
		dts.addDocumentToDTS(linkBase);
		
		// Create a new resource
		XBRLResource reference = new XBRLResource(lb,ReferenceLinkbase.lbResource,true);
		
		// add new elements to the reference resource
		XMLFragment part1 = new XMLFragment(dts);
		part1.setNodeName(new QName(XBRL.XBRL_ref_2006_02_27_NS,"Name"));
		part1.setValue(new StringValue("IAS"));
		reference.addChild(part1);
		
		XMLFragment part2 = new XMLFragment(dts);
		part2.setNodeName(new QName(XBRL.XBRL_ref_2006_02_27_NS,"Number"));
		part2.setValue(new StringValue("4"));
		reference.addChild(part2);
		
		XMLFragment part3 = new XMLFragment(dts);
		part3.setNodeName(new QName(XBRL.XBRL_ref_2006_02_27_NS,"Paragraph"));
		part3.setValue(new StringValue("14"));
		reference.addChild(part3);
		
		// Set the reference role type
		XBRLRoleType refRole = dts.getRoleType(ReferenceLinkbase.reference_role_URI).firstElement();
		reference.setRoleType(refRole);
		

		// get the concept_reference arcrole
		XBRLArcroleType concept_reference_arcrole = dts.getArcroleType(ReferenceLinkbase.concept_reference_arcrole_URI).firstElement();
		// Create the relationship linking the concept with the resource
		XBRLRelationship.make(lb, ReferenceLinkbase.standard_arc, conceptA, reference, concept_reference_arcrole, null, true);
		
		dts.save(/*true*/);
	}
}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
