/*
 *  2009 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2009 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 05/05/2009
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.net.URI;

import javax.xml.namespace.QName;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRL;
import com.ihr.xbrl.om.XBRLAttribute;
import com.ihr.xbrl.om.exLinks.CalculationLinkbase;
import com.ihr.xbrl.om.exLinks.LabelLinkbase;
import com.ihr.xbrl.om.exLinks.XBRLExtendedLink;
import com.ihr.xbrl.om.exLinks.XBRLLabelResource;
import com.ihr.xbrl.om.exLinks.XBRLLinkbase;
import com.ihr.xbrl.om.exLinks.XBRLRelationship;
import com.ihr.xbrl.om.taxonomy.XBRLArcroleType;
import com.ihr.xbrl.om.taxonomy.XBRLItem;
import com.ihr.xbrl.om.taxonomy.XBRLRoleType;
import com.ihr.xbrl.om.taxonomy.XBRLTaxonomy;
import com.ihr.xbrl.om.taxonomy.XBRLXSDTypeDefinition;
import com.ihr.xbrl.om.taxonomy.XMLElementDefinition;

/**
 * @author Ignacio
 *
 */
public class SampleCreateTaxonomyProgramatically {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		try {
			DTSContainer dts = DTSContainer.newEmptyContainer();
			
			// Due to the reason we will not do dts.load(...) of any other taxonomies 
			// we can call recompile now so the elements defined in the XBRL schemas 
			// are available for further schema validation.
			dts.recompile();
			
			XBRLTaxonomy tx = new XBRLTaxonomy(dts, true);
			
			String targetNamespace = "http://www.example.com/newTaxonomy";
			tx.setURI(new URI("taxonomy.xsd"));
			tx.setTargetNamespace(targetNamespace);
			tx.getDocumentRootNode().addNamespace("tx", targetNamespace);
			
			// This is required when new documents are created programatically because the load method
			// automatically adds all discovered documents to the DTS but new documents created are not
			// added automatically. Documents are indexed in the DTS by the document key. The document key
			// is build from the document URI. so call this method after setting a URI to the new document.
			dts.addDocumentToDTS(tx);
			
			/*
			XBRLXSDTypeDefinition monetaryType = dts.getTypeDefinition(new QName(XBRL.XBRL_instanceNS,"monetaryItemType"));
			*/
			XBRLXSDTypeDefinition monetaryType = 
				new XBRLXSDTypeDefinition(
						dts,
						new QName(XBRL.XBRL_instanceNS,"monetaryItemType")
				);
			
			QName itemSg = new QName(XBRL.XBRL_instanceNS,"item");
			
			XBRLItem conceptA = new XBRLItem(
					tx,
					"conceptA",
					null,
					monetaryType,
					itemSg,
					XBRLItem.PeriodType.INSTANT,
					XBRLItem.BalanceType.NO_BALANCE,
					false,
					true,
					XMLElementDefinition.UNDEFINED,
					XMLElementDefinition.UNDEFINED,
					null);
			
			// Create the linkbase container and attach it to the taxonomy
			XBRLLinkbase linkBase = new XBRLLinkbase(dts, true);
			// we need a role type for any extended link. Get the standard role from the DTS
			XBRLRoleType roleLink = dts.getRoleType(XBRLExtendedLink.standard_role_URI).firstElement();
			// Create a new presentation linkbase in the DTS
			LabelLinkbase lb = new LabelLinkbase(linkBase, roleLink);
			
			// add a label to the concept definition
			XBRLRoleType labelRole = dts.getRoleType(LabelLinkbase.label_role_URI).firstElement();
			XBRLLabelResource label = new XBRLLabelResource(lb, "en", "Label text for concept A", labelRole, true);
			
			XBRLArcroleType conceptLabelArt = dts.getArcroleType(LabelLinkbase.concept_label_arcrole_URI).firstElement();
			
			XBRLRelationship.make(
					lb,
					LabelLinkbase.standard_arc,
					conceptA,
					label,
					conceptLabelArt,
					null,
					true);
			
			// put a uri for the new labels linkbase
			linkBase.setURI(new URI("labels.xml"));
			dts.addDocumentToDTS(linkBase);
			
            // Total Sales
            XBRLItem item1 = new XBRLItem(
                        tx, // XBRLTaxonomy parent taxonomy
                        "totalSales", // String concept name
                        null, // Default : String default value, can be null
                        monetaryType, // type : XBRLXSDTypeDefinition (xbrli based type)
                        itemSg, // substitutionGroup : QName (item or based on item)
                        XBRLItem.PeriodType.DURATION, // Period: UNDEFINED_PERIOD_TYPE | INSTANT | DURATION | FOREVER 
                        XBRLItem.BalanceType.NO_BALANCE, // Balance: DEBIT | CREDIT | NO_BALANCE 
                        false, //Abstract
                        true, //Nillable
                        XMLElementDefinition.UNDEFINED, // Block
                        XMLElementDefinition.UNDEFINED, // Final
                        null // String fixed value
                        );
            item1.setId("tx_totalSales");

            // Sales from malaysia
            XBRLItem item2 = new XBRLItem(
                        tx, // XBRLTaxonomy parent taxonomy
                        "salesFromMalaysia", // String concept name
                        null, // Default : String default value, can be null
                        monetaryType, // type : XBRLXSDTypeDefinition (xbrli based type)
                        itemSg, // substitutionGroup : QName (item or based on item)
                        XBRLItem.PeriodType.DURATION, // Period: UNDEFINED_PERIOD_TYPE | INSTANT | DURATION | FOREVER 
                        XBRLItem.BalanceType.NO_BALANCE, // Balance: DEBIT | CREDIT | NO_BALANCE 
                        false, //Abstract
                        true, //Nillable
                        XMLElementDefinition.UNDEFINED, // Block
                        XMLElementDefinition.UNDEFINED, // Final
                        null // String fixed value
                        );
            item2.setId("tx_salesFromMalaysia");

            XBRLArcroleType conceptCalculationArt = dts.getArcroleType(CalculationLinkbase.summation_item_arcrole_URI).firstElement();
            
			XBRLLinkbase linkBase2 = new XBRLLinkbase(dts, true);
			// Create a new presentation linkbase in the DTS
			CalculationLinkbase cLinkBase = new CalculationLinkbase(linkBase2, roleLink);
			// attach the presentation linkbase to the linkbase container
			// Create the linkbase container and attach it to the taxonomy
			linkBase2.setURI(new URI("calc.xml"));
			dts.addDocumentToDTS(linkBase2);
			
			XBRLRelationship.make(cLinkBase, CalculationLinkbase.standard_arc, item1, item2, conceptCalculationArt, new XBRLAttribute[] {new XBRLAttribute(new QName("","weight"), "1.0")}, true);
            
            // This automatically re-links the linkbases to the taxonomy schema 
            tx.getLinkbases();
            
			// Alternativally you can manually add the linkbases to the taxonomy before saving.
            // This may be interesting if you wish to add a linkbase to the DTS that has no locatos
            // pointing to concepts in the schema
			/*
			tx.addLinkbase(linkBase2);
			tx.addLinkbase(linkBase);
			*/			
            
            // Save the whole DTS
            dts.save(/*true*/);
            
			System.exit(0);
			
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
