/*
 *  2009 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2009 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 05/05/2009
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.net.URI;

import javax.xml.namespace.QName;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRL;
import com.ihr.xbrl.om.taxonomy.XBRLImport;
import com.ihr.xbrl.om.taxonomy.XBRLTaxonomy;
import com.ihr.xbrl.om.taxonomy.XBRLTuple;
import com.ihr.xbrl.om.taxonomy.XBRLXSDTypeDefinition;
import com.ihr.xbrl.om.taxonomy.XMLElementDefinition;

/**
 * Example about how to create a tuple definition in a taxonomy.
 * 
 * @author Ignacio
 *
 */
public class SampleCreateTaxonomyProgramaticallyWithTupleBasedOnExternalType {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		try {
			DTSContainer dts = DTSContainer.newEmptyContainer();
			
			// Due to the reason we will not do dts.load(...) of any other taxonomies 
			// we can call recompile now so the elements defined in the XBRL schemas 
			//are available for further schema validation.
			//dts.recompile();
			
			URI uTxTypes = new URI("types2.xsd");
			XBRLTaxonomy typesTx = XBRLTaxonomy.LateLoadXBRLTaxonomy(dts,uTxTypes,true,null);			
			dts.loadIntoConfiguration(typesTx/*,false*/);
			dts.recompile();
			
			// Obtain the type from the DTS
			XBRLXSDTypeDefinition tupleType = dts.getTypeDefinition(
					new QName(typesTx.getTargetNamespace(),"accountingEntriesComplexType"));
			
			// Creates the new taxonomy
			XBRLTaxonomy tx = new XBRLTaxonomy(dts, true);
			
			
			// populate the taxonomy with default minimum information
			String targetNamespace = "http://www.example.com/newTaxonomyWithTuple";
			tx.setURI(new URI("taxonomyWithTupleExternalType.xsd"));
			tx.setTargetNamespace(targetNamespace);
			tx.getDocumentRootNode().addNamespace("tx", targetNamespace);
			
			// This is required when new documents are created programatically because the load method
			// automatically adds all discovered documents to the DTS but new documents created are not
			// added automatically. Documents are indexed in the DTS by the document key. The document key
			// is build from the document URI. so call this method after setting a URI to the new document.
			dts.addDocumentToDTS(tx);
						
			XBRLTaxonomy xbrlTx = dts.getTaxonomyByNamespace(XBRL.XBRL_instanceNS).next();
			tx.addImport(new XBRLImport(xbrlTx,XBRL.XBRL_instanceNS,XBRL.XBRL_instanceURL));
			tx.addImport(new XBRLImport(typesTx,typesTx.getTargetNamespace(),typesTx.getDocumentURI().toString()));
			
			// Create a tuple with the content model defined above
			XBRLTuple t = new XBRLTuple(
					tx,"SampleTuple",XBRLTuple.substitutionGroup,false,false,
					XMLElementDefinition.UNDEFINED,XMLElementDefinition.UNDEFINED,
					null);
			t.setType(tupleType);
			
            // Save the whole DTS
            dts.save(/*true*/);
            
			System.exit(0);
			
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
