/*
 *  2009 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2009 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 11/08/2009
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.io.File;
import java.net.URI;

import javax.xml.namespace.QName;

import net.sf.saxon.om.NamespaceBinding;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRL;
import com.ihr.xbrl.om.errors.XBRLValidationException;
import com.ihr.xbrl.om.taxonomy.XBRLItem;
import com.ihr.xbrl.om.taxonomy.XBRLTaxonomy;
import com.ihr.xbrl.om.taxonomy.XBRLTuple;
import com.ihr.xbrl.om.taxonomy.XBRLXSDTypeDefinition;
import com.ihr.xbrl.om.taxonomy.XMLElementDefinition;
import com.ihr.xbrl.om.taxonomy.XSDAttributeDefinition;
import com.ihr.xbrl.om.taxonomy.XSDComposition;
import com.ihr.xbrl.om.taxonomy.XSDElementRef;
import com.ihr.xbrl.util.XmlFormater;

/**
 * @author Ignacio
 *
 */
public class SampleCreateTuple {
	
	public static final String targetNamespace = "http://www.example.com/newTaxonomyWithTuple";
	
	// make them static in order to allow using them from functions.
	private static XBRLTaxonomy tx = null;
	private static XBRLXSDTypeDefinition monetaryType = null;

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		try {
			DTSContainer dts = DTSContainer.newEmptyContainer();
			
			// Due to the reason we will not do dts.load(...) of any other taxonomies 
			// we can call recompile now so the elements defined in the XBRL schemas 
			//are available for further schema validation.
			dts.recompile();
			
			// Create the taxonomy first.
			tx = new XBRLTaxonomy(dts, true);
			
			// populate the taxonomy with default minimum information
			tx.setURI(new URI("taxonomyWithTuple2.xsd"));
			tx.setTargetNamespace(targetNamespace);
			tx.getDocumentRootNode().addNamespace("ni", targetNamespace);
			
			// This is required when new documents are created programatically because the load method
			// automatically adds all discovered documents to the DTS but new documents created are not
			// added automatically. Documents are indexed in the DTS by the document key. The document key
			// is build from the document URI. so call this method after setting a URI to the new document.
			dts.addDocumentToDTS(tx);
			
			// Obtain the monetaryItemType from the DTS in order to define two monetary items
			monetaryType = dts.getTypeDefinition(
					new QName(XBRL.XBRL_instanceNS,"monetaryItemType"));
			
			// Create conceptX (item)
			XBRLItem conceptX = createConcept("conceptX");
			
			// Create conceptY (item)
			XBRLItem conceptY = createConcept("conceptY");
			
			// Create conceptZ (item)
			XBRLItem conceptZ = createConcept("conceptZ");
			
			// Create conceptA (item)
			XBRLItem conceptA = createConcept("conceptA");
			
			// Create conceptB (item)
			XBRLItem conceptB = createConcept("conceptB");
			
			// Create conceptC (item)
			XBRLItem conceptC = createConcept("conceptC");
			
			// Create conceptP (item)
			XBRLItem conceptP = createConcept("conceptP");
			
			// Create conceptQ (item)
			XBRLItem conceptQ = createConcept("conceptQ");
			
			// Create conceptR (item)
			XBRLItem conceptR = createConcept("conceptR");
			
			// Create the tuple content model. In this case it is a SEQUENCE
			XSDComposition comp = new XSDComposition(XSDComposition.SEQUENCE,1,1);
			
			// Add conceptA and conceptB to the content model of the tuple
			comp.add(new XSDElementRef(conceptX,1,XSDComposition.UNBOUNDED));
			comp.add(new XSDElementRef(conceptY,1,1));
			comp.add(new XSDElementRef(conceptZ,1,1));
			XSDComposition choice = new XSDComposition(XSDComposition.CHOICE,1,10);
			comp.add(choice);
			choice.add(new XSDElementRef(conceptA,1,1));
			choice.add(new XSDElementRef(conceptB,1,1));
			choice.add(new XSDElementRef(conceptC,1,1));

			// Note this composition is not legal at this place according to the XML Schema specification
			XSDComposition all = new XSDComposition(XSDComposition.SEQUENCE,1,1);
			choice.add(all);
			all.add(new XSDElementRef(conceptP,1,1));
			all.add(new XSDElementRef(conceptQ,1,1));
			all.add(new XSDElementRef(conceptR,1,1));
			
			// Best practices recommends to define an attribute called id of xs:ID type
			comp.addAttribute(new XSDAttributeDefinition("id","optional",new QName(XBRL.XML_schemaNS,"ID","xs"),null,null,null,null,null));
			
			// Create a tuple with the content model defined above
			XBRLTuple t = new XBRLTuple(
					tx,"TestTuple",XBRLTuple.substitutionGroup,false,true,
					XMLElementDefinition.UNDEFINED,XMLElementDefinition.UNDEFINED,
					comp);
			
			t.setId(t.getPrefix(targetNamespace)+"_"+t.getName());
			
            // Save the whole DTS
            dts.save(/*true*/);
            
            DTSContainer dts2 = DTSContainer.newEmptyContainer();            
            dts2.load(new File("taxonomyWithTuple2.xsd").toURI());
            XBRLTuple copyOfTuple = (XBRLTuple)dts2.getConcept(new QName(targetNamespace,"TestTuple"));
            XSDComposition composition2 = copyOfTuple.getComposition();
            
            NamespaceBinding xsNs = new NamespaceBinding("", XBRL.XML_schemaNS);
            System.out.println(new String(XmlFormater.format(composition2.generateElement(dts2, xsNs))));
            
			System.exit(0);
			
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Helper function to create new concepts. Only the concept name is required. All
	 * other parameters are static
	 * 
	 * @param name
	 * @return
	 * @throws XBRLValidationException
	 */
	private static XBRLItem createConcept(String name) throws XBRLValidationException {
		return new XBRLItem(
				tx,
				name,
				null,
				monetaryType,
				XBRLItem.substitutionGroup,
				XBRLItem.PeriodType.INSTANT,
				XBRLItem.BalanceType.NO_BALANCE,
				false,
				true,
				XMLElementDefinition.UNDEFINED,
				XMLElementDefinition.UNDEFINED,
				null);		
	}
}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
