/*
 *  2009 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2009 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 28/07/2009
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.net.URI;

import javax.xml.namespace.QName;

import net.sf.saxon.value.QNameValue;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRL;
import com.ihr.xbrl.om.XMLFragment;
import com.ihr.xbrl.om.taxonomy.XBRLItem;
import com.ihr.xbrl.om.taxonomy.XBRLTaxonomy;
import com.ihr.xbrl.om.taxonomy.XBRLTuple;
import com.ihr.xbrl.om.taxonomy.XBRLXSDTypeDefinition;
import com.ihr.xbrl.om.taxonomy.XMLElementDefinition;

/**
 * @author Ignacio
 *
 */
public class SampleGenerateComplexType1 {
	
	/*
<complexType name="accountingEntriesComplexType">
    <complexContent>
        <restriction base="anyType">
            <sequence>
                <element ref="tx:documentInfo" minOccurs="0" maxOccurs="1"/>
                <element ref="tx:entityInformation" minOccurs="0" maxOccurs="1"/>
                <element ref="tx:entryHeader" minOccurs="0" maxOccurs="unbounded"/>
            </sequence>
            <attribute name="id" type="ID"/>
        </restriction>
    </complexContent>
</complexType>
	 */

	/**
	 * @param args
	 */
	public static void main(String[] args) throws Exception {
		
		DTSContainer dts = DTSContainer.newEmptyContainer();
		
		// Due to the reason we will not do dts.load(...) of any other taxonomies 
		// we can call recompile now so the elements defined in the XBRL schemas 
		//are available for further schema validation.
		dts.recompile();
		
		XBRLTaxonomy tx = new XBRLTaxonomy(dts, true);
		
		String targetNamespace = "http://www.example.com/newTaxonomy";
		tx.setURI(new URI("taxonomyWithLocalType.xsd"));
		tx.setTargetNamespace("http://www.example.com/newTaxonomy");
		tx.getDocumentRootNode().addNamespace("tx", targetNamespace);

		XBRLXSDTypeDefinition stringItemType = 
			new XBRLXSDTypeDefinition(
					dts,
					new QName(XBRL.XBRL_instanceNS,"stringItemType")
			);
		
		QName itemSg = new QName(XBRL.XBRL_instanceNS,"item");
		
		XBRLItem documentInfo = new XBRLItem(
				tx,
				"documentInfo",
				null,
				stringItemType,
				itemSg,
				XBRLItem.PeriodType.DURATION,
				XBRLItem.BalanceType.NO_BALANCE,
				false,
				true,
				XMLElementDefinition.UNDEFINED,
				XMLElementDefinition.UNDEFINED,
				null);
		
		XBRLItem entityInformation = new XBRLItem(
				tx,
				"entityInformation",
				null,
				stringItemType,
				itemSg,
				XBRLItem.PeriodType.DURATION,
				XBRLItem.BalanceType.NO_BALANCE,
				false,
				true,
				XMLElementDefinition.UNDEFINED,
				XMLElementDefinition.UNDEFINED,
				null);
		
		XBRLItem entryHeader = new XBRLItem(
				tx,
				"entryHeader",
				null,
				stringItemType,
				itemSg,
				XBRLItem.PeriodType.DURATION,
				XBRLItem.BalanceType.NO_BALANCE,
				false,
				true,
				XMLElementDefinition.UNDEFINED,
				XMLElementDefinition.UNDEFINED,
				null);
		
		// Generate the complexType structure using XMLFragment objects
		QName typeName = new QName(tx.getTargetNamespace(),"accountingEntriesComplexType",tx.getPrefix(tx.getTargetNamespace()));
		
		XMLFragment complexType = new XMLFragment(dts);
		complexType.setNodeName(new QName(XBRL.XML_schemaNS,"complexType"));
		complexType.addAttribute(new QName("name"), typeName.getLocalPart());
		
		XMLFragment complexContent = new XMLFragment(dts);
		complexContent.setNodeName(new QName(XBRL.XML_schemaNS,"complexContent"));
		complexType.addChild(complexContent);
		
		XMLFragment restriction = new XMLFragment(dts);
		restriction.setNodeName(new QName(XBRL.XML_schemaNS,"restriction"));
		restriction.addAttribute(new QName("base"), "anyType");
		complexContent.addChild(restriction);
		
		XMLFragment sequence = new XMLFragment(dts);
		sequence.setNodeName(new QName(XBRL.XML_schemaNS,"sequence"));
		restriction.addChild(sequence);
		
		XMLFragment element1 = new XMLFragment(dts);
		element1.setNodeName(new QName(XBRL.XML_schemaNS,"element"));
		QName q = documentInfo.getQName();
		element1.addAttribute(new QName("ref"), new QNameValue(q.getPrefix(),q.getNamespaceURI(),q.getLocalPart()).getStringValue());
		element1.addAttribute(new QName("minOccurs"), "0");
		element1.addAttribute(new QName("maxOccurs"), "1");
		sequence.addChild(element1);		
		
		XMLFragment element2 = new XMLFragment(dts);
		element2.setNodeName(new QName(XBRL.XML_schemaNS,"element"));
		q = entityInformation.getQName();
		element2.addAttribute(new QName("ref"), new QNameValue(q.getPrefix(),q.getNamespaceURI(),q.getLocalPart()).getStringValue());
		element2.addAttribute(new QName("minOccurs"), "0");
		element2.addAttribute(new QName("maxOccurs"), "1");
		sequence.addChild(element2);
		
		XMLFragment element3 = new XMLFragment(dts);
		element3.setNodeName(new QName(XBRL.XML_schemaNS,"element"));
		q = entryHeader.getQName();
		element3.addAttribute(new QName("ref"), new QNameValue(q.getPrefix(),q.getNamespaceURI(),q.getLocalPart()).getStringValue());
		element3.addAttribute(new QName("minOccurs"), "0");
		element3.addAttribute(new QName("maxOccurs"), "unbounded");
		sequence.addChild(element3);
		
		XMLFragment attribute = new XMLFragment(dts);
		attribute.setNodeName(new QName(XBRL.XML_schemaNS,"attribute"));
		attribute.addAttribute(new QName("name"), "id");
		attribute.addAttribute(new QName("type"), new QNameValue("xs",XBRL.XML_schemaNS,"ID").getStringValue());
		restriction.addChild(attribute);
		
		XBRLTuple t = new XBRLTuple(tx,"sampleTuple",XBRLTuple.substitutionGroup,Boolean.FALSE,
				Boolean.FALSE,XMLElementDefinition.UNDEFINED,XMLElementDefinition.UNDEFINED,
				null);
		XBRLXSDTypeDefinition tupleType = new XBRLXSDTypeDefinition(dts, typeName);
		t.setType(tupleType);
		
		tx.addGlobalType(complexType);
		
		tx.save(/*true*/);
		
		System.exit(0);
	}
}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
