/*
 *  2009 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2009 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 12/05/2009
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.Iterator;

import javax.xml.namespace.QName;

import net.sf.saxon.value.StringValue;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XMLFragment;
import com.ihr.xbrl.om.errors.DTSDiscoveryException;
import com.ihr.xbrl.om.errors.XBRLSerializationException;
import com.ihr.xbrl.om.exLinks.DTSBase;
import com.ihr.xbrl.om.exLinks.LabelLinkbase;
import com.ihr.xbrl.om.exLinks.XBRLExtendedLink;
import com.ihr.xbrl.om.exLinks.XBRLRelationship;
import com.ihr.xbrl.om.taxonomy.XBRLArcroleType;
import com.ihr.xbrl.om.taxonomy.XBRLRoleType;
import com.rs.lic.LicenseVerificationException;

/**
 * This test assumes that the code for SampleCreateTaxonomyProgramatically has
 * been executed in advance and there is a taxonomy in the same directory of the
 * execution of the Java virtual machine for this main class.
 * 
 * @author Ignacio
 *
 */
public class SampleModifyLabels {

	/**
	 * No arguments are used 
	 * 
	 * @param args
	 */
	public static void main(String[] args) {
		try {
			DTSContainer dts = DTSContainer.newEmptyContainer();
			dts.load(new URI("taxonomy.xsd"));
			
			XMLFragment el = dts.getConcept(new QName("http://www.example.com/newTaxonomy","conceptA"));
			
			// let's find the a label resource
			XBRLArcroleType arcrole = dts.getArcroleType(LabelLinkbase.concept_label_arcrole_URI).firstElement();
			XBRLRoleType roleLink = dts.getRoleType(XBRLExtendedLink.standard_role_URI).firstElement();
			DTSBase base = dts.getBase(LabelLinkbase.lbType, roleLink);
			Iterator<XBRLRelationship> iterR = el.getFromRelationships(arcrole, base);
			if (!iterR.hasNext()) {
				System.out.println("The concept has no labels");
				System.exit(-1);
			}			
			XBRLRelationship rel = iterR.next();
			XMLFragment to = rel.getTo();
			to.setValue(new StringValue("New label value"));

			/* alternativelly you can make sure it is a label resource by doing
			if (to instanceof XBRLLabelResource) {
				XBRLLabelResource label = (XBRLLabelResource)to;				
				label.setValue(new StringValue("New label value"));
			}
			*/
			
			// save the linkbase again
			dts.save(/*true*/);
			
			System.exit(0);
			
		} catch (DTSDiscoveryException e) {
			e.printStackTrace();
		} catch (LicenseVerificationException e) {
			e.printStackTrace();
		} catch (URISyntaxException e) {
			e.printStackTrace();
		} catch (XBRLSerializationException e) {
			e.printStackTrace();
		}
	}

}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
