/*
 *  2009 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2009 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 26/06/2009
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.io.File;
import java.util.Iterator;
import java.util.Properties;

import javax.xml.namespace.QName;

import com.ihr.xbrl.dts.LabelsProvider;
import com.ihr.xbrl.dts.LabelsProviderInstance;
import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XMLFragment;
import com.ihr.xbrl.om.exLinks.DefinitionLinkbase;
import com.ihr.xbrl.om.exLinks.PresentationLinkbase;
import com.ihr.xbrl.om.exLinks.XBRLExtendedLink;
import com.ihr.xbrl.om.exLinks.XBRLRelationship;
import com.ihr.xbrl.om.formula.FContext;
import com.ihr.xbrl.om.instance.XBRLContext;
import com.ihr.xbrl.om.instance.XBRLFact;
import com.ihr.xbrl.om.instance.XBRLInstance;
import com.ihr.xbrl.om.taxonomy.XBRLRoleType;
import com.ihr.xbrl.om.taxonomy.XMLElementDefinition;
import com.ihr.xbrl.om.xdt.XDTHypercubeUse;
import com.ihr.xbrl.util.TextTools;

/**
 * Sample, a new instance document is read form the file received as a parameter.
 * Shows all facts properties in the instance.
 * 
 * @author Ignacio
 */
public class SampleReadInstanceFactsProps {

	/**
	 * @param args
	 */
	public static void main(String[] args) throws Exception {		
		if (args.length == 0) {
			args = new String[1];
			args[0] = "instance.xbrl";
		}
		Properties pros = new Properties();
		DTSContainer dts = DTSContainer.newEmptyContainer(pros);
		XBRLInstance instance = (XBRLInstance)dts.load(new File(args[0]).toURI());
		LabelsProvider lp = LabelsProviderInstance.getLabelsProvider();
		
		// mostrar en la salida:
		// concepto|valor|idContexto|periodo|rol|{dimension}, {miembro}
		Iterator<XBRLFact> iterF = instance.getInstanceRootNode().iterator();
		while (iterF.hasNext()) {
			XBRLFact f = iterF.next();
			XMLElementDefinition el = f.getElementDefinition();
			String strIdContext = f.getAttributeValue(new QName("contextRef"));
			XBRLContext ctx = instance.getContextById(strIdContext);
			
			String strConcept = el.getName();
			String strValue = valueToString(f.getStringValue());
			String strPeriod = ctx.getPeriod().getFriendlyName(XMLFragment.PERIOD_TYPE | XMLFragment.PERIOD_VALUE, lp);
			String strRole = roleToString(el, lp);
			String strDims = dimToString(ctx, el, lp);
			
			System.out.println(strConcept+"|"+strValue+"|"+strIdContext+"|"+strPeriod+"|"+strRole+"|"+strDims);
		}
	}

	private static String valueToString(String stringValue) {
		String value = stringValue.trim();
		value = value.replace('\r', ' ');
		value = value.replace('\n', ' ');
		value = value.replace('\t', ' ');
		value = TextTools.normalizeSpace(value);
		if (value.length() > 50) {
			value = value.substring(0, 50);
			value += "...";
		}
		return value;
	}

	private static String roleToString(XMLElementDefinition el, LabelsProvider lp) {
		final boolean doElementLabel = true;
		boolean first = true;
		StringBuffer sb = new StringBuffer();
		Iterator<XBRLRelationship> iterR = el.getFromRelationships();
		while (iterR.hasNext()) {
			XBRLRelationship rel = iterR.next();
			XBRLExtendedLink exLink = null;
			if (rel.getParentXBRL() instanceof PresentationLinkbase)
				exLink = rel.getParentXBRL();
//			else if (rel.getParentXBRL() instanceof CalculationLinkbase)
//				exLink = rel.getParentXBRL();
			else if (rel.getParentXBRL() instanceof DefinitionLinkbase)
				exLink = rel.getParentXBRL();
			if (exLink != null) {
				if (!first)
					sb.append(", ");
				first = false;
				XBRLRoleType rt = exLink.getRoleType();
				sb.append(rt.getFriendlyName(doElementLabel ? XMLFragment.ELEMENT_LABEL : 0, lp));
			}
		}
		iterR = el.getToRelationships();
		while (iterR.hasNext()) {
			XBRLRelationship rel = iterR.next();
			XBRLExtendedLink exLink = null;
			if (rel.getParentXBRL() instanceof PresentationLinkbase)
				exLink = rel.getParentXBRL();
//			else if (rel.getParentXBRL() instanceof CalculationLinkbase)
//				exLink = rel.getParentXBRL();
			else if (rel.getParentXBRL() instanceof DefinitionLinkbase)
				exLink = rel.getParentXBRL();
			if (exLink != null) {
				if (!first)
					sb.append(", ");
				first = false;
				XBRLRoleType rt = exLink.getRoleType();
				sb.append(rt.getFriendlyName(doElementLabel ? XMLFragment.ELEMENT_LABEL : 0, lp));
			}
		}
		return sb.toString();
	}

	private static String dimToString(XBRLContext ctx, XMLElementDefinition el, LabelsProvider lp) {
		StringBuffer sb = new StringBuffer();
		FContext fc = new FContext(el, ctx);
		if (fc.getDimensions().size() > 0) {
			boolean first = true;
			for (QName dim : fc.getDimensions()) {
				if (!first)
					sb.append(", ");
				first = false;
				XMLFragment elDim = ctx.getContainer().getConcept(dim);
				sb.append(elDim.getFriendlyName(XMLFragment.CONCEPT_LABEL, lp));
				sb.append(": ");
				XMLFragment frDimValue = fc.getDimension(dim);
				if (FContext.isExplicitDimensionContainer(frDimValue)) {
					QName dimValue = frDimValue.getQNameValue();
					XMLFragment elValue = ctx.getContainer().getConcept(dimValue);
					sb.append(elValue.getFriendlyName(XMLFragment.CONCEPT_LABEL, lp));
				} else if (FContext.isTypedDimensionContainer(frDimValue)) {
					XMLFragment dimValue = frDimValue.getElements().next();
					sb.append(dimValue.toString());				
				}
			}
		}
		if (fc.getNonDimensionalInformation(XDTHypercubeUse.SCENARIO).size() > 0) {
			boolean first = true;
			for (XMLFragment f : fc.getNonDimensionalInformation(XDTHypercubeUse.SCENARIO)) {
				if (!first)
					sb.append(", ");
				first = false;
				sb.append(f.toString());
			}
		}
		if (fc.getNonDimensionalInformation(XDTHypercubeUse.SEGMENT).size() > 0) {
			boolean first = true;
			for (XMLFragment f : fc.getNonDimensionalInformation(XDTHypercubeUse.SEGMENT)) {
				if (!first)
					sb.append(", ");
				first = false;
				sb.append(f.toString());
			}
		}
		return sb.toString();
	}
}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
