/*
 *  2009 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2009 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 26/06/2009
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.io.File;
import java.util.Iterator;
import java.util.Properties;

import org.apache.log4j.Appender;
import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Layout;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.WriterAppender;
import org.apache.log4j.spi.ErrorHandler;
import org.apache.log4j.spi.Filter;
import org.apache.log4j.spi.LoggingEvent;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.instance.XBRLFact;
import com.ihr.xbrl.om.instance.XBRLInstance;
import com.ihr.xbrl.om.instance.XBRLContext;

/**
 * Sample, a new instance document is read form the file received as a parameter
 * After the instance object is obtained, the user can start working with it
 * 
 * @author Ignacio
 */
@SuppressWarnings("unused")
public class SampleReadInstanceShowLog {

	/**
	 * @param args
	 */
	public static void main(String[] args) throws Exception {
		if (args.length == 0) {
			args = new String[1];
			args[0] = "instance.xbrl";
		}
		Properties props = new Properties();
		props.setProperty(DTSContainer.LOG_LEVEL, "ALL");				
		DTSContainer dts = DTSContainer.newEmptyContainer(props);
		
		// now we can, programatically, capture the output sent to the logger if we need to... note this can be
		// configured in the log4j.xml file too.
		Logger log = Logger.getLogger(DTSContainer.LOGGER_NAME);
		log.addAppender(new LocalAppender());
		
		XBRLInstance instance = (XBRLInstance)dts.load(new File(args[0]).toURI());
		// ... rest of the application code goes here
		
		Iterator<XBRLContext> iterC = instance.getInstanceRootNode().getContexts();
		while (iterC.hasNext()) {
			System.out.println("Context Id:"+iterC.next().getId());
		}
		
		Iterator<XBRLFact> iterF = instance.getInstanceRootNode().iterator();
		while (iterF.hasNext()) {
			XBRLFact f = iterF.next();
			// ... do dome work with the fact
		}
		
		
	}

	static class LocalAppender extends WriterAppender {
		
		@Override
		public void append(LoggingEvent event) {
			System.err.println( event.getMessage().toString() );
		}

	}
	
	
}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
