/*
 *  2009 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2009 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 26/06/2009
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.io.File;
import java.net.URI;
import java.util.Iterator;
import java.util.Properties;

import javax.xml.transform.URIResolver;

import org.apache.log4j.Level;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.SimpleMatchURIResolver;
import com.ihr.xbrl.om.instance.XBRLFact;
import com.ihr.xbrl.om.instance.XBRLInstance;
import com.ihr.xbrl.om.instance.XBRLContext;
import com.ihr.xbrl.om.resolvers.HasCreateDTSUriResolver;

/**
 * An instance document is referring to a taxonomy in a site that does not exists and a folder
 * that also does not exists.
 * 
 * But we are going to configure the processor to dynamically map those URLs to good ones.
 * 
 * This sample shows how to:
 * 
 *  - use the HasCreateDTSUriResolver interface
 *  - use the DTS_RESOLVER DTSContainer property
 *  - use the SimpleMatchURIResolver class
 * 
 * @author Ignacio
 */
@SuppressWarnings("unused")
public class SampleReadInstanceUsingCustomMappings {

	/**
	 * @param args
	 */
	public static void main(String[] args) throws Exception {		

		// this is the local file uri.
		final URI fileUri = new File("instance_map_schemas.xbrl").toURI();
		
		HasCreateDTSUriResolver createNewResolver = new HasCreateDTSUriResolver() {
			@Override
			public void createAndRegisterDTSUriResolver(DTSContainer dts) {
				
				// obtain current resolver from the DTS.
				URIResolver currentResolver = dts.getDTSURIResolver();
				
				// create the Simple Match URI Resolver instance (Open the sample instance to see why this is here)
				SimpleMatchURIResolver newUriResolver = new SimpleMatchURIResolver(
						"http://www.nonexistentwebsitewithxbrldata.com/this_site_and_folder_does_not_exists_but_will_be_mapped_locally/", 
						fileUri.resolve(".").toString(), 
						currentResolver);
				
				dts.installDTSURIResolver(newUriResolver);
			}
		};
				
		Properties props = new Properties();
		// we use put rather than setProperty because the value is not a string.
		props.put(DTSContainer.DTS_RESOLVER, createNewResolver);		
		DTSContainer dts = DTSContainer.newEmptyContainer(props);
		
		// now we can simple open and validate the instance
		XBRLInstance instance = (XBRLInstance)dts.load(fileUri);
		// ... rest of the application code goes here
		
		Iterator<XBRLContext> iterC = instance.getInstanceRootNode().getContexts();
		while (iterC.hasNext()) {
			System.out.println("Context Id:"+iterC.next().getId());
		}
		
		Iterator<XBRLFact> iterF = instance.getInstanceRootNode().iterator();
		while (iterF.hasNext()) {
			XBRLFact f = iterF.next();
			// ... do dome work with the fact
		}
		
		
	}

}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
