/*
 *  2009 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2009 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 18/08/2009
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.io.File;
import java.io.FileWriter;
import java.net.URI;
import java.net.URISyntaxException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;

import org.jdom.Attribute;
import org.jdom.DocType;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XMLFragment;
import com.ihr.xbrl.om.exLinks.LabelLinkbase;
import com.ihr.xbrl.om.exLinks.ReferenceLinkbase;
import com.ihr.xbrl.om.exLinks.XBRLExtendedLink;
import com.ihr.xbrl.om.exLinks.XBRLRelationship;
import com.ihr.xbrl.om.exLinks.XBRLResource;
import com.ihr.xbrl.om.taxonomy.XBRLArcroleType;
import com.ihr.xbrl.om.taxonomy.XBRLRoleType;
import com.ihr.xbrl.om.taxonomy.XMLElementDefinition;

/**
 * Reads a DTS passed in as a parameter, then
 * produces an html page with a table that contains
 * all concepts defined in the DTS, all concept labels
 * and all concept references un multiple columns.
 * 
 * @author Ignacio
 *
 */
public class SampleReadTaxonomyCreateConceptsList {
	private static final String XHTMLNS = "http://www.w3.org/1999/xhtml";

	/**
	 * @param args
	 */
	public static void main(String[] args) throws Exception {
		if (args.length < 1)
			System.out.println("First parameter must be the DTS Entry point to load");
		
		String outputFile = "Concepts list.xhtml";
		
		/*
		 * Create a new empty DTS Container
		 */
		DTSContainer dts = DTSContainer.newEmptyContainer();
		
		/*
		 * Open the taxonomy DTS from a local file. The DTS is discovered now.
		 */
		URI uris[] = new URI[args.length];
		int i=0;
		for (String arg : args) {
			try {
				uris[i] = new URI(arg);
			} catch (URISyntaxException e) {
				uris[i] = new File(arg).toURI();
			}
			i++;
		}
		dts.load(uris);
				
		XBRLArcroleType concept_label_arcrole = dts.getArcroleType(LabelLinkbase.concept_label_arcrole_URI).firstElement();			 
		XBRLArcroleType concept_reference_arcrole = dts.getArcroleType(ReferenceLinkbase.concept_reference_arcrole_URI).firstElement();
		
		Hashtable<XBRLRoleType,Integer> uniqueLabelRoleTypes = new Hashtable<XBRLRoleType,Integer>();
		Hashtable<XBRLRoleType,Integer> uniqueReferenceRoleTypes = new Hashtable<XBRLRoleType,Integer>();
		Iterator<XMLElementDefinition> iterE = dts.getConcepts();
		while (iterE.hasNext()) {
			XMLElementDefinition el = iterE.next();
			// relationships:
			Iterator<XBRLRelationship> iterFR = el.getFromRelationships();
//			boolean bIsNewConcept = true;
			Hashtable<XBRLRoleType,Integer> uniqueLabelRoleTypesOnThisElement = new Hashtable<XBRLRoleType,Integer>();
			Hashtable<XBRLRoleType,Integer> uniqueReferenceRoleTypesOnThisElement = new Hashtable<XBRLRoleType,Integer>();
			while (iterFR.hasNext()) {
				XBRLRelationship rel = iterFR.next();				
				if (rel.getArcroleType().equals(concept_label_arcrole)) {
					XMLFragment to = rel.getTo();
					if (to instanceof XBRLResource) {
						XBRLRoleType rt = ((XBRLResource)to).getRoleType();
						if (rt == null)
							rt = dts.getRoleType(XBRLExtendedLink.standard_role_URI).firstElement();						
						Integer counter = uniqueLabelRoleTypesOnThisElement.get(rt);
						if (counter == null) {
							uniqueLabelRoleTypesOnThisElement.put(rt, 1);
						} else {
							++counter;
							uniqueLabelRoleTypesOnThisElement.put(rt, counter);
						}
					}
				} else
				if (rel.getArcroleType().equals(concept_reference_arcrole)) {
					XMLFragment to = rel.getTo();
					if (to instanceof XBRLResource) {
						XBRLRoleType rt = ((XBRLResource)to).getRoleType();
						if (rt == null)
							rt = dts.getRoleType(XBRLExtendedLink.standard_role_URI).firstElement();
						Integer counter = uniqueReferenceRoleTypesOnThisElement.get(rt);
						if (counter == null) {
							uniqueReferenceRoleTypesOnThisElement.put(rt, 1);
						} else {
							++counter;
							uniqueReferenceRoleTypesOnThisElement.put(rt, counter);
						}
					}
				}
			}
			for (Map.Entry<XBRLRoleType, Integer> ent : uniqueLabelRoleTypesOnThisElement.entrySet()) {
				Integer maxCounter = uniqueLabelRoleTypes.get(ent.getKey());
				if (maxCounter == null)
					uniqueLabelRoleTypes.put(ent.getKey(),ent.getValue());
				else {
					if (ent.getValue() > maxCounter) {
						uniqueLabelRoleTypes.put(ent.getKey(),ent.getValue());
					}
				}
			}
			for (Map.Entry<XBRLRoleType, Integer> ent : uniqueReferenceRoleTypesOnThisElement.entrySet()) {
				Integer maxCounter = uniqueReferenceRoleTypes.get(ent.getKey());
				if (maxCounter == null)
					uniqueReferenceRoleTypes.put(ent.getKey(),ent.getValue());
				else {
					if (ent.getValue() > maxCounter) {
						uniqueReferenceRoleTypes.put(ent.getKey(),ent.getValue());
					}
				}
			}
		}
		
		// The output will be sent to a JDom document
		Document page = new Document();
		page.setDocType(new DocType("html","-//W3C//DTD XHTML 1.0 Strict//EN","http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd"));
		Element html = new Element("html",XHTMLNS);
		page.addContent(html);
		Element head = new Element("head",XHTMLNS);
		html.addContent(head);
		Element title = new Element("title",XHTMLNS);
		title.setText("Table of concepts, labels and references, Prepared by Reporting Standard S.L. XBRL Software");
		head.addContent(title);
//		Element style = new Element("style",XHTMLNS);
//		head.addContent(style);
//		Attribute styleType = new Attribute("type","text/css");
//		style.setAttribute(styleType);
//		style.setText("td.testOK { background: lightgreen; }\ntd.testFail { background: red; }");			
		Element body = new Element("body",XHTMLNS);
		html.addContent(body);
		Element pageH2 = new Element("h2",XHTMLNS);
		pageH2.setText("Table of concepts, labels and references, Prepared by Reporting Standard S.L. XBRL Software");
		body.addContent(pageH2);
		Date d = new Date(System.currentTimeMillis());
		SimpleDateFormat dtf = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSZ");
		Element header3 = new Element("h3",XHTMLNS);
		header3.setText("Date executed: "+dtf.format(d));
		body.addContent(header3);
		Element outTable = new Element("table",XHTMLNS);
		Attribute border1 = new Attribute("border","1");
		outTable.setAttribute(border1);
		body.addContent(outTable);
		
		Element tr00 = new Element("tr",XHTMLNS);
		outTable.addContent(tr00);
		
		Element th0 = new Element("th",XHTMLNS);
		tr00.addContent(th0);
		th0.setAttribute("style", "background: orange");
		th0.setText("Element namespace");
		
		Element th00 = new Element("th",XHTMLNS);
		tr00.addContent(th00);
		th00.setAttribute("style", "background: orange");
		th00.setText("Element name (XBRL)");
		
		Element th01 = new Element("th",XHTMLNS);
		tr00.addContent(th01);
		th01.setAttribute("style", "background: orange");
		th01.setText("Element prefix");
		
		for (Map.Entry<XBRLRoleType,Integer> ent : uniqueLabelRoleTypes.entrySet()) {
			for (int col = 0; col < ent.getValue(); col++) {
				Element thlX = new Element("th",XHTMLNS);
				tr00.addContent(thlX);
				thlX.setAttribute("style", "background: yellow");
				thlX.setText("Label for role "+ent.getKey().getRoleAndArcroleURI());
			}
		}
		for (Map.Entry<XBRLRoleType,Integer> ent : uniqueReferenceRoleTypes.entrySet()) {
			for (int col = 0; col < ent.getValue(); col++) {
				Element thlX = new Element("th",XHTMLNS);
				tr00.addContent(thlX);
				thlX.setAttribute("style", "background: lightblue");
				thlX.setText("Label for reference "+ent.getKey().getRoleAndArcroleURI());
			}
		}
		
		// show data
		iterE = dts.getConcepts();
		while (iterE.hasNext()) {
			XMLElementDefinition el = iterE.next();
			
			Element tr01 = new Element("tr",XHTMLNS);
			outTable.addContent(tr01);
			
			Element td00 = new Element("td",XHTMLNS);
			tr01.addContent(td00);
			td00.setText(el.getQName().getNamespaceURI());
			
			Element td01 = new Element("td",XHTMLNS);
			tr01.addContent(td01);
			td01.setText(el.getName());
			
			Element td02 = new Element("td",XHTMLNS);
			tr01.addContent(td02);
			td02.setText(el.getQName().getPrefix());
			
			for (Map.Entry<XBRLRoleType,Integer> ent : uniqueLabelRoleTypes.entrySet()) {
				for (int col = 0; col < ent.getValue(); col++) {					
					Element thlX = new Element("td",XHTMLNS);
					tr01.addContent(thlX);
					Iterator<XBRLRelationship> iterR = el.getFromRelationships(concept_label_arcrole);
					int onCol = 0;
					boolean bDone = false;
					while (iterR.hasNext() && !bDone) {
						XBRLRelationship rel = iterR.next();
						XMLFragment to = rel.getTo();
						if (to instanceof XBRLResource) {
							XBRLRoleType rt = ((XBRLResource) to).getRoleType();
							if (rt == null)
								rt = dts.getRoleType(XBRLExtendedLink.standard_role_URI).firstElement();
							if (rt.equals(ent.getKey())) {
								if (onCol < col) {
									onCol++;
									continue; // find next relationships for the same role
								}
								thlX.setText(to.getStringValue());
								bDone = true;
							}
						}
					}
				}
			}
			
			for (Map.Entry<XBRLRoleType,Integer> ent : uniqueReferenceRoleTypes.entrySet()) {
				for (int col = 0; col < ent.getValue(); col++) {					
					Element thlX = new Element("td",XHTMLNS);
					tr01.addContent(thlX);								
					Iterator<XBRLRelationship> iterR = el.getFromRelationships(concept_reference_arcrole);
					int onCol = 0;
					boolean bDone = false;
					while (iterR.hasNext() && !bDone) {
						XBRLRelationship rel = iterR.next();
						XMLFragment to = rel.getTo();
						if (to instanceof XBRLResource) {
							XBRLRoleType rt = ((XBRLResource) to).getRoleType();
							if (rt == null)
								rt = dts.getRoleType(XBRLExtendedLink.standard_role_URI).firstElement();
							if (rt.equals(ent.getKey())) {
								if (onCol < col) {
									onCol++;
									continue; // find next relationships for the same role
								}
								thlX.setText(to.getStringValue());								
								bDone = true;
							}
						}
					}					
				}
			}
			

		}
		
		// produce the output to stdout or file
		Format ft = Format.getPrettyFormat();
		ft.setEncoding("ISO-8859-1");
		XMLOutputter outputter = new XMLOutputter(ft);
		String s = outputter.outputString(page);
		if (outputFile != null) {
			File f = new File(outputFile);
			FileWriter fw = new FileWriter(f);
			fw.write(s);
			fw.flush();
			fw.close();
		}
//		else {
//			System.out.println(s);
//		}
	}

}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
