/*
 *  2009 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2009 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 11/08/2009
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.io.File;

import javax.xml.namespace.QName;

import net.sf.saxon.om.NamespaceBinding;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRL;
import com.ihr.xbrl.om.taxonomy.XBRLTuple;
import com.ihr.xbrl.om.taxonomy.XSDComposition;
import com.ihr.xbrl.om.taxonomy.XSDElementRef;
import com.ihr.xbrl.om.taxonomy.XSDElementWildcard;
import com.ihr.xbrl.om.taxonomy.XSDParticle;
import com.ihr.xbrl.util.XmlFormater;

/**
 * @author Ignacio
 *
 */
public class SampleReadTupleType {
	
	public static final String targetNamespace = "http://www.example.com/newTaxonomyWithTuple";
	
	/**
	 * @param args
	 */
	public static void main(String[] args) throws Exception {
        DTSContainer dts = DTSContainer.newEmptyContainer();            
        dts.load(new File("taxonomyWithTuple2.xsd").toURI());
        XBRLTuple t = (XBRLTuple)dts.getConcept(new QName(targetNamespace,"TestTuple"));
        XSDComposition comp = t.getComposition();
        
        // Output 1 - generate the output from the internal structures
        NamespaceBinding xsNs = new NamespaceBinding("", XBRL.XML_schemaNS);
        System.out.println(new String(XmlFormater.format(comp.generateElement(dts, xsNs))));
        
        // Output 2 - traverse the exposed interface of internal structures
        System.out.println(printComposition(comp, 0));
        
		System.exit(0);
	}

	/**
	 * @param comp
	 */
	private static String printComposition(XSDComposition comp,int indent) {
		StringBuffer sb = new StringBuffer();
		for (int i=0; i<indent; ++i)
			sb.append(' ');
		switch (comp.getType()) {
		case XSDComposition.ALL:
			sb.append("Composition ALL");
			break;
		case XSDComposition.CHOICE:
			sb.append("Composition CHOICE");
			break;
		case XSDComposition.SEQUENCE:
			sb.append("Composition SEQUENCE");
			break;
		}
		sb.append(" minOccurs=");
		sb.append(comp.getMinoccurs());
		sb.append(" maxOccurs=");
		int max = comp.getMaxoccurs();
		sb.append(max == XSDComposition.UNBOUNDED ? "Unbounded" : max);
		
		sb.append("\n");
		for (int i=0; i<indent; ++i)
			sb.append(' ');
		sb.append("Content:\n");
		
		for (XSDParticle particle: comp) {
			if (particle instanceof XSDElementRef) {
				XSDElementRef ref = (XSDElementRef)particle;
				for (int i=0; i<indent; ++i)
					sb.append(' ');
				sb.append("Element Ref ");
				sb.append(ref.getElement().getName());
				sb.append(" minOccurs=");
				sb.append(ref.getMinoccurs());
				sb.append(" maxOccurs=");
				max = ref.getMaxoccurs();
				sb.append(max == XSDComposition.UNBOUNDED ? "Unbounded" : max);
			} else if (particle instanceof XSDComposition) {
				sb.append(printComposition((XSDComposition)particle, 2+indent));
			} else if (particle instanceof XSDElementWildcard) {
				// not imlemented
			}
			sb.append("\n");
		}
		
		return sb.toString();
	}

	
}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
