/*
 *  2009 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2009 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 26/06/2009
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.lang.management.ManagementFactory;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.DTSContainer.Status;
import com.ihr.xbrl.om.XBRLEvent;
import com.ihr.xbrl.om.instance.XBRLInstance;

/**
 * Sample, reads and validates all inline Xbrl files from a directory
 * 
 * Discovery time:
 * Time to load the source file into the DTSContainer. This time includes conversion from IXBRL to XBRL
 *
 * Load time:
 * This is the time required to load all related files into the DTSContainer. Taxonomy files and linkbase files
 *
 * Validation time:
 * This is the time required to execute the validation chain. This is the entire chain of processors defined in the
 * validation chain. Actually this is XBRL 2.1 processor only. In order to include additional processors it is required
 * to add properties to line 266 of this code DTSContainer dts = DTSContainer.newEmptyContainer(null); and set the
 * DTSContainer.PROCESSORS_SEQUENCE to the list of processors to use.
 *
 * Total time:
 * This is just adding up previous values
 *
 * Conversion time:
 * This is the time required to convert IXBRL to XBRL. this time is already included in the Discovery time.
 */
public class SampleReadiXbrl {

	private static boolean bMultithreadMode = false;
	private static final String fileExtension = ".html"; //".xml";
	//private static File directory = new File("C:\\xbrl\\Accounts_Bulk_Data-2017-04-29\\");
	private static File directory = new File("ixbrl-tests");

	private static final String sSep = "|";
	private static List<String> resultLines;
	
	private static ExecutorService exs;
	
	private static class ChronometerListener implements PropertyChangeListener {

		File fd;
		int i;
		
		// esta es la secuencia esperada, el resto se ignoran
		long t1_DISCOVERY = 0L;
		long t2_IXBRL_START = 0L;
		long t3_IXBRL_END = 0L;
		long t4_DISCOVERED = 0L;
		long t5_SCHEMA_VALIDATION = 0L;
		long t6_INFOSET = 0L;
		long t7_LOADED = 0L;
		long t8_VALIDATION = 0L;
		long t9_LOADED = 0L;
		
		public ChronometerListener(File fd, int i) {
			this.fd = fd;
			this.i = i;
		}

		@Override
		public void propertyChange(PropertyChangeEvent evt) {
			if (evt.getPropertyName().equals(XBRLEvent.STATUS_CHANGED)) {
				Status st = (Status)evt.getNewValue();
				if (st.equals(Status.ERROR)) {
					t1_DISCOVERY = 0L;
					//System.out.println(fd.getName()+" ERROR");
				} else if (st.equals(Status.DISCOVERY)) {
					t1_DISCOVERY = System.currentTimeMillis();
					//System.out.println(fd.getName()+" DISCOVERY");
				} else if (st.equals(Status.DISCOVERED)) {
					t4_DISCOVERED = System.currentTimeMillis();
					//System.out.println(fd.getName()+" DISCOVERED");
				} else if (st.equals(Status.SCHEMAVALIDATION)) {
					t5_SCHEMA_VALIDATION = System.currentTimeMillis();
					//System.out.println(fd.getName()+" SCHEMAVALIDATION");
				} else if (st.equals(Status.INFOSET)) {
					t6_INFOSET = System.currentTimeMillis();
					//System.out.println(fd.getName()+" INFOSET");
				} else if (st.equals(Status.LOADED)) {
					if (t7_LOADED == 0)
						t7_LOADED = System.currentTimeMillis();
					else
						t9_LOADED = System.currentTimeMillis();
					//System.out.println(fd.getName()+" LOADED");
				}
				else if (st.equals(Status.VALIDATION)) {
					t8_VALIDATION = System.currentTimeMillis();
					//System.out.println(fd.getName()+" VALIDATION");
				}
			} else if (evt.getPropertyName().equals(XBRLEvent.IXBRL_START)) {
				t2_IXBRL_START = System.currentTimeMillis();
				//System.out.println(fd.getName()+" IXBRL_START");
			} else if (evt.getPropertyName().equals(XBRLEvent.IXBRL_END)) {
				t3_IXBRL_END = System.currentTimeMillis();
				//System.out.println(fd.getName()+" IXBRL_END");
			}
		}
		
		void println() {
			if (t1_DISCOVERY == 0L || t4_DISCOVERED == 0L || t5_SCHEMA_VALIDATION == 0L ||
				t6_INFOSET == 0L || t7_LOADED == 0L || t8_VALIDATION == 0L || t9_LOADED == 0L)
				return;
			
			StringBuffer sbResultLine = new StringBuffer();
			sbResultLine.append(String.valueOf(i));
			sbResultLine.append(sSep);
			sbResultLine.append(fd.getName());
			sbResultLine.append(sSep);
			sbResultLine.append(fd.length());
			sbResultLine.append(sSep);
			sbResultLine.append(t9_LOADED-t1_DISCOVERY);  // Total time
			sbResultLine.append(sSep);
			sbResultLine.append(t4_DISCOVERED-t1_DISCOVERY);  // Discovery time
			sbResultLine.append(sSep);
			sbResultLine.append(t7_LOADED-t4_DISCOVERED);  // Load time
			sbResultLine.append(sSep);
			sbResultLine.append(t9_LOADED-t8_VALIDATION);  // Validation time
			sbResultLine.append(sSep);
			sbResultLine.append(t3_IXBRL_END-t2_IXBRL_START);  // Conversion to XBRL time
			sbResultLine.append(" Thread ID: " +  Thread.currentThread().getId());
			System.out.println(sbResultLine.toString());
			resultLines.add(sbResultLine.toString());
		}
	}
	
//	private static String getWorkstationId() throws UnknownHostException {
//        InetAddress ip = InetAddress.getLocalHost();
//        return ip.getHostName();
//	}
//	
	/**
	 * @param args
	 * @throws Throwable 
	 */
	public static void main(String[] args) throws Throwable {		
		resultLines = new ArrayList<String>();
//		String workstationId = getWorkstationId();
    	long memorySize = ((com.sun.management.OperatingSystemMXBean) ManagementFactory.getOperatingSystemMXBean()).getTotalPhysicalMemorySize();
    	memorySize /= 1000L;  // KBytes
    	memorySize /= 1000L;  // MBytes
    	memorySize /= 1000L;  // GBytes
    	long tStart = 0L;
		long tEnd = 0L;
		tStart = System.currentTimeMillis();
		
    	StringBuffer sbExtraInfo = new StringBuffer("#");
    	sbExtraInfo.append("JAVA_VERSION=");
    	sbExtraInfo.append(Runtime.class.getPackage().getImplementationVersion());
    	sbExtraInfo.append(", ");
    	sbExtraInfo.append("PROCESSOR_IDENTIFIER=");
    	sbExtraInfo.append(System.getenv("PROCESSOR_IDENTIFIER"));
    	sbExtraInfo.append(", ");
    	sbExtraInfo.append("PROCESSOR_ARCHITECTURE=");
    	sbExtraInfo.append(System.getenv("PROCESSOR_ARCHITECTURE"));
    	sbExtraInfo.append(", ");
    	sbExtraInfo.append("NUMBER_OF_PROCESSORS=");
    	int maxThreads = Runtime.getRuntime().availableProcessors();
    	maxThreads = 10;
    	if (maxThreads >Runtime.getRuntime().availableProcessors() || maxThreads<1)
    		maxThreads = Runtime.getRuntime().availableProcessors();
    	System.out.println("# of Threads used:"+maxThreads);
    	exs = Executors.newFixedThreadPool(maxThreads);
    	sbExtraInfo.append(maxThreads);
//    	sbExtraInfo.append(", ");
//    	sbExtraInfo.append("MAC_ADDRESS=");
//    	sbExtraInfo.append(workstationId);
    	sbExtraInfo.append(", ");
    	sbExtraInfo.append("TOTAL_PHYSICAL_MEMORY=");
    	sbExtraInfo.append(memorySize);
    	sbExtraInfo.append(" GB");
		resultLines.add(sbExtraInfo.toString());

		resultLines.add("#All times in milliseconds and file length in bytes");

		StringBuffer sbResultLine = new StringBuffer("#");
		sbResultLine.append("Number");
		sbResultLine.append(sSep);
		sbResultLine.append("Filename");
		sbResultLine.append(sSep);
		sbResultLine.append("Length");
		sbResultLine.append(sSep);
		sbResultLine.append("Total time");
		sbResultLine.append(sSep);
		sbResultLine.append("Discovery time");
		sbResultLine.append(sSep);
		sbResultLine.append("Load time");
		sbResultLine.append(sSep);
		sbResultLine.append("Validation time");
		sbResultLine.append(sSep);
		sbResultLine.append("Conversion time");
		resultLines.add(sbResultLine.toString());
		
		int i = 1;
		processDirectory(directory, i);

		exs.shutdown();
		exs.awaitTermination(Long.MAX_VALUE, TimeUnit.DAYS);

//		Path file = Paths.get(directory.getAbsolutePath(), "SampleReadIXbrl_"+workstationId+".txt");
//		Files.write(file, resultLines, Charset.forName("UTF-8"));
		tEnd = System.currentTimeMillis();
		System.out.println("Overall process time: "+(tEnd - tStart));
		System.out.println("AVG time per file: "+((tEnd - tStart)/ 10));
		System.out.println("AVG time per Thread: "+((tEnd - tStart)/ maxThreads));
		
	}

	private static void processDirectory(File directory, int i) {
		File[] dirFiles = directory.listFiles();
		if (dirFiles != null)
			for (final File fd : dirFiles) {
				if ((fd.getName().endsWith(fileExtension))) {
					try {
						if (bMultithreadMode) {
							final int thisTestNumner = i;
							exs.execute(new Runnable() {
								@Override
								public void run() {
									try {
										processIXbrlFile(fd, thisTestNumner);
									} catch (Throwable e) {
										e.printStackTrace();
									}
								}});
						} else {
							processIXbrlFile(fd, i);
						}
					} catch (Throwable e) {
						e.printStackTrace();
					} finally {
						i++;
					}
				}
			}
	}

	private static void processIXbrlFile(File fd, int i) throws Throwable {
		ChronometerListener chrono = new ChronometerListener(fd, i);
		DTSContainer dts = DTSContainer.newEmptyContainer(null);
		dts.addPropertyChangeListener(XBRLEvent.STATUS_CHANGED, chrono);
		dts.addPropertyChangeListener(XBRLEvent.IXBRL_START, chrono);
		dts.addPropertyChangeListener(XBRLEvent.IXBRL_END, chrono);
		//dts.log.setLevel(Level.INFO);
		XBRLInstance instance = (XBRLInstance)dts.load(fd.toURI());
		/*XBRLValidationReport rep = */dts.validate(instance);
//		if (rep.getExceptionsCount() > 0)
//			for (XBRLValidationException ex : GeneralIterable.asIterator(rep.getExceptions()))
//				System.out.println("Excepcin de validacin: \n " + ex.getMessage() +" \n");
//		if (rep.getInconsistenciesCount() > 0)
//			for (XBRLInconsistency inconsistency : GeneralIterable.asIterator(rep.getInconsistencies()))
//				System.out.println("Mensaje inconsistencia: " + inconsistency.getLocalizedMessage());
		chrono.println();
		
		if (fd.getName().endsWith(".html"))
			instance.save();
	}
}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
