/*
 *  2009 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2009 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 26/06/2009
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.io.File;

import org.apache.log4j.Level;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRLValidationReport;
import com.ihr.xbrl.om.errors.XBRLInconsistency;
import com.ihr.xbrl.om.errors.XBRLValidationException;
import com.ihr.xbrl.om.instance.XBRLInstance;
import com.ihr.xbrl.util.GeneralIterable;

/**
 * Sample, reads all iXbrl files from a directory
 */
public class SampleReadiXbrl2 {

	private static final String fExtension = ".html";
	//private static final String fExtension = ".xml";
	
	/**
	 * @param args
	 * @throws Throwable 
	 */
	public static void main(String[] args) throws Throwable {		
		File d = new File("ixbrl-tests");
		processDirectory(d);
	}

	private static void processDirectory(File d) {
		int i = 1;
		File[] dirFiles = d.listFiles();
		if (dirFiles != null)
			for (File fd : dirFiles) {
				if (!fd.isDirectory()) {
//					processDirectory(fd);
//				else {
					if ((fd.getName().endsWith(fExtension)))
						try {
							processIXbrlFile(fd, i++);
						} catch (Throwable e) {
							e.printStackTrace();
						}
				}
			}
	}

	private static void processIXbrlFile(File fd, int i) throws Throwable {
		System.out.println(i+" processIXbrlFile: "+fd.getAbsolutePath());
		DTSContainer dts = DTSContainer.newEmptyContainer(null);
		dts.log.setLevel(Level.INFO);
		XBRLInstance instance = (XBRLInstance)dts.load(fd.toURI());
		XBRLValidationReport rep = dts.validate(instance);
		if (rep.getExceptionsCount() > 0) {
			for (XBRLValidationException ex : GeneralIterable.asIterator(rep.getExceptions()))
				//if (!(ex instanceof XDTValidationException))
				System.out.println("Excepcin de validacin: \n " + ex.getMessage() +" \n");
		}
		if (rep.getInconsistenciesCount() > 0)
			for (XBRLInconsistency inconsistency : GeneralIterable.asIterator(rep.getInconsistencies()))
				System.out.println("Mensaje inconsistencia:  " + inconsistency.getLocalizedMessage());
		System.out.println(i+" processIXbrlFile: "+fd.getAbsolutePath()+" END");
	}
}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
