/*
 *  2009 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2009 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 26/06/2009
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.lang.management.ManagementFactory;
import java.net.URI;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import com.ihr.xbrl.om.ixbrl.IXBRLTransformerFactory;

/**
 * Sample, reads and validates all inline Xbrl files from a directory
 * 
 * Discovery time:
 * Es el tiempo que se tarda en la carga y descubrimiento del DTS del informe (incluye la conversin de IXBRL a XBRL).
 * Es el primer paso. El siguiente paso es crear y cargar el informe XBRL propiamente dicho.
 *
 * Load time:
 * Es el tiempo que se tarda en crear y cargar el informe XBRL. Es el segundo paso. El siguiente paso es la validacin XBRL.
 *
 * Validation time:
 * Es el tiempo que se tarda en validar el XBRL. Es el tercer paso.
 *
 * Total time:
 * Es la suma de los tres tiempos anteriores
 *
 * Conversion time:
 * Es el tiempo que se tarda en convertir el IXBRL en XBRL. Este tiempo est incluido en el primer paso, es decir,
 * en el "Discovery time" y a su vez contiene el tiempo de validacin IXBRL que piden para el informe. De momento no podemos
 * disociar este tiempo.
 */
public class SampleReadiXbrl3 {

	private static boolean bMultithreadMode = false;
	private static final String fileExtension = ".html"; //".xml";
	//private static File directory = new File("C:\\xbrl\\Accounts_Bulk_Data-2017-04-29\\");
	private static String inDirectory = "inline-1.1-conformance";
	private static File directory = new File(inDirectory);
	private static String outDirectory = "xbrl-temp-xslt-1.1";

	private static final String sSep = "|";
	private static List<String> resultLines;
	
	private static ExecutorService exs;
	
//	private static class ChronometerListener implements PropertyChangeListener {
//
//		File fd;
//		int i;
//		
//		// esta es la secuencia esperada, el resto se ignoran
//		long t1_DISCOVERY = 0L;
//		long t2_IXBRL_START = 0L;
//		long t3_IXBRL_END = 0L;
//		long t4_DISCOVERED = 0L;
//		long t5_SCHEMA_VALIDATION = 0L;
//		long t6_INFOSET = 0L;
//		long t7_LOADED = 0L;
//		long t8_VALIDATION = 0L;
//		long t9_LOADED = 0L;
//		
//		public ChronometerListener(File fd, int i) {
//			this.fd = fd;
//			this.i = i;
//		}
//
//		@Override
//		public void propertyChange(PropertyChangeEvent evt) {
//			if (evt.getPropertyName().equals(XBRLEvent.STATUS_CHANGED)) {
//				Status st = (Status)evt.getNewValue();
//				if (st.equals(Status.ERROR)) {
//					t1_DISCOVERY = 0L;
//					//System.out.println(fd.getName()+" ERROR");
//				} else if (st.equals(Status.DISCOVERY)) {
//					t1_DISCOVERY = System.currentTimeMillis();
//					//System.out.println(fd.getName()+" DISCOVERY");
//				} else if (st.equals(Status.DISCOVERED)) {
//					t4_DISCOVERED = System.currentTimeMillis();
//					//System.out.println(fd.getName()+" DISCOVERED");
//				} else if (st.equals(Status.SCHEMAVALIDATION)) {
//					t5_SCHEMA_VALIDATION = System.currentTimeMillis();
//					//System.out.println(fd.getName()+" SCHEMAVALIDATION");
//				} else if (st.equals(Status.INFOSET)) {
//					t6_INFOSET = System.currentTimeMillis();
//					//System.out.println(fd.getName()+" INFOSET");
//				} else if (st.equals(Status.LOADED)) {
//					if (t7_LOADED == 0)
//						t7_LOADED = System.currentTimeMillis();
//					else
//						t9_LOADED = System.currentTimeMillis();
//					//System.out.println(fd.getName()+" LOADED");
//				}
//				else if (st.equals(Status.VALIDATION)) {
//					t8_VALIDATION = System.currentTimeMillis();
//					//System.out.println(fd.getName()+" VALIDATION");
//				}
//			} else if (evt.getPropertyName().equals(XBRLEvent.IXBRL_START)) {
//				t2_IXBRL_START = System.currentTimeMillis();
//				//System.out.println(fd.getName()+" IXBRL_START");
//			} else if (evt.getPropertyName().equals(XBRLEvent.IXBRL_END)) {
//				t3_IXBRL_END = System.currentTimeMillis();
//				//System.out.println(fd.getName()+" IXBRL_END");
//			}
//		}
//		
//		void println() {
//			if (t1_DISCOVERY == 0L || t4_DISCOVERED == 0L || t5_SCHEMA_VALIDATION == 0L ||
//				t6_INFOSET == 0L || t7_LOADED == 0L || t8_VALIDATION == 0L || t9_LOADED == 0L)
//				return;
//			
//			StringBuffer sbResultLine = new StringBuffer();
//			sbResultLine.append(String.valueOf(i));
//			sbResultLine.append(sSep);
//			sbResultLine.append(fd.getName());
//			sbResultLine.append(sSep);
//			sbResultLine.append(fd.length());
//			sbResultLine.append(sSep);
//			sbResultLine.append(t9_LOADED-t1_DISCOVERY);  // Total time
//			sbResultLine.append(sSep);
//			sbResultLine.append(t4_DISCOVERED-t1_DISCOVERY);  // Discovery time
//			sbResultLine.append(sSep);
//			sbResultLine.append(t7_LOADED-t4_DISCOVERED);  // Load time
//			sbResultLine.append(sSep);
//			sbResultLine.append(t9_LOADED-t8_VALIDATION);  // Validation time
//			sbResultLine.append(sSep);
//			sbResultLine.append(t3_IXBRL_END-t2_IXBRL_START);  // Conversion to XBRL time
//			sbResultLine.append(" Thread ID: " +  Thread.currentThread().getId());
//			System.out.println(sbResultLine.toString());
//			resultLines.add(sbResultLine.toString());
//		}
//	}
	
//	private static String getWorkstationId() throws UnknownHostException {
//        InetAddress ip = InetAddress.getLocalHost();
//        return ip.getHostName();
////        
////        NetworkInterface network = NetworkInterface.getByInetAddress(ip);
////        byte[] mac = network.getHardwareAddress();
////        StringBuilder sbMacAddress = new StringBuilder();
////        for (int i = 0; i < mac.length; i++)
////        	sbMacAddress.append(String.format("%02X%s", mac[i], (i < mac.length - 1) ? "-" : ""));     
////		return sbMacAddress.toString();
//	}
	
	/**
	 * @param args
	 * @throws Throwable 
	 */
	public static void main(String[] args) throws Throwable {		
		resultLines = new ArrayList<String>();
//		String workstationId = getWorkstationId();
    	long memorySize = ((com.sun.management.OperatingSystemMXBean) ManagementFactory.getOperatingSystemMXBean()).getTotalPhysicalMemorySize();
    	memorySize /= 1000L;  // KBytes
    	memorySize /= 1000L;  // MBytes
    	memorySize /= 1000L;  // GBytes
    	long tStart = 0L;
		long tEnd = 0L;
		tStart = System.currentTimeMillis();
		
    	StringBuffer sbExtraInfo = new StringBuffer("#");
    	sbExtraInfo.append("JAVA_VERSION=");
    	sbExtraInfo.append(Runtime.class.getPackage().getImplementationVersion());
    	sbExtraInfo.append(", ");
    	sbExtraInfo.append("PROCESSOR_IDENTIFIER=");
    	sbExtraInfo.append(System.getenv("PROCESSOR_IDENTIFIER"));
    	sbExtraInfo.append(", ");
    	sbExtraInfo.append("PROCESSOR_ARCHITECTURE=");
    	sbExtraInfo.append(System.getenv("PROCESSOR_ARCHITECTURE"));
    	sbExtraInfo.append(", ");
    	sbExtraInfo.append("NUMBER_OF_PROCESSORS=");
    	int maxThreads = Runtime.getRuntime().availableProcessors();
    	maxThreads = 10;
    	if (maxThreads >Runtime.getRuntime().availableProcessors() || maxThreads<1)
    		maxThreads = Runtime.getRuntime().availableProcessors();
    	System.out.println("# of Threads used:"+maxThreads);
    	exs = Executors.newFixedThreadPool(maxThreads);
    	sbExtraInfo.append(maxThreads);
//    	sbExtraInfo.append(", ");
//    	sbExtraInfo.append("MAC_ADDRESS=");
//    	sbExtraInfo.append(workstationId);
    	sbExtraInfo.append(", ");
    	sbExtraInfo.append("TOTAL_PHYSICAL_MEMORY=");
    	sbExtraInfo.append(memorySize);
    	sbExtraInfo.append(" GB");
		resultLines.add(sbExtraInfo.toString());

		resultLines.add("#All times in milliseconds and file length in bytes");

		StringBuffer sbResultLine = new StringBuffer("#");
		sbResultLine.append("Number");
		sbResultLine.append(sSep);
		sbResultLine.append("Filename");
		sbResultLine.append(sSep);
		sbResultLine.append("Length");
		sbResultLine.append(sSep);
		sbResultLine.append("Total time");
		sbResultLine.append(sSep);
		sbResultLine.append("Discovery time");
		sbResultLine.append(sSep);
		sbResultLine.append("Load time");
		sbResultLine.append(sSep);
		sbResultLine.append("Validation time");
		sbResultLine.append(sSep);
		sbResultLine.append("Conversion time");
		resultLines.add(sbResultLine.toString());
		
		processDirectory(directory, 1);

		exs.shutdown();
		exs.awaitTermination(Long.MAX_VALUE, TimeUnit.DAYS);

//		Path file = Paths.get(directory.getAbsolutePath(), "SampleReadIXbrl_"+workstationId+".txt");
//		Files.write(file, resultLines, Charset.forName("UTF-8"));
		tEnd = System.currentTimeMillis();
		System.out.println("Overall process time: "+(tEnd - tStart));
		System.out.println("AVG time per file: "+((tEnd - tStart)/ 10));
		System.out.println("AVG time per Thread: "+((tEnd - tStart)/ maxThreads));
		
	}

	private static int processDirectory(File directory, int i) {
		File[] dirFiles = directory.listFiles();
		if (dirFiles != null)
			for (final File fd : dirFiles) {
				if (fd.isDirectory())
					i = processDirectory(fd, i);
				else {
					if ((fd.getName().endsWith(fileExtension))) {
						try {
							if (bMultithreadMode) {
								final int thisTestNumner = i;
								exs.execute(new Runnable() {
									@Override
									public void run() {
										try {
											processIXbrlFile(fd, thisTestNumner);
										} catch (Throwable e) {
//											e.printStackTrace();
										}
									}});
							} else {
								processIXbrlFile(fd, i);
							}
						} catch (Throwable e) {
							//System.err.println(e.getMessage());
						} finally {
							i++;
						}
					}
				}
			}
		return i;
	}

	private static void processIXbrlFile(File file, int i) throws Throwable {
//		ChronometerListener chrono = new ChronometerListener(fd, i);
//		DTSContainer dts = DTSContainer.newEmptyContainer(null);
//		dts.addPropertyChangeListener(XBRLEvent.STATUS_CHANGED, chrono);
//		dts.addPropertyChangeListener(XBRLEvent.IXBRL_START, chrono);
//		dts.addPropertyChangeListener(XBRLEvent.IXBRL_END, chrono);
//		//dts.log.setLevel(Level.INFO);
//		XBRLInstance instance = (XBRLInstance)dts.load(fd.toURI());
//		/*XBRLValidationReport rep = */dts.validate(instance);
////		if (rep.getExceptionsCount() > 0)
////			for (XBRLValidationException ex : GeneralIterable.asIterator(rep.getExceptions()))
////				System.out.println("Excepcin de validacin: \n " + ex.getMessage() +" \n");
////		if (rep.getInconsistenciesCount() > 0)
////			for (XBRLInconsistency inconsistency : GeneralIterable.asIterator(rep.getInconsistencies()))
////				System.out.println("Mensaje inconsistencia: " + inconsistency.getLocalizedMessage());
//		chrono.println();
		
		System.out.println("Processing: "+file.getName());
		boolean isError = file.getName().startsWith("FAIL");
		IXBRLTransformerFactory factory = new IXBRLTransformerFactory();
		Transformer tr = factory.newTransformer();
		FileReader fr = new FileReader(file);
		Source ss = new StreamSource(fr);
		ss.setSystemId(file.toURI().toString());		
		Path outPath = Paths.get(URI.create(file.toURI().toString().replace(inDirectory, outDirectory).replace(".html",".xbrl")));
		outPath.toFile().getParentFile().mkdirs();
		FileOutputStream fout = new FileOutputStream(outPath.toFile());
		Result rr = new StreamResult(fout);
		try {
			tr.transform(ss, rr);
			if (!isError)
				System.out.println("OK - Check output file!");
			else
				System.out.println("ERROR - Error not detected");
		} catch (TransformerException e) {
			if (isError)
				System.out.println("OK - Error detected : "+e.getMessage());
			else
				System.out.println("ERROR - Should be OK : "+e.getMessage());
		}
		
		
//		if (fd.getName().endsWith(".html"))
//			instance.save();
	}
}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
