/*
 *  2010 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2010 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 09/03/2010
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.Iterator;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.errors.DTSDiscoveryException;
import com.rs.lic.LicenseVerificationException;

/**
 * Load a DTS and report all files in the DTS after DTS discovery rules has been executed
 * 
 * @author Ignacio
 *
 */
public class SampleReportFilesInDTS {

	/**
	 * @param args
	 * @throws LicenseVerificationException 
	 * @throws DTSDiscoveryException 
	 * @throws URISyntaxException 
	 * @throws MalformedURLException 
	 */
	public static void main(String[] args) throws DTSDiscoveryException, LicenseVerificationException, URISyntaxException, MalformedURLException {

		if (args.length == 0) {
			System.err.println("Uso: SampleReportFiles {URL Starting Point} [{URL Starting Point} ...]");
			System.exit(-1);
		}
		
		URI startingPoints[] = new URI[args.length];
		int i=0;
		for (String s : args) {
			if (s.startsWith("http:"))
				startingPoints[i] = (new URL(s)).toURI();
			else
				startingPoints[i] = new File(s).toURI();
			i++;
		}
		
		DTSContainer dts = DTSContainer.newEmptyContainer();
		dts.load(startingPoints);
		
		System.out.println("Entry point(s):");
		for (URI entryP : dts.getRootURIs()) {
			System.out.println(" * "+entryP.toString());
		}
		System.out.println("\nTaxonomies loaded in the DTS:");
		Iterator<URI> iterU1 = dts.getTaxonomyURIs();
		while (iterU1.hasNext()) {
			URI u = iterU1.next();
			System.out.println(" * "+u.toString());
		}
		
		System.out.println("\nLinkbases loaded in the DTS:");
		Iterator<URI> iterU2 = dts.getLinkbaseURIs();
		while (iterU2.hasNext()) {
			URI u = iterU2.next();
			System.out.println(" * "+u.toString());
		}
	}

}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
