/*
 *  2008 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2008 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 26/01/2008
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.net.URI;
import java.util.Iterator;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XMLFragment;
import com.ihr.xbrl.om.exLinks.FootnoteLinkbase;
import com.ihr.xbrl.om.exLinks.XBRLExtendedLink;
import com.ihr.xbrl.om.exLinks.XBRLRelationship;
import com.ihr.xbrl.om.exLinks.XBRLResource;
import com.ihr.xbrl.om.instance.XBRLInstance;
import com.ihr.xbrl.om.taxonomy.XBRLArcroleType;

/**
 * 
 * Demonstration about how an application can create an instance document
 * and include a footnote for a fact item.
 * 
 * @author Ignacio
 *
 */
public class SampleSeeFootnotesAndContinueDownToFacts {
	
	public static void main(String args[]) throws Exception {
		
		// Loads the sample taxonomy
		DTSContainer dts = DTSContainer.newEmptyContainer();
		URI uri = new URI("instance.xbrl");
		XBRLInstance instance = (XBRLInstance)dts.load(uri);
		
		// Obtain the arcrole type, it will be used later
		XBRLArcroleType art = dts.getArcroleType(FootnoteLinkbase.fact_footnote_arcrole_URI).firstElement();

		// Iterate through the footnote extended links
		Iterator<XBRLExtendedLink> iterL = instance.getDocumentRootNode().getExtendedLinks();
		while (iterL.hasNext()) {
			XBRLExtendedLink footnotesLink = iterL.next();

			// Now iterate through the resources. Note this will return also the 
			// orphaned footnotes if any exists
			Iterator<XBRLResource> iterR = footnotesLink.getResources();
			while (iterR.hasNext()) {
				XBRLResource footnote = iterR.next();
				
				// Obtains relationships where this footnote is target and
				// filter the result for fact-footnote arcrole type
				Iterator<XBRLRelationship> iterE = footnote.getToRelationships(art);
				if (!iterE.hasNext()) {
					System.out.println("Footnote \""+footnote.getStringValue()+"\" is not linked with facts.");
				} else {
					while (iterE.hasNext()) {
						XBRLRelationship rel = iterE.next();						
						XMLFragment fact = rel.getFrom();
						
						System.out.println("Footnote \""+footnote.getStringValue()+"\" is linked with fact \""+fact.getNodeName());
					}
				}
			}
		}		
		System.exit(0);
	}
}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
