/*
 *  2008 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2008 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 26/01/2008
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.net.URI;
import java.util.HashSet;
import java.util.Iterator;

import javax.xml.namespace.QName;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XMLFragment;
import com.ihr.xbrl.om.exLinks.DTSBase;
import com.ihr.xbrl.om.exLinks.FootnoteLinkbase;
import com.ihr.xbrl.om.exLinks.XBRLRelationship;
import com.ihr.xbrl.om.instance.XBRLFact;
import com.ihr.xbrl.om.instance.XBRLInstance;
import com.ihr.xbrl.om.taxonomy.XBRLArcroleType;
import com.ihr.xbrl.om.taxonomy.XBRLRoleType;
import com.ihr.xbrl.om.taxonomy.XBRLUsedOn;

/**
 * 
 * Demonstration about how an application can create an instance document
 * and include a footnote for a fact item.
 * 
 * @author Ignacio
 *
 */
public class SampleShowFactsAndFootnoteInformation {
	
	public static void main(String args[]) throws Exception {
		
		// Loads the sample taxonomy
		final DTSContainer dts = DTSContainer.newEmptyContainer();
		URI uri = new URI("instance.xbrl");
		XBRLInstance instance = (XBRLInstance)dts.load(uri);
		
		class ClsPackQNameAndRole {
			QName extendedLinkQname;
			HashSet<XBRLRoleType> vRoles = new HashSet<XBRLRoleType>();
			
			public Iterator<DTSBase>iterator() {
				return new ClsIteratorHelper();
			}
			
			class ClsIteratorHelper implements Iterator<DTSBase> {
				
				Iterator<XBRLRoleType> iterR = vRoles.iterator();

				/* (non-Javadoc)
				 * @see java.util.Iterator#hasNext()
				 */
				@Override
				public boolean hasNext() {
					return iterR.hasNext();
				}

				/* (non-Javadoc)
				 * @see java.util.Iterator#next()
				 */
				@Override
				public DTSBase next() {
					return dts.getBase(extendedLinkQname, iterR.next());
				}

				/* (non-Javadoc)
				 * @see java.util.Iterator#remove()
				 */
				@Override
				public void remove() {
					
				}
				
			}
		}
		
		final HashSet<ClsPackQNameAndRole> hs = new HashSet<ClsPackQNameAndRole>();
		XBRLArcroleType art = dts.getArcroleType(FootnoteLinkbase.fact_footnote_arcrole_URI).firstElement();

		Iterator<XBRLUsedOn> iterOn = art.getUsedOn();
		while (iterOn.hasNext()) {
			XBRLUsedOn on = iterOn.next();
			ClsPackQNameAndRole c = new ClsPackQNameAndRole();
			c.extendedLinkQname = on.getQName();
			
			Iterator<XBRLRoleType> iterR = dts.getBaseRoles(c.extendedLinkQname);
			while (iterR.hasNext())
				c.vRoles.add(iterR.next());
			
			hs.add(c);
		}
		
		class IterX implements Iterator<DTSBase> {
			
			Iterator<ClsPackQNameAndRole> iterC = hs.iterator();
			Iterator<DTSBase> current = null;
			boolean hasNext = false;
			DTSBase base = null;
			
			private IterX() {
				advance();
			}
			
			private void advance() {
				if (!iterC.hasNext()) {
					hasNext = false;
					return;
				}
				current = iterC.next().iterator();
				advance2();
			}
			
			private void advance2() {
				if (!current.hasNext()) {
					advance();
					return;
				}
				base = current.next();
				hasNext = true;
			}

			/* (non-Javadoc)
			 * @see java.util.Iterator#hasNext()
			 */
			@Override
			public boolean hasNext() {
				return hasNext;
			}

			/* (non-Javadoc)
			 * @see java.util.Iterator#next()
			 */
			@Override
			public DTSBase next() {
				DTSBase retVal = base;
				advance2();
				return retVal;
			}

			/* (non-Javadoc)
			 * @see java.util.Iterator#remove()
			 */
			@Override
			public void remove() {
				// TODO Auto-generated method stub
				
			}
			
		}
		
		if (hs.size() > 0) {
			Iterator<XBRLFact> iterF = instance.getInstanceRootNode().iterator();
			while (iterF.hasNext()) {
				XBRLFact fact = iterF.next();

				IterX iter = new IterX();
				while (iter.hasNext()) {
					Iterator<XBRLRelationship> iterR = fact.getFromRelationships(art, iter.next());
					while (iterR.hasNext()) {
						XBRLRelationship r = iterR.next();
						
						XMLFragment f = r.getTo();
						
						System.out.println("Fact \""+fact.getElementDefinition().getQName()+"\" has note \""+f.getStringValue()+"\"");
					}
				}
			}
		}
		
		System.exit(0);
	}
}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
