/*
 *  2010 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2010 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 13/05/2010
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.io.File;
import java.util.Iterator;
import java.util.Properties;
import java.util.Vector;

import com.ihr.xbrl.dts.LabelsProvider;
import com.ihr.xbrl.dts.LabelsProviderInstance;
import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.errors.DTSDiscoveryException;
import com.ihr.xbrl.om.errors.XBRLValidationException;
import com.ihr.xbrl.om.taxonomy.XMLElementDefinition;
import com.ihr.xbrl.util.GeneralIterable;
import com.rs.lic.LicenseVerificationException;

/**
 * This small application loads a DTS, then it shows presentation tree in all presentation extended links
 * 
 * @author Ignacio
 *
 */
public class SampleShowLabels {
	
	static Vector<String> languages = new Vector<String>();
	
	static {
		languages.add("en");
	}

	/**
	 * @param args
	 * @throws LicenseVerificationException 
	 * @throws DTSDiscoveryException 
	 * @throws XBRLValidationException 
	 */
	public static void main(String[] args) throws LicenseVerificationException, DTSDiscoveryException, XBRLValidationException {
		
		if (args.length == 0) {
			args = new String[1];
			args[0] = "new_taxonomy.xsd";
		}
		
		Properties props = new Properties();		
		DTSContainer dts = DTSContainer.newEmptyContainer(props);
		dts.load(new File(args[0]).toURI());
		
		// show all concept labels
		
		// On this example we will use the simple LabelsProvider. 
		// created by the API. 
		// There are more than one LabelsProvider and you can provide your own
		// and register it to the LabelsProviderInstance using
		// LabelsProviderInstance.setLabelsProvider(lp);
		// static method
		LabelsProvider lp = LabelsProviderInstance.getLabelsProvider();
		
		Iterator<XMLElementDefinition> iterEl = dts.getConcepts();
		for (XMLElementDefinition el : GeneralIterable.asIterator(iterEl)) {
			
			// There are 4 methods in the LabelsProvider interface.
			// lp.getLabel(QName conceptQName)
			// lp.getLabel(XBRLResource res)
			// lp.getLabel(XMLElementDefinition el)
			// lp.getLabel(XMLElementDefinition, XBRLRoleType preferredRole)
			System.out.println(lp.getLabel(el));
		}
	}

}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
