package com.ihr.xbrl.sample;

import java.net.URI;
import java.util.Iterator;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XMLFragment;
import com.ihr.xbrl.om.errors.DTSDiscoveryException;
import com.ihr.xbrl.om.exLinks.LabelLinkbase;
import com.ihr.xbrl.om.exLinks.XBRLRelationship;
import com.ihr.xbrl.om.taxonomy.XBRLArcroleType;
import com.ihr.xbrl.om.taxonomy.XMLElementDefinition;
import com.rs.lic.LicenseVerificationException;

/**
 * Extractor del cdigo XPointer de una etiqueta de un concepto
 * XBRL de una taxonoma.
 * 
 * TODO: En futuras versiones se puede modificar para que muestre todos los
 * XPoineters de las labels de un concepto o las referencias de un concepto.
 * 
 * @author Ignacio
 *
 */
public class XPointerX {

	/**
	 * @param args
	 * @throws LicenseVerificationException 
	 * @throws DTSDiscoveryException 
	 */
	public static void main(String[] args) throws DTSDiscoveryException, LicenseVerificationException {
		
		String etiqueta = "";
		
		// el primer parmetro ser la etiqueta a buscar.
		// acuerdense de ponerla entre comillas si tiene espacios
		if (args.length > 0)
			etiqueta = args[0];
		
		// Cargamos la taxonoma a partir del fichero CORE.
		// Esto incluye las etiquetas en espaol y las referencias, pero niguna link de presentacin
		URI txURI = URI.create("http://www.svs.cl/cl/fr/ci/2008-10-31/cl-ci_cor_2008-10-31.xsd");
		
		// cargamos la taxonoma en el modelo de objetos
		DTSContainer dts = DTSContainer.newEmptyContainer();
		dts.load(txURI);
		
		// Obtenemos el arcrole concept-label
		XBRLArcroleType concept_label = dts.getArcroleType(LabelLinkbase.concept_label_arcrole_URI).firstElement();
		
		// Bucle por todos los conceptos de la taxonoma.
		// Puede ser lento si hay muchos conceptos. Se puede acceder a un
		// concepto de forma inmediata si se conoce el QName del concepto.
		// pero en este caso vamos a aceder slo por el texto del label.
		Iterator<XMLElementDefinition> iterC = dts.getConcepts();
		while (iterC.hasNext()) {
			XMLElementDefinition el = iterC.next();
			
			/*
			 *  El puede ser de tipo item o tupla. A efectos
			 *  de buscar etiquetas no nos importa, lo que  
			 *  nos importa es que tiene relaciones de tipo
			 *  concept_label
			 */
			Iterator<XBRLRelationship> iterR = el.getFromRelationships(concept_label);
			while (iterR.hasNext()) {
				XBRLRelationship rel = iterR.next();
				
				/*
				 * rel es una relacin del tipo concept_label ahora vamos a
				 * obtener el elemento DESTINO de la relacin
				 */
				XMLFragment fr = rel.getTo();
				
				/*
				 * Dado que la relacin es de tipo concept_label, el DESTINO es un
				 * recurso de tipo XBRLLabel
				 */
				// No es necesario ya que getStringValue es el mismo mtodo para XBRLresource y para XBRLLabelResource
//				XBRLLabelResource label = (XBRLLabelResource)fr;
				
//				try {
				String labelString = fr.getStringValue();
				/* 
				 * Aqu comprobamos si coincide la equiqueta que hemos
				 * usado como parmetro con la etiqueta que tiene el concepto
				 */
				if (labelString.equals(etiqueta)) {
					// Si coincide, imprimimos el concepto, la etiqueta
					// y el XPointer de la etiqueta.
					StringBuffer sb = new StringBuffer();
					sb.append("El concepto: ");
					sb.append(el.getQName().getPrefix());
					sb.append(":");
					sb.append(el.getQName().getLocalPart());
					sb.append("\nEtiqueta encontrada: ");
					sb.append(etiqueta);
					sb.append("\nXPointer: ");
					if (fr.getXPointerShorthandScheme() != null) {
						sb.append(fr.getXPointerShorthandScheme());
					} else {
						sb.append(fr.getXPointerElementScheme());
					}
					sb.append("\n");
					
					/*
					 * Imprimimos el resultado 
					 */
					System.out.println(sb);
				}
//				} catch (XBRLSerializationException e) {
//					// Se producira si hay un error en el archivo XML.
//					// dado que se trata de taxonomas es muy poco probable
//					// que este error se produzca
//					e.printStackTrace();
//				}
			}
		}
	}
}
