/*
 *  2011 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2011 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 20/11/2011
 */
/**
 * 
 */
package com.rs.xbrl.demo.library;

import java.util.HashMap;

import com.ihr.xbrl.om.formula.XBRLFPRSCustomFunctionLibraryAbstract;
import com.ihr.xbrl.om.formula.functions.XFFunction;

import net.sf.saxon.expr.StaticProperty;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.functions.StandardFunction.Entry;
import net.sf.saxon.om.Sequence;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.trans.XPathException;
import net.sf.saxon.type.BuiltInAtomicType;
import net.sf.saxon.type.ItemType;
import net.sf.saxon.type.Type;
import net.sf.saxon.value.SequenceType;

/**
 * @author Ignacio
 *
 */
public class CustomFunctionLibrary extends XBRLFPRSCustomFunctionLibraryAbstract {


	private static final String FUNCTIONLIBRARYURI = "http://www.reportingstandarc.com/custom-library/2011";
	private static final String FUNCTIONLIBRARYPREFIX = "rs-lib";

	/* (non-Javadoc)
	 * @see com.ihr.xbrl.om.formula.XBRLRSCustomFunctionLibraryAbstract#getFunctionLibraryURI()
	 */
	@Override
	public String getFunctionLibraryURI() {
		return FUNCTIONLIBRARYURI;
	}

	/* (non-Javadoc)
	 * @see com.ihr.xbrl.om.formula.XBRLRSCustomFunctionLibraryAbstract#getFunctionLibraryPrefix()
	 */
	@Override
	public String getFunctionLibraryPrefix() {
		return FUNCTIONLIBRARYPREFIX;
	}

	/* (non-Javadoc)
	 * @see com.ihr.xbrl.om.formula.XBRLRSCustomFunctionLibraryAbstract#getFunction(java.lang.String, int)
	 */
	@Override
	public Entry getFunction(String local, int arity) {
		return ConformanceFormulasCustomFunctionsRegistry.getFunction(local, arity);
	}

	/**
	 * This class contains the implementation for the custom function PDxEV
	 * @author Ignacio
	 *
	 */
	public static class sEqualTwoContexts extends XFFunction {
		
		/* (non-Javadoc)
		 * @see net.sf.saxon.expr.Expression#iterate(net.sf.saxon.expr.XPathContext)
		 */
		@Override
		public SequenceIterator iterate(XPathContext context) throws XPathException {			
			// TODO: obtain the function description
			
			return super.iterate(context);
		}
	}

	/**
	 * Contains the custom function definitions and methods to register them and access them.
	 * 
	 * @author Ignacio
	 *
	 */
	private static abstract class ConformanceFormulasCustomFunctionsRegistry {
		
	    private static HashMap<String,Entry> functionTable = new HashMap<String,Entry>(200);

	    /**
	     * All functions are declared here
	     */
	    static {
	        // XBRL Item Types are not available
	        Entry e;
	        e = register("EqualContexts", sEqualTwoContexts.class, 0, 1, 1, BuiltInAtomicType.BOOLEAN, StaticProperty.EXACTLY_ONE);
	        	arg(e, 0, Type.NODE_TYPE, StaticProperty.ALLOWS_ONE_OR_MORE, null);
	        	
// to add another function just declare it here and implement the function expression in a Java class	        	
//	        e = register("PDxEV", FUNCTIONLIBRARYURI_PDxEV.class, 0, 2, 2, Type.NODE_TYPE, StaticProperty.ALLOWS_ONE_OR_MORE);
//	        	arg(e, 0, Type.NODE_TYPE, StaticProperty.EXACTLY_ONE, null);
//	        	arg(e, 0, Type.NODE_TYPE, StaticProperty.EXACTLY_ONE, null);
	        	
	    }
	    
	    /**
	     * Register a system function in the table of function details.
	     * @param name the function name
	     * @param implementationClass the class used to implement the function
	     * @param opcode identifies the function when a single class implements several functions
	     * @param minArguments the minimum number of arguments required
	     * @param maxArguments the maximum number of arguments allowed
	     * @param itemType the item type of the result of the function
	     * @param cardinality the cardinality of the result of the function
	     * @return the entry describing the function. The entry is incomplete, it does not yet contain information
	     * about the function arguments.
	    */
	    private static Entry register( String name,
	                                 Class<? extends XFFunction> implementationClass,
	                                 int opcode,
	                                 int minArguments,
	                                 int maxArguments,
	                                 ItemType itemType,
	                                 int cardinality ) {
	        Entry e = makeEntry(name, implementationClass, opcode, minArguments, maxArguments, itemType, cardinality);
	        functionTable.put(name, e);
	        return e;
	    }

	    /**
	     * Make a table entry describing the signature of a function, with a reference to the implementation class.
	     * @param name the function name
	     * @param implementationClass the class used to implement the function
	     * @param opcode identifies the function when a single class implements several functions
	     * @param minArguments the minimum number of arguments required
	     * @param maxArguments the maximum number of arguments allowed
	     * @param itemType the item type of the result of the function
	     * @param cardinality the cardinality of the result of the function
	     * @return the entry describing the function. The entry is incomplete, it does not yet contain information
	     * about the function arguments.
	     */
	    public static Entry makeEntry(String name, Class<? extends XFFunction> implementationClass, int opcode,
	                                   int minArguments, int maxArguments, ItemType itemType, int cardinality) {
	        Entry e = new Entry();
	        int hash = name.indexOf('#');
	        if (hash < 0) {
	            e.name = name;
	        } else {
	            e.name = name.substring(0, hash);
	        }
	        e.implementationClass = implementationClass;
	        e.opcode = opcode;
	        e.minArguments = minArguments;
	        e.maxArguments = maxArguments;
	        e.itemType = itemType;
	        e.cardinality = cardinality;
	        if (maxArguments > 100) {
	            // special case for concat()
	            e.argumentTypes = new SequenceType[1];
	            e.resultIfEmpty = new Sequence[1];
	        } else {
	            e.argumentTypes = new SequenceType[maxArguments];
	            e.resultIfEmpty = new Sequence[maxArguments];
	        }
	        return e;
	    }

	    /**
	     * Add information to a function entry about the argument types of the function
	     * @param e the entry for the function
	     * @param a the position of the argument, counting from zero
	     * @param type the item type of the argument
	     * @param cardinality the cardinality of the argument
	     * @param resultIfEmpty the value returned by the function if an empty sequence appears as the value,
	     * when this result is unaffected by any other arguments
	     */
	    // jmf public static void arg(Entry e, int a, ItemType type, int cardinality, Value<?> resultIfEmpty) {
	    public static void arg(Entry e, int a, ItemType type, int cardinality, Sequence resultIfEmpty) {
	        try {
	            e.argumentTypes[a] = SequenceType.makeSequenceType(type, cardinality);
	            e.resultIfEmpty[a] = resultIfEmpty;
	        } catch (ArrayIndexOutOfBoundsException err) {
	            System.err.println("Internal Saxon error: Can't set argument " + a + " of " + e.name);
	        } 
	    }    
		
	    /**
	     * Get the table entry for the function with a given name
	     * @param name the name of the function. This may be an unprefixed local-name for functions in the
	     * system namespace, or may use the conventional prefix "saxon:" in the case of Saxon extension functions
	     * that are specially recognized
	     * @return if the function name is known, an Entry containing information about the function. Otherwise,
	     * null
	     */
	    public static Entry getFunction(String name, int arity) {
	        // try first for an entry of the form name#arity
	        Entry e = (Entry)functionTable.get(name + '#' + arity);
	        if (e != null) {
	            return e;
	        }
	        // try for a generic entry
	        return (Entry)functionTable.get(name);
	    }
	}
}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
