/*
 *  2011 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2011 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 14/09/2011
 */
/**
 * 
 */
package com.rs.xbrl.demo.validation;

import com.ihr.xbrl.om.XBRLPlugInProcessor;
import com.ihr.xbrl.om.XBRLProcessorFactoryBase;
import com.ihr.xbrl.om.errors.XBRLValidationException;
import com.rs.lic.LicenseVerificationException;

/**
 * Factory class to create instances of the DemoValidator.
 * 
 * The Factory class name is registered to the XBRL Processor using a file called
 * RSXBRLMANIFEST/xbrlprocessors.properties
 * 
 * At the startup of Reporting Standard XBRL Processor all files in available JAR files with that name
 * are investigated for processors to be registered.
 * 
 * The factories are registered so new processors can be created when they are needed.
 * 
 * @author Ignacio
 *
 */
public class DemoValidatorFactory extends XBRLProcessorFactoryBase {

	/* (non-Javadoc)
	 * @see com.ihr.xbrl.om.XBRLProcessorFactory#create()
	 */
	@Override
	public XBRLPlugInProcessor create() throws LicenseVerificationException, XBRLValidationException {
		
		// Called by the XBRL processor when a new instance of the DemoProcessor shall be created
		return DemoValidator.create(theDTS);
	}

	/* (non-Javadoc)
	 * @see com.ihr.xbrl.om.XBRLProcessorFactory#getProcessorDescription()
	 */
	@Override
	public String getProcessorDescription() {
		// Access to the Processor description this factory can create
		return DemoValidator.ProcessorDescription;
	}

	/* (non-Javadoc)
	 * @see com.ihr.xbrl.om.XBRLProcessorFactory#getProcessorName()
	 */
	@Override
	public String getProcessorName() {
		// Access to the Processor name this factory can create
		return DemoValidator.PROCESSORNAME;
	}

}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
